{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_802dot11header;

interface

uses base_structures,
     definition_ip_addresses,
     class_parser_802dot11header,
     class_assembly_baseclass;


type t802dot11headerAssembler = class (tFrameAssembler)

                                constructor create (const IsOverCAPWAP : boolean = false);

                                private
                                       priv_Version          : byte;
                                       priv_FrameType        : t802dot11FrameType;
                                       priv_FrameSubType     : t802dot11FrameSubType;
                                       priv_HeaderFlags      : t802dot11HeaderFCFFlagSet;
                                       priv_Duration         : word;

                                       priv_receiveraddr     : tMACAddress;
                                       priv_destinationaddr  : tMACAddress;
                                       priv_transmitteraddr  : tMACAddress;
                                       priv_BSSId            : tMACAddress;
                                       priv_sourceaddr       : tMACAddress;
                                       priv_FragmentNumber   : word;
                                       priv_SequenceNumber   : word;

                                       priv_isOverCAPWAP     : boolean;

                                       procedure setHeaderFlags (const value : t802dot11HeaderFCFFlagSet);
                                       procedure setFrameSubType (const value : t802dot11FrameSubType);

                                       procedure AssembleManagementValues;
                                       procedure AssembleControlValues;
                                       procedure AssembleDataValues;

                                       function getFrameTypeValue : byte;
                                       function getSubTypeValue : byte;
                                       function getFlags : byte;

                                protected
                                         // called from base class from GetAssembly

                                         procedure Assemble; override;
                                         procedure DetermineAssemblySize; override;

                                public
                                      property Version : byte read priv_version write priv_Version;
                                      property FrameType : t802dot11FrameType read priv_FrameType write priv_FrameType;

                                      property FrameSubType : t802dot11FrameSubType read priv_FrameSubType write priv_FrameSubType;
                                      property HeaderFlags : t802dot11HeaderFCFFlagSet read priv_HeaderFlags write setHeaderFlags;
                                      property Duration : word read priv_Duration write priv_Duration;

                                      property ReceiverAddress : tMACAddress read priv_receiveraddr write priv_receiveraddr;
                                      property DestinationAddress : tMACAddress read priv_destinationaddr write priv_destinationaddr;
                                      property TransmitterAddress : tMACAddress read priv_transmitteraddr write priv_transmitteraddr;
                                      property SourceAddress : tMACAddress read priv_sourceaddr write priv_sourceaddr;
                                      property BSSId : tMACAddress read priv_BSSId write priv_BSSId;
                                      property FragmentNumber : word read priv_fragmentnumber write priv_fragmentnumber;
                                      property SequenceNumber : word read priv_SequenceNumber write priv_SequenceNumber;

                                      procedure CopyFromParser (Parser : t802dot11headerParser);
                           end;


implementation

uses SysUtils,
     toolkit;


(* ------------------------------------------- *)
{ t802dot11headerAssembler }
(* ------------------------------------------- *)

constructor t802dot11headerAssembler.create (const IsOverCAPWAP : boolean = false);

begin
     inherited create;

     priv_Version := 0;
     priv_FrameType := ft802None;
     priv_FrameSubType := mgmtftNone;
     priv_HeaderFlags := [];
     priv_Duration := 0;

     priv_receiveraddr.initialize;
     priv_destinationaddr.initialize;
     priv_transmitteraddr.initialize;
     priv_BSSId.initialize;
     priv_sourceaddr.initialize;
     priv_FragmentNumber := 0;
     priv_SequenceNumber := 0;

     priv_isOverCAPWAP := IsOverCAPWAP;
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.DetermineAssemblySize;

begin
     case priv_FrameType of
      ft802None       : begin
                             prot_length := 0;
                        end;
      ft802Management : begin
                             prot_length := sizeOf (t802dot11MgmtDataHeaderStruct);
                        end;
      ft802Control    : begin
                             prot_length := sizeOf (t802dot11CtrlHeaderStruct);
                        end;
      ft802Data       : begin
                             prot_length := sizeOf (t802dot11MgmtDataHeaderStruct);
                        end;
     end;
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.setHeaderFlags (const value: t802dot11HeaderFCFFlagSet);

var flag  : t802dot11HeaderFCFFlag;
    count : integer;

begin
     // Make sure only one directional flag is present

     count := 0;

     for flag in value do
         if flag in [fcfIBSS, fcfToAP, fcfFromAP, fcfWDS] then
            inc (count);

     assert (count = 1, 'Invalid flag set; must have exactly one directional flag (t802dot11headerAssembler.setHeaderFlags)');

     priv_HeaderFlags := value;
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.setFrameSubType (const value: t802dot11FrameSubType);

begin
     assert (not (priv_frametype = ft802None), 'Cannot set 802.11 frame subtype for undefined frame type (t802dot11headerAssembler.setSubType)');

     case priv_FrameType of
      ft802Management : if value in [mgmtftNone, mgmtftAssocReq, mgmtftAssocResp, mgmtftReassocReq, mgmtftReassocResp, mgmtftProbeReq, mgmtftProbeResp, mgmtftBeacon, mgmtftATIM, mgmtftDisassoc, mgmtftAuth, mgmtftDeAuth, mgmtftAction, mgmtftActionNoAck] then
                           priv_FrameSubType := value
                        else
                            raise Exception.Create('Invalid 802.11 subtype (t802dot11headerAssembler.setSubType)');
      ft802Control    : if value in [ctrlftNone, ctrlWrapper, ctrlftBlockAckReq, ctrlftBlockAck, ctrlftPSPoll, ctrlftRTS, ctrlftCTS, ctrlftACK, ctrlftCFend, ctrlftCFEndAck] then
                           priv_FrameSubType := value
                        else
                            raise Exception.Create('Invalid 802.11 subtype (t802dot11headerAssembler.setSubType)');
      ft802Data       : if value in [dftNone, dftData, dftDataCFAck, dftDataCFPoll, dftDataCFAckCFPoll, dftNull, dftCFAck, dftCFPoll, dftCFAckCFPoll, dftQoSData, dftQoSDataCFAck, dftQoSDataCFPoll, dftQoSCFAckCFPoll, dftQoSNull, dftQoSCFPollNoData, dftQoSCFAckNoData] then
                           priv_FrameSubType := value
                        else
                            raise Exception.Create('Invalid 802.11 subtype (t802dot11headerAssembler.setSubType)');
     end;
end;

(* ------------------------------------------- *)

function t802dot11headerAssembler.getFrameTypeValue : byte;

begin
     case priv_frametype of
      ft802Management : result := 0;
      ft802Control    : result := 1;
      ft802Data       : result := 2;
     end;
end;

(* ------------------------------------------- *)

function t802dot11headerAssembler.getSubTypeValue : byte;

var help : byte;

begin
     assert (not (priv_frametype = ft802None), 'Cannot set 802.11 frame subtype for undefined frame type (t802dot11headerAssembler.setSubType)');

     case priv_FrameType of
      ft802Management : begin
                             case priv_FrameSubType of
                              mgmtftAssocReq     : help := 0;
                              mgmtftAssocResp    : help := 1;
                              mgmtftReassocReq   : help := 2;
                              mgmtftReassocResp  : help := 3;
                              mgmtftProbeReq     : help := 4;
                              mgmtftProbeResp    : help := 5;
                              mgmtftBeacon       : help := 8;
                              mgmtftATIM         : help := 9;
                              mgmtftDisassoc     : help := 10;
                              mgmtftAuth         : help := 11;
                              mgmtftDeAuth       : help := 12;
                              mgmtftAction       : help := 13;
                              mgmtftActionNoAck  : help := 14;
                             else
                                 raise Exception.Create('Invalid 802.11 subtype (t802dot11headerAssembler.setSubType)');
                             end
                        end;
      ft802Control    : begin
                             case priv_FrameSubType of
                              ctrlftNone         : help := 0;
                              ctrlWrapper        : help := 7;
                              ctrlftBlockAckReq  : help := 8;
                              ctrlftBlockAck     : help := 9;
                              ctrlftPSPoll       : help := 10;
                              ctrlftRTS          : help := 11;
                              ctrlftCTS          : help := 12;
                              ctrlftACK          : help := 13;
                              ctrlftCFend        : help := 14;
                              ctrlftCFEndAck     : help := 15;
                             else
                                 raise Exception.Create('Invalid 802.11 subtype (t802dot11headerAssembler.setSubType)');
                             end;
                        end;
      ft802Data       : begin
                             case priv_FrameSubType of
                              dftData            : help := 0;
                              dftDataCFAck       : help := 1;
                              dftDataCFPoll      : help := 2;
                              dftDataCFAckCFPoll : help := 3;
                              dftNull            : help := 4;
                              dftCFAck           : help := 5;
                              dftCFPoll          : help := 6;
                              dftCFAckCFPoll     : help := 7;
                              dftQoSData         : help := 8;
                              dftQoSDataCFAck    : help := 9;
                              dftQoSDataCFPoll   : help := 10;
                              dftQoSCFAckCFPoll  : help := 11;
                              dftQoSNull         : help := 12;
                              dftQoSCFPollNoData : help := 14;
                              dftQoSCFAckNoData  : help := 15;
                             else
                                 raise Exception.Create('Invalid 802.11 subtype (t802dot11headerAssembler.setSubType)');
                             end;
                        end;
     end;

     result := help;
end;

(* ------------------------------------------- *)

function t802dot11headerAssembler.getFlags : byte;

var direction : byte;
    other     : byte;
    flag  : t802dot11HeaderFCFFlag;

begin
     other := 0;

     for flag in priv_HeaderFlags do
         begin
              case flag of
               fcfIBSS          : direction := 0;
               fcfToAP          : direction := 1;
               fcfFromAP        : direction := 2;
               fcfWDS           : direction := 3;
               fcfMoreFragments : other := (other or 2);
               fcfRetry         : other := (other or 4);
               fcfPwrMgmt       : other := (other or 8);
               fcfMoreData      : other := (other or 16);
               fcfProtected     : other := (other or 32);
               fcfOrdered       : other := (other or 64);
              end;
         end;

     result := (direction or other shl 2);
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.AssembleControlValues;

var pData : p802dot11CtrlHeaderStruct;
    bHelp : byte;

begin
     pData := p802dot11CtrlHeaderStruct (prot_data);

     // Construct Version, Type and Subtype byte

     bHelp := priv_Version;
     bHelp := bHelp or (getFrameTypeValue shl 2);
     bHelp := bHelp or (getSubTypeValue shl 4);

     if priv_isOverCAPWAP then
        begin
             // FCS bytes are swapped (recreating the buggy implementation of CAPWAP)

             pData^.HeaderFrameControl := getFlags;
             pData^.HeaderFCFFlags := bHelp;
        end
     else
         begin
             pData^.HeaderFrameControl := bHelp;
             pData^.HeaderFCFFlags := getFlags;
         end;

     pData^.HeaderDuration := priv_Duration;
     pData^.ReceiverAddress1 := priv_receiveraddr.RangeBytes;

     assert (sizeOf (t802dot11CtrlHeaderStruct) <= prot_length, 'Memory offset check mismatch (t802dot11ManagementAssembler.AssembleControlValues)')
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.AssembleDataValues;

var pData : p802dot11MgmtDataHeaderStruct;
    bHelp : byte;

begin
     pData := p802dot11MgmtDataHeaderStruct (prot_data);

     // Construct Version, Type and Subtype byte

     bHelp := priv_Version;
     bHelp := bHelp or (getFrameTypeValue shl 2);
     bHelp := bHelp or (getSubTypeValue shl 4);

     if priv_isOverCAPWAP then
        begin
             // FCS bytes are swapped (recreating the buggy implementation of CAPWAP)

             pData^.HeaderFrameControl := getFlags;
             pData^.HeaderFCFFlags := bHelp;
        end
     else
         begin
             pData^.HeaderFrameControl := bHelp;
             pData^.HeaderFCFFlags := getFlags;
         end;

     pData^.HeaderDuration := priv_Duration;
     pData^.HeaderAddress1 := priv_receiveraddr.RangeBytes;
     pData^.HeaderAddress2 := priv_transmitteraddr.RangeBytes;
     pData^.HeaderAddress3 := priv_sourceaddr.RangeBytes;
     pData^.HeaderFragmentSeq := (priv_FragmentNumber or (priv_SequenceNumber shl 4));

     assert (sizeOf (t802dot11MgmtDataHeaderStruct) <= prot_length, 'Memory offset check mismatch (t802dot11ManagementAssembler.AssembleDataValues)')
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.AssembleManagementValues;

var pData : p802dot11MgmtDataHeaderStruct;
    bHelp : byte;

begin
     pData := p802dot11MgmtDataHeaderStruct (prot_data);

     // Construct Version, Type and Subtype byte

     bHelp := priv_Version;
     bHelp := bHelp or (getFrameTypeValue shl 2);
     bHelp := bHelp or (getSubTypeValue shl 4);

     if priv_isOverCAPWAP then
        begin
             // FCS bytes are swapped (recreating the buggy implementation of CAPWAP)

             pData^.HeaderFrameControl := getFlags;
             pData^.HeaderFCFFlags := bHelp;
        end
     else
         begin
             pData^.HeaderFrameControl := bHelp;
             pData^.HeaderFCFFlags := getFlags;
         end;

     pData^.HeaderDuration := priv_Duration;
     pData^.HeaderAddress1 := priv_receiveraddr.RangeBytes;
     pData^.HeaderAddress2 := priv_transmitteraddr.RangeBytes;
     pData^.HeaderAddress3 := priv_sourceaddr.RangeBytes;
     pData^.HeaderFragmentSeq := (priv_FragmentNumber or (priv_SequenceNumber shl 4));

     assert (sizeOf (t802dot11MgmtDataHeaderStruct) <= prot_length, 'Memory offset check mismatch (t802dot11ManagementAssembler.AssembleManagementValues)')
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.Assemble;

var pData : pByte;
    bHelp : byte;

begin
     assert (not (priv_frametype = ft802None), 'Invalid 802.11 frame type (t802dot11headerAssembler.Assemble)');

     case priv_frametype of
      ft802Management : AssembleManagementValues;
      ft802Control    : AssembleControlValues;
      ft802Data       : AssembleDataValues;
     end;
end;

(* ------------------------------------------- *)

procedure t802dot11headerAssembler.CopyFromParser (Parser : t802dot11headerParser);

begin
     priv_Version := Parser.Version;
     priv_FrameType := Parser.FrameType;
     priv_FrameSubType := Parser.FrameSubType;
     priv_HeaderFlags := Parser.HeaderFlags;
     priv_Duration := Parser.Duration;

     priv_receiveraddr := Parser.ReceiverAddress;
     priv_destinationaddr := Parser.DestinationAddress;
     priv_transmitteraddr := Parser.TransmitterAddress;
     priv_BSSId  := Parser.BSSId;
     priv_sourceaddr := Parser.SourceAddress;
     priv_FragmentNumber := Parser.FragmentNumber;
     priv_SequenceNumber := Parser.SequenceNumber;
end;

(* ------------------------------------------- *)

end.
