{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_802dot11management;

interface

uses base_structures,
     class_parser_802dot11management,
     class_assembly_baseclass;


type t802dot11ManagementAssembler = class (tFrameAssembler)

                                          constructor create (const SubType : t802dot11FrameSubType; WriteUnknownParams : boolean = false);

                                          private
                                                 priv_subtype            : t802dot11FrameSubType;

                                                 priv_Timestamp          : uint64;
                                                 priv_BeaconInterval     : word;                                 // in milliseconds, usually 100
                                                 priv_Capabilities       : t802dot11MgmtCapabilitiesSet;

                                                 priv_TaggedParams       : t802dot11TaggedParamList;

                                                 priv_WriteUnknownParams : boolean;

                                                 function CapabilitiesToWord : word;

                                                 /// <summary>
                                                 ///   Assembles a beacon frame
                                                 ///   structure and returns
                                                 ///   the offset for any
                                                 ///   following tagged
                                                 ///   parameters
                                                 /// </summary>
                                                 function AssembleBeaconFrame : integer;
                                                 /// <summary>
                                                 ///   Assembles the tagged
                                                 ///   parameter set starting
                                                 ///   at the given offset.
                                                 ///   Returns the offset of
                                                 ///   the next byte after the
                                                 ///   set.
                                                 /// </summary>
                                                 function AssembleTaggedParameters (StartOffset : integer) : integer;

                                          protected
                                                   // called from base class from GetAssembly

                                                   procedure DetermineAssemblySize; override;
                                                   procedure Assemble; override;

                                          public
                                                property SubType : t802dot11FrameSubType read priv_SubType;
                                                property Timestamp : uint64 read priv_Timestamp write priv_Timestamp;
                                                property BeaconInterval : word read priv_BeaconInterval write priv_BeaconInterval;
                                                property Capabilities : t802dot11MgmtCapabilitiesSet read priv_Capabilities write priv_Capabilities;
                                                property TaggedParameters : t802dot11TaggedParamList read priv_TaggedParams write priv_TaggedParams;

                                                procedure AddTaggedParameter (const Parameter : t802dot11TaggedParam);

                                                procedure CopyFromParser (Parser : t802dot11ManagementParser; const CopyUnknownParameters : boolean = true);
                                     end;


implementation

uses toolkit,
     common_assembly_helpers;

(* ------------------------------------------- *)
{ t802dot11ManagementAssembler }
(* ------------------------------------------- *)

constructor t802dot11ManagementAssembler.create (const SubType : t802dot11FrameSubType; WriteUnknownParams : boolean = false);

begin
     inherited create;

     priv_SubType := SubType;
     priv_WriteUnknownParams := WriteUnknownParams;

     priv_Timestamp := 0;
     priv_BeaconInterval := 0;
     priv_Capabilities := [];

     setLength (priv_TaggedParams, 0);
end;

(* ------------------------------------------- *)

procedure t802dot11ManagementAssembler.DetermineAssemblySize;

var i : integer;

begin
     case priv_SubType of
      mgmtftAssocReq    : prot_length := 0;
      mgmtftAssocResp   : prot_length := 0;
      mgmtftReassocReq  : prot_length := 0;
      mgmtftReassocResp : prot_length := 0;
      mgmtftProbeReq    : prot_length := 0;
      mgmtftProbeResp   : prot_length := 0;
      mgmtftBeacon      : prot_length := sizeOf (tBeaconFixedParams);
      mgmtftATIM        : prot_length := 0;
      mgmtftDisassoc    : prot_length := 0;
      mgmtftAuth        : prot_length := 0;
      mgmtftDeAuth      : prot_length := 0;
      mgmtftAction      : prot_length := 0;
      mgmtftActionNoAck : prot_length := 0;
     end;

     assert (prot_length > 0, 'Unknown subtype, cannot assemble (t802dot11ManagementAssembler.DetermineAssemblySize)');

     // add tagged parameter length, plus type and len byte
     
     for i := 1 to length (priv_Taggedparams) do
         prot_length := prot_length + priv_Taggedparams[i-1].ParamLen + 2;        
end;

(* ------------------------------------------- *)

procedure t802dot11ManagementAssembler.AddTaggedParameter (const Parameter: t802dot11TaggedParam);

var len : integer;

begin
     len := length (priv_TaggedParams);
     setLength (priv_TaggedParams, len + 1);
     priv_TaggedParams[len] := Parameter;
end;

(* ------------------------------------------- *)

function t802dot11ManagementAssembler.CapabilitiesToWord : word;

var flag  : t802dot11MgmtCapability;
    value : word;

begin
     value := 0;

     for flag in priv_Capabilities do
         begin
              case flag of
               capESS                : value := (value or 1);
               capIBSS               : value := (value or 2);
               capCFPollable         : value := (value or 4);
               capCFPollRequest      : value := (value or 8);
               capPrivacy            : value := (value or 16);
               capShortPreamble      : value := (value or 32);
               capPBCC               : value := (value or 64);
               capChannelAgility     : value := (value or 128);
               capSpectrumManagement : value := (value or 256);
               // bit 9 is reserved  : value := (value or 512);
               capShortSlotTime      : value := (value or 1024);
               capAutoPowerSafe      : value := (value or 2048);
               capRadioMeasurement   : value := (value or 4096);
               capDSSSOFDM           : value := (value or 8192);
               capDelayedBlockAck    : value := (value or 16384);
               capImmediateBlockAck  : value := (value or 32768);
              end;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function t802dot11ManagementAssembler.AssembleBeaconFrame : integer;

var pData : pBeaconFixedParams;

begin
     pData := pBeaconFixedParams (prot_data);

     pData^.Timestamp := priv_timestamp;
     pData^.BeaconInterval := priv_beaconinterval;
     pData^.Capabilities := CapabilitiesToWord;

     result := sizeOf (tBeaconFixedParams);
end;

(* ------------------------------------------- *)

function t802dot11ManagementAssembler.AssembleTaggedParameters (StartOffset : integer) : integer;

var localoffset : integer;
    writeoffset : integer;
    taggedparam : t802dot11TaggedParam;
    i,j         : integer;

begin
     localoffset := 0;

     for i := 1 to length (priv_TaggedParams) do
         begin
              taggedparam := priv_TaggedParams[i-1];
              writeoffset := StartOffset + LocalOffset;

              // Write type and length

              pByte (prot_data + WriteOffset)^ := taggedparam.ParamType;
              pByte (prot_data + WriteOffset + 1)^ := taggedparam.ParamLen;

              inc (WriteOffset, 2);

              // write value

              case taggedparam.ParamType of
               cSSIDParameterSet       : begin
                                              assert (TaggedParam.ParamLen = length (TaggedParam.SSID), 'TaggedParameter length check mismatch (t802dot11ManagementAssembler.AssembleTaggedParameters)');

                                              WriteStringToOffset (prot_Data, TaggedParam.SSID, WriteOffset);
                                         end;
               cSupportedRates         : begin
                                              assert (TaggedParam.ParamLen = length (TaggedParam.SupportedRates), 'TaggedParameter length check mismatch (t802dot11ManagementAssembler.AssembleTaggedParameters)');

                                              for j := 0 to (length (TaggedParam.SupportedRates) - 1) do
                                                  pByte (prot_Data + WriteOffset + j)^ := TaggedParam.SupportedRates[j];
                                         end;
               cDSParameterSet         : begin
                                              pByte (prot_Data + WriteOffset)^ := TaggedParam.CurrentChannel;
                                         end;
               cTrafficIndicationMap   : begin
                                              p802dot11TrafficIndicationMap (prot_Data + WriteOffset)^ := TaggedParam.TrafficIndicationMap;
                                         end;
               cCountryInformation     : begin
                                              p802dot11CountryInformation (prot_Data + WriteOffset)^ := TaggedParam.CountryInformation;
                                         end;
               cERPInformation         : begin
                                              pByte (prot_Data + WriteOffset)^ := TaggedParam.ERPInformation;
                                         end;
               cExtendedSupportedRates : begin
                                              assert (TaggedParam.ParamLen = length (TaggedParam.ExtendedSupportedRates), 'TaggedParameter length check mismatch (t802dot11ManagementAssembler.AssembleTaggedParameters)');

                                              for j := 0 to (length (TaggedParam.ExtendedSupportedRates) - 1) do
                                                  pByte (prot_Data + WriteOffset + j)^ := TaggedParam.ExtendedSupportedRates[j];
                                         end;
              else
                  if priv_WriteUnknownParams then
                     begin
                          assert (TaggedParam.ParamLen = length (TaggedParam.UnknownParam), 'TaggedParameter length check mismatch (t802dot11ManagementAssembler.AssembleTaggedParameters)');

                          for j := 0 to (length (TaggedParam.UnknownParam) - 1) do
                              pByte (prot_Data + WriteOffset + j)^ := TaggedParam.UnknownParam[j];
                     end;
              end;

              LocalOffset := LocalOffset + 2 + TaggedParam.ParamLen;
         end;

     assert (Startoffset + localoffset = prot_length, 'Memory offset check mismatch (t802dot11ManagementAssembler.AssembleTaggedParameters)');

     result := Startoffset + localoffset;
end;

(* ------------------------------------------- *)

procedure t802dot11ManagementAssembler.Assemble;

var Offset : integer;

begin
     case priv_SubType of
      mgmtftAssocReq    : ;
      mgmtftAssocResp   : ;
      mgmtftReassocReq  : ;
      mgmtftReassocResp : ;
      mgmtftProbeReq    : ;
      mgmtftProbeResp   : ;
      mgmtftBeacon      : Offset := AssembleBeaconFrame;
      mgmtftATIM        : ;
      mgmtftDisassoc    : ;
      mgmtftAuth        : ;
      mgmtftDeAuth      : ;
      mgmtftAction      : ;
      mgmtftActionNoAck : ;
     end;

     AssembleTaggedParameters (Offset);
end;

(* ------------------------------------------- *)

procedure t802dot11ManagementAssembler.CopyFromParser (Parser: t802dot11ManagementParser; const CopyUnknownParameters : boolean = true);

var i : integer;

begin
     priv_Timestamp := Parser.Timestamp;
     priv_BeaconInterval := Parser.BeaconInterval;
     priv_Capabilities := Parser.Capabilities;

     for i := 1 to length (Parser.TaggedParameters) do
         begin
              if (not Parser.TaggedParameters[i-1].isUnknown) or CopyUnknownParameters then
                 AddTaggedParameter (Parser.TaggedParameters[i-1]);
         end;
end;

(* ------------------------------------------- *)

end.
