{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_icmpv4;

interface

uses base_structures,
     base_checksums,
     definition_ip_addresses,
     class_parser_icmpv4,
     class_assembly_baseclass;


type tICMPv4Assembler = class (tFrameAssembler)

                              constructor create (QuotedPayload: tFrameChunk); overload;
                              constructor create; overload;

                              Destructor Destroy; override;

                              private
                                     priv_type          : byte;
                                     priv_code          : byte;
                                     priv_checksum      : word;
                                     priv_payload       : tFrameChunk;
                                     priv_payloadoffset : integer;
                                     priv_keepcrc       : boolean;
                                     priv_forcebadcrc   : boolean;

                                     // Optional values

                                     priv_echo_identifier        : word;                   // stored in Big Endian
                                     priv_echo_sequence          : word;                   // stored in Big Endian
                                     priv_next_hop_mtu           : word;
                                     priv_redirect_gwy           : tIPv4Address;
                                     priv_OriginalDatagramLength : word;

                                     // Internal values

                                     procedure setChecksum (value : word);

                                     procedure setPayloadFromFrameChunk (Chunk : tFrameChunk);

                              protected
                                       // called from base class from GetAssembly

                                       procedure DetermineAssemblySize; override;
                                       procedure Assemble; override;

                              public
                                    property ICMPType : byte read priv_type write priv_type;
                                    property ICMPCode : byte read priv_code write priv_code;
                                    property Checksum : word read priv_checksum write setChecksum;
                                    property EchoIdentifier : word read priv_echo_identifier write priv_echo_identifier;
                                    property EchoSequence : word read priv_echo_sequence write priv_echo_sequence;
                                    property NextHopMTU : word read priv_next_hop_mtu write priv_next_hop_mtu;
                                    property RedirectGwy : tIPv4Address read priv_redirect_gwy write priv_redirect_gwy;
                                    property OriginalDatagramLength : word read priv_OriginalDatagramLength write priv_OriginalDatagramLength;
                                    property Payload  : tFrameChunk read priv_payload write setPayloadFromFrameChunk;
                                    property ForceBadCRC : boolean read priv_forcebadcrc write priv_forcebadcrc;

                                    procedure setPayloadFromRawbuffer (buffer : tRawBuffer; Len : integer);
                                    procedure clearPayload;

                                    procedure CopyFromParser (Parser : tICMPv4Parser);
                        end;


implementation

uses SysUtils,
     toolkit;

(* ------------------------------------------- *)
{ tICMPv4Assembler }
(* ------------------------------------------- *)

constructor tICMPv4Assembler.create;

begin
     inherited;

     priv_echo_identifier := 0;
     priv_echo_sequence := 0;
     priv_next_hop_mtu := 0;
     priv_redirect_gwy.initialize;
     priv_OriginalDatagramLength := 0;

     priv_payload.initialize;
end;

(* ------------------------------------------- *)

constructor tICMPv4Assembler.create (QuotedPayload: tFrameChunk);

begin
     inherited create;

     priv_echo_identifier := 0;
     priv_echo_sequence := 0;
     priv_next_hop_mtu := 0;
     priv_redirect_gwy.initialize;
     priv_OriginalDatagramLength := 0;

     priv_payload.initialize;
     priv_payload.copyFrom (QuotedPayload);
end;

(* ------------------------------------------- *)

destructor tICMPv4Assembler.Destroy;

begin
     if priv_payload.Len > 0 then
        priv_payload.free;

     inherited;
end;

(* ------------------------------------------- *)

procedure tICMPv4Assembler.DetermineAssemblySize;

begin
     case priv_type of
       0,
       8 : prot_length := sizeOf (tICMPv4Struct) + 4 + priv_payload.Len;
       3,
       5,
      11 : prot_length := sizeOf (tICMPv4Struct) + 4;
     else
         prot_length := sizeOf (tICMPv4Struct);
     end;
end;

(* ------------------------------------------- *)

procedure tICMPv4Assembler.setChecksum (value: word);

begin
     priv_checksum := value;
     priv_keepcrc := true;
end;

(* ------------------------------------------- *)

procedure tICMPv4Assembler.setPayloadFromFrameChunk (Chunk: tFrameChunk);

begin
     if chunk.Len > 0 then
        priv_payload.copyFrom (Chunk)
     else
         priv_payload.len := 0;
end;

(* ------------------------------------------- *)

procedure tICMPv4Assembler.setPayloadFromRawbuffer (buffer: tRawBuffer; Len: integer);

begin
     if len > 0 then
        priv_payload.copyFrom (buffer, len)
     else
         priv_payload.Len := 0;
end;

(* ------------------------------------------- *)

procedure tICMPv4Assembler.Assemble;

var pData       : pICMPv4Struct;
    pEchoParams : pEchoStruct;
    HelpLong    : pLongword;
    HelpWord    : pWord;
    HelpByte    : pByte;
    Help        : word;
    hdrBlock    : tFrameChunk;
    crcbuf      : tBufferWords;
    payloadbuf  : tBufferWords;
    i           : integer;
    offset      : integer;

begin
     pData := pICMPv4Struct (prot_data);

     pData^.icmp_type := priv_type;
     pData^.icmp_code := priv_code;

     offset := sizeOf (tICMPv4Struct);

     case priv_type of
       0,
       8 : begin
                // Echo Request/Reply

                pEchoParams := pEchoStruct (prot_data + sizeOf (tICMPv4Struct));

                pEchoParams^.echo_identifier := Swap16 (priv_echo_identifier);
                pEchoParams^.echo_sequence := Swap16 (priv_echo_sequence);

                // Copy Echo Payload

                offset := offset + sizeOf (tEchoStruct);

                if priv_payload.Len > 0 then
                   begin
                        HelpByte := pByte (prot_data + offset);

                        move (priv_payload.Data^, HelpByte^, priv_payload.Len);

                        offset := offset + priv_payload.Len;
                   end;
           end;
       3 : begin
                // Destination Unreachable - check if we need to add MTU for "Fragmentation Needed"

                if priv_code = 4 then
                   begin
                        inc (offset, 2);

                        HelpWord := pWord (prot_data + offset);
                        HelpWord^ := Swap16 (priv_next_hop_mtu);

                        inc (offset, 2);
                   end
                else
                    inc (offset, 4);
           end;
       5 : begin
                // Redirect

                HelpLong := pLongword (prot_data + offset);
                HelpLong^ := Swap32 (priv_redirect_gwy.ValueAsLong);

                inc (offset, 4);
           end;
      11 : begin
                // TTL exceeded, as extended in rfc4884

                inc (offset);

                HelpByte := pByte (prot_data + offset);
                Help := priv_OriginalDatagramLength div 4;

                HelpByte^:= Help;

                inc (offset);

                HelpWord := pWord (prot_data + offset);
                HelpWord^ := 0;

                inc (offset, 2);
           end;
     end;

     assert (offset = prot_length, 'Memory offset check mismatch (tICMPv4Assembler.Assemble)');

     // Calculate the checksum

     if not priv_keepcrc then
        begin
             hdrBlock.Data := prot_data;
             hdrBlock.Len := prot_length;

             crcbuf := hdrBlock.getAsBufferWords;

             // For Echo Request/Reply, the payload is part of ICMP and contained in hdrBlock, otherwise its a quote

             if (priv_type = 0) or (priv_type = 8) then
                setLength (payloadbuf, 0)
             else
                 payloadbuf := priv_payload.getAsBufferWords;

             priv_checksum := getICMPv4Checksum (crcbuf, payloadbuf);

             if priv_forcebadcrc then
                priv_checksum := getBadCRC (priv_checksum);
        end;

     pData^.Checksum := swap16 (priv_checksum);
end;

(* ------------------------------------------- *)

procedure tICMPv4Assembler.clearPayload;

var len : integer;

begin
     // replace payload with zero buffer

     if priv_payload.Len > 0 then
        begin
             len := priv_payload.Len;
             priv_payload.free;

             priv_payload.CreateZeroBuffer(len);
        end;
end;

(* ------------------------------------------- *)

procedure tICMPv4Assembler.CopyFromParser (Parser: tICMPv4Parser);

begin
     priv_type := Parser.ICMPType;
     priv_code := Parser.ICMPCode;
     priv_checksum := Parser.Checksum;

     // Optional values

     case priv_type of
       0,
       8 : begin
                priv_echo_identifier := Parser.EchoIdentifier;
                priv_echo_sequence := Parser.EchoSequence;

                setPayloadFromRawbuffer (Parser.Payload, Parser.PayloadLength);
           end;
       3 : begin
                priv_next_hop_mtu := Parser.NextHopMTU;
           end;
       5 : begin
                priv_redirect_gwy := Parser.RedirectGwy;
           end;
      11 : begin
                priv_OriginalDatagramLength := Parser.OriginalDatagramLength;
           end;
     end;
end;

(* ------------------------------------------- *)

end.
