{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_ipv4;

interface

uses base_structures,
     base_checksums,
     definition_ip_addresses,
     class_parser_ipv4,
     class_assembly_baseclass;

type tIPv4Assembler  = class (tFrameAssembler)

                             {$REGION 'Documentation'}
                             ///	<param name="ForceTotalLength">
                             ///	  In case of a truncated packet this optional
                             ///	  parameter holds the length that will be
                             ///	  written as IP total length to the IP header
                             ///	</param>
                             {$ENDREGION}
                             constructor create (const PayloadLength : integer; const ForceTotalLength : integer = 0);

                             private
                                    priv_ihl             : byte;    // just for internal use
                                    priv_versionihl      : byte;
                                    priv_tos             : byte;
                                    priv_totallength     : word;
                                    priv_payloadlen      : integer; // from create, tells us what the chunk of data is that we're transporting
                                    priv_forcedtotallen  : word;
                                    priv_identification  : word;
                                    priv_flagsandoffset  : word;
                                    priv_ttl             : byte;
                                    priv_protocol        : byte;
                                    priv_checksum        : word;
                                    priv_source          : tIPv4Address;
                                    priv_destination     : tIPv4Address;
                                    priv_currentroute    : tIPv4Address;
                                    priv_optionlist      : tIPv4OptionList;
                                    priv_keepcrc         : boolean;
                                    priv_forcebadcrc     : boolean;

                                    procedure SetIPv4Flags (const Value: tIPFlagSet);
                                    procedure SetIPv4Offset (const Value: word);
                                    procedure setIHL (const Value: byte);
                                    procedure setVersion (const Value: byte);
                                    procedure setManualCRC(const Value: word);

                                    procedure writeSrcRouteOption (const opt : tIPv4Option; var Offset : integer);
                                    procedure writeTimestampOption (const opt : tIPv4Option; var Offset : integer);
                                    procedure writeRecordRouteOption (const opt : tIPv4Option; var Offset : integer);
                                    procedure writeRouterAlertOption (const opt : tIPv4Option; var Offset : integer);
                                    procedure writeIPv4OptionList (OptionList : tIPv4OptionList; Offset : integer);


                             protected
                                      // called from base class from GetAssembly

                                      procedure DetermineAssemblySize; override;
                                      procedure Assemble; override;

                             public
                                   property Version : byte write setVersion;
                                   property IHL : byte read priv_IHL;
                                   property TOS : byte write priv_tos;
                                   property TotalLength : word write priv_forcedtotallen;           // If this is set, we need to skip calculation and force it instead
                                   property Identification : word write priv_identification;
                                   property Flags : tIPFlagSet write SetIPv4Flags;
                                   property Offset : word write SetIPv4Offset;
                                   property TTL : byte write priv_ttl;
                                   property Protocol : byte write priv_protocol;
                                   property Checksum : word write setManualCRC;
                                   property Source : tIPv4Address write priv_source;
                                   property Destination : tIPv4Address write priv_destination;
                                   property OptionList : tIPv4OptionList read priv_optionlist write priv_optionlist;
                                   property CurrentRoute : tIPv4Address read priv_currentroute write priv_currentroute;

                                   property DoNotRecalcuateCRC : boolean read priv_keepcrc write priv_keepcrc;
                                   property ForceBadCRC : boolean read priv_forcebadcrc write priv_forcebadcrc;

                                   procedure AddOptionRecord (const Value : tIPv4Option);

                                   {$REGION 'Documentation'}
                                   ///	<summary>
                                   ///	  copies all values from a parser object
                                   ///	  to the assembler, including the
                                   ///	  existing checksum
                                   ///	</summary>
                                   {$ENDREGION}
                                   procedure CopyFromParser (Parser : tIPv4Parser);
                       end;


implementation

uses Toolkit;

type pIPv4OptionTLP = ^tIPv4OptionTLP;
     tIPv4OptionTLP = record
                            optiontype : byte;
                            optionlen  : byte;
                            optionptr  : byte;
                      end;

(* ------------------------------------------- *)
{ tEthernetAssembler }
(* ------------------------------------------- *)

constructor tIPv4Assembler.create (const PayloadLength : integer; const ForceTotalLength : integer = 0);

begin
     inherited create;

     priv_ihl := 20;
     priv_versionihl := $45;
     priv_tos := 0;
     priv_forcedtotallen := ForceTotalLength;
     priv_payloadlen := PayloadLength;
     priv_identification := 0;
     priv_flagsandoffset := 0;
     priv_ttl := 0;
     priv_protocol := 0;
     priv_checksum := 0;
     priv_source.initialize;
     priv_destination.initialize;
     priv_currentroute.initialize;
     setLength (priv_optionlist, 0);
     priv_keepcrc := false;
     priv_forcebadcrc := false;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.DetermineAssemblySize;

var i : integer;

begin
     prot_length := 20;

     for i := 1 to length (priv_optionlist) do
         inc (prot_length, priv_optionlist[i-1].OptionLength);
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.setIHL (const Value: byte);

var help : byte;

begin
     priv_ihl := Value;         // we keep this for the assembly process

     help := ((Value div 4) and $f);

     // Clear all existing IHL bits, then set the new ones

     priv_versionihl := (($f0 and priv_versionihl) or help);
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.setVersion (const Value: byte);

var help : byte;

begin
     help := ((Value shl 4) and $f0);

     // Clear all existing version bits, then set the new ones

     priv_versionihl := (($f and priv_versionihl) or help);
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.SetIPv4Flags (const Value: tIPFlagSet);

begin
     // First, unset all flags

     priv_flagsandoffset := priv_flagsandoffset and $1fff;

     // Enable flags if necessary

     if ipfreserved in Value then
        priv_flagsandoffset := priv_flagsandoffset or $8000;

     if ipfDontFragment in Value then
        priv_flagsandoffset := priv_flagsandoffset or $4000;

     if ipfMoreFragments in Value then
        priv_flagsandoffset := priv_flagsandoffset or $2000;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.SetIPv4Offset (const Value: word);

var help : word;

begin
     help := ((Value DIV 8) AND $1fff);       // measured in units of 8 octets

     priv_flagsandoffset := priv_flagsandoffset or help;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.setManualCRC (const Value: word);

begin
     priv_keepcrc := true;
     priv_checksum := Value;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.AddOptionRecord (const Value: tIPv4Option);

var count : integer;

begin
     count := length (priv_optionlist);
     setlength (priv_optionlist, count + 1);

     priv_optionlist[count] := Value;

     // Increase header length - if it is the first option, we add one byte for the final EOL

     if count = 0 then
        setIHL (priv_ihl + Value.OptionLength + 1)
     else
         setIHL (priv_ihl + Value.OptionLength);
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.writeRecordRouteOption (const opt: tIPv4Option; var Offset: integer);

var pHelpTLP    : pIPv4OptionTLP;
    pByteHelp   : pByte;
    pLongHelp   : pLongword;
    i           : integer;

begin
     // Write Type, Length, Pointer

     pHelpTLP := pIPv4OptionTLP (prot_data + Offset);
     pHelpTLP^.optiontype := opt.getOptionType;
     pHelpTLP^.optionlen := opt.OptionLength;
     pHelpTLP^.optionptr := opt.OptionPointer;

     // Move offset for route records

     inc (Offset, sizeOf (tIPv4OptionTLP));

     // Write route records

     for i := 1 to length (opt.RouteRecords) do
         begin
              pLongHelp := pLongword (prot_data + Offset);
              pLongHelp^ := swap32 (opt.RouteRecords[i-1].valueAsLong);

              inc (Offset, 4);
         end;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.writeSrcRouteOption (const opt: tIPv4Option; var Offset: integer);

var pHelpTLP    : pIPv4OptionTLP;
    pByteHelp   : pByte;
    pLongHelp   : pLongword;
    i           : integer;

begin
     // Write Type, Length, Pointer

     pHelpTLP := pIPv4OptionTLP (prot_data + Offset);
     pHelpTLP^.optiontype := opt.getOptionType;
     pHelpTLP^.optionlen := opt.OptionLength;
     pHelpTLP^.optionptr := opt.OptionPointer;

     // Move offset for route records

     inc (Offset, sizeOf (tIPv4OptionTLP));

     // Write route records

     for i := 1 to length (opt.RouteRecords) do
         begin
              pLongHelp := pLongword (prot_data + Offset);
              pLongHelp^ := swap32 (opt.RouteRecords[i-1].valueaslong);

              inc (Offset, 4);
         end;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.writeTimestampOption (const opt: tIPv4Option; var Offset: integer);

var pHelpTLP     : pIPv4OptionTLP;
    pByteHelp    : pByte;
    pTimeRecHelp : pIPv4TimestampRecord;
    i            : integer;

begin
     // Write Type, Length, Pointer

     pHelpTLP := pIPv4OptionTLP (prot_data + Offset);
     pHelpTLP^.optiontype := opt.getOptionType;
     pHelpTLP^.optionlen := opt.OptionLength;
     pHelpTLP^.optionptr := opt.OptionPointer;

     // Move offset of Overflow byte

     inc (Offset, sizeOf (tIPv4OptionTLP));

     pByteHelp := pByte (prot_data + Offset);
     pByteHelp^ := opt.OptionOverflow;

     // Move offset for route records

     inc (Offset);

     // Write route records

     for i := 1 to length (opt.RouteRecords) do
         begin
              pTimeRecHelp := pIPv4TimestampRecord (prot_data + Offset);
              pTimeRecHelp^.address := swap32 (opt.TimestampRecords[i-1].address.valueaslong);
              pTimeRecHelp^.timestamp := swap32 (opt.TimestampRecords[i-1].timestamp);

              inc (Offset, sizeOf (tIPv4TimestampRecord));
         end;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.writeRouterAlertOption (const opt : tIPv4Option; var Offset : integer);

var pHelpTLP     : pIPv4OptionTLP;

begin
     // Write Type, Length, Pointer

     pHelpTLP := pIPv4OptionTLP (prot_data + Offset);
     pHelpTLP^.optiontype := opt.getOptionType;
     pHelpTLP^.optionlen := opt.OptionLength;
     pHelpTLP^.optionptr := opt.OptionPointer;

     pWord (prot_Data + Offset + 2)^ := opt.OptionWord;
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.writeIPv4OptionList (OptionList: tIPv4OptionList; Offset: integer);

var opt    : tIPv4Option;
    pHelp  : pByte;
    count  : integer;
    i      : integer;

begin
     count := length (OptionList);

     if count > 0 then
        begin
             for i := 1 to count do
                 begin
                      opt := OptionList[i-1];

                      // TODO: write option record to buffer

                      case opt.OptionNumber of
                       IPV4_OPTION_EOL              : begin
                                                           // Write end of opt

                                                           pHelp := pByte (prot_data + offset);
                                                           pHelp^ := 0;

                                                           inc (offset);
                                                      end;
                       IPV4_OPTION_LOOSE_SRC_ROUTE,
                       IPV4_OPTION_STRICT_SRC_ROUTE : begin
                                                           writeSrcRouteOption (opt, offset);
                                                      end;
                       IPV4_OPTION_TIMESTAMP        : begin
                                                           writeTimestampOption (opt, offset);
                                                      end;
                       IPV4_OPTION_RECORD_ROUTE     : begin
                                                           writeRecordRouteOption (opt, offset);
                                                      end;
                       IPV4_OPTION_ROUTER_ALERT     : begin
                                                           writeRouterAlertOption (opt, offset);
                                                      end;
                      end;

                      inc (offset, opt.OptionLength);
                 end;
        end;

     assert (offset = prot_length, 'Memory offset check mismatch (tIPv4Assembler.Assemble)');
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.Assemble;

var pData  : pIPv4Struct;
    offset : integer;
    hw     : tIPv4HeaderWords;

begin
     pData := pIPv4Struct (prot_data);

     setIHL (prot_length);

     pData^.VersionIHL := priv_versionihl;
     pData^.TOS := priv_tos;

     // Determine IP total length from Header Length plus payload length

     priv_totallength := priv_IHL + priv_payloadlen;

     // If we have a truncated payload set the length that was issued on create

     if priv_forcedtotallen = 0 then
        pData^.TotalLength := swap16 (priv_totallength)
     else
         pData^.TotalLength := swap16 (priv_forcedtotallen);

     pData^.Identification := swap16 (priv_identification);
     pData^.FlagsOffset := swap16 (priv_flagsandoffset);
     pData^.TTL := priv_ttl;
     pData^.Protocol := priv_protocol;
     pData^.source := swap32 (priv_source.ValueAsLong);

     if priv_currentroute.ValueAsLong = 0 then
        pData^.destination := swap32 (priv_destination.ValueAsLong)
     else
         pData^.destination := swap32 (priv_currentroute.ValueAsLong);

     // Handle Options

     if priv_ihl > 20 then
        writeIPv4OptionList (priv_optionlist, 20);

     // Calculate the checksum

     if not priv_keepcrc then
        begin
             hw := tIPv4Parser.getHeaderWordsFromOffset(prot_data, 0, priv_ihl);
             priv_checksum := getIPv4Checksum (hw, priv_ihl);

             if priv_forcebadcrc then
                priv_checksum := getBadCRC (priv_checksum);
        end;

     pData^.Checksum := swap16 (priv_checksum);
end;

(* ------------------------------------------- *)

procedure tIPv4Assembler.CopyFromParser (Parser: tIPv4Parser);

var option : tIPv4Option;
    i      : integer;

begin
     priv_ihl := Parser.IHL;
     setIHL (Parser.IHL);
     priv_tos := Parser.TOS;
//   priv_forcedtotallen := Parser.TotalLength;               // We set it as forced value because it could be a truncated packet
     priv_identification := Parser.Identification;
     SetIPv4Flags (Parser.Flags);
     SetIPv4Offset (Parser.Offset);
     priv_ttl := Parser.TTL;
     priv_protocol := Parser.Protocol;
     priv_checksum := Parser.Checksum;
     priv_source := Parser.Source;
     priv_destination := Parser.Destination;
     priv_currentroute := Parser.CurrentRoute;

     // Copy option records

     for i := 1 to length (Parser.OptionList) do
         begin
              option.copyFrom (Parser.OptionList[i-1]);

              AddOptionRecord (option);
         end;
end;

(* ------------------------------------------- *)

end.


