{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_rtps;

interface

uses base_structures,
     class_parser_rtps,
     class_assembly_baseclass;


type tRTPSAssembler = class (tFrameAssembler)

                            constructor create;

                            private
                                   priv_MajorVersion  : byte;
                                   priv_MinorVersion  : byte;
                                   priv_VendorID      : word;
                                   priv_GUIDHostID    : longword;
                                   priv_GUIDAppID     : longword;
                                   priv_GUIDCounter   : longword;
                                   priv_submessages   : tRTPSSubMessageList;
                                   priv_isPing        : boolean;

                                   function RTPSFlagsToByte (Flagset : tRTPSSubMessageFlagSet) : byte;

                            public
                                  property MajorVersion : byte read priv_MajorVersion write priv_MajorVersion;
                                  property MinorVersion : byte read priv_MinorVersion write priv_MinorVersion;
                                  property VendorID : word read priv_VendorID write priv_VendorID;
                                  property GUIDHostID : longword read priv_GUIDHostID write priv_GUIDHostID;
                                  property GUIDAppID : longword read priv_GUIDAppID write priv_GUIDAppID;
                                  property GUIDCounter : longword read priv_GUIDCounter write priv_GUIDCounter;
                                  property SubMessages : tRTPSSubMessageList read priv_submessages write priv_submessages;
                                  property IsPing : boolean read priv_isPing write priv_isPing;

                            protected
                                     // called from base class from GetAssembly

                                     procedure DetermineAssemblySize; override;
                                     procedure Assemble; override;

                            public
                                  procedure AddSubMessage (const SubMessage : tRTPSSubMessage);

                                  procedure CopyFromParser (Parser : tRTPSParser);
                       end;


implementation

uses common_parse_helpers,
     common_assembly_helpers,
     toolkit;

(* ------------------------------------------- *)
{ tRTPSAssembler }
(* ------------------------------------------- *)

constructor tRTPSAssembler.create;

begin
     inherited create;

     priv_MajorVersion := 0;
     priv_MinorVersion := 0;
     priv_VendorID := 0;
     priv_GUIDHostID := 0;
     priv_GUIDAppID := 0;
     priv_GUIDCounter := 0;

     priv_isPing := false;

     setLength (priv_submessages, 0);
end;

(* ------------------------------------------- *)

procedure tRTPSAssembler.DetermineAssemblySize;

var i   : integer;

begin
     if priv_isPing then
        prot_length := sizeOf (tRTPSStruct) + length (RTPS_PING)
     else
         prot_length := sizeOf (tRTPSStruct) + sizeOf (tRTPSGUIDStruct);

//   prot_length := sizeOf (tRTPSStruct);

     for i := 1 to length (priv_submessages) do
         prot_length := prot_length + priv_submessages[i-1].Available + 4;              // 4 bytes for the tRTPSSubMessage id, flags, len
end;

(* ------------------------------------------- *)

function tRTPSAssembler.RTPSFlagsToByte (Flagset: tRTPSSubMessageFlagSet): byte;

var flag : tRTPSSubMessageFlag;

begin
     result := 0;

     for flag in Flagset do
         case flag of
          rtpsfEndianess : result := (result or 1);
          rtpsfTimestamp : result := (result or 2);
         end;
end;

(* ------------------------------------------- *)

procedure tRTPSAssembler.AddSubMessage (const SubMessage: tRTPSSubMessage);

var len  : integer;
    i    : integer;

begin
     // Add to the list

     len := length (priv_submessages);
     setLength (priv_submessages, len + 1);
     priv_submessages[len] := SubMessage;
end;

(* ------------------------------------------- *)

procedure tRTPSAssembler.Assemble;

var pRTPSHeader    : pRTPSStruct;
    pGUID          : pRTPSGUIDStruct;
    pSubMessageHdr : pRTPSSubMessageStruct;
    pDataStruct    : pRTPSDataMessageStruct;
    pEncapHeader   : pRTPSDataMessageEncapStruct;
    pTLVHeader     : pTLVWordWordHeader;
    wHelp          : word;
    i64help        : uint64;
    pHelpByte      : pByte;
    submsg         : tRTPSSubMessage;
    dataentry      : tRTPSSerializedDataEntry;
    offset         : integer;
    count          : integer;
    len            : integer;
    i,j,k          : integer;

begin
     pRTPSHeader := pRTPSStruct (prot_data);

     // write "RTPS" magic

     pRTPSHeader^.Protocol[0] := $52;
     pRTPSHeader^.Protocol[1] := $54;
     pRTPSHeader^.Protocol[2] := $50;
     pRTPSHeader^.Protocol[3] := $53;

     // Calcualte combined version

     wHelp := (priv_MajorVersion shl 8) + priv_MinorVersion;

     pRTPSHeader^.Version := swap16 (wHelp);
     pRTPSHeader^.VendorID := swap16 (priv_VendorID);

     offset := sizeOf (tRTPSStruct);

     if priv_IsPing then
        begin
             writeStringToOffset (prot_data, RTPS_PING, offset);

             inc (offset, length (RTPS_PING));
        end
     else
         begin
              pGUID := pRTPSGUIDStruct (prot_data + offset);

              pGUID^.GUIDPrefixHostID := swap32 (priv_GUIDHostID);
              pGUID^.GUIDPrefixAppID := swap32 (priv_GUIDAppID);
              pGUID^.GUIDPrefixCounter := swap32 (priv_GUIDCounter);

              offset := offset + sizeOf (tRTPSGUIDStruct);

              // Write Submessages

              for i := 1 to Length (priv_submessages) do
                  begin
                       submsg := priv_submessages[i-1];

                       pSubMessageHdr := pRTPSSubMessageStruct (prot_data + offset);

                       pSubMessageHdr^.MessageID := submsg.MessageID;
                       pSubMessageHdr^.Flags := RTPSFlagsToByte (submsg.Flags);
                       pSubMessageHdr^.Len := submsg.Len;

                       inc (offset, 4);

                       // Check if we have to write a DATA submessage with the Writer types we want to keep

                       if (submsg.MessageID = RTPS_ID_DATA) then
                          begin
                               // first, write the DATA header

                               pDataStruct := pRTPSDataMessageStruct (prot_data + offset);
                               pDataStruct^.ExtraFlags := submsg.DataMessage.ExtraFlags;
                               pDataStruct^.LenQoS := submsg.DataMessage.LenQoS;
                               pDataStruct^.ReaderEntity := swap32 (submsg.DataMessage.ReaderEntity);
                               pDataStruct^.WriterEntity := swap32 (submsg.DataMessage.WriterEntity);

                               i64help := submsg.DataMessage.WriterSeq shr 32;
                               pDataStruct^.WriterSeqHi := i64help;

                               i64help := (submsg.DataMessage.WriterSeq and $FFFFFFFF);
                               pDataStruct^.WriterSeqLo := i64help;

                               Offset := Offset + sizeOf (tRTPSDataMessageStruct);

                               // Write QoS stuff if needed

                               if (pSubMessageHdr^.Flags and $2) = $2 then
                                  begin
                                       count := length (submsg.DataMessage.InlineQoSList);

                                       // Need to write the data sub elements one by one

                                       for j := 1 to count do
                                           begin
                                                dataentry := submsg.DataMessage.InlineQoSList[j-1];

                                                pTLVHeader := pTLVWordWordHeader (prot_data + offset);

                                                pTLVHeader^.OptionType := dataentry.DataType;
                                                pTLVHeader^.OPtionLength := dataentry.DataLength;

                                                inc (offset, sizeOf (tTLVWordWordHeader));

                                                if dataentry.DataType <> RTPS_PID_SENTINEL then
                                                   begin
                                                        // write the data blob

                                                        len := length (dataentry.Data);

                                                        for k := 1 to len do
                                                            begin
                                                                 pHelpByte := pByte (prot_data + offset);
                                                                 pHelpByte^ := dataentry.Data[k-1];

                                                                 inc (offset);
                                                            end;
                                                   end;
                                           end;
                                  end;

                               // Write encap stuff

                               pEncapHeader := pRTPSDataMessageEncapStruct (prot_data + Offset);

                               if rtpsfEndianess in submsg.Flags then
                                  pEncapHeader^.EncapsulationKind := swap16 (submsg.DataMessage.EncapsulationKind)
                               else
                                   pEncapHeader^.EncapsulationKind := submsg.DataMessage.EncapsulationKind;

                               pEncapHeader^.EncapsulationOpts := submsg.DataMessage.EncapsulationOpts;

                               inc (offset, sizeof (tRTPSDataMessageEncapStruct));

                               // Write the data(r) and data(w) portions

                               if ((submsg.DataMessage.WriterEntity = $3c2) or (submsg.DataMessage.WriterEntity = $4c2) or (submsg.DataMessage.WriterEntity = $100c2)) and (length (submsg.DataMessage.DataList) > 0) then
                                  begin
                                       count := length (submsg.DataMessage.DataList);

                                       // Need to write the data sub elements one by one

                                       for j := 1 to count do
                                           begin
                                                dataentry := submsg.DataMessage.DataList[j-1];

                                                pTLVHeader := pTLVWordWordHeader (prot_data + offset);

                                                pTLVHeader^.OptionType := dataentry.DataType;
                                                pTLVHeader^.OPtionLength := dataentry.DataLength;

                                                inc (offset, sizeOf (tTLVWordWordHeader));

                                                // write the data blob

                                                len := length (dataentry.Data);

                                                for k := 1 to len do
                                                    begin
                                                         pHelpByte := pByte (prot_data + offset);
                                                         pHelpByte^ := dataentry.Data[k-1];

                                                         inc (offset);
                                                    end;
                                           end;
                                  end
                               else
                                   begin
                                        // write data(p)

                                        len := length (submsg.DataMessage.Data);

                                        for j := 1 to len do
                                            begin
                                                 pHelpByte := pByte (prot_data + offset);
                                                 pHelpByte^ := submsg.DataMessage.Data[j-1];

                                                 inc (offset);
                                            end;
                                   end;
                          end
                       else
                           begin
                                // Just write the data blob

                                len := length (submsg.data);

                                for j := 1 to len do
                                    begin
                                         pHelpByte := pByte (prot_data + offset);
                                         pHelpByte^ := submsg.data[j-1];

                                         inc (offset);
                                    end;
                           end;
                  end;
         end;

     assert (offset = prot_length, 'Memory offset check mismatch (tRTPSAssembler.Assemble)');
end;

(* ------------------------------------------- *)

procedure tRTPSAssembler.CopyFromParser (Parser: tRTPSParser);

var len : integer;
    msg : tRTPSSubMessage;
    i   : integer;

begin
     priv_MajorVersion := Parser.MajorVersion;
     priv_MinorVersion := Parser.MinorVersion;
     priv_VendorID := Parser.VendorID;
     priv_GUIDHostID := Parser.GUIDHostID;
     priv_GUIDAppID := Parser.GUIDAppID;
     priv_GUIDCounter := Parser.GUIDCounter;

     priv_isPing := Parser.IsPing;

     if not priv_isPing then
        begin
             len := length (Parser.SubMessages);

             for i := 1 to len do
                 begin
                      msg := Parser.SubMessages[i-1];

                      AddSubMessage (msg);
                 end;
        end;
end;

(* ------------------------------------------- *)

end.
