{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_tcp;

interface

uses base_structures,
     base_checksums,
     definition_tcp_structures,
     class_parser_tcp,
     class_assembly_baseclass;

type tTCPOptionFormat = (ofOriginal, ofWindows);
     tTCPAssembler  = class (tFrameAssembler)

                            {$REGION 'Documentation'}
                            ///	<summary>
                            ///	  The TCP header needs to know what payload
                            ///	  will be following next to be able to
                            ///	  calculate the CRC correctly. It also needs
                            ///	  the IP pseudo header information for the same
                            ///	  reason
                            ///	</summary>
                            ///	<param name="IPv4PseudoHeader">
                            ///	  IPv4 Pseudo header
                            ///	</param>
                            ///	<param name="TCPPayload">
                            ///	  Assembly containing the payload that is
                            ///	  following the TCP header
                            ///	</param>
                            {$ENDREGION}
                            constructor create (IPv4PseudoHeader : tIPv4PseudoHeader; TCPPayload : tFrameChunk); overload;
                            constructor create (IPv4PseudoHeader : tIPv4PseudoHeader); overload;

                            {$REGION 'Documentation'}
                            ///	<summary>
                            ///	  The TCP header needs to know what payload
                            ///	  will be following next to be able to
                            ///	  calculate the CRC correctly. It also needs
                            ///	  the IP pseudo header information for the same
                            ///	  reason
                            ///	</summary>
                            ///	<param name="IPv6PseudoHeader">
                            ///	  IPv6 Pseudo header
                            ///	</param>
                            ///	<param name="TCPPayload">
                            ///	  Assembly containing the payload that is
                            ///	  following the TCP header
                            ///	</param>
                            {$ENDREGION}
                            constructor create (IPv6PseudoHeader : tIPv6PseudoHeader; TCPPayload : tFrameChunk); overload;
                            constructor create (IPv6PseudoHeader : tIPv6PseudoHeader); overload;

                            Destructor Destroy; override;

                            private
                                   priv_Sourceport         : word;
                                   priv_Destinationport    : word;
                                   priv_Sequence           : tTCPSequence;
                                   priv_Acknowledge        : tTCPSequence;
                                   priv_flags              : tTCPFlagSet;
                                   priv_Window             : word;
                                   priv_checksum           : word;
                                   priv_UrgentPointer      : word;

                                   priv_optionset          : tTCPOptionSet;
                                   priv_exactoptionlist    : tTCPOptionList;         // exact order of options, including duplicates
                                   priv_sackranges         : tTCPSackRanges;
                                   priv_mss                : word;
                                   priv_ScaleFactor        : tTCPScaleFactorOption;
                                   priv_Timestamp          : tTCPTimeStampOption;

                                   priv_payload            : tFrameChunk;
                                   priv_IPLayerVersion     : byte;
                                   priv_IPv4PseudoHeader   : tIPv4PseudoHeader;
                                   priv_IPv6PseudoHeader   : tIPv6PseudoHeader;
                                   priv_keepcrc            : boolean;
                                   priv_forcebadcrc        : boolean;

                                   priv_optionformat       : tTCPOptionFormat;

                                   priv_mpathoptionlist    : tTCPMPathOptionList;
                                   priv_fastopenoptionlist : tTCPFastOpenOptionList;
                                   priv_paddinglen         : integer;
                                   priv_optionpad          : tTCPOptionPad;

                                   priv_originalhdrlen     : word;                   // used in case TCP was truncated, cutting options

                                   procedure initialize;

                                   {$REGION 'Documentation'}
                                   ///	<summary>
                                   ///	  Creates the actual 16 bit header value
                                   ///	  containing Offset and flags from the
                                   ///	  distinct values that are set via the
                                   ///	  properties
                                   ///	</summary>
                                   ///	<returns>
                                   ///	  Correctly swapped 16bit value that
                                   ///	  goes into the TCP header
                                   ///	</returns>
                                   {$ENDREGION}
                                   function setDataOffsetFlags : word;

                                   function CalculateHeaderLengthInOriginalOrder : integer;
                                   function CalculateHeaderLengthWindowsOptionsStyle : integer;

                                   procedure setMaximumSegmentSize(const Value: word);
                                   procedure setSACKSupported(const Value: boolean);
                                   function getSACKSupported: boolean;
                                   procedure setScaleFactor(const Value: tTCPScaleFactorOption);
                                   procedure setTimestamp(const Value: tTCPTimeStampOption);

                                   {$REGION 'Documentation'}
                                   ///	<summary>
                                   ///	  writes a NOP Option at the given offset
                                   ///	</summary>
                                   ///	<param name="Offset">
                                   ///	  Offset to write the NOP at; will be
                                   ///	  increased and returned
                                   ///	</param>
                                   {$ENDREGION}
                                   procedure writeNOP (var Offset : integer);
                                   procedure writeMSS (var Offset : integer);
                                   procedure writeScaling (var Offset : integer);
                                   procedure writeSACKPermitted (var offset : integer);
                                   procedure writeSACKBlocks (var offset : integer);
                                   procedure writeTimestamp (var offset : integer);
                                   procedure writeEOL (var offset : integer);
                                   procedure WritePadding (var offset : integer);

                                   /// <summary>
                                   ///   Writes the TCP options in the order as
                                   ///   they were set (usually via copy from
                                   ///   parser)
                                   /// </summary>
                                   /// <param name="Offset">
                                   ///   Position at which the options start
                                   /// </param>
                                   /// <returns>
                                   ///   Length of the option block just
                                   ///   written. Add this to the TCP header
                                   ///   length (which is 20 Bytes) to get the
                                   ///   total TCP header length.
                                   /// </returns>
                                   function writeTCPOptionListInOriginalOrder (Offset : integer) : integer;
                                   {$REGION 'Documentation'}
                                   ///	<summary>
                                   ///	  Mimicswriting TCP options as the
                                   ///	  windowsTCP kernel does
                                   ///	</summary>
                                   ///	<param name="Offset">
                                   ///	  Position at which theoptions start
                                   ///	</param>
                                   ///	<returns>
                                   ///	  Length of the option block just
                                   ///	  written. Add this to theTCP header
                                   ///	  length (which is 20 Bytes) to get the
                                   ///	  total TCP header length.
                                   ///	</returns>
                                   {$ENDREGION}
                                   function writeTCPOptionListWindowsStyle (Offset : integer) : integer;

                                   procedure writeTCPOptionList (Offset : integer);

                                   procedure setManualCRC(const Value: word);

                                   procedure writeMPathOption (var Offset: integer; Option : tTCPMultiPathOption);
                                   procedure writeFastOpenOption (var Offset: integer; Option : tTCPFastOpenOption);

                            protected
                                     // called from base class from GetFrameChunk

                                     procedure DetermineAssemblySize; override;
                                     procedure Assemble; override;

                            public
                                  property SourcePort : word read priv_Sourceport write priv_Sourceport;
                                  property DestinationPort : word read priv_DestinationPort write priv_Destinationport;
                                  property SequenceNumber : tTCPSequence read priv_Sequence write priv_Sequence;
                                  property AcknowledgeNumber : tTCPSequence read priv_Acknowledge write priv_Acknowledge;
                                  property Flags : tTCPFlagSet read priv_flags write priv_flags;
                                  property WindowSize : word read priv_Window write priv_Window;
                                  property Checksum : word read priv_checksum write setManualCRC;
                                  property DoNotRecalculateCRC : boolean read priv_keepcrc write priv_keepcrc;
                                  property ForceBadCRC : boolean read priv_forcebadcrc write priv_forcebadcrc;
                                  property UrgentPointer : word read priv_UrgentPointer write priv_UrgentPointer;

                                  {$REGION 'Documentation'}
                                  ///	<summary>
                                  ///	  This property is read only because
                                  ///	  Header Length is determined by the
                                  ///	  header contents, not set arbitrarily
                                  ///	</summary>
                                  {$ENDREGION}
                                  property HeaderLength : integer read prot_Length;

                                  property TCPOptionsSet : tTCPOptionSet read priv_optionset;
                                  property MaximumSegmentSize : word read priv_mss write setMaximumSegmentSize;
                                  property ScaleFactor : tTCPScaleFactorOption read priv_ScaleFactor write setScaleFactor;
                                  property Timestamp : tTCPTimeStampOption read priv_Timestamp write setTimestamp;
                                  property SackPermitted : boolean read getSACKSupported write setSACKSupported;
                                  property TCPOptionFormat : tTCPOptionFormat read priv_optionformat write priv_optionformat;

                                  {$REGION 'Documentation'}
                                  ///	<summary>
                                  ///	  Allows adding SACK ranges, without
                                  ///	  limiting the amount of ranges. The TCP
                                  ///	  header will still not have more ranges
                                  ///	  than fit, usually a maximum of 5
                                  ///	  ranges. All others will be ignored on
                                  ///	  assembly.
                                  ///	</summary>
                                  {$ENDREGION}
                                  procedure AddSACKRange (const NewRange : tTCPSACKEdges);

                                  property MultiPathOptionList : tTCPMPathOptionList read priv_mpathoptionlist;

                                  {$REGION 'Documentation'}
                                  ///	<summary>
                                  ///	  copies all values from a parser object
                                  ///	  to the assembler, including the
                                  ///	  existing checksum
                                  ///	</summary>
                                  {$ENDREGION}
                                  procedure CopyFromParser (Parser : tTCPParser);
                      end;


implementation

uses Math,
     SiAuto,
     definition_ip_addresses,
     Toolkit;

type pTCPOptionTL = ^tTCPOptionTL;
     tTCPOptionTL = record
                           optiontype : byte;
                           optionlen  : byte;
                      end;

(* ------------------------------------------- *)
{ tTCPAssembler }
(* ------------------------------------------- *)

constructor tTCPAssembler.create (IPv4PseudoHeader: tIPv4PseudoHeader; TCPPayload: tFrameChunk);

begin
     inherited create;

     priv_IPLayerVersion := 4;
     priv_IPv4PseudoHeader := IPv4PseudoHeader;
     priv_keepcrc := false;
     priv_forcebadcrc := false;

     priv_payload := TCPPayload;

     initialize;
end;

(* ------------------------------------------- *)

constructor tTCPAssembler.create (IPv4PseudoHeader: tIPv4PseudoHeader);

begin
     inherited create;

     priv_IPLayerVersion := 4;
     priv_IPv4PseudoHeader := IPv4PseudoHeader;
     priv_forcebadcrc := false;
     priv_keepcrc := false;

     priv_payload.Len := 0;
     priv_payload.Data := nil;

     initialize;
end;

(* ------------------------------------------- *)

constructor tTCPAssembler.create (IPv6PseudoHeader: tIPv6PseudoHeader; TCPPayload: tFrameChunk);

begin
     inherited create;

     priv_IPLayerVersion := 6;
     priv_IPv6PseudoHeader := IPv6PseudoHeader;
     priv_forcebadcrc := false;
     priv_keepcrc := false;

     priv_payload := TCPPayload;

     initialize;
end;

(* ------------------------------------------- *)

constructor tTCPAssembler.create (IPv6PseudoHeader: tIPv6PseudoHeader);

begin
     inherited create;

     priv_IPLayerVersion := 6;
     priv_IPv6PseudoHeader := IPv6PseudoHeader;
     priv_forcebadcrc := false;
     priv_keepcrc := false;

     priv_payload.Len := 0;
     priv_payload.Data := nil;

     initialize;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.DetermineAssemblySize;

begin
     prot_length := 0;

     case priv_optionformat of
      ofOriginal : prot_length := CalculateHeaderLengthInOriginalOrder;
      ofWindows  : prot_length := CalculateHeaderLengthWindowsOptionsStyle;
     end;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.initialize;

begin
     priv_optionformat := ofWindows;               // We only have that one at the moment, but you never know :-)

     priv_optionset := [];
     setLength (priv_exactoptionlist, 0);

     setLength (priv_sackranges, 0);
     priv_mss := 0;
     priv_ScaleFactor.ShiftCount := 0;
     priv_ScaleFactor.Multiplier := 0;
     priv_Timestamp.tsval := 0;
     priv_Timestamp.tsecho := 0;

     priv_originalhdrlen := 0;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.setManualCRC (const Value: word);

begin
     priv_keepcrc := true;
     priv_forcebadcrc := false;
     priv_checksum := Value;
end;

(* ------------------------------------------- *)

destructor tTCPAssembler.Destroy;

begin
     setLength (priv_sackranges, 0);

     inherited;
end;

(* ------------------------------------------- *)

function tTCPAssembler.setDataOffsetFlags : word;

var value : word;
    help  : word;
    len   : integer;

begin
     len := prot_length;

     // Check if options were truncated and force the full option length to be included in that case
     // priv_originalhdrlen is always 0 except when copying from parser

     if priv_originalhdrlen > 0 then
        begin
             if (priv_originalhdrlen > 20) and (prot_length = 20) then
                len := priv_originalhdrlen
        end;

     assert (len > 0, 'TCP header length is zero (tTCPAssembler.setDataOffsetFlags)');

     value := getWordFromTCPFlagSet (priv_flags);

     // Add the offset bits for header length

     help := (len div 4) shl 12;
     value := value or help;

     result := swap16 (value);
end;

(* ------------------------------------------- *)

function tTCPAssembler.CalculateHeaderLengthInOriginalOrder: integer;

var Len        : integer;
    remaining  : integer;
    blockcount : integer;
    i          : integer;
    help       : integer;

begin
     Len := 20;

     for i := 1 to length (priv_exactoptionlist) do
         begin
              case priv_exactoptionlist[i-1] of
               tcpEOL                : inc (Len, 1);
               tcpNOP                : inc (Len, 1);
               tcpMSS                : inc (Len, 4);
               tcpScaleFactor        : inc (Len, 3);
               tcpSACKPermitted      : inc (Len, 2);
               tcpTimestamp          : inc (Len, 10);
               tcpCC                 : inc (Len, 6);
               tcpCCNEW              : inc (Len, 6);
               tcpCCECHO             : inc (Len, 6);
              end;
         end;

     // Multipath options

     if length (priv_mpathoptionlist) > 0 then
        begin
             for i := 1 to length (priv_mpathoptionlist) do
                 inc (Len, priv_mpathoptionlist[i-1].optionlength);
        end;

     // FastOpen options

     if length (priv_fastopenoptionlist) > 0 then
        begin
             for i := 1 to length (priv_fastopenoptionlist) do
                 inc (Len, priv_fastopenoptionlist[i-1].CookieLength + 2);
        end;

     // SACK Blocks

     if length (priv_sackranges) > 0 then
        begin
             // calculate how many block we could write
             // the +2 is for the SACK block header bytes

             remaining := 60 - (Len + 2);
             blockcount := min ((remaining div 8), length (priv_sackranges));

             inc (Len, (2 + (8 * blockcount)));

             // truncate SACK ranges if necessary

             if blockcount < length (priv_sackranges) then
                begin
                     setLength (priv_sackranges, blockcount);

                     // TODO: add warning using MSG_TCP_OPTION_SACK_BLOCKS_TRUNCATED
                end;
        end;

     // Check 32 bit alignment

     help := len mod 4;

     if help > 0 then
        begin
             priv_paddinglen := (4 - help);

             // Create padding bytes unless we already have them

             if not (length (priv_optionpad) = priv_paddinglen) then
                begin
                     setLength (priv_optionpad, priv_paddinglen);

                     for i := 0 to priv_paddinglen - 1 do
                         priv_optionpad[i] := 0;
                end;

             // Adjust len if required

             if not (help = 0) then
                len := ((len div 4) + 1) * 4;
        end
     else
         begin
              priv_paddinglen := 0;
              setLength (priv_optionpad, priv_paddinglen);
         end;

     result := Len;
end;

(* ------------------------------------------- *)

function tTCPAssembler.CalculateHeaderLengthWindowsOptionsStyle : integer;

var Len        : integer;
    remaining  : integer;
    blockcount : integer;
    i          : integer;
    help       : integer;

begin
     Len := 20;

     // MSS

     if tcpMSS in priv_Optionset then
        inc (Len, 4);

     if tcpScaleFactor in priv_OptionSet then
        inc (Len, 4);

     // SACK permitted

     if tcpSACKPermitted in priv_OptionSet then
        begin
             if not (tcpTimestamp in priv_optionset) then
                inc (Len, 4)
             else
                 inc (Len, 2);
        end;

     // Timestamp

     if tcpTimestamp in priv_OptionSet then
        begin
             // we need to pad if no SACK permitted option was written before

             if not (tcpSACKPermitted in priv_optionset) then
                inc (Len, 12)
             else
                 inc (Len, 10);
        end;

     // SACK Blocks

     if length (priv_sackranges) > 0 then
        begin
             // calculate how many block we could write
             // the +4 is for the two NOPs and the SACK block header bytes

             remaining := 60 - (Len + 4);
             blockcount := min ((remaining div 8), length (priv_sackranges));

             inc (Len, (4 + (8 * blockcount)));
        end;

     // Multipath options

     if length (priv_mpathoptionlist) > 0 then
        begin
             for i := 1 to length (priv_mpathoptionlist) do
                 inc (Len, priv_mpathoptionlist[i-1].optionlength);
        end;

     // FastOpen options

     if length (priv_fastopenoptionlist) > 0 then
        begin
             for i := 1 to length (priv_fastopenoptionlist) do
                 inc (Len, priv_fastopenoptionlist[i-1].CookieLength + 2);
        end;

     // Check 32 bit alignment

     help := len mod 4;

     if help > 0 then
        begin
             priv_paddinglen := (4 - help);

             // Create padding bytes unless we already have them

             if not (length (priv_optionpad) = priv_paddinglen) then
                begin
                     setLength (priv_optionpad, priv_paddinglen);

                     for i := 0 to priv_paddinglen - 1 do
                         priv_optionpad[i] := 0;
                end;

             // Adjust len if required

             if not (help = 0) then
                len := ((len div 4) + 1) * 4;
        end
     else
         begin
              priv_paddinglen := 0;
              setLength (priv_optionpad, priv_paddinglen);
         end;

     result := Len;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.setMaximumSegmentSize (const Value: word);

var count : integer;

begin
     priv_optionset := priv_optionset + [tcpMSS];

     count := length (priv_exactoptionlist);
     setLength (priv_exactoptionlist, count + 1);
     priv_exactoptionlist[count] := tcpMSS;

     priv_mss := Value;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.setScaleFactor (const Value: tTCPScaleFactorOption);

var count : integer;

begin
     priv_optionset := priv_optionset + [tcpScalefactor];

     count := length (priv_exactoptionlist);
     setLength (priv_exactoptionlist, count + 1);
     priv_exactoptionlist[count] := tcpScalefactor;

     priv_ScaleFactor := Value;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.setTimestamp (const Value: tTCPTimeStampOption);

var count : integer;

begin
     priv_optionset := priv_optionset + [tcpTimestamp];

     count := length (priv_exactoptionlist);
     setLength (priv_exactoptionlist, count + 1);
     priv_exactoptionlist[count] := tcpTimestamp;

     priv_Timestamp := Value;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.setSACKSupported (const Value: boolean);

var count : integer;

begin
     priv_optionset := priv_optionset + [tcpSACKPermitted];

     count := length (priv_exactoptionlist);
     setLength (priv_exactoptionlist, count + 1);
     priv_exactoptionlist[count] := tcpSACKPermitted;
end;

(* ------------------------------------------- *)

function tTCPAssembler.getSACKSupported: boolean;

begin
     result := tcpSACKPermitted in priv_optionset;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.AddSACKRange (const NewRange: tTCPSACKEdges);

var count : integer;

begin
     priv_optionset := priv_optionset + [tcpSACK];

     count := length (priv_exactoptionlist);
     setLength (priv_exactoptionlist, count + 1);
     priv_exactoptionlist[count] := tcpSACK;

     count := length (priv_sackranges);
     setLength (priv_sackranges, count);          // since it's starting a 0 count makes it one item larger
     priv_sackranges[count] := NewRange;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeEOL (var offset: integer);

begin
     pByte (prot_data + Offset)^ := 0;

     inc (Offset, 1);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeNOP (var Offset: integer);

var pOptionHdr : pTCPOptionTL;

begin
     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 1;
     pOptionHdr^.optionlen := 1;

     inc (Offset, 1);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeMSS (var Offset: integer);

var pOptionHdr : pTCPOptionTL;
    pHelpWord  : pWord;

begin
     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 2;
     pOptionHdr^.optionlen := 4;

     pHelpWord := pWord (prot_data + Offset + 2);
     pHelpWord^ := swap16 (priv_mss);

     inc (Offset, 4);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeSACKBlocks (var offset: integer);

var pOptionHdr : pTCPOptionTL;
    pSackBlock : pTCPSackOption;
    remaining  : integer;
    blockcount : integer;
    len        : integer;
    i          : integer;

begin
     // First: calculate how many SACKs we will write -
     // its the min of blocks we have and how many fit in the
     // remaining option bytes

     remaining := 60 - (offset + 2);
     blockcount := min ((remaining div 8), length (priv_sackranges));

     // Now write the header

     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 5;
     pOptionHdr^.optionlen := 2 + (blockcount * 8);

     inc (offset, 2);

     // And the blocks

     for i := 1 to blockcount do
         begin
              pSackBlock := pTCPSackOption (prot_data + offset);
              pSackBlock^.leftEdge := swap32 (priv_sackranges[i-1].leftEdge.Value);
              pSackBlock^.rightEdge := swap32 (priv_sackranges[i-1].rightEdge.Value);

              inc (offset, 8)
         end;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeSACKPermitted (var offset: integer);

var pOptionHdr : pTCPOptionTL;
    pHelpByte  : pByte;

begin
     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 4;
     pOptionHdr^.optionlen := 2;

     inc (Offset, 2);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeScaling (var Offset: integer);

var pOptionHdr : pTCPOptionTL;
    pHelpByte  : pByte;

begin
     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 3;
     pOptionHdr^.optionlen := 3;

     pHelpByte := pByte (prot_data + Offset + 2);
     pHelpByte^ := priv_ScaleFactor.ShiftCount;

     inc (Offset, 3);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeTimestamp (var offset: integer);

var pOptionHdr : pTCPOptionTL;
    pTimestamp : pTCPTimeStampOption;

begin
     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 8;
     pOptionHdr^.optionlen := 10;

     pTimestamp := pTCPTimeStampOption (prot_data + offset + 2);
     pTimestamp^.tsval := swap32 (priv_Timestamp.tsval);
     pTimestamp^.tsecho := Swap32 (priv_Timestamp.tsecho);

     inc (offset, 10);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeMPathOption (var Offset: integer; Option : tTCPMultiPathOption);

var pOptionHdr   : pTCPOptionTL;
    pHelpWord    : pWord;
    pHelpLong    : pLongword;
    HelpLong     : Longword;
    pHelpUint64  : pUINT64;
    pHelpIPv6Adr : pIPv6AddressByteStruct;
    pHelpByte    : pByte;
    HelpByte     : Byte;
    OptionDelta  : integer;

begin
     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 30;
     pOptionHdr^.optionlen := Option.optionlength;

     case Option.subType of
      mtcpCapable            : begin
                                    HelpByte := 0;            // Subtype is 0
                                    HelpByte := HelpByte or (Option.TCPVersion);

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^:= HelpByte;

                                    // Flags

                                    if tmpChecksumReq in Option.MPathFlags then
                                       HelpByte := $80
                                    else
                                        HelpByte := 0;

                                    if tmpExtensibility in Option.MPathFlags then
                                       HelpByte := HelpByte or $40;

                                    if tmpCryptoUseHMAC in Option.MPathFlags then
                                       HelpByte := HelpByte or $1;

                                    pHelpByte := pByte (prot_data + Offset + 3);
                                    pHelpByte^:= HelpByte;

                                    // Senders Key

                                    pHelpUint64 := pUINT64 (prot_data + Offset + 4);
                                    pHelpUInt64^ := Swap64 (option.SendersKey);

                                    // Receivers Key

                                    if option.ReceiversKey <> 0 then
                                       begin
                                            pHelpUint64 := pUINT64 (prot_data + Offset + 12);
                                            pHelpUInt64^ := Swap64 (option.ReceiversKey);
                                       end;
                               end;
      mtcpJoinConnection     : begin
                                    HelpByte := 1 shl 4;         // Subtype is 1

                                    // Flags

                                    if tmpBackup in option.MPathFlags then
                                       HelpByte := HelpByte or $1;

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^:= HelpByte;

                                    // Address ID

                                    pHelpByte := pByte (prot_data + Offset + 3);
                                    pHelpByte^ := Option.AddressID;

                                    // Tokens

                                    pHelpLong := pLongword (prot_data + Offset + 4);
                                    pHelpLong^ := swap32 (Option.ReceiversToken);

                                    pHelpLong := pLongword (prot_data + Offset + 8);
                                    pHelpLong^ := swap32 (Option.SendersRndNr);
                               end;
      mtcpDataSequenceSignal : begin
                                    HelpByte := 2 shl 4;

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^ := HelpByte;

                                    // Flags

                                    if tmpDataFin in Option.MPathFlags then
                                       HelpByte := $10
                                    else
                                        HelpByte := 0;

                                    if tmpDataSeq8Octets in Option.MPathFlags then
                                       HelpByte := HelpByte or $8;

                                    if tmpDNSetcPresent in Option.MPathFlags then
                                       HelpByte := HelpByte or $4;

                                    if tmpDataACK8Octets in Option.MPathFlags then
                                       HelpByte := HelpByte or $2;

                                    if tmpDataACKpresent in Option.MPathFlags then
                                       HelpByte := HelpByte or $1;

                                    pHelpByte := pByte (prot_data + Offset + 3);
                                    pHelpByte^ := HelpByte;

                                    // ACK and Sequence

                                    if tmpDataACKpresent in Option.MPathFlags then
                                       begin
                                            if tmpDataACK8Octets in Option.MPathFlags then
                                               begin
                                                    pHelpUInt64 := pUINT64 (prot_data + Offset + 4);
                                                    pHelpUint64^ := swap64 (Option.ACKNumber);

                                                    OptionDelta := 8;
                                               end
                                            else
                                                begin
                                                    pHelpLong := pLongword (prot_data + Offset + 4);
                                                    HelpLong := Option.ACKNumber;
                                                    pHelpLong^ := swap32 (HelpLong);

                                                    OptionDelta := 4;
                                                end;
                                       end;

                                    if tmpDNSetcPresent in Option.MPathFlags then
                                       begin
                                            if tmpDataSeq8Octets in Option.MPathFlags then
                                               begin
                                                    pHelpUInt64 := pUINT64 (prot_data + Offset + 4 + OptionDelta);
                                                    pHelpUint64^ := swap64 (Option.SEQNumber);
                                               end
                                            else
                                                begin
                                                    pHelpLong := pLongword (prot_data + Offset + 4 + OptionDelta);
                                                    HelpLong := Option.SEQNumber;
                                                    pHelpLong^ := swap32 (HelpLong);
                                                end;
                                       end;
                               end;
      mtcpAddAddress         : begin
                                    HelpByte := 3 shl 4;

                                    // Version

                                    HelpByte := HelpByte or Option.IPVersion;

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^ := HelpByte;

                                    // Address ID

                                    pHelpByte := pByte (prot_data + Offset + 3);
                                    pHelpByte^ := option.AddressID;

                                    case Option.IPVersion of
                                     4 : begin
                                              pHelpLong := pLongword (prot_data + Offset + 4);
                                              pHelpLong^ := swap32 (Option.IPv4Address.ValueAsLong);

                                              OptionDelta := 4;
                                         end;
                                     6 : begin
                                              pHelpIPv6Adr := pIPv6AddressByteStruct (prot_data + Offset + 4);
                                              pHelpIPv6Adr^ := Option.IPv6Address.ValueAsByteStruct;

                                              OptionDelta := 16;
                                         end;
                                    end;

                                    if Option.TCPPort <> 0 then
                                       begin
                                            pHelpWord := pWord (prot_data + Offset + 4 + OptionDelta);
                                            pHelpWord^ := swap16 (Option.TCPPort);
                                       end;
                               end;
      mtcpRemovAddress       : begin
                                    HelpByte := 4 shl 4;

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^ := HelpByte;

                                    // Address ID

                                    pHelpByte := pByte (prot_data + Offset + 3);
                                    pHelpByte^ := Option.AddressID;

                                    // TODO: there could be more than one address ID in this option
                               end;
      mtcpSubflowPrio        : begin
                                    HelpByte := 5 shl 4;

                                    if tmpBackup in Option.MPathFlags then
                                       HelpByte := HelpByte or $1;

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^ := HelpByte;
                               end;
      mtcpFallback           : begin
                                    HelpByte := 6 shl 4;

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^ := HelpByte;

                                    // Data Sequence Number

                                    pHelpUInt64 := pUINT64 (prot_data + Offset + 4);
                                    pHelpUInt64^ := Option.DataSequenceNr;
                               end;
      mtcpFastClose          : begin
                                    HelpByte := 7 shl 4;

                                    pHelpByte := pByte (prot_data + Offset + 2);
                                    pHelpByte^ := HelpByte;

                                    // Receivers Key

                                    pHelpUInt64 := pUINT64 (prot_data + Offset + 4);
                                    pHelpUInt64^ := Option.ReceiversKey;
                               end;
     end;

     inc (Offset, Option.optionlength);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeFastOpenOption (var Offset: integer; Option: tTCPFastOpenOption);

var pOptionHdr   : pTCPOptionTL;
    i            : integer;

begin
     pOptionHdr := pTCPOptionTL (prot_data + Offset);

     pOptionHdr^.optiontype := 34;
     pOptionHdr^.optionlen := Option.CookieLength + 2;

     // write cookie, if any

     if Option.CookieLength > 0 then
        for i := 0 to Option.CookieLength do
            pByte (prot_data + Offset + 2 + i)^ := Option.Cookie[i];

     inc (Offset, Option.CookieLength + 2);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.WritePadding (var offset : integer);

var i : integer;

begin
     for i := 0 to priv_paddinglen - 1 do
         pByte (prot_data + offset + i)^ := priv_optionpad[i];

     inc (offset, priv_paddinglen);
end;

(* ------------------------------------------- *)

function tTCPAssembler.writeTCPOptionListInOriginalOrder (Offset: integer): integer;

var OriginalOffset : integer;
    MPathOption    : tTCPMultiPathOption;
    FastOpenOption : tTCPFastOpenOption;
    i, j           : integer;

begin
     // Store the starting offset for later option size calculation

     OriginalOffset := Offset;

     for i := 1 to length (priv_exactoptionlist) do
         begin
              case priv_exactoptionlist[i-1] of
               tcpEOL                : begin
                                            writeEOL (Offset);
                                       end;
               tcpNOP                : begin
                                            writeNOP (Offset);
                                       end;
               tcpMSS                : begin
                                            writeMSS (Offset);
                                       end;
               tcpScaleFactor        : begin
                                            writeScaling (Offset);
                                       end;
               tcpSACKPermitted      : begin
                                            writeSACKPermitted (Offset);
                                       end;
               tcpSACK               : begin
                                            writeSACKBlocks (Offset);
                                       end;
               tcpTimestamp          : begin
                                            writeTimestamp (Offset);
                                       end;
               tcpCC                 : begin
                                            inc (Offset, 6);        // TODO: implement CC option
                                       end;
               tcpCCNEW              : begin
                                            inc (Offset, 6);        // TODO: implement CCNEW option
                                       end;
               tcpCCECHO             : begin
                                            inc (Offset, 6);        // TODO: implement CCECHO option
                                       end;
               tcpMPATH              : begin
                                            for j := 0 to length (priv_mpathoptionlist) - 1 do
                                                begin
                                                     MPathOption := priv_mpathoptionlist[j];

                                                     WriteMPathOption (Offset, MPathOption);
                                                end;
                                       end;
               tcpFastOpen           : begin
                                            for j := 0 to length (priv_fastopenoptionlist) - 1 do
                                                begin
                                                     FastOpenOption := priv_fastopenoptionlist[j];

                                                     writeFastOpenOption (Offset, FastOpenOption);
                                                end;
                                       end;
              end;
         end;

     if priv_paddinglen > 0 then
        WritePadding (Offset);

     // return length of option set

     result := Offset - OriginalOffset;
end;

(* ------------------------------------------- *)

function tTCPAssembler.writeTCPOptionListWindowsStyle (Offset: integer) : integer;

var OriginalOffset : integer;
    MPathOption    : tTCPMultiPathOption;
    FastOpenOption : tTCPFastOpenOption;
    i              : integer;

begin
     // Store the starting offset for later option size calculation

     OriginalOffset := Offset;

     // MSS

     if tcpMSS in priv_Optionset then
        writeMSS (Offset);

     if tcpScaleFactor in priv_OptionSet then
        begin
             writeNOP (Offset);
             writeScaling (Offset);
        end;

     // SACK permitted

     if tcpSACKPermitted in priv_OptionSet then
        begin
             // if no timestamp follows we need to pad

             if not (tcpTimestamp in priv_optionset) then
                begin
                     writeNOP (Offset);
                     writeNOP (Offset);
                end;

             writeSACKPermitted (Offset);
        end;

     // Timestamp

     if tcpTimestamp in priv_OptionSet then
        begin
             // we need to pad if no SACK permitted option was written before

             if not (tcpSACKPermitted in priv_optionset) then
                begin
                     writeNOP (Offset);
                     writeNOP (Offset);
                end;

             writeTimestamp (Offset);
        end;

     // SACK Blocks

     if length (priv_sackranges) > 0 then
        begin
             writeNOP (Offset);
             writeNOP (Offset);

             writeSACKBlocks (Offset);
        end;

     // Multipath Options

     for i := 1 to length (priv_mpathoptionlist) do
         begin
              MPathOption := priv_mpathoptionlist[i-1];

              WriteMPathOption (Offset, MPathOption);
         end;

     // FastOpenOptions

     for i := 1 to length (priv_fastopenoptionlist) do
         begin
              FastOpenOption := priv_fastopenoptionlist[i-1];

              writeFastOpenOption (Offset, FastOpenOption);
         end;

     if priv_paddinglen > 0 then
        WritePadding (Offset);

     // return length of option set

     result := Offset - OriginalOffset;
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.writeTCPOptionList (Offset : integer);

var optionlen : integer;

begin
     assert (priv_paddinglen = length (priv_optionpad), 'Padding byte length mismatch (tTCPAssembler.writeTCPOptionList)');

     case priv_optionformat of
      ofOriginal : optionlen := writeTCPOptionListInOriginalOrder (Offset);
      ofWindows  : optionlen := writeTCPOptionListWindowsStyle (Offset);
     end;

     assert (20 + OptionLen = prot_length, 'Memory offset check mismatch (tTCPAssembler.writeTCPOptionList)');
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.Assemble;

var pData    : pTCPStruct;
    hdrBlock : tFrameChunk;
    crcBlock : tFrameChunk;
    crcbuf   : tBufferWords;

begin
     pData := pTCPStruct (prot_data);

     pData^.Sourceport := swap16 (priv_Sourceport);
     pData^.Destinationport := swap16 (priv_Destinationport);
     pData^.Sequence := swap32 (priv_Sequence.Value);
     pData^.Acknowledge := swap32 (priv_Acknowledge.Value);
     pData^.DataOffsetFlags := setDataOffsetFlags;
     pData^.Window := swap16 (priv_Window);
     pData^.UrgentPointer := swap16 (priv_UrgentPointer);

     // Handle Options

     if priv_optionset <> [] then
        writeTCPOptionList (20);

     // Calculate the checksum

     if not priv_keepcrc then
        begin
             // Concatenate TCP header chunk and payload chunk

             hdrBlock.Data := prot_data;
             hdrBlock.Len := prot_length;

//           SiMain.LogDebug ('TCP Assemly - hdrBlock: "' + hdrBlock.getAsHEXString(',') + '"');

             crcBlock := mergeFrameChunks (hdrBlock, priv_payload);
             crcbuf := crcBlock.getAsBufferWords;

             // since the TCP header could have a different length now (depending on the options we write)
             // we might need to adjust the L4 length value of the pseudo header

             priv_IPv4PseudoHeader.L4Length := (prot_length + priv_payload.Len);
             priv_IPv6PseudoHeader.L4Length := (prot_length + priv_payload.Len);

             case priv_IPLayerVersion of
              4 : priv_checksum := getTCPChecksum (priv_IPv4PseudoHeader, crcbuf);
              6 : priv_checksum := getTCPChecksum (priv_IPv6PseudoHeader, crcbuf);
             end;

             if priv_forcebadcrc then
                priv_checksum := GetBadCRC (priv_checksum);

             crcBlock.free;
        end;

     pData^.CRC := swap16 (priv_checksum);
end;

(* ------------------------------------------- *)

procedure tTCPAssembler.CopyFromParser (Parser: tTCPParser);

var count  : integer;
    i      : integer;

begin
     priv_Sourceport := Parser.SourcePort;
     priv_Destinationport := Parser.DestinationPort;
     priv_Sequence := Parser.SequenceNumber;
     priv_Acknowledge := Parser.AcknowledgeNumber;
     priv_flags := Parser.Flags;
     priv_Window := Parser.WindowSize;
     priv_checksum := Parser.Checksum;
     priv_keepcrc := true;
     priv_UrgentPointer := Parser.UrgentPointer;

     priv_optionformat := ofOriginal;
//   priv_optionformat := ofWindows;
     priv_optionset := Parser.TCPOptionsSet;

     // copy exact option order list

     count := length (Parser.ExactOptionList);
     setLength (priv_exactoptionlist, count);

     for i := 0 to count - 1 do
         priv_exactoptionlist[i] := Parser.ExactOptionList[i];

     priv_mss := Parser.MSS;
     priv_ScaleFactor := Parser.Scalefactor;
     priv_Timestamp := Parser.Timestamp;

     // SACK ranges

     count := length (Parser.SACKRanges);

     setlength (priv_sackranges, count);

     for i := 0 to count - 1 do
         begin
              priv_sackranges[i].leftEdge := Parser.SACKRanges[i].leftEdge;
              priv_sackranges[i].rightEdge := Parser.SACKRanges[i].rightEdge;
         end;

     // Multipath TCP Options

     count := length (Parser.MultiPathOptionList);
     setlength (priv_mpathoptionlist, count);

     for i := 0 to count - 1 do
         priv_mpathoptionlist[i].CopyFrom (Parser.MultiPathOptionList[i]);

     // FastOpen options

     count := length (Parser.FastOpenOptionList);
     setlength (priv_fastopenoptionlist, count);

     for i := 0 to count - 1 do
         priv_fastopenoptionlist[i].CopyFrom (Parser.FastOpenOptionList[i]);

     // Padding

     priv_paddinglen := Parser.PaddingLength;
     setLength (priv_optionpad, priv_paddinglen);

     for i := 0 to priv_paddinglen - 1 do
         priv_optionpad[i] := Parser.PaddingBytes[i];

     // Transfer the original header length, in case we need to indicate truncated options

     priv_originalhdrlen := Parser.HeaderLength;
end;

(* ------------------------------------------- *)

end.
