{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_udp;

interface

uses base_structures,
     class_parser_udp,
     base_checksums,
     class_assembly_baseclass;

type tUDPAssembler  = class (tFrameAssembler)

                                 {$REGION 'Documentation'}
                                 ///	<param name="IPv4PseudoHeader">
                                 ///	  IPv4 Pseudoheader as needed for CRC
                                 ///	  calculation
                                 ///	</param>
                                 ///	<param name="UDPPayload">
                                 ///	  UDP payload as word array, for CRC
                                 ///	  calculation
                                 ///	</param>
                                 {$ENDREGION}
                                 constructor create (IPv4PseudoHeader : tIPv4PseudoHeader; UDPPayload : tFrameChunk); overload;
                                 constructor create (IPv4PseudoHeader : tIPv4PseudoHeader); overload;
                                 {$REGION 'Documentation'}
                                 ///	<param name="IPv6PseudoHeader">
                                 ///	  IPv6 Pseudoheader as needed for CRC
                                 ///	  calculation
                                 ///	</param>
                                 ///	<param name="UDPPayload">
                                 ///	  UDP payload as word array, for CRC
                                 ///	  calculation
                                 ///	</param>
                                 {$ENDREGION}
                                 constructor create (IPv6PseudoHeader : tIPv6PseudoHeader; UDPPayload : tFrameChunk); overload;
                                 constructor create (IPv6PseudoHeader : tIPv6PseudoHeader); overload;

                                 private
                                        priv_sourceport       : word;
                                        priv_destinationport  : word;
                                        priv_totallength      : word;
                                        priv_crc              : word;

                                        priv_payload          : tFrameChunk;

                                        priv_IPLayerVersion   : byte;
                                        priv_IPv4PseudoHeader : tIPv4PseudoHeader;
                                        priv_IPv6PseudoHeader : tIPv6PseudoHeader;
                                        priv_keepcrc          : boolean;
                                        priv_forcebadcrc      : boolean;

                                        procedure setManualCRC (const Value: word);

                                 protected
                                          // called from base class from GetAssembly

                                          procedure DetermineAssemblySize; override;
                                          procedure Assemble; override;

                                 public
                                       property SourcePort : word read priv_sourceport write priv_sourceport;
                                       property DestinationPort : word read priv_destinationport write priv_destinationport;
                                       property TotalLength : word read priv_totallength write priv_totallength;
                                       property Checksum : word read priv_crc write setManualCRC;

                                       property DoNotRecalculateCRC : boolean read priv_keepcrc write priv_keepcrc;
                                       property ForceBadCRC : boolean read priv_forcebadcrc write priv_forcebadcrc;

                                       procedure CopyFromParser (Parser : tUDPParser);
                           end;


implementation

uses SiAuto,
     toolkit;

(* ------------------------------------------- *)
{ tUDPAssembler }
(* ------------------------------------------- *)

constructor tUDPAssembler.create (IPv4PseudoHeader : tIPv4PseudoHeader; UDPPayload : tFrameChunk);

begin
     inherited create;

     priv_totallength := IPv4PseudoHeader.L4Length;

     priv_IPLayerVersion := 4;
     priv_IPv4PseudoHeader := IPv4PseudoHeader;

     priv_payload := UDPPayload;

     priv_forcebadcrc := false;
     priv_keepcrc := false;
end;

(* ------------------------------------------- *)

constructor tUDPAssembler.create (IPv4PseudoHeader : tIPv4PseudoHeader);

begin
     inherited create;

     priv_totallength := IPv4PseudoHeader.L4Length;

     priv_IPLayerVersion := 4;
     priv_IPv4PseudoHeader := IPv4PseudoHeader;

     priv_payload.Len := 0;
     priv_payload.Data := nil;

     priv_forcebadcrc := false;
     priv_keepcrc := false;
end;

(* ------------------------------------------- *)

constructor tUDPAssembler.create (IPv6PseudoHeader : tIPv6PseudoHeader; UDPPayload : tFrameChunk);

begin
     inherited create;

     priv_totallength := IPv6PseudoHeader.L4Length;

     priv_IPLayerVersion := 6;
     priv_IPv6PseudoHeader := IPv6PseudoHeader;

     priv_forcebadcrc := false;
     priv_keepcrc := false;

     priv_payload := UDPPayload;
end;

(* ------------------------------------------- *)

constructor tUDPAssembler.create (IPv6PseudoHeader : tIPv6PseudoHeader);

begin
     inherited create;

     priv_totallength := IPv6PseudoHeader.L4Length;

     priv_IPLayerVersion := 6;
     priv_IPv6PseudoHeader := IPv6PseudoHeader;

     priv_payload.Len := 0;
     priv_payload.Data := nil;

     priv_forcebadcrc := false;
     priv_keepcrc := false;
end;

(* ------------------------------------------- *)

procedure tUDPAssembler.DetermineAssemblySize;

begin
     prot_length := sizeOf (tUDPStruct);
end;

(* ------------------------------------------- *)

procedure tUDPAssembler.setManualCRC (const Value: word);

begin
     priv_keepcrc := true;
     priv_forcebadcrc := false;
     priv_crc := Value;
end;

(* ------------------------------------------- *)

procedure tUDPAssembler.Assemble;

var pData    : pUDPStruct;
    hdrBlock : tFrameChunk;
    crcBlock : tFrameChunk;
    crcbuf   : tBufferWords;

begin
//   SiMain.EnterMethod('tUDPAssembler.Assemble');

     pData := pUDPStruct (prot_data);

     pData^.sourceport := swap16 (priv_sourceport);
     pData^.destinationport := swap16 (priv_destinationport);
     pData^.length := swap16 (priv_totallength);

     // Create CRC

     if not priv_keepcrc then
        begin
             // Concatenate UDP header chunk and payload chunk

             hdrBlock.Data := prot_data;
             hdrBlock.Len := prot_length;

             crcBlock := mergeFrameChunks (hdrBlock, priv_payload);
             crcbuf := crcBlock.getAsBufferWords;

             case priv_IPLayerVersion of
              4 : priv_crc := getUDPChecksum (priv_IPv4PseudoHeader, crcbuf);
              6 : priv_crc := getUDPChecksum (priv_IPv6PseudoHeader, crcbuf);
             end;

             if priv_forcebadcrc then
                priv_crc := getBadCRC (priv_crc);
        end;

     // Write CRC

     pData^.crc := swap16 (priv_crc);

//   SiMain.LeaveMethod('tUDPAssembler.Assemble');
end;

(* ------------------------------------------- *)

procedure tUDPAssembler.CopyFromParser(Parser: tUDPParser);

begin
     priv_sourceport := Parser.SourcePort;
     priv_destinationport := Parser.DestinationPort;

     priv_crc := Parser.Checksum;
     priv_keepcrc := true;
end;

(* ------------------------------------------- *)

end.
