{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit base_checksums;

interface

uses base_structures,
     definition_ip_addresses;

type tIPv4HeaderWords = array [1..30] of word;       // equals 15 * 32 bit, which is the maximum header size
     pIPv4HeaderWords = ^tIPv4HeaderWords;

     tIPv4PseudoHeader = record
                               Source      : longword;
                               Destination : longword;
                               Protocol    : word;
                               L4Length    : word;

                               function GetSourceLowWord : word;
                               function GetSourceHighWord : word;
                               function GetDestinationLowWord : word;
                               function GetDestinationHighWord : word;

                               function getDebug : string;
                         end;

     tIPv6PseudoHeader = record
                               Source      : tIPv6AddressWordStruct;
                               Destination : tIPv6AddressWordStruct;
                               L4Length    : longword;                // IP payload length minus extension header size
                               NextHeader  : byte;                    // 6 or 17, not extension headers, see rfc 2460

                               function GetLengthHighWord : word;
                               function getLengthLowWord : word;
                               function GetNextHeaderWord : word;

                               function getDebug : string;
                         end;

     function getBadCRC (GoodCRC : word) : word;

     {$REGION 'Documentation'}
     ///	<param name="Buffer">
     ///	  Byte Pointer toframe data buffer
     ///	</param>
     ///	<param name="Offset">
     ///	  offset at which thedata to be converted starts
     ///	</param>
     ///	<param name="Length">
     ///	  length of the data to be converted
     ///	</param>
     ///	<returns>
     ///	  Buffer contentas BigEndian word array; usually needed forchecksum
     ///	  calculation
     ///	</returns>
     {$ENDREGION}
     function getIPv4Checksum (IPHeader : tIPv4HeaderWords; HeaderLen : integer) : word;
     function getTCPChecksum (IPv4Header : tIPv4PseudoHeader; TCPSection : tBufferWords) : word; overload;
     function getTCPChecksum (IPv6Header : tIPv6PseudoHeader; TCPSection : tBufferWords) : word; overload;
     function getUDPChecksum (IPv4Header : tIPv4PseudoHeader; UDPSection : tBufferWords) : word; overload;
     function getUDPChecksum (IPv6Header : tIPv6PseudoHeader; UDPSection : tBufferWords) : word; overload;
     function getICMPv4Checksum (ICMPSection : tBufferWords; Payload : tBufferWords) : word;
     function getICMPv6Checksum (IPv6Header : tIPv6PseudoHeader; ICMPSection : tBufferWords; Payload : tBufferWords) : word;


implementation

uses SysUtils,
     toolkit,
     SiAuto;

const OFFSET_IPV4_HDRCHKSUM = 6;          // the sixt word contains the IPv4 checksum
      OFFSET_TCP_HDRCHKSUM = 9;           // the ninth word contains the TCP checksum
      OFFSET_UDP_HDRCHKSUM = 4;           // the fourth word contains the UDP checksum
      OFFSET_ICMP_HDRCHKSUM = 2;          // the second word contains the ICMP checksum

(* ------------------------------------------- *)
{ tIPv4PseudoHeader }
(* ------------------------------------------- *)

function tIPv4PseudoHeader.GetDestinationHighWord: word;

var help : longword;

begin
     help := (Destination and $ffff0000) shr 16;
     result := help;
end;

(* ------------------------------------------- *)

function tIPv4PseudoHeader.GetDestinationLowWord: word;

begin
     result := (Destination and $ffff);
end;

(* ------------------------------------------- *)

function tIPv4PseudoHeader.GetSourceHighWord: word;

var help : longword;

begin
     help := (Source and $ffff0000) shr 16;
     result := help;
end;

(* ------------------------------------------- *)

function tIPv4PseudoHeader.GetSourceLowWord: word;

begin
     result := (Source and $ffff);
end;

(* ------------------------------------------- *)

function tIPv4PseudoHeader.getDebug: string;

var value : string;
    src  : tIPv4Address;
    dst  : tIPv4Address;

begin
     src.ValueAsLong := Source;
     dst.ValueAsLong := Destination;

     value := src.ValueAsString + ',' + dst.ValueAsString + ',' + IntTostr (Protocol) + ',' + IntTostr (L4Length);

     result := value;
end;

(* ------------------------------------------- *)
{ tIPv6PseudoHeader }
(* ------------------------------------------- *)

function tIPv6PseudoHeader.GetLengthHighWord: word;

var help : longword;

begin
     help := (L4Length and $ffff0000) shr 16;
     result := help;
end;

(* ------------------------------------------- *)

function tIPv6PseudoHeader.getLengthLowWord: word;

begin
     result := (L4Length and $ffff);
end;

(* ------------------------------------------- *)

function tIPv6PseudoHeader.GetNextHeaderWord: word;

begin
     result := NextHeader;
end;

(* ------------------------------------------- *)

function tIPv6PseudoHeader.getDebug: string;

var value : string;
    src   : tIPv6Address;
    dst   : tIPv6Address;

begin
     src.ValueAsWordStruct := Source;
     dst.ValueAsWordStruct := Destination;

     value := src.ValueAsString + ',' + dst.ValueAsString + ',' + IntTostr (NextHeader) + ',' + IntTostr (L4Length);

     result := value;
end;

(* ------------------------------------------- *)
{ functions }
(* ------------------------------------------- *)

function getBadCRC (GoodCRC : word) : word;

var value : word;

begin
     if GoodCRC < 65535 then
        value := GoodCRC + 1
     else
         value := 0;

     result := value;
end;

(* ------------------------------------------- *)

function getIPv4Checksum (IPHeader : tIPv4HeaderWords; HeaderLen : integer) : word;

var temp  : longword;
    help  : longword;
    i     : integer;

begin
     temp := 0;

     // Add all 16 bit words, but keeping it in the 16 bit range

     for i := 1 to (HeaderLen Div 2) do
         begin
              // Ignore the checksum word

              if i <> OFFSET_IPV4_HDRCHKSUM then
                 temp := temp + IPHeader[i];
         end;

     // If we have a carry beyond 16 bits we add it

     while temp > $ffff do
           begin
                help := (temp shr 16);

                temp := (temp and $ffff) + help;
           end;

     // Get one's completent

     temp := ($ffffffff - temp);
     temp := temp and $0000ffff;   // clear the higher two bytes in the longword;

{
     SiMain.LogVerbose ('IPv4 CRC calculated: 0x' + IntToHex (temp, 2));
}

     result := temp;
end;

(* ------------------------------------------- *)

function getTCPChecksum (IPv4Header : tIPv4PseudoHeader; TCPSection : tBufferWords) : word;

var temp  : longword;
    help  : longword;
    count : integer;
    i     : integer;

begin
     // Add pseudoheader values

     temp := IPv4Header.GetSourceHighWord + IPv4Header.GetSourceLowWord;
     temp := temp + IPv4Header.GetDestinationHighWord + IPv4Header.GetDestinationLowWord;
     temp := temp + IPv4Header.Protocol;
     temp := temp + IPv4Header.L4Length;

     count := length (TCPSection);

     for i := 1 to count do
         begin
              // Ignore the checksum word

              if i <> OFFSET_TCP_HDRCHKSUM then
                 temp := temp + TCPSection[i-1];
         end;

     // If we have a carry beyond 16 bits we add it

     while temp > $ffff do
           begin
                help := (temp shr 16);

                temp := (temp and $ffff) + help;
           end;

     // Get one's completent

     temp := ($ffffffff - temp);
     temp := temp and $0000ffff;   // clear the higher two bytes in the longword;

{
     SiMain.LogVerbose ('CRC Buffer: ' + CRLF + getBufferWordsAsHexString (TCPSection));
     SiMain.LogVerbose ('IPv4 PseudoHeader: ' + CRLF + IPv4Header.getDebug);
     SiMain.LogVerbose ('TCP CRC calculated: 0x' + IntToHex (temp, 2));
}

     result := temp;
end;

(* ------------------------------------------- *)

function getTCPChecksum (IPv6Header : tIPv6PseudoHeader; TCPSection : tBufferWords) : word;

var temp  : longword;
    help  : longword;
    count : integer;
    i     : integer;

begin
     // Workaround for Extension Headers giving us a Protocol type of 0 - we know this must be TCP

     IPv6Header.NextHeader := 6;

     // Add pseudoheader values

     temp := 0;

     for i := 0 to 7 do
         temp := temp + IPv6Header.Source[i] + IPv6Header.Destination[i];          // Instead of adding all source words first and then destinations we do it all at once

     temp := temp + IPv6Header.GetLengthHighWord + IPv6Header.getLengthLowWord;
     temp := temp + IPv6Header.GetNextHeaderWord;

     // Add TCP section

     count := length (TCPSection);

     for i := 1 to count do
         begin
              // Ignore the checksum word

              if i <> OFFSET_TCP_HDRCHKSUM then
                 temp := temp + TCPSection[i-1];
         end;

     // If we have a carry beyond 16 bits we add it

     while temp > $ffff do
           begin
                help := (temp shr 16);

                temp := (temp and $ffff) + help;
           end;

     // Get one's completent

     temp := ($ffffffff - temp);
     temp := temp and $0000ffff;   // clear the higher two bytes in the longword;

{
     SiMain.LogVerbose ('CRC Buffer: ' + CRLF + getBufferWordsAsHexString (TCPSection));
     SiMain.LogVerbose ('IPv4 PseudoHeader: ' + CRLF + IPv6Header.getDebug);
     SiMain.LogVerbose ('TCP CRC calculated: 0x' + IntToHex (temp, 2));
}

     result := temp;
end;

(* ------------------------------------------- *)

function getUDPChecksum (IPv4Header : tIPv4PseudoHeader; UDPSection : tBufferWords) : word;

var temp  : longword;
    help  : longword;
    count : integer;
    i     : integer;

begin
     // Add pseudoheader values

     temp := IPv4Header.GetSourceHighWord + IPv4Header.GetSourceLowWord;
     temp := temp + IPv4Header.GetDestinationHighWord + IPv4Header.GetDestinationLowWord;
     temp := temp + IPv4Header.Protocol;
     temp := temp + IPv4Header.L4Length;

     count := length (UDPSection);

     for i := 1 to count do
         begin
              // Ignore the checksum word

              if i <> OFFSET_UDP_HDRCHKSUM then
                 temp := temp + UDPSection[i-1];
         end;

     // If we have a carry beyond 16 bits we add it

     while temp > $ffff do
           begin
                help := (temp shr 16);

                temp := (temp and $ffff) + help;
           end;

     // Get one's completent

     temp := ($ffffffff - temp);
     temp := temp and $0000ffff;   // clear the higher two bytes in the longword;

     // Since 0 is reserved for "no checksum computed, it will be replaced with $ffff if
     // we actually get it from the calculation
     // see https://ask.wireshark.org/questions/13962/can-tcp-checksum-be-zero0

     if temp = 0 then
        temp := $ffff;
{
     SiMain.LogVerbose ('CRC Buffer: ' + CRLF + getBufferWordsAsHexString (UDPSection));
     SiMain.LogVerbose ('IPv4 PseudoHeader: ' + CRLF + IPv4Header.getDebug);
     SiMain.LogVerbose ('UDP CRC calculated: 0x' + IntToHex (temp, 2));
}

     result := temp;
end;

(* ------------------------------------------- *)

function getUDPChecksum (IPv6Header : tIPv6PseudoHeader; UDPSection : tBufferWords) : word;

var temp  : longword;
    help  : longword;
    count : integer;
    i     : integer;

begin
     // Workaround for Extension Headers giving us a Protocol type of 0 - we know this must be UDP

     IPv6Header.NextHeader := 17;

     // Add pseudoheader values

     temp := 0;

     for i := 0 to 7 do
         temp := temp + IPv6Header.Source[i] + IPv6Header.Destination[i];          // Instead of adding all source words first and then destinations we do it all at once

     temp := temp + IPv6Header.GetLengthHighWord + IPv6Header.getLengthLowWord;
     temp := temp + IPv6Header.GetNextHeaderWord;

     // Add UDP section

     count := length (UDPSection);

     for i := 1 to count do
         begin
              // Ignore the checksum word

              if i <> OFFSET_UDP_HDRCHKSUM then
                 temp := temp + UDPSection[i-1];
         end;

     // If we have a carry beyond 16 bits we add it

     while temp > $ffff do
           begin
                help := (temp shr 16);

                temp := (temp and $ffff) + help;
           end;

     // Get one's completent

     temp := ($ffffffff - temp);
     temp := temp and $0000ffff;   // clear the higher two bytes in the longword;

{
     SiMain.LogVerbose ('CRC Buffer: ' + CRLF + getBufferWordsAsHexString (UDPSection));
     SiMain.LogVerbose ('IPv4 PseudoHeader: ' + CRLF + IPv6Header.getDebug);
     SiMain.LogVerbose ('UDP CRC calculated: 0x' + IntToHex (temp, 2));
}

     result := temp;
end;

(* ------------------------------------------- *)

function getICMPv4Checksum (ICMPSection : tBufferWords; Payload : tBufferWords) : word;

var count : integer;
    temp  : longword;
    help  : longword;
    i     : integer;

begin
     count := length (ICMPSection);

     temp := 0;

     for i := 1 to count do
         begin
              // Ignore the checksum word

              if i <> OFFSET_ICMP_HDRCHKSUM then
                 temp := temp + ICMPSection[i-1];
         end;

     // Calculate payload

     for i := 1 to length (Payload) do
         temp := temp + Payload[i-1];

     // If we have a carry beyond 16 bits we add it

     while temp > $ffff do
           begin
                help := (temp shr 16);

                temp := (temp and $ffff) + help;
           end;

     // Get one's completent

     temp := ($ffffffff - temp);
     temp := temp and $0000ffff;   // clear the higher two bytes in the longword;
{
     SiMain.LogVerbose ('CRC Buffer: ' + CRLF + getBufferWordsAsHexString (ICMPSection));
     SiMain.LogVerbose ('ICMPv4 CRC calculated: 0x' + IntToHex (temp, 2));
}
     result := temp;
end;

(* ------------------------------------------- *)

function getICMPv6Checksum (IPv6Header : tIPv6PseudoHeader; ICMPSection : tBufferWords; Payload : tBufferWords) : word;

var count : integer;
    temp  : longword;
    help  : longword;
    i     : integer;

begin
     // Workaround for Extension Headers giving us a Protocol type of 0 - we know this must be ICMPv6

     IPv6Header.NextHeader := 58;

     // Add pseudoheader values

     temp := 0;

     for i := 0 to 7 do
         temp := temp + IPv6Header.Source[i] + IPv6Header.Destination[i];          // Instead of adding all source words first and then destinations we do it all at once

     temp := temp + IPv6Header.GetLengthHighWord + IPv6Header.getLengthLowWord;
     temp := temp + IPv6Header.GetNextHeaderWord;

     count := length (ICMPSection);

     for i := 1 to count do
         begin
              // Ignore the checksum word

              if i <> OFFSET_ICMP_HDRCHKSUM then
                 temp := temp + ICMPSection[i-1];
         end;

     // Calculate payload

     for i := 1 to length (Payload) do
         temp := temp + Payload[i-1];

     // If we have a carry beyond 16 bits we add it

     while temp > $ffff do
           begin
                help := (temp shr 16);

                temp := (temp and $ffff) + help;
           end;

     // Get one's completent

     temp := ($ffffffff - temp);
     temp := temp and $0000ffff;   // clear the higher two bytes in the longword;
{
     SiMain.LogVerbose ('CRC Buffer: ' + CRLF + getBufferWordsAsHexString (ICMPSection));
     SiMain.LogVerbose ('ICMPv6 CRC calculated: 0x' + IntToHex (temp, 2));
}
     result := temp;
end;

(* ------------------------------------------- *)

end.
