{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_dirparser;

// This unit defines a base object for finding and processing files in a directory.
// For each file the processFile routine is called.
//
// You can set the IncludeSubDirs property to false to exclude subdirectories
// You can set the ProcessFileInDir property to force processing the file while its subdirectory is scanned
// If you add entries to the ValidExtensions stringlist, only matching files will be added
//
// Don't forget to overload the processFile procedure or the execute method will
// fail with an abstract error!!!

interface

uses System.Classes;

type tOnAddFile = procedure (Sender : tObject; const Filename : string) of object;
     tOnScanDirectory = procedure (Sender : tObject; const Path : string) of object;

     tDirParser   = class (TObject)

                          constructor create; overload;
                          constructor create (const BaseDirectory : string; const FileMask : string = '*.*'; const ScanRecursive : boolean = true); overload;

                          private

                          protected
                                   EventOnAddFile       : tOnAddFile;
                                   EventOnScanDirectory : tOnScanDirectory;

                                   prot_BaseDir             : string;
                                   prot_FileMask            : string;
                                   prot_Recursive           : boolean;
                                   prot_ProcessPerDir       : boolean;
                                   prot_validExtensions     : tStringList;
                                   prot_FileList            : tStringlist;
                                   prot_IgnorePathList      : tStringlist;

                                   procedure setBaseDir (const Path : string);
                                   procedure setFileMask (const Mask : string);
                                   procedure GetFilesInDirectory;
                                   function getFileCount : integer;

                                   procedure addFileToList (const Filename : widestring); virtual;
                                   procedure processFile (const Filename : widestring); virtual;
                          public
                                property BaseDirectory : string read prot_BaseDir write setBaseDir;
                                property FileMask : string read prot_FileMask write setFileMask;
                                property ValidExtensions : tStringList read prot_validExtensions write prot_validExtensions;
                                property FileCount : integer read getFileCount;
                                property IncludeSubDirs : boolean read prot_Recursive write prot_Recursive;
                                property ProcessFileInDir : boolean read prot_ProcessPerDir write prot_ProcessPerDir;

                                property OnAddFile : tOnAddFile read EventOnAddFile write EventOnAddFile;
                                property OnScanDirectory : tOnScanDirectory read EventOnScanDirectory write EventOnScanDirectory;

                                procedure execute;
                                function getFileName (index : integer) : widestring;
                                procedure BulkSetValidExtensions (Extensions, Separator : string; IncludesDot : boolean);

                                destructor Destroy; override;
                    end;


implementation

uses
     System.SysUtils,
     Toolkit;




(* --------------------------------- *)
{ tDirParser }
(* --------------------------------- *)

constructor tDirParser.create;

begin
     inherited;

     prot_FileList := tStringList.Create;
     prot_validextensions := tStringlist.Create;
     prot_IgnorePathList := tStringlist.Create;

     prot_BaseDir := '';
     prot_FileMask := '';
     prot_Recursive := true;
     prot_ProcessPerDir := false;
end;

(* --------------------------------- *)

constructor tDirParser.create (const BaseDirectory, FileMask: string; const ScanRecursive: boolean);

begin
     inherited create;

     prot_FileList := tStringList.Create;
     prot_validextensions := tStringlist.Create;
     prot_IgnorePathList := tStringlist.Create;

     prot_BaseDir := IncludeTrailingPathDelimiter (BaseDirectory);
     prot_FileMask := FileMask;
     prot_Recursive := ScanRecursive;
     prot_ProcessPerDir := false;
end;

(* --------------------------------- *)

destructor tDirParser.Destroy;

begin
     // do what is neccessary

     prot_FileList.Free;
     prot_validextensions.Free;

     inherited;
end;

(* --------------------------------- *)

procedure tDirParser.setBaseDir(const Path: string);

begin
     prot_BaseDir := Path;
end;

(* --------------------------------- *)

procedure tDirParser.setFileMask(const Mask: string);

begin
     prot_FileMask := Mask;
end;

(* --------------------------------- *)

// Default Behaviour: just add the file. This can be overloaded if neccessary

procedure tDirParser.addFileToList (const Filename: widestring);

var sExt : string;

begin
     if assigned (EventOnAddFile) then
        EventOnAddFile (self, Filename);

     if prot_validextensions.Count = 0 then
        begin
             prot_FileList.Add(Filename);
        end
     else
         begin
              sExt := rightstring (ExtractFileExt (Filename), '.');

              if prot_validExtensions.IndexOf (sExt) > -1 then
                 prot_FileList.Add(Filename);
         end;
end;

(* --------------------------------- *)

// Default Behaviour: Do nothing. This can be overloaded if neccessary

procedure tDirParser.processFile(const Filename: widestring);

begin

end;

(* --------------------------------- *)

procedure tDirParser.GetFilesInDirectory;

   procedure ScanDir(const Directory: String);

   var SR: TSearchRec;

   begin
        if assigned (EventOnScanDirectory) then
           EventOnScanDirectory (self, Directory);

        if FindFirst (Directory + prot_FileMask, faAnyFile - faDirectory, SR) = 0 then
           try
              repeat
                    // Add the file to the list

                    addFileToList (Directory + SR.Name);

                    // If a processing of the file is required while its directory is processed
                    // we call the processFile routine from here

                    if prot_ProcessPerDir then
                       processFile (Directory + SR.Name);

              until FindNext(SR) <> 0;
           finally
                  FindClose(SR);
           end;

           if prot_Recursive then
              begin
                   if FindFirst(Directory + '*.*', faDirectory, SR) = 0 then
                      try
                         repeat
                               if ((SR.attr and faDirectory) = faDirectory) and (SR.Name <> '.') and (SR.Name <> '..') then
                                  ScanDir(Directory + SR.Name + '\');
                         until FindNext(SR) <> 0;
                      finally
                             FindClose(SR);
                      end;
              end;
   end;

begin
     prot_FileList.BeginUpdate;

     try
        prot_FileList.Clear;

        if prot_BaseDir = '' then
           Exit;

        prot_BaseDir := IncludeTrailingPathDelimiter (prot_BaseDir);

        ScanDir(prot_BaseDir);
     finally
            prot_FileList.EndUpdate;
     end;
end;

(* --------------------------------- *)

function tDirParser.getFileCount: integer;

begin
     result := prot_FileList.Count;
end;

(* --------------------------------- *)

function tDirParser.getFileName(index: integer): widestring;

begin
     result := '';

     if (index > 0) and (index <= prot_FileList.Count) then
        result := prot_FileList[index - 1];
end;

(* --------------------------------- *)

procedure tDirParser.BulkSetValidExtensions (Extensions, Separator : string; IncludesDot : boolean);

var sHelp : string;
    i     : integer;

begin
     prot_validExtensions.Clear;

     for i := 1 to numEntries (Extensions, Separator) do
         begin
              sHelp := getEntry (Extensions, Separator, i);

              if IncludesDot then
                 sHelp := stringReplace (sHelp, '.', '', [rfReplaceAll]);

              prot_validExtensions.Add(sHelp)
         end;
end;

(* --------------------------------- *)

procedure tDirParser.execute;

var i : integer;

begin
     if prot_FileMask = '' then
        prot_FileMask := '*.*';

     if prot_BaseDir = '' then
        prot_BaseDir := GetCurrentDir;

     GetFilesInDirectory;

     // If all files should be processed after collection of the filenames we do it here

     if not prot_ProcessPerDir then
        for i := 0 to prot_FileList.Count - 1 do
            begin
                 processFile (prot_FileList[i]);
            end;
end;

(* --------------------------------- *)

end.
