{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit definition_interface_mapping_types;

interface

type tAbsoluteFileInterface = record
                                    FileIndex      : integer;
                                    InterfaceIndex : longword;

                                    class operator equal (a, b : tAbsoluteFileInterface) : boolean;
                                    class operator notequal (a, b : tAbsoluteFileInterface) : boolean;

                                    procedure initialize; overload;
                                    procedure initialize (const InitialFileIndex : integer; const InitialInterfaceindex : longword); overload;
                              end;

     tInterfaceIndexList = array of integer;

     tInterfaceFileIDEntry = record
                                  FileIndex      : integer;
                                  InterfaceFlags : longword;              // bits 0-31 are set to 1 if the interface is used for the current endpoint/conversation

                                  class operator equal (a, b : tInterfaceFileIDEntry) : boolean;
                                  class operator notequal (a, b : tInterfaceFileIDEntry) : boolean;

                                  procedure initialize; overload;
                                  procedure initialize (const ValueAsString : string); overload;
                                  {$REGION 'Documentation'}
                                  /// <summary>
                                  ///   Initializes an interface entry with a
                                  ///   file index and an interface index
                                  /// </summary>
                                  /// <param name="InitFileIndex">
                                  ///   file index
                                  /// </param>
                                  /// <param name="InitInterfaceIndex">
                                  ///   interface index (not a flag value!)
                                  /// </param>
                                  {$ENDREGION}
                                  procedure initialize (const InitFileIndex : integer; const InitInterfaceIndex : longword); overload;
                                  procedure initialize (const AbsoluteIndex : tAbsoluteFileInterface); overload;
                                  procedure MarkInterfaceIndexActive (const Index : integer);
                                  function toString : string;
                                  function GetInterfaceIndexList : tInterfaceIndexList;
                            end;
     tInterfaceFileIDList = array of tInterfaceFileIDEntry;

     function InterfaceListToStr (List : tInterfaceFileIDList) : string;
     function StrToInterfaceList (Value : string) : tInterfaceFileIDList;
     procedure copyInterfaceEntryList (const SourceList : tInterfaceFileIDList; var TargetList : tInterfaceFileIDList);
     {$REGION 'Documentation'}
     /// <summary>
     ///   Merges two lists containing file indexes and the interfaces that are
     ///   used in that file. Those lists are used to mark which endpoint or
     ///   conversation uses what interface in what file
     /// </summary>
     /// <param name="List1">
     ///   List of InterfaceFileIDEntries
     /// </param>
     /// <param name="List2">
     ///   List of InterfaceFileIDEntries
     /// </param>
     {$ENDREGION}
     function mergeInterfaceLists (List1, List2 : tInterfaceFileIDList) : tInterfaceFileIDList;
     {$REGION 'Documentation'}
     /// <summary>
     ///   Merges an InterfaceEntry into an existing InterfaceEntryList
     /// </summary>
     /// <param name="List">
     ///   List of InterfaceEntries, containing FileIndex and an InterfaceIndexMap
     /// </param>
     /// <param name="Entry">
     ///   Entry that will be added to the list, or updated if already in the list
     /// </param>
     {$ENDREGION}
     procedure mergeInterfaceEntryToList (var List : tInterfaceFileIDList; const Entry : tInterfaceFileIDEntry);
     function FileIndexExistsInList (const List : tInterfaceFileIDList; const FileIndex : integer) : boolean;
     function InterfaceListsAreIdentical (List1, List2 : tInterfaceFileIDList) : boolean;

implementation

uses SysUtils,
     SiAuto,
     toolkit;

(* ------------------------------------------- *)
{ tAbsoluteFileInterface }
(* ------------------------------------------- *)

class operator tAbsoluteFileInterface.equal (a, b: tAbsoluteFileInterface): boolean;

begin
     result := (a.FileIndex = b.FileIndex) and (a.InterfaceIndex = b.InterfaceIndex);
end;

(* ------------------------------------------- *)

class operator tAbsoluteFileInterface.notequal (a, b: tAbsoluteFileInterface): boolean;

begin
     result := not (a=b);
end;

(* ------------------------------------------- *)

procedure tAbsoluteFileInterface.initialize;

begin
     initialize (0, 0);
end;

(* ------------------------------------------- *)

procedure tAbsoluteFileInterface.initialize (const InitialFileIndex : integer; const InitialInterfaceindex : longword);

begin
     FileIndex := InitialFileIndex;
     InterfaceIndex := InitialInterfaceIndex;
end;

(* ------------------------------------------- *)
{ tInterfaceFileIDList }
(* ------------------------------------------- *)

procedure tInterfaceFileIDEntry.initialize;

begin
     FileIndex := 0;
     InterfaceFlags := 0;
end;

(* ------------------------------------------- *)

procedure tInterfaceFileIDEntry.initialize (const ValueAsString: string);

var sub   : string;
    count : integer;
    ifid  : string;
    intid : integer;
    i     : integer;

begin
     if numEntries (ValueAsString, '/') > 0 then
        begin
             sub := trim (getEntry (ValueAsString, '/', 1));
             tryStrToInt (sub, FileIndex);

             sub := getEntry (ValueAsString, '/', 2);

             if FileIndex > 0 then
                begin
                     // Get all interface IDs (comma separated)

                     count := numEntries (sub, ',');

                     for i := 1 to count do
                         begin
                              ifid := getEntry (sub, ',', i);
                              tryStrToInt (ifid, intid);

                              // flag interface as active

                              InterfaceFlags := InterfaceFlags or (1 shl intid);
                         end;
                end
             else
                 initialize;
        end
     else
         begin
              // Get all interface IDs (comma separated)

              count := numEntries (sub, ',');

              for i := 1 to count do
                  begin
                       ifid := getEntry (sub, ',', i);
                       tryStrToInt (ifid, intid);

                       // flag interface as active

                       InterfaceFlags := InterfaceFlags or (1 shl intid);
                  end;
         end;
end;

(* ------------------------------------------- *)

procedure tInterfaceFileIDEntry.initialize (const InitFileIndex: integer; const InitInterfaceIndex: longword);

begin
     FileIndex := InitFileIndex;
     InterfaceFlags := 0 or (1 shl InitInterfaceIndex);
end;

(* ------------------------------------------- *)

procedure tInterfaceFileIDEntry.initialize (const AbsoluteIndex: tAbsoluteFileInterface);

begin
     Initialize (AbsoluteIndex.FileIndex, AbsoluteIndex.InterfaceIndex);
end;

(* ------------------------------------------- *)

class operator tInterfaceFileIDEntry.equal (a, b: tInterfaceFileIDEntry): boolean;

begin
     result := (a.FileIndex = b.FileIndex) and (a.InterfaceFlags = b.InterfaceFlags);
end;

(* ------------------------------------------- *)

class operator tInterfaceFileIDEntry.notequal (a, b: tInterfaceFileIDEntry): boolean;

begin
     result := not (a=b);
end;

(* ------------------------------------------- *)

function tInterfaceFileIDEntry.toString: string;

var value : string;
    test  : longword;
    len   : integer;
    i     : integer;

begin
     value := '';

     // Store all interface IDs (comma separated)

     for i := 0 to (sizeof (InterfaceFlags) * 8) - 1 do
         begin
              test := (1 shl i);

              if InterfaceFlags and test  = test then
                 value := value + IntToStr (i) + ','
         end;

     // Return without the last comma

     len := length (value);

     if len > 0 then
        value := copy (value, 1, len - 1);

     result := value;
end;

(* ------------------------------------------- *)

function tInterfaceFileIDEntry.GetInterfaceIndexList: tInterfaceIndexList;

var list  : tInterfaceIndexList;
    i     : integer;
    count : integer;
    test  : integer;

begin
     count := 0;
     setLength (list, count);

     for i := 0 to (sizeOf (InterfaceFlags) * 8) - 1 do
         begin
              test := (1 shl i);

              if (InterfaceFlags and test) = test then
                 begin
                      inc (count);
                      setLength (list, count);

                      list[count-1] := i;
                 end;
         end;

     result := list;
end;

(* ------------------------------------------- *)

procedure tInterfaceFileIDEntry.MarkInterfaceIndexActive (const Index: integer);

begin
     InterfaceFlags := InterfaceFlags or (1 shl Index);
end;

(* ------------------------------------------- *)

procedure copyInterfaceEntryList (const SourceList : tInterfaceFileIDList; var TargetList : tInterfaceFileIDList);

var len : integer;
    i   : integer;

begin
     len := length (SourceList);
     setLength (TargetList, len);

     for i := 1 to len do
         TargetList[i-1] := SourceList[i-1];
end;

(* ------------------------------------------- *)

function InterfaceListToStr (List : tInterfaceFileIDList) : string;

var len   : integer;
    i     : integer;
    value : string;
    entry : tInterfaceFileIDEntry;

begin
     value := '';
     len := length (List);

     // If only one file, return it, otherwise concatenate all files separated by semicolon and add the file indexes

     if len = 1 then
        value := list[0].toString
     else
         begin
              for i := 1 to len do
                  begin
                       if i < len then
                          value := value + IntToStr (list[i-1].FileIndex) + '/' + list[i-1].toString + ';'
                       else
                           value := value + IntToStr (list[i-1].FileIndex) + '/' + list[i-1].toString
                  end;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function StrToInterfaceList (Value : string) : tInterfaceFileIDList;

var count : integer;
    entry : string;
    sub   : string;
    list  : tInterfaceFileIDList;
    i,j   : integer;

begin
     count := numEntries (Value, ';');

     setLength (list, count);

     // Get the entries per file and set them

     for i := 1 to count do
         begin
              entry := getEntry (value, ';', i);

              list[i-1].initialize (entry);
         end;

     result := list;
end;

(* ------------------------------------------- *)

procedure mergeInterfaceEntryToList (var List : tInterfaceFileIDList; const Entry : tInterfaceFileIDEntry);

var found  : boolean;
    len    : integer;
    i      : integer;

begin
     // Search file id and merge interfaces - we do this the slow way by iterating through the list
     // because I have no patience to do something fancy, and this should not be called that often anyway
     // If it turns out to be too slow I'll improve it later.
     // TODO: improve search routine if possible

     len := length (List);

     i := 0;
     found := false;

     // find out if we have the file entry already

     while not found and (i < len) do
           begin
                found := List[i].FileIndex = Entry.FileIndex;

                if not found then
                   inc (i);
           end;

     if not found then
        begin
             inc (len);

             setLength (List, len);
             List[len-1].FileIndex := Entry.FileIndex;
             List[len-1].InterfaceFlags := Entry.InterfaceFlags;
        end
     else
         List[i].InterfaceFlags := List[i].InterfaceFlags or Entry.InterfaceFlags;
end;

(* ------------------------------------------- *)

function mergeInterfaceLists (List1, List2 : tInterfaceFileIDList) : tInterfaceFileIDList;

var list   : tInterfaceFileIDList;
    entry1 : tInterfaceFileIDEntry;
    entry2 : tInterfaceFileIDEntry;
    entry  : tInterfaceFileIDEntry;
    len1   : integer;
    len2   : integer;
    i,j    : integer;

begin
//   SiMain.EnterMethod ('InterfaceMappingTypes.mergeInterfaceLists');

     len1 := length (List1);
     len2 := length (List2);

     // Check if both lists only contain 1 file, and if both are for the same file, we can do this real quick

     if ((len1 and len2) = 1) and (List1[0].FileIndex = List2[0].FileIndex) then
        begin
             // If interfaces are the same, just return the list, otherwise merge interfaces

             if List1[0].InterfaceFlags = List2[0].InterfaceFlags then
                list := List1
             else
                 begin
                      entry1 := List1[0];
                      entry2 := List2[0];

                      // Set all Interface flags to active where at least one of both entries has it

                      entry.FileIndex := entry1.FileIndex;
                      entry.InterfaceFlags := entry1.InterfaceFlags or entry2.InterfaceFlags;

                      // Return a list with a single entry

                      setLength (list, 1);
                      list[0] := entry;
                 end;
        end
     else
         begin
              for i := 1 to len2 do
                  begin
                       entry2 := List2[i-1];

                       mergeInterfaceEntryToList (List1, entry2);
                  end;

              // Set the expanded list as result

              list := list1;
         end;

     result := list;

//   SiMain.LeaveMethod ('InterfaceMappingTypes.mergeInterfaceLists');
end;

(* ------------------------------------------- *)

function FileIndexExistsInList (const List : tInterfaceFileIDList; const FileIndex : integer) : boolean;

var count      : integer;
    isRelevant : boolean;
    i          : integer;

begin
     SiMain.EnterMethod ('InterfaceMappingTypes.FileIndexExistsInList');

     count := length (List);

     SiMain.LogDebug ('InterfaceFileIDList has ' + count.ToString + ' entries');

     if count > 0 then
        begin
             isRelevant := false;

             i := 0;

             while not isRelevant and (i < count) do
                   begin
                        inc (i);

                        if List[i-1].FileIndex = FileIndex then
                           begin
                                isRelevant := true;

                                SiMain.LogDebug ('Match found at entry #' + i.ToString);
                           end;
                   end;
{
             for i := 1 to count do
                 begin
                      if List[i-1].FileIndex = FileIndex then
                         isRelevant := true;
                 end;
}
        end
     else
         isRelevant := true;

     result := isRelevant;

     SiMain.LeaveMethod ('InterfaceMappingTypes.FileIndexExistsInList');
end;

(* ------------------------------------------- *)

function InterfaceListsAreIdentical (List1, List2 : tInterfaceFileIDList) : boolean;

var len1  : integer;
    len2  : integer;
    i     : integer;

begin
     len1 := length (List1);
     len2 := length (List2);

     // Check if both lists only contain 1 file, and if both are for the same file, we can do this real quick

     if ((len1 and len2) = 1) then
        result := (List1[0].FileIndex = List2[0].FileIndex)
     else
         if len1 <> len2 then
            result := false
         else
             begin
                  result := True;

                  i := 0;

                  while (i < len1) and result do
                        begin
                             inc (i);

                             if not FileIndexExistsInList (List2, List1[i-1].FileIndex) then
                                result := false;
                        end;
{
                  for i := 1 to len1 do
                      if not FileIndexExistsInList (List2, List1[i-1].FileIndex) then
                         result := False;
}
             end;
end;

(* ------------------------------------------- *)

end.
