{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_filter;

interface

uses Generics.Collections,
     base_structures,
     class_rawframe,
     definition_filter_structures,
     class_frameprocessor_baseclass,
     class_parsemaster;


type
     tFrameProcessorFilter = class (tFrameProcessor)

                                   constructor create (const Name : string; const AcceptAllOnEmptyList : boolean = true); overload;
                                   /// <summary>
                                   ///   Creates a filtering frameprocessor.
                                   /// </summary>
                                   /// <param name="Name">
                                   ///   Name of the processor, in case there
                                   ///   are multiple (this is mostly for
                                   ///   debugging purposes)
                                   /// </param>
                                   /// <param name="FilterList">
                                   ///   List of filters against which frames
                                   ///   are matched. This list is copied to a
                                   ///   freshly created internal list class
                                   /// </param>
                                   /// <param name="AcceptAllOnEmptyList">
                                   ///   if true, all frames match if the
                                   ///   filterlist is empty, otherwise none
                                   ///   match
                                   /// </param>
                                   constructor create (const Name : string; FilterList : tFilterSet; const AcceptAllOnEmptyList : boolean = true); overload;

                                   private
                                          priv_filterlist         : tFilterSet;
                                          priv_tuplelist          : tDictionary<string, tFrameFilter>;
                                          priv_nontuples          : tList<tFrameFilter>;

                                          priv_OnlyTupleFilters   : boolean;
                                          priv_autolimitparser    : boolean;
                                          priv_emptylistacceptall : boolean;
                                          priv_parsemode          : tParseProcessMode;

                                          priv_statistics         : tFilterStatistics;

                                          {$REGION 'Documentation'}
                                          /// <summary>
                                          ///   This routine determines the
                                          ///   parse mode used for running the
                                          ///   parsemaster on each frame,
                                          ///   depending on what filters are
                                          ///   defined. E.g. if there is only
                                          ///   an ethernet filter we don't
                                          ///   need to parse anything above
                                          ///   layer2
                                          /// </summary>
                                          {$ENDREGION}
                                          procedure determineParseMode;

                                          procedure setFilterlist (FilterList : tFilterSet);

                                          function FrameMatchesMACFilter (ParsedFrame : tParseMaster; Filter : tMACAddressFilter) : boolean;
                                          function FrameMatchesIPv4Address (ParsedFrame : tParseMaster; Filter : tIPv4AddressFilter) : boolean;
                                          function FrameMatchesIPv4Subnet (ParsedFrame : tParseMaster; Filter : tIPv4AddressFilter) : boolean;
                                          function FrameMatchesIPv4Filter (ParsedFrame : tParseMaster; Filter : tIPv4AddressFilter) : boolean;
                                          function FrameMatchesIPv6Address (ParsedFrame : tParseMaster; Filter : tIPv6AddressFilter) : boolean;
                                          function FrameMatchesIPv6Subnet (ParsedFrame : tParseMaster; Filter : tIPv6AddressFilter) : boolean;
                                          function FrameMatchesIPv6Filter (ParsedFrame : tParseMaster; Filter : tIPv6AddressFilter) : boolean;
                                          function FrameMatchesICMPFilter (ParsedFrame : tParseMaster; Filter : tICMPFilter) : boolean;
                                          function FrameMatchesPortOrTCPFilter (ParsedFrame : tParseMaster; PortFilter : tPortFilter; TCPFilter : tTCPFilter) : boolean;

                                          function FrameMatchesFilter (ParsedFrame : tParseMaster; Filter : tFrameFilter) : boolean;
                                          function FullFilterMatchExists (Parser : tParseMaster; var Frame : tRawFrame) : boolean;
                                          function TupleFilterMatchExists (Parser : tParseMaster; var Frame : tRawFrame) : boolean;
                                          function FilterMatchExists (var Frame : tRawFrame) : boolean;

                                   public
                                         property FilterList : tFilterSet read priv_filterlist write setFilterlist;
                                         property AutoLimitParser : boolean read priv_autolimitparser write priv_autolimitparser;
                                         property AcceptAllOnEmptyFilterlist : boolean read priv_emptylistacceptall write priv_emptylistacceptall;
                                         property Statistics : tFilterStatistics read priv_statistics;

                                         {$REGION 'Documentation'}
                                         /// <summary>
                                         ///   Checks if the passed frame
                                         ///   matches any of the filters in
                                         ///   the internal filter list. If
                                         ///   so the frame object is
                                         ///   returned as result, otherwise
                                         ///   it is discarded and set to nil
                                         /// </summary>
                                         /// <param name="Frame">
                                         ///   Frame to be checked against
                                         ///   the filter list
                                         /// </param>
                                         {$ENDREGION}
                                         function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                                         procedure AddFilter (Filter : tFrameFilter);
                                         function FilterCount : integer;

                                         destructor Destroy; override;
                             end;

implementation

uses SiAuto,
     SysUtils,
     Math,
     definition_ip_addresses,
     definition_tcp_structures,
     definition_5tuple,
     common_parse_helpers,
     class_parser_baseclass,
     class_parser_ethernet,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parser_icmpv4,
     class_parser_icmpv6,
     class_parser_tcp,
     class_parser_udp;

(* ------------------------------------------- *)
{ tFrameProcessorFilter }
(* ------------------------------------------- *)

constructor tFrameProcessorFilter.create (const Name : string; const AcceptAllOnEmptyList : boolean = true);

begin
     inherited create (Name);

     priv_parsemode := ppmParseAll;
     priv_autolimitparser := true;
     priv_emptylistacceptall := priv_emptylistacceptall;

     priv_tuplelist := tDictionary<string, tFrameFilter>.create;
     priv_nontuples := tList<tFrameFilter>.create;

     priv_filterlist := tFilterSet.create;
     priv_OnlyTupleFilters := false;

     priv_statistics.initialize;
end;

(* ------------------------------------------- *)

constructor tFrameProcessorFilter.create (const Name: string; FilterList: tFilterSet; const AcceptAllOnEmptyList : boolean = true);
{
var filter : tFrameFilter;
    tuple  : t5Tuple;
    len    : integer;
    key    : string;
    i      : Integer;
}
begin
     inherited create (Name);

     priv_parsemode := ppmParseAll;
     priv_autolimitparser := true;
     priv_emptylistacceptall := priv_emptylistacceptall;

     priv_tuplelist := tDictionary<string, tFrameFilter>.create;
     priv_nontuples := tList<tFrameFilter>.create;

     priv_statistics.initialize;

     priv_filterlist := tFilterSet.create;
     setFilterList (filterlist);
{
     priv_filterlist.AddFilterSet (Filterlist);

     priv_OnlyTupleFilters := true;

     filter := priv_filterlist.getFirst;

     // Build fast filters for tuple filtering

     while not filter.isUnspecified do
         begin
              tuple := filter.as5Tuple;

              if tuple.isUndefined then
                 begin
                      // As soon as we have at least one filter that can't be matched
                      // via tuple, set the TupleFilterMode to false

                      priv_OnlyTupleFilters := false;

                      // Add the filter to the "slow" filter list

                      priv_nontuples.Add (filter);
                 end
              else
                  begin
                       key := tuple.DictionaryKey;

                       if not priv_tuplelist.ContainsKey (key) then
                          priv_tuplelist.add (key, filter);
                  end;

              filter := priv_filterlist.getNext;
         end;
}
     determineParseMode;
end;

(* ------------------------------------------- *)

destructor tFrameProcessorFilter.Destroy;

begin
     priv_filterlist.Free;
     priv_tuplelist.Free;
     priv_nontuples.Free;

     inherited;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorFilter.setFilterlist (FilterList : tFilterSet);

var filter : tFrameFilter;
    tuple  : t5Tuple;
    len    : integer;
    key    : string;
    i      : Integer;

begin
     priv_filterlist.Clear;

     priv_filterlist.AddFilterSet (Filterlist);

     priv_OnlyTupleFilters := true;

     filter := priv_filterlist.getFirst;

     // Build fast filters for tuple filtering

     while not filter.isUnspecified do
         begin
              tuple := filter.as5Tuple;

              if tuple.isUndefined then
                 begin
                      // As soon as we have at least one filter that can't be matched
                      // via tuple, set the TupleFilterMode to false

                      priv_OnlyTupleFilters := false;

                      // Add the filter to the "slow" filter list

                      priv_nontuples.Add (filter);
                 end
              else
                  begin
                       key := tuple.DictionaryKey;

                       if not priv_tuplelist.ContainsKey (key) then
                          priv_tuplelist.add (key, filter);
                  end;

              filter := priv_filterlist.getNext;
         end;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorFilter.determineParseMode;

var filter : tFrameFilter;
    layer  : integer;
    len    : integer;
    i      : integer;

begin
     // if auto limiting is enabled we check how far the filter goes (but that will exclude some
     // special occurences like ICMP quotes containing IP layers etc.)

     if priv_autolimitparser then
        begin
             layer := 2;

             filter := priv_filterlist.getFirst;

             while not filter.isUnspecified do
                 begin
                      // If we find an active IP filter, set layer to at least 3

                      if (filter.IPv4Filter.AddressFilterType <> ftIgnore) or (filter.IPv6Filter.AddressFilterType <> ftIgnore) then
                         layer := max (layer, 3);

                      if (filter.ICMPFilter.ICMPFilterType <> IFTIgnore) then
                         layer := max (layer, 3);

                      if (filter.PortFilter.PortFilterType <> ftIgnore) or (filter.TCPFilter.TCPFilterType <> ftIgnore) then
                         layer := max (layer, 4);

                      filter := priv_filterlist.getNext;
                 end;

             // Set new parse mode

             case layer of
              2 : priv_parsemode := ppmUpToLayer2;
              3 : priv_parsemode := ppmUpToLayer3;
              4 : priv_parsemode := ppmUpToLayer4;
             else
                 priv_parsemode := ppmParseAll;
             end;
        end
     else
         priv_parsemode := ppmParseAll;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorFilter.AddFilter (Filter: tFrameFilter);

var len : integer;

begin
     priv_filterlist.AddFilter (Filter);

     determineParseMode;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FilterCount: integer;

begin
     result := priv_filterlist.count;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesMACFilter (ParsedFrame : tParseMaster; Filter: tMACAddressFilter): boolean;

var EthParser : tEthernetParser;
    boFound   : boolean;
    i         : Integer;

begin
     boFound := false;

     if (Filter.AddressFilterType = ftSingle) and (ParsedFrame.EtherLayerCount > 0) then
        begin
             if ParsedFrame.EtherLayerCount = 1 then
                begin
                     // Only one ethernet layer means we can match it quickly

                     case Filter.AddressScope of
                      fsAny         : boFound := (ParsedFrame.MACAddressPair.Source = Filter.SingleAddress) or (ParsedFrame.MACAddressPair.Destination = Filter.SingleAddress);
                      fsSource      : boFound := (ParsedFrame.MACAddressPair.Source = Filter.SourceAddress);
                      fsDestination : boFound := (ParsedFrame.MACAddressPair.Destination = Filter.DestinationAddress);
                      fsBoth        : boFound := (ParsedFrame.MACAddressPair.Source = Filter.SourceAddress) and (ParsedFrame.MACAddressPair.Destination = Filter.DestinationAddress) or
                                                 (ParsedFrame.MACAddressPair.Source = Filter.DestinationAddress) and (ParsedFrame.MACAddressPair.Destination = Filter.SourceAddress);
                     end;
                end
             else
                 begin
                      // Scan all parsers if they are ethernet parsers and try to find a match

                      for i := 1 to ParsedFrame.ParserCount do
                          if (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_ETHERNET) or (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_TRANS_ETHER_BRIDGE) then
                             begin
                                  EthParser := ParsedFrame.getParserByIndex(i) as tEthernetParser;

                                  case Filter.AddressScope of
                                   fsAny         : if (EthParser.SourceAddress = Filter.SingleAddress) or (EthParser.DestinationAddress = Filter.SingleAddress) then
                                                      boFound := true;

                                   fsSource      : if (EthParser.SourceAddress = Filter.SourceAddress) then
                                                      boFound := true;

                                   fsDestination : if (EthParser.DestinationAddress = Filter.DestinationAddress) then
                                                      boFound := true;

                                   fsBoth        : if ((EthParser.SourceAddress = Filter.SourceAddress) and (EthParser.DestinationAddress = Filter.DestinationAddress)) or
                                                      ((EthParser.SourceAddress = Filter.DestinationAddress) and (EthParser.DestinationAddress = Filter.SourceAddress)) then
                                                      boFound := true;
                                  end;
                             end;
                 end;
        end
     else
         boFound := true;

     result := boFound;

     SiMain.LogDebug ('Frame matches MAC filter: ' + result.ToString(true));
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesIPv4Address (ParsedFrame: tParseMaster; Filter: tIPv4AddressFilter): boolean;

var IPParser  : tIPv4Parser;
    boFound   : boolean;
    i         : Integer;

begin
     boFound := false;

     if Filter.AddressFilterType = ftSingle then
        begin
             if (ParsedFrame.IPLayerCount = 1) then
                begin
                     if ParsedFrame.LastIPLayerVersion = 4 then
                        begin
                             // Only one IPv4 layer means we can match it quickly

                             case Filter.AddressScope of
                              fsAny         : boFound := (ParsedFrame.IPv4SocketPair.Source.IPAddress = Filter.SingleAddress) or (ParsedFrame.IPv4SocketPair.Destination.IPAddress = Filter.SingleAddress);
                              fsSource      : boFound := (ParsedFrame.IPv4SocketPair.Source.IPAddress = Filter.SourceAddress);
                              fsDestination : boFound := (ParsedFrame.IPv4SocketPair.Destination.IPAddress = Filter.DestinationAddress);
                              fsBoth        : boFound := (ParsedFrame.IPv4SocketPair.Source.IPAddress = Filter.SourceAddress) and (ParsedFrame.IPv4SocketPair.Destination.IPAddress = Filter.DestinationAddress) or
                                                         (ParsedFrame.IPv4SocketPair.Source.IPAddress = Filter.DestinationAddress) and (ParsedFrame.IPv4SocketPair.Destination.IPAddress = Filter.SourceAddress);
                             end;
                        end
                end
             else
                 begin
                      // Scan all parsers if they are IPv4 parsers and try to find a match

                      for i := 1 to ParsedFrame.ParserCount do
                          if (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_IPv4_ENCAPSULATION) or (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_IPv4) then
                             begin
                                  IPParser := ParsedFrame.getParserByIndex(i) as tIPv4Parser;

                                  case Filter.AddressScope of
                                   fsAny         : if (IPParser.Source = Filter.SingleAddress) or (IPParser.Destination = Filter.SingleAddress) then
                                                      boFound := true;

                                   fsSource      : if (IPParser.Source = Filter.SourceAddress) then
                                                      boFound := true;

                                   fsDestination : if (IPParser.Destination = Filter.DestinationAddress) then
                                                      boFound := true;

                                   fsBoth        : if ((IPParser.Source = Filter.SourceAddress) and (IPParser.Destination = Filter.DestinationAddress)) or
                                                      ((IPParser.Source = Filter.DestinationAddress) and (IPParser.Destination = Filter.SourceAddress)) then
                                                      boFound := true;
                                  end;
                             end;
                 end;
        end
     else
         boFound := true;

     result := boFound;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesIPv4Subnet (ParsedFrame: tParseMaster; Filter: tIPv4AddressFilter): boolean;

var IPParser  : tIPv4Parser;
    boFound   : boolean;
    i         : Integer;

begin
     boFound := false;

     if Filter.AddressFilterType = ftSingle then
        begin
             if (ParsedFrame.IPLayerCount = 1) then
                begin
                     if ParsedFrame.LastIPLayerVersion = 4 then
                        begin
                             // Only one IPv4 layer means we can match it quickly

                             case Filter.AddressScope of
                              fsAny         : boFound := (ParsedFrame.IPv4SocketPair.Source.IPAddress.BelongsToSubnet (Filter.SingleAddress)) or (ParsedFrame.IPv4SocketPair.Destination.IPAddress.BelongsToSubnet (Filter.SingleAddress));
                              fsSource      : boFound := (ParsedFrame.IPv4SocketPair.Source.IPAddress.BelongsToSubnet (Filter.SourceAddress));
                              fsDestination : boFound := (ParsedFrame.IPv4SocketPair.Destination.IPAddress.BelongsToSubnet (Filter.DestinationAddress));
                              fsBoth        : boFound := (ParsedFrame.IPv4SocketPair.Source.IPAddress.BelongsToSubnet (Filter.SourceAddress)) and (ParsedFrame.IPv4SocketPair.Destination.IPAddress.BelongsToSubnet (Filter.DestinationAddress)) or
                                                         (ParsedFrame.IPv4SocketPair.Source.IPAddress.BelongsToSubnet (Filter.DestinationAddress)) and (ParsedFrame.IPv4SocketPair.Destination.IPAddress.BelongsToSubnet (Filter.DestinationAddress));
                             end;
                        end
                end
             else
                 begin
                      // Scan all parsers if they are IPv4 parsers and try to find a match

                      for i := 1 to ParsedFrame.ParserCount do
                          if (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_IPv4_ENCAPSULATION) or (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_IPv4) then
                             begin
                                  IPParser := ParsedFrame.getParserByIndex(i) as tIPv4Parser;

                                  case Filter.AddressScope of
                                   fsAny         : if (IPParser.Source.BelongsToSubnet (Filter.SingleAddress)) or (IPParser.Destination.BelongsToSubnet (Filter.SingleAddress)) then
                                                      boFound := true;

                                   fsSource      : if (IPParser.Source.BelongsToSubnet (Filter.SourceAddress)) then
                                                      boFound := true;

                                   fsDestination : if (IPParser.Destination.BelongsToSubnet (Filter.DestinationAddress)) then
                                                      boFound := true;

                                   fsBoth        : if ((IPParser.Source.BelongsToSubnet (Filter.SourceAddress)) and (IPParser.Destination.BelongsToSubnet (Filter.DestinationAddress))) or
                                                      ((IPParser.Source.BelongsToSubnet (Filter.DestinationAddress)) and (IPParser.Destination.BelongsToSubnet (Filter.SourceAddress))) then
                                                      boFound := true;
                                  end;
                             end;
                 end;
        end
     else
         boFound := true;

     result := boFound;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesIPv4Filter (ParsedFrame : tParseMaster; Filter: tIPv4AddressFilter): boolean;

var value : boolean;

begin
     case filter.AddressScope of
      fsAny         : begin
                           if filter.SingleAddress.MaskBitCount = 32 then
                              value := FrameMatchesIPv4Address (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv4Subnet (ParsedFrame, Filter);
                      end;
      fsSource      : begin
                           if filter.SourceAddress.MaskBitCount = 32 then
                              value := FrameMatchesIPv4Address (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv4Subnet (ParsedFrame, Filter);
                      end;
      fsDestination : begin
                           if filter.DestinationAddress.MaskBitCount = 32 then
                              value := FrameMatchesIPv4Address (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv4Subnet (ParsedFrame, Filter);
                      end;
      fsBoth        : begin
                           if filter.SourceAddress.MaskBitCount = 32 then
                              value := FrameMatchesIPv4Address (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv4Subnet (ParsedFrame, Filter);

                           if value then
                              begin
                                   if filter.DestinationAddress.MaskBitCount = 32 then
                                      value := value and FrameMatchesIPv4Address (ParsedFrame, Filter)
                                   else
                                       value := value and FrameMatchesIPv4Subnet (ParsedFrame, Filter);
                              end;
                      end;
     end;

     result := value;

     SiMain.LogDebug ('Frame matches IPv4 filter: ' + result.ToString(true));
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesIPv6Address (ParsedFrame: tParseMaster; Filter: tIPv6AddressFilter): boolean;

var IPParser  : tIPv6Parser;
    boFound   : boolean;
    i         : Integer;

begin
     boFound := false;

     if Filter.AddressFilterType = ftSingle then
        begin
             if (ParsedFrame.IPLayerCount = 1) then
                begin
                     if ParsedFrame.LastIPLayerVersion = 6 then
                        begin
                             // Only one IPv6 layer means we can match it quickly

                             case Filter.AddressScope of
                              fsAny         : boFound := (ParsedFrame.IPv6SocketPair.Source.IPAddress = Filter.SingleAddress) or (ParsedFrame.IPv6SocketPair.Destination.IPAddress = Filter.SingleAddress);
                              fsSource      : boFound := (ParsedFrame.IPv6SocketPair.Source.IPAddress = Filter.SourceAddress);
                              fsDestination : boFound := (ParsedFrame.IPv6SocketPair.Destination.IPAddress = Filter.DestinationAddress);
                              fsBoth        : boFound := (ParsedFrame.IPv6SocketPair.Source.IPAddress = Filter.SourceAddress) and (ParsedFrame.IPv6SocketPair.Destination.IPAddress = Filter.DestinationAddress);
                             end;
                        end
                end
             else
                 begin
                      // Scan all parsers if they are ethernet parsers and try to find a match

                      for i := 1 to ParsedFrame.ParserCount do
                          if (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_TYPE_IPV6) or (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_IPv6) then
                             begin
                                  IPParser := ParsedFrame.getParserByIndex(i) as tIPv6Parser;

                                  case Filter.AddressScope of
                                   fsAny         : if (IPParser.Source = Filter.SingleAddress) or (IPParser.Destination = Filter.SingleAddress) then
                                                      boFound := true;

                                   fsSource      : if (IPParser.Source = Filter.SourceAddress) then
                                                      boFound := true;

                                   fsDestination : if (IPParser.Destination = Filter.DestinationAddress) then
                                                      boFound := true;

                                   fsBoth        : if ((IPParser.Source = Filter.SourceAddress) and (IPParser.Destination = Filter.DestinationAddress)) or
                                                      ((IPParser.Source = Filter.DestinationAddress) and (IPParser.Destination = Filter.SourceAddress)) then
                                                      boFound := true;
                                  end;
                             end;
                 end;
        end
     else
         boFound := true;

     result := boFound;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesIPv6Subnet (ParsedFrame: tParseMaster; Filter: tIPv6AddressFilter): boolean;

var IPParser  : tIPv6Parser;
    boFound   : boolean;
    i         : Integer;

begin
     boFound := false;

     if Filter.AddressFilterType = ftSingle then
        begin
             if (ParsedFrame.IPLayerCount = 1) then
                begin
                     if ParsedFrame.LastIPLayerVersion = 6 then
                        begin
                             // Only one IPv6 layer means we can match it quickly

                             case Filter.AddressScope of
                              fsAny         : boFound := (ParsedFrame.IPv6SocketPair.Source.IPAddress.BelongsToSubnet (Filter.SingleAddress)) or (ParsedFrame.IPv6SocketPair.Destination.IPAddress.BelongsToSubnet (Filter.SingleAddress));
                              fsSource      : boFound := (ParsedFrame.IPv6SocketPair.Source.IPAddress.BelongsToSubnet (Filter.SourceAddress));
                              fsDestination : boFound := (ParsedFrame.IPv6SocketPair.Destination.IPAddress.BelongsToSubnet (Filter.DestinationAddress));
                              fsBoth        : boFound := (ParsedFrame.IPv6SocketPair.Source.IPAddress.BelongsToSubnet (Filter.SourceAddress)) and (ParsedFrame.IPv6SocketPair.Destination.IPAddress.BelongsToSubnet (Filter.DestinationAddress));
                             end;
                        end
                end
             else
                 begin
                      // Scan all parsers if they are ethernet parsers and try to find a match

                      for i := 1 to ParsedFrame.ParserCount do
                          if (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_TYPE_IPV6) or (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_IPv6) then
                             begin
                                  IPParser := ParsedFrame.getParserByIndex(i) as tIPv6Parser;

                                  case Filter.AddressScope of
                                   fsAny         : if (IPParser.Source.BelongsToSubnet (Filter.SingleAddress)) or (IPParser.Destination.BelongsToSubnet (Filter.SingleAddress)) then
                                                      boFound := true;

                                   fsSource      : if (IPParser.Source.BelongsToSubnet (Filter.SourceAddress)) then
                                                      boFound := true;

                                   fsDestination : if (IPParser.Destination.BelongsToSubnet (Filter.DestinationAddress)) then
                                                      boFound := true;

                                   fsBoth        : if ((IPParser.Source.BelongsToSubnet (Filter.SourceAddress)) and (IPParser.Destination.BelongsToSubnet (Filter.DestinationAddress))) or
                                                      ((IPParser.Source.BelongsToSubnet (Filter.DestinationAddress)) and (IPParser.Destination.BelongsToSubnet (Filter.SourceAddress))) then
                                                      boFound := true;
                                  end;
                             end;
                 end;
        end
     else
         boFound := true;

     result := boFound;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesIPv6Filter (ParsedFrame : tParseMaster; Filter: tIPv6AddressFilter): boolean;

var value : boolean;

begin
     case filter.AddressScope of
      fsAny         : begin
                           if filter.SingleAddress.IsPrefixInformation then
                              value := FrameMatchesIPv6Subnet (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv6Address (ParsedFrame, Filter);
                      end;
      fsSource      : begin
                           if filter.SourceAddress.IsPrefixInformation then
                              value := FrameMatchesIPv6Subnet (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv6Address (ParsedFrame, Filter);
                      end;
      fsDestination : begin
                           if filter.DestinationAddress.IsPrefixInformation then
                              value := FrameMatchesIPv6Subnet (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv6Address (ParsedFrame, Filter);
                      end;
      fsBoth        : begin
                           if filter.SourceAddress.IsPrefixInformation then
                              value := FrameMatchesIPv6Subnet (ParsedFrame, Filter)
                           else
                               value := FrameMatchesIPv6Address (ParsedFrame, Filter);

                           if filter.DestinationAddress.IsPrefixInformation then
                              value := value and FrameMatchesIPv6Subnet (ParsedFrame, Filter)
                           else
                               value := value and FrameMatchesIPv6Address (ParsedFrame, Filter);
                      end;
     end;

     result := value;

     SiMain.LogDebug ('Frame matches IPv6 filter: ' + result.ToString(true));
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesICMPFilter (ParsedFrame: tParseMaster; Filter: tICMPFilter): boolean;

var ICMPv4Parser : tICMPv4Parser;
    ICMPv6Parser : tICMPv6Parser;
    bofound      : boolean;
    i            : integer;

begin
     boFound := false;

     if (ppICMPv4 in ParsedFrame.ParsedProtocols) or (ppICMPv6 in ParsedFrame.ParsedProtocols) then
        begin
             case Filter.ICMPFilterType of
              IFTReject : begin
                               // I know this doesn't make much sense since boFound is false already, but just to have all ducks in a row...

                               if (ppICMPv4 in ParsedFrame.ParsedProtocols) or (ppICMPv6 in ParsedFrame.ParsedProtocols) then
                                  boFound := false;
                          end;
              IFTAny    : begin
                               boFound := true;
                          end;
              IFTType   : begin
                               if (ppICMPv4 in ParsedFrame.ParsedProtocols) then
                                  begin
                                       i := 1;

                                       ICMPv4Parser := ParsedFrame.getParserByProtocol (ppICMPv4, i) as tICMPv4Parser;

                                       while not boFound and assigned (ICMPv4Parser) do
                                             begin
                                                  inc (i);
                                                  ICMPv4Parser := ParsedFrame.getParserByProtocol (ppICMPv4, i) as tICMPv4Parser;

                                                  boFound := ICMPv4Parser.ICMPType = Filter.ICMPType;
                                             end;
                                  end;

                               if not boFound and (ppICMPv6 in ParsedFrame.ParsedProtocols) then
                                  begin
                                       i := 1;

                                       ICMPv6Parser := ParsedFrame.getParserByProtocol (ppICMPv6, i) as tICMPv6Parser;

                                       while not boFound and assigned (ICMPv6Parser) do
                                             begin
                                                  inc (i);
                                                  ICMPv6Parser := ParsedFrame.getParserByProtocol (ppICMPv6, i) as tICMPv6Parser;

                                                  boFound := ICMPv6Parser.ICMPType = Filter.ICMPType;
                                             end;
                                  end;
                          end;
             end;
        end
     else
         begin
              // this frame isn't ICMP, so unless reject ICMP it's not a match

              if (Filter.ICMPFilterType = IFTIgnore) or (Filter.ICMPFilterType = IFTReject) then
                 boFound := true
              else
                  boFound := false;
         end;

     result := boFound;

     SiMain.LogDebug ('Frame matches ICMP filter: ' + result.ToString(true));
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesPortOrTCPFilter (ParsedFrame : tParseMaster; PortFilter : tPortFilter; TCPFilter : tTCPFilter): boolean;

var TCPParser  : tTCPParser;
    UDPParser  : tUDPParser;
    ParserType : tHeaderType;
    boFound    : boolean;
    count      : integer;
    flag       : tTCPFlags;
    i          : Integer;

begin
     boFound := false;

     if (PortFilter.PortFilterType = ftSingle) or (TCPFilter.TCPFilterType = ftSingle) then
        begin
             i := 0;
             count := ParsedFrame.ParserCount;

             while not boFound and (i < count) do
                   begin
                        inc (i);
                        ParserType := ParsedFrame.getParserTypeByIndex (i);

                        // Check UDP stuff, which means ports

                        if (ParserType = NEXT_HEADER_UDP) and ((PortFilter.Layer4Protocol = l4TCPUDP) or (PortFilter.Layer4Protocol = l4UDP)) then
                           begin
                                UDPParser := ParsedFrame.getParserByIndex(i) as tUDPParser;

                                case PortFilter.PortScope of
                                 fsAny         : if (UDPParser.SourcePort = PortFilter.SinglePort) or (UDPParser.DestinationPort = PortFilter.SinglePort) then
                                                    boFound := true;
                                 fsSource      : if UDPParser.SourcePort = PortFilter.SourcePort then
                                                    boFound := true;
                                 fsDestination : if UDPParser.DestinationPort = PortFilter.DestinationPort then
                                                    boFound := true;
                                 fsBoth        : if ((UDPParser.SourcePort = PortFilter.SourcePort) and (UDPParser.DestinationPort = PortFilter.DestinationPort)) or
                                                    ((UDPParser.SourcePort = PortFilter.DestinationPort) and (UDPParser.DestinationPort = PortFilter.SourcePort)) then
                                                    boFound := true;
                                end;
                           end;

                        // Check TCP stuff, which means ports and specific TCP stuff like flags

                        if (not boFound) and (ParserType = NEXT_HEADER_TCP) and ((PortFilter.Layer4Protocol = l4TCPUDP) or (PortFilter.Layer4Protocol = l4TCP)) then
                           begin
                                TCPParser := ParsedFrame.getParserByIndex(i) as tTCPParser;

                                case PortFilter.PortScope of
                                 fsAny         : if (TCPParser.SourcePort = PortFilter.SinglePort) or (TCPParser.DestinationPort = PortFilter.SinglePort) then
                                                    boFound := true;
                                 fsSource      : if TCPParser.SourcePort = PortFilter.SourcePort then
                                                    boFound := true;
                                 fsDestination : if TCPParser.DestinationPort = PortFilter.DestinationPort then
                                                    boFound := true;
                                 fsBoth        : if ((TCPParser.SourcePort = PortFilter.SourcePort) and (TCPParser.DestinationPort = PortFilter.DestinationPort)) or
                                                    ((TCPParser.SourcePort = PortFilter.DestinationPort) and (TCPParser.DestinationPort = PortFilter.SourcePort)) then
                                                    boFound := true;
                                end;

                                // Check TCP stuff

                                if not boFound then
                                   begin
                                        for flag in TCPParser.Flags do
                                            if flag in TCPFilter.Flags then
                                               boFound := true;
                                   end;

                                // TODO: TCP Handshake Filter, needs TCP expert first
                           end;

                        // ICMP filtering is pretty dumb right now

                        if not (not boFound) and (ParsedFrame.getParserTypeByIndex (i) = NEXT_HEADER_ICMP) and ((PortFilter.Layer4Protocol = l4Any) or (PortFilter.Layer4Protocol = l4ICMP)) then
                           boFound := true;
                 end
        end
     else
         boFound := true;

     result := boFound;

     SiMain.LogDebug ('Frame matches TCP or Port filter: ' + result.ToString(true));
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FrameMatchesFilter (ParsedFrame : tParseMaster; Filter : tFrameFilter): boolean;

begin
     SiMain.EnterMethod ('tFrameProcessorFilter.FrameMatchesFilter');

     result := false;

     result := FrameMatchesMACFilter (ParsedFrame, Filter.MACFilter) and
               FrameMatchesIPv4Filter (ParsedFrame, Filter.IPv4Filter) and
               FrameMatchesIPv6Filter (ParsedFrame, Filter.IPv6Filter) and
               FrameMatchesICMPFilter (ParsedFrame, Filter.ICMPFilter) and
               FrameMatchesPortOrTCPFilter (ParsedFrame, Filter.PortFilter, filter.TCPFilter);

     SiMain.LeaveMethod ('tFrameProcessorFilter.FrameMatchesFilter');
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.TupleFilterMatchExists (Parser : tParseMaster; var Frame: tRawFrame): boolean;

var Filter  : tFrameFilter;
    help    : tSocketLayerEntry;
    tuple   : t5Tuple;
    boMatch : boolean;
    key     : string;
    count   : integer;
    i       : integer;

begin
     boMatch := False;

     // Find out how many tuples we have, and check them

     count := length (Parser.IPLayers);
     i := 0;

     while (i < count) and not boMatch do
           begin
                help := Parser.IPLayers[i];

                case Parser.IPLayers[i].IPVersion of
                 4 : tuple.initialize (help.IPv4SocketPair, help.Layer4NextProtocol);
                 6 : tuple.initialize (help.IPv6SocketPair, help.Layer4NextProtocol);
                end;

                // Check if we have that tuple in our filter dictionary

                key := tuple.DictionaryKey;

                if priv_tuplelist.ContainsKey (key) then
                   begin
                        filter := priv_tuplelist.items[key];

                        boMatch := True;
                        Frame.FilterLabel := Filter.FilterLabel;
                   end
                else
                    inc (i);
           end;

     result := boMatch;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FullFilterMatchExists (Parser : tParseMaster; var Frame: tRawFrame): boolean;

var Filter      : tFrameFilter;
    boMatch     : boolean;
    i           : integer;

begin
     boMatch := false;

     if priv_nontuples.Count > 0 then
        begin
             i := 0;

             repeat
                   filter := priv_nontuples[i];

                   if FrameMatchesFilter (Parser, Filter) then
                      begin
                           boMatch := true;

                           Frame.FilterLabel := Filter.FilterLabel;
                      end
                   else
                       inc (i);

             until boMatch or (i >= priv_nontuples.Count);
        end;

     result := boMatch;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.FilterMatchExists (var Frame: tRawFrame): boolean;

var match  : boolean;

begin
     inc (priv_statistics.TotalCompares);

     // Create parser if not already existing and handed over from preceding processor

     if not assigned (prot_ParseMaster) then
        prot_ParseMaster := tParseMaster.create (Frame, priv_parsemode);

     match := TupleFilterMatchExists (prot_ParseMaster, Frame);

     if not match then
        begin
             // Check for full filter match if we are not sure that we only need to match tuples

             if not priv_OnlyTupleFilters then
                match := FullFilterMatchExists (prot_ParseMaster, Frame);
        end
     else
         inc (priv_statistics.TupleHits);

     if match then
        inc (priv_statistics.TotalHits)
     else
         inc (priv_statistics.TotalMisses);

     result := match;
end;

(* ------------------------------------------- *)

function tFrameProcessorFilter.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var len : integer;

begin
//   SiMain.EnterMethod ('tFrameProcessorFilter.getProcessedFrame');

     // if there is no filter at all, the accept all parameter comes into play

     if priv_filterlist.count = 0 then
        begin
             if priv_emptylistacceptall then
                begin
                     SiMain.LogDebug ('Frame #' + Frame.FileFrameID.ToString + ' passed filter processor because of emtpy list -> accept all');

                     // Frame is an automatch, so we pass it back

                     inc (prot_framespassed);

                     result := Frame;
                end
             else
                 begin
                      SiMain.LogDebug ('Frame #' + Frame.FileFrameID.ToString + ' discarded by filter processor because of emtpy list -> discard all');

                      // Frame is filtered away, so we discard it

                      inc (prot_framesfiltered);

                      DiscardFrame (Frame);
                      result := nil;
                 end;
        end
     else
         begin
              // We have filters, so let's check if at least one matches

              if FilterMatchExists (Frame) then
                 begin
                      SiMain.LogDebug ('Frame #' + Frame.FileFrameID.ToString + ' passed filter processor by filter match');

                      // Frame is a match, so we pass it back

                      inc (prot_framespassed);

                      result := Frame;
                 end
              else
                  begin
                       SiMain.LogDebug ('Frame #' + Frame.FileFrameID.ToString + ' discarded by filter processor with no filter match');

                       // Frame is filtered away, so we discard it

                       inc (prot_framesfiltered);
                       DiscardFrame (Frame);

                       result := nil;
                  end;
         end;

//   SiMain.LeaveMethod ('tFrameProcessorFilter.getProcessedFrame');
end;

(* ------------------------------------------- *)

end.

