{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_fixethertypes;

interface

uses base_structures,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorFixEthertypes = class (tFrameProcessor)

                                    private
                                           function getRawFrameFromFrameChunk (Frame : tRawFrame; PayLoad : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;

                                    public
                                          /// <summary>
                                          ///   Checks all ethertypes and fixes
                                          ///   them in a heuristic manner by
                                          ///   doing sanity checks of the
                                          ///   upper layers
                                          /// </summary>
                                          /// <param name="Frame">
                                          ///   Frame to be processed
                                          /// </param>
                                          function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                              end;

implementation

uses SiAuto,
     SysUtils,
     class_parser_baseclass,
     common_parse_helpers,
     base_checksums,
     class_parser_802dot1q,
     class_parser_ethernet,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parsemaster,
     class_assembly_ethernet;

(* ------------------------------------------- *)
{ tFrameProcessorFixEthertypes }
(* ------------------------------------------- *)

function tFrameProcessorFixEthertypes.getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs

     WireDetails.LinkType := Frame.LinkType;
     WireDetails.WireSize := Frame.WireSize - FrameLengthModifier;
     WireDetails.CaptureSize := Payload.Len;
     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     // Transfer Packet direction

//   WireDetails.PacketDirection := pdUnknown;

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := prot_framesprocessed;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tFrameProcessorFixEthertypes.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var NewFrame     : tRawFrame;
    EthParser    : tEthernetParser;
    EthAssembler : tEthernetAssembler;
    VLANParser   : t802dot1qParser;
    Payload      : tFrameChunk;
    EthChunk     : tFrameChunk;
    MergeChunk   : tFrameChunk;
    isEthernet   : boolean;
    TPID         : word;
    NewEthertype : word;

begin
     // Get the Ethernet header first (if at all)

     if not assigned (prot_ParseMaster) then
        prot_ParseMaster := tParseMaster.create(Frame, ppmParseAll);

     // This only makes sense if we have at least an Ethernet header

     isEthernet := (ppEthernet in prot_Parsemaster.ParsedProtocols) and ((prot_ParseMaster.getParserTypeByIndex (1) = NEXT_HEADER_ETHERNET) or (prot_ParseMaster.getParserTypeByIndex (1) = NEXT_HEADER_TRANS_ETHER_BRIDGE));

     // Check if parsing did not return anything else except maybe VLAN tags

     if isEthernet and (prot_Parsemaster.ParsedProtocols - [ppEthernet,ppVLANTag] = []) then
        begin
             NewEthertype := 0;

             // now we know we only have Ethernet and maybe VLAN tag(s)

             EthParser := prot_ParseMaster.getParserByIndex (1) as tEthernetParser;

             Payload := prot_ParseMaster.RemainingPayloadCopy;

             if EthParser.EtherTypeLength = NEXT_HEADER_8021_Q then
                begin
                     VLANParser := prot_ParseMaster.getParserByIndex (2) as t802dot1qParser;

                     // TODO: fix VLAN tag TPIDs


                end
             else
                 begin
                      // May have to fix the Ethertype, so let's try some find some characteristic bytes. First, we check if we have well known
                      // Ethertypes following a VLAN tag, which is simple, but we need at least 4 bytes to check (obviously)

                      if Payload.Len > 3 then
                         begin
                              TPID := Payload.getWord (2, true);

                              case TPID of
                               NEXT_HEADER_8021_Q,
                               NEXT_HEADER_802_QINQ_Outer,
                               NEXT_HEADER_802_QINQ_Outer_OLD1,
                               NEXT_HEADER_802_QINQ_Outer_OLD2,
                               NEXT_HEADER_802_QINQ_Outer_OLD3  : begin
                                                                       // Ethertype should be a 0x88a8 VLAN tag

                                                                       NewEthertype := NEXT_HEADER_802_QINQ_Outer;
                                                                  end;
                               NEXT_HEADER_SNAP,
                               NEXT_HEADER_ARP,
                               NEXT_HEADER_RARP,
                               NEXT_HEADER_MPLS_UNICAST,
                               NEXT_HEADER_MPLS_MULTICAST,
                               NEXT_HEADER_IPv4,
                               NEXT_HEADER_TYPE_IPV6            : begin
                                                                       // Ethertype should be a 0x8100 VLAN tag

                                                                       NewEthertype := NEXT_HEADER_8021_Q;
                                                                  end;
                              end;
                         end;
                 end;

             // If we have a better Ethertype, rebuild the frame

             if NewEthertype > 0 then
                begin
                     EthAssembler := tEthernetAssembler.create;
                     EthAssembler.CopyFromParser (EthParser);
                     EthAssembler.EtherTypeLength := NewEtherType;

                     // Build new Ethernet chunk

                     EthChunk := EthAssembler.GetAssembly;
                     EthAssembler.Free;

                     // Merge Ethernet and Payload

                     MergeChunk := mergeFrameChunks (EthChunk, Payload);

                     EthChunk.free;
{
                     if prot_ParseMaster.HasEthernetFCS then                                              for this we need to cut the last 4 bytes, which I don't do right now
                        MergeChunk.AttachCRC32 (not prot_ParseMaster.EthernetFSCCorrect);
}
                     // Create New Frame

                     NewFrame := getRawFrameFromFrameChunk (Frame, MergeChunk, Frame.FrameSize - MergeChunk.Len);

                     // Discard Chunk and old Frame

                     MergeChunk.free;

                     Frame.free;
                     Frame := nil;

                     // Set frame to modified to force discarding the parsemaster object

                     prot_framewasmodified := true;
                end;
        end
     else
         begin
              // Give up and pass the frame

              NewFrame := Frame;

         end;

     result := NewFrame;
end;

(* ------------------------------------------- *)

end.
