{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_mpls;

interface

uses base_structures,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorMPLS = class (tFrameProcessor)

                                    private
                                           function getRawFrameFromFrameChunk (Frame : tRawFrame; PayLoad : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;

                                    public
                                          /// <summary>
                                          ///   Removes all MPLS shims from the
                                          ///   frame if found, and moves the
                                          ///   tunneled content to the
                                          ///   beginning of the frame.
                                          /// </summary>
                                          /// <param name="Frame">
                                          ///   Frame to be processed
                                          /// </param>
                                          function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                              end;

implementation

uses SiAuto,
     SysUtils,
     class_parser_baseclass,
     common_parse_helpers,
     base_checksums,
     class_parser_MPLS,
     class_parser_ethernet,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parsemaster,
     class_assembly_ethernet;

(* ------------------------------------------- *)
{ tFrameProcessorMPLS }
(* ------------------------------------------- *)

function tFrameProcessorMPLS.getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs

     WireDetails.LinkType := Frame.LinkType;
     WireDetails.WireSize := Frame.WireSize - FrameLengthModifier;
     WireDetails.CaptureSize := Payload.Len;
     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     // Transfer Packet direction

//   WireDetails.PacketDirection := pdUnknown;

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := prot_framesprocessed;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tFrameProcessorMPLS.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var
//  ParseMaster  : tParseMaster;
    ParseEntry   : tParseEntry;
    NewFrame     : tRawFrame;
    HeaderType   : tHeaderType;
    EthParser    : tEthernetParser;
    EthAssembler : tEthernetAssembler;
    GenParser    : tFrameParser;
    FirstChunk   : tFrameChunk;
    SecondChunk  : tFrameChunk;
    MergeChunk   : tFrameChunk;
    found        : boolean;
    FirstShim    : integer;
    LastShim     : integer;
    FirstOffset  : integer;
    LastOffset   : integer;
    i            : integer;

begin
//   SiMain.EnterMethod ('tFrameProcessorMPLS.getProcessedFrame');

     if not assigned (prot_ParseMaster) then
        prot_ParseMaster := tParseMaster.create(Frame, ppmParseAll);

     // Find the MPLS Shims

     FirstShim := 0;
     found := false;

     repeat
           inc (FirstShim);
           HeaderType := prot_ParseMaster.getParserTypeByIndex (FirstShim);

           if (HeaderType = NEXT_HEADER_MPLS_UNICAST) or (HeaderType = NEXT_HEADER_MPLS_UNICAST) then
              found := true;

     until found or (FirstShim = prot_ParseMaster.ParserCount);

     if found and (FirstShim > 1) and (FirstShim < prot_ParseMaster.ParserCount) then
        begin
             FirstOffset := 0;
             LastOffset := 0;

             // Find last shim

             LastShim := FirstShim;

             while (prot_ParseMaster.getParserTypeByIndex (LastShim + 1) = NEXT_HEADER_MPLS_UNICAST) or (prot_ParseMaster.getParserTypeByIndex (LastShim + 1) = NEXT_HEADER_MPLS_UNICAST)  do
                   inc (LastShim);

             // Determine what follows the shim(s), because it determines what we have to cut away

             if LastShim < prot_ParseMaster.ParserCount then
                begin
                     GenParser := prot_ParseMaster.getParserByIndex (LastShim + 1);
                     LastOffset := GenParser.FrameOffset;

                     HeaderType := prot_ParseMaster.getParserTypeByIndex (LastShim + 1);

                     case HeaderType of
                      NEXT_HEADER_ETHERNET,
                      NEXT_HEADER_TRANS_ETHER_BRIDGE  : begin
                                                             // we have Ethernet over MPLS, so cut away the first Ethernet header

                                                             EthParser := prot_ParseMaster.getParserByIndex (LastShim + 1) as tEthernetParser;
                                                             FirstOffset := EthParser.FrameOffset;

                                                             MergeChunk := Frame.FrameDataAsChunk (FirstOffset);
                                                        end;
                      NEXT_HEADER_IPv4_ENCAPSULATION,
                      NEXT_HEADER_IPv4,
                      NEXT_HEADER_IPV6,
                      NEXT_HEADER_TYPE_IPV6           : begin
                                                             // Just cut after Ethernet

                                                             EthParser := prot_ParseMaster.getParserByIndex (FirstShim - 1) as tEthernetParser;
                                                             EthAssembler := tEthernetAssembler.create;
                                                             EthAssembler.CopyFromParser (EthParser);

                                                             // Update Ethertype

                                                             if (HeaderType = NEXT_HEADER_IPv4_ENCAPSULATION) or (HeaderType = NEXT_HEADER_IPv4) then
                                                                EthAssembler.EtherTypeLength := NEXT_HEADER_IPv4
                                                             else
                                                                 EthAssembler.EtherTypeLength := NEXT_HEADER_TYPE_IPV6;

                                                             // Build new Ethernet chunk

                                                             FirstChunk := EthAssembler.GetAssembly;
                                                             EthAssembler.Free;

                                                             // Get the part starting right after the MPLS Shims

                                                             SecondChunk := Frame.FrameDataAsChunk (LastOffset);

                                                             // Merge Ethernet and IP

                                                             MergeChunk := mergeFrameChunks (FirstChunk, SecondChunk);

                                                             FirstChunk.free;
                                                             SecondChunk.free;
                                                        end;
                     end;
                end;

                if prot_ParseMaster.HasEthernetFCS then
                   MergeChunk.AttachCRC32 (not prot_ParseMaster.EthernetFSCCorrect);

                // Create New Frame

                NewFrame := getRawFrameFromFrameChunk (Frame, MergeChunk, Frame.FrameSize - MergeChunk.Len);

                // Discard Chunk and old Frame

                MergeChunk.free;

                Frame.free;
                Frame := nil;

                // Set frame to modified to force discarding the parsemaster object

                prot_framewasmodified := true;
        end
     else
         // Give up and pass the frame

         NewFrame := Frame;

     // Free up the parsemaster

//   ParseMaster.free;

     result := NewFrame;

//   SiMain.LeaveMethod ('tFrameProcessorMPLS.getProcessedFrame');
end;

(* ------------------------------------------- *)

end.
