{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_dupechecker;

interface

uses SysUtils,
     Contnrs,
     base_structures,
     class_rawframe;

type tDupeChecker = class (tObject)

                          constructor create (CheckMethod : tDupeCheckMethod; MaxFrames, MaxTime : integer);
                          procedure free;

                          private
                                 FrameList_History     : array [0..maxInterfaces - 1] of tObjectList;  // Blast from the Past: Frames we use to compare new frames to for dupe checking
                                 iMaxHistorySize       : integer;
                                 iMaxTimeDelta         : integer;

                                 DupeCheckMode         : tDupeCheckMethod;
                          public
                                property DupeCheckMethod : tDupeCheckMethod read DupeCheckMode write DupeCheckMode;
                                property MaxHistorySize : integer read iMaxHistorySize write iMaxHistorySize;
                                property MaxTimeDelta : integer read iMaxTimeDelta write iMaxTimeDelta;

                                function FrameIsDuplicate (FrameToTest : tRawFrame) : boolean;

                    end;


implementation

(* ------------------------------------------- *)
{ tDupeChecker }
(* ------------------------------------------- *)

constructor tDupeChecker.create (CheckMethod : tDupeCheckMethod; MaxFrames, MaxTime : integer);

begin
     inherited create;

     DupeCheckMethod := CheckMethod;
     iMaxHistorySize := MaxFrames;
     iMaxTimeDelta := MaxTime;
end;

(* ------------------------------------------- *)

procedure tDupeChecker.free;

var i : integer;

begin
     for i := 0 to (maxInterfaces - 1) do
         begin
              if assigned (FrameList_History[i]) then
                 begin
                      FrameList_History[i].Clear;
                      FrameList_History[i].free;
                      FrameList_History[i] := nil;
                 end;
         end;

     inherited free;
end;

(* ------------------------------------------- *)

function tDupeChecker.FrameIsDuplicate (FrameToTest: tRawFrame): boolean;

var boKeepSearching  : boolean;
    i                : integer;
    HistoryFrame     : tRawFrame;
    iDeltaTime       : int64;
    currentInterface : longword;
    boDupeFound      : boolean;

begin
     boDupeFound := false;

     currentInterface := FrameToTest.InterfaceID;

     // Check if we need to create a new dupe history

     if not assigned (FrameList_History[currentInterface]) then
        FrameList_History[currentInterface] := tObjectList.Create (false);          // the list may not own the frames or it would destroy valid frames when we kill the list on free

     // Only test if we have at least one frame to compare the new frame to, and the new frame has more than 0 bytes

     if not ((FrameList_History[currentInterface].Count = 0) or (FrameToTest.FrameSize = 0)) then
        begin
             boKeepSearching := true;
             i := 0;

             // Compare against remaining frames in the history list

             while boKeepSearching and (i < FrameList_History[currentInterface].Count) do
                   begin
                        inc (i);

                        HistoryFrame := (FrameList_History[currentInterface][i-1] as tRawFrame);

                        // If the CRC32 matches the history frame, we consider it a Dupe or investigate further

                        if (HistoryFrame.CRC32 = FrameToTest.CRC32) then
                           begin
                                case DupeCheckMode of
                                 dcFrameCRC       : begin
                                                         boDupeFound := true;
                                                    end;
                                 dcFullFrame      : begin
                                                         // We need an additional channel check, and if it matches, a full frame byte compare

                                                         if FrameToTest.FrameDataIsEqualTo (HistoryFrame) then
                                                            boDupeFound := true;
                                                    end;
                                end;
                           end;

                        if boDupeFound then
                           begin
                                // Mark the search as complete

                                boKeepSearching := false;

                                // Remove the matching frame from the history as it gets replaced by the current new frame
                                // That way we can detect multiple dupes of the same frame without running out of the time
                                // boundary

                                FrameList_History[currentInterface].Delete(i);
                           end
                        else
                            begin
                                 // Now we need to find out if we need to continue or if an abort condition is met

                                 iDeltaTime := (FrameToTest.Ticks - HistoryFrame.Ticks);

                                 // Check if the Search Time Window is exceeded

                                 if iDeltaTime > iMaxTimeDelta then
                                    boKeepSearching := false;
                            end;
                   end;
        end;

     // Add the current frame to the list

     FrameList_History[currentInterface].Add (FrameToTest);

     // Check if we need to remove the oldest frame

     if (FrameList_History[currentInterface].Count > iMaxHistorySize) then
        FrameList_History[currentInterface].Delete(1);
end;

(* ------------------------------------------- *)

end.
