{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_interfacestatistics;

interface

uses Contnrs,
     base_structures;

type tStatisticsRecord = class (tObject)

                               constructor create (InterfaceID : tInterfaceID);
                               procedure free;

                               private
                                      priv_interfaceID         : tInterfaceID;
                                      priv_referencetime       : tTimingDetails;             // Time this statistics refers to
                                      priv_comment             : string;                     // Comment
                                      priv_starttime           : tTimingDetails;             // Time in which the capture started
                                      priv_endtime             : tTimingDetails;             // Time in which the capture ended
                                      priv_packetsreceived     : uint64;                     // Number of packets received from the physical interface starting from the beginning of the capture
                                      priv_packetsdropped      : uint64;                     // Number of packets dropped by the interface due to lack of resources starting from the beginning of the capture
                                      priv_filteraccepted      : uint64;                     // Number of packets accepted by filter starting from the beginning of the capture
                                      priv_osdropped           : uint64;                     // Number of packets dropped by the operating system starting from the beginning of the capture
                                      priv_usrdelivered        : uint64;                     // Number of packets delivered to the user starting from the beginning of the capture. The value contained in this field can be different from the value 'isb_filteraccept - isb_osdrop' because some packets could still lay in the OS buffers when the capture ended.

                               public
                                     property InterfaceIndex : tInterfaceID read priv_interfaceID write priv_interfaceID;
                                     property ReferenceTime : tTimingDetails read priv_referencetime write priv_referencetime;
                                     property Comment : string read priv_comment write priv_comment;
                                     property StartTime : tTimingDetails read priv_starttime write priv_starttime;
                                     property EndTime : tTimingDetails read priv_endtime write priv_endtime;
                                     property PacketsReceived : uint64 read priv_packetsreceived write priv_packetsreceived;
                                     property PacketsDropped : uint64 read priv_packetsdropped write priv_packetsdropped;
                                     property FilterAccepted : uint64 read priv_filteraccepted write priv_filteraccepted;
                                     property OSDropped : uint64 read priv_osdropped write priv_osdropped;
                                     property UserDelivered : uint64 read priv_usrdelivered write priv_usrdelivered;

                                     procedure copyTo (DestinationRecord : tStatisticsRecord);
                         end;

     tStatisticsRecordList  = class (tObject)

                                    constructor create (DiscardEntriesOnDestory : boolean);
                                    procedure free;

                                    private
                                           priv_statisticslist : tObjectList;

                                           function getItemCount : integer;

                                    protected

                                    public
                                          property Count : integer read getItemCount;

                                          procedure copyTo (DestinationRecordList : tStatisticsRecordList);
                                          procedure addEntry (Entry : tStatisticsRecord);
                                          function getEntry (index : integer) : tStatisticsRecord;
                                          procedure clear;
                              end;


implementation

(* ------------------------------------------- *)
{ tStatisticsRecord }
(* ------------------------------------------- *)

constructor tStatisticsRecord.create(InterfaceID: tInterfaceID);

begin
     inherited create;

     priv_interfaceID := InterfaceID;
end;

(* ------------------------------------------- *)

procedure tStatisticsRecord.free;

begin
     inherited free;
end;

(* ------------------------------------------- *)

procedure tStatisticsRecord.copyTo (DestinationRecord: tStatisticsRecord);

begin
     DestinationRecord.ReferenceTime := priv_referencetime;
     DestinationRecord.Comment := priv_comment;
     DestinationRecord.StartTime := priv_starttime;
     DestinationRecord.EndTime := priv_endtime;
     DestinationRecord.PacketsReceived := priv_packetsreceived;
     DestinationRecord.PacketsDropped := priv_packetsdropped;
     DestinationRecord.FilterAccepted := priv_filteraccepted;
     DestinationRecord.OSDropped := priv_osdropped;
     DestinationRecord.UserDelivered := priv_usrdelivered;
end;

(* ------------------------------------------- *)
{ tStatisticsRecordList }
(* ------------------------------------------- *)

constructor tStatisticsRecordList.create (DiscardEntriesOnDestory : boolean);

begin
     inherited create;

     priv_statisticslist := tObjectList.Create (DiscardEntriesOnDestory);
end;

(* ------------------------------------------- *)

procedure tStatisticsRecordList.free;

begin
     inherited free;
end;

(* ------------------------------------------- *)

procedure tStatisticsRecordList.addEntry(Entry: tStatisticsRecord);

begin
     priv_statisticslist.Add (Entry);
end;

(* ------------------------------------------- *)

procedure tStatisticsRecordList.clear;

begin
     priv_statisticslist.Clear;
end;

(* ------------------------------------------- *)

function tStatisticsRecordList.getEntry(index: integer): tStatisticsRecord;

begin
     result := priv_statisticslist.Items[index -1] as tStatisticsRecord;
end;

(* ------------------------------------------- *)

function tStatisticsRecordList.getItemCount: integer;

begin
     result := priv_statisticslist.Count;
end;

(* ------------------------------------------- *)

procedure tStatisticsRecordList.copyTo (DestinationRecordList: tStatisticsRecordList);

var entry    : tStatisticsRecord;
    newentry : tStatisticsRecord;
    i        : integer;

begin
     for i := 0 to priv_statisticslist.Count - 1 do
         begin
              entry := priv_statisticslist[i] as tStatisticsRecord;

              newentry := tStatisticsRecord.create (entry.InterfaceIndex);
              entry.copyTo(newentry);
         end;
end;

(* ------------------------------------------- *)

end.
