{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_loader_naicap;

interface

uses base_structures,
     class_rawframe,
     class_traceloader;

type tCAP_FrameTime = record
                            lTimeLow  : longword;                            // lower 32 bits of time stamp
                            lTimeHigh : longword;                            // upper 32 bits of time stamp
                      end;

     tTickRec = record
                      MicroSeconds      : int64;
                      NanoDeltaSeconds  : int64;
                end;


     tCAPLoader   = class (tTraceLoader)

                          private
                                 iMajorVersion      : byte;
                                 iMinorVersion      : byte;
                                 lwStartTime        : longword;
                                 i64StartTick       : int64;                 // Tick value of the first frame in the file
                                 iTotalFramesInFile : longint;               // Number of total frames in the current trace file
                                 iFileFormat        : byte;                  // CAP File Format Version Identifier
                                 StartTimeStamp     : int64;

                                 function CAPTimetoTicks (const FrameIndex : integer; ThisFrameTime : tCAP_FrameTime) : tTickRec;
                                 function getStartTicksFromHeader (const unixtime : longword) : int64;
                                 function CAPErrorFlagsToGenericFormat (ErrorFlag : integer) : tErrorFlagsSet;

                          protected
                                   function thisFileFormat : tCaptureFileType; override;
                                   function ParseFrameFromOffset (const FrameIndex : integer; var FileOffset: integer; var DataPointer: pByte): tRawFrame;  override;
                          public
                                procedure readTraceFileheader; override;
                                procedure readBatchFromDisk; override;
                                function readSingleFrameFromDisk (const FrameIndex : integer) : tRawFrame; override;
                                function readSingleFrameByOffset (const Offset : integer; FrameIndex : integer = -1) : tRawFrame; override;
                    end;


implementation

uses SysUtils,
     Math,
     DateUtils,
     class_interfacelist;


const CAP_FILE_UNSUPPORTED    = 0;
      CAP_FILE_NETXRAY_1_0    = 1;
      CAP_FILE_NETXRAY_1_1    = 2;
      CAP_FILE_NETXRAY_2_00x  = 3;

      // Ticks per Second Tables

      NUM_NETXRAY_TIMEUNITS   = 3;                                       // # of entries in the TpS array
      NUM_NETXRAY_TIMEUNITS_GIGPOD = 3;
      NUM_NETXRAY_TIMEUNITS_OTHERPOD = 3;
      NUM_NETXRAY_TIMEUNITS_OTHERPOD2 = 3;
      NUM_NETXRAY_TIMEUNITS_GIGPOD2 = 3;

      TpS : array [1..3] of double = (1e6, 1193000.0, 1193182.0);        // # of ticks that equal 1 second, in version 002.xxx files
                                                                         // other than Ethernet captures with a captype other than CAPTYPE_NDIS
                                                                         // the index into this array is hdr.timeunit

      TpS_gigpod : array [1..3] of double = (1e9, 0.0, 31250000.0);      // Table of time units for Ethernet captures with
                                                                         // captype ETH_CAPTYPE_GIGPOD. 0.0 means "unknown".

      TpS_gigpod2 : array [1..3] of double = (1e9, 0.0, 20000000.0);     // Table of time units for Ethernet captures with captype ETH_CAPTYPE_GIGPOD2

      TpS_otherpod : array [1..3] of double = (1e6, 0.0, 1250000.0);     // Table of time units for Ethernet captures with captype ETH_CAPTYPE_OTHERPOD

      TpS_otherpod2 : array [1..3] of double = (1e6, 0.0, 0.0);          // Table of time units for Ethernet captures with captype ETH_CAPTYPE_OTHERPOD2

      // CAP Error Absolutes

      CAP_Error_None           = 0;
      CAP_Error_BadCRC         = 1;
      CAP_Error_Jabber         = 2;
      CAP_Error_Runt           = 3;
      CAP_Error_Fragment       = 4;
      CAP_Error_Oversize       = 5;
      CAP_Error_Collision      = 6;
      CAP_Error_Alignment      = 7;
      CAP_Error_Unknown        = 8;


type // CAP File structures (NAI/NetGen Sniffer Pro Format)

     pCAP_FileHeader  = ^tCAP_FileHeader;
     tCAP_FileHeader  = packed record
                             sVersion      : array [0..7] of ansichar;   // Version Number
                             iStartTime    : longword;                   // UNIX time when capture started
                             iNumFrames    : longword;                   // number of frames in trace
                             iFileSize     : longword;                   // Size of the Capture File
                             iStartOffset  : longword;                   // Offset of the first frame in capture
                             iEndOffset    : longword;                   // Offset of the last frame in capture
                             unknown1      : array [0..2] of longword;   // unknown
                             iNetwork      : byte;                       // datalink type
                             iNetworkPlus  : byte;                       // is usually 0, in which case iNetwork byte is an NDIS network type value - 1.
                             unknown2      : array [0..1] of byte;       // XXX - is this the upper 3 bytes of the datalink type?
                             iTimeUnit     : shortint;                   // encodes length of a tick
                             unknown3      : array [0..2] of byte;       // XXX - is this the upper 3 bytes of the time units?
                             iTimeLow      : longword;                   // lower 32 bits of time stamp of capture start
                             iTimeHi       : longword;                   // upper 32 bits of time stamp of capture star
                             lwLinespeed   : longword;                   // speed of network, in bits/second
                             unknown4      : array [0..11] of byte;      // other stuff
                             iRealtick     : longword;                   // in version 2, units of the timestamp

                             unknown5      : array[0..3] of byte;
                             capType       : byte;                       // Capture Type
//                           unknown6      : array[0..2] of byte;
                             unknown6      : byte;
                             ChannelCount  : byte;                       // Number of Channels in a capture (FDX = 2)
                             unknown7      : byte;
                             unknown8      : array[0..3] of byte;
                             hdlc_captype  : byte;                       // WAN HDLC subsub_captype
                             unknown9      : array[0..2] of byte;
                             unknown10     : array[0..15] of byte;
                             unknown11     : array[0..13] of byte;
                             TimezoneHours : smallint;                   // timezone hours [at least for version 2.2..];
                                                                         //  positive values = west of UTC:
                                                                         //  negative values = east of UTC:
                                                                         //  e.g. +5 is American Eastern
                                                                         // [Does not appear to be adjusted for DST ]


//                           unknown5      : array [0..47] of byte;      // other unknown stuff
                       end;

     pCAP_V1Frameheader = ^tCAP_V1FrameHeader;
     tCAP_V1FrameHeader = packed record
                                rTime        : tCAP_FrameTime;
                                iFrameSize   : word;                 // frame length (= on the wire)
                                iCaptureSize : word;                 // captured bytes
                                unknown1     : array [0..15] of byte;
                          end;

     pCAP_V2Frameheader = ^tCAP_V2FrameHeader;
     tCAP_V2FrameHeader = packed record
                                rTime        : tCAP_FrameTime;
                                iFrameSize   : word;                     // frame length (= on the wire)
                                iCaptureSize : word;                     // captured bytes
                                unknown1     : array [0..7] of byte;
                                iErrorFlag   : byte;
                                unknown2     : array [0..1] of byte;
                                LocalCapture : byte;                     // if 1 the frame was transmitted from the card captured on
                                Channel      : byte;                     // for multichannel captures (FDX, Trunks)
                                BladeIndex   : byte;                     // Blades Number
                                unknown3     : array [0..1] of byte;
                                SlicedSize   : word;                     // Sometimes Sniffer writes more bytes to the file than the slice sice
                                                                         // was, probably for performance reasons. This value is the number of
                                                                         // bytes that were really captured from the write.
                                unknown4     : array [0..9] of byte;
                          end;


(* ------------------------------------------- *)
{ tCAPLoader }
(* ------------------------------------------- *)

function tCAPLoader.thisFileFormat : tCaptureFileType;

begin
     result := cfCAP;
end;

(* ------------------------------------------- *)

// This function converts the frame time found in .CAP files to
// the number of microseconds ("Tick") since midnight.

function tCAPLoader.CAPTimetoTicks (const FrameIndex : Integer; ThisFrameTime : tCAP_FrameTime) : tTickRec;

var t     : int64;
    m     : int64;
    n     : int64;                 // nanoseconds (not including microsecs)

begin
     try
        t := ThisFrameTime.lTimeLow + ThisFrameTime.lTimeHigh * 4294967296;
        m := trunc (t / dTraceTimeUnit * 1000000);

        // If the resolution is in nanoseconds, determine the delta value

        // TODO: handle dTraceTimeUnit 2e7

        if (dTraceTimeUnit = 1e9) then
           n := t - (m * 1000)
        else
        if (dTraceTimeUnit = 20000000.0) then
           n := trunc (t / dTraceTimeUnit * 1000000000) - (m * 1000)
        else
            n := 0;

        if n < 0 then
           n := 0;

        result.MicroSeconds := (i64StartTick + m);
        result.NanoDeltaSeconds := n;

     except
           ReportMessage (msMajor, 'Error converting NAICAP timestame', 'tCAPLoader.CAPTimetoTicks', FrameIndex);
     end;
end;

(* ------------------------------------------- *)

// This function calculates the base tick from the unix start time in the trace file header

function tCAPLoader.getStartTicksFromHeader (const unixtime : longword) : int64;

var t : int64;

begin
     // 86400 seconds are one day, so we take the rest
     // and get it to Microseconds

     t := (unixtime + (iUTCDelta * 3600)) mod 86400;
     result := t * 1000000;
end;

(* ------------------------------------------- *)

function tCAPLoader.CAPErrorFlagsToGenericFormat (ErrorFlag: integer): tErrorFlagsSet;

var value : tErrorFlagsSet;

begin
     value := [];

     case ErrorFlag of
      CAP_Error_BadCRC   : result := result + [efBadCRC];
      CAP_Error_Jabber   : result := result + [efBadPreamble];
      CAP_Error_Runt,
      CAP_Error_Fragment : result := result + [efPacketTooShort];
      CAP_Error_Oversize : result := result + [efPacketTooLong];
      CAP_Error_Collision : result := result + [efCollision];
      CAP_Error_Alignment : result := result + [efAlignment];
      CAP_Error_Unknown : result := result + [efUnknown];
     end;

     result := value;
end;

(* ------------------------------------------- *)

procedure tCAPLoader.readTraceFileheader;

var sHelp          : string;
    i              : integer;
    helpdate       : tDateTime;
    pHeader        : pCAP_Fileheader;
    InterfaceEntry : tCaptureInterface;
    BladeIndex     : integer;

begin
     iBatchFileOffset := 0;                  // This keeps track of where we are in the whole mapped file

     sHelp := '';
     pBatchPosition := pByte (pData);

     if pData = nil then
        raise EInOutError.Create('Error: memory file mapping failed with pData=NIL (tCAPLoader.readTraceFileheader)');

     // Check File Length

     if iFileSize < length (CAP_MagicString) then
        raise EInOutError.Create ('Error: Sniffer CAP file too short (tCAPLoader.readTraceFileheader).');

     // 1. Check if the Magic String has the correct length and matches

     for i := 0 to (length (CAP_Magicstring) - 1) do
         begin
              sHelp := sHelp + chr (pData^[i]);

              inc (iBatchFileOffset);
              inc (pBatchPosition);
         end;

     if sHelp <> CAP_MagicString then
        begin
             raise EInOutError.Create('Error: Not a Sniffer CAP file (tCAPLoader.readTraceFileheader).');

             exit;
        end;

     // Check File Length

     if iFileSize < (iBatchFileOffset + sizeOf (tCAP_FileHeader)) then
        begin
             raise EInOutError.Create ('Error: Sniffer CAP file too short (tCAPLoader.readTraceFileheader).');

             exit;
        end;

    // 2. Now we read the CAP file header

    pHeader := pCAP_Fileheader (pBatchPosition);

    inc (iBatchFileOffset, sizeOf (tCAP_FileHeader));
    inc (pBatchPosition, sizeOf (tCAP_FileHeader));

    // 3. Do some checks/calculations on the header data

    // UTC Delta (if the one we have is 0 we take the one specified in the file header)

    if iUTCDelta = 0 then
       begin
            // Extract the start date of the trace to determine what Bias we have (daylight saving time or not)

            HelpDate := UnixToDateTime (pHeader^.iStartTime);

//          iUTCDelta := (pHeader^.TimezoneHours * - 1);       // disbled 17.04.2014 because in summer time the result for timezone berlin was +3
            iUTCDelta := iUTCDelta + (TimeZoneHandler.GetBiasForDate (HelpDate) DIV 60 * -1);
       end;

    // - CAP File Version  ('0', '0', '2', '.', '0', '0', '2', #0)

    try
       iMajorVersion := StrToInt (pHeader^.sVersion[2]);
       iMinorVersion := StrToInt (pHeader^.sVersion[6]);
    except
          iFileFormat := CAP_FILE_UNSUPPORTED;
    end;

    case iMajorVersion of
     1 : if iMinorVersion = 0 then
            iFileFormat := CAP_FILE_NETXRAY_1_0
         else
             iFileFormat := CAP_FILE_NETXRAY_1_1;
     2 : iFileFormat := CAP_FILE_NETXRAY_2_00x;
    end;

    if (iMajorVersion < 2) or (iFileFormat = CAP_FILE_UNSUPPORTED) then
       begin
             raise EInOutError.Create ('Error: Sniffer CAP file version not supported (tCAPLoader.readTraceFileheader).');

             exit;
       end;

    lwFramesInFile := pHeader^.iNumFrames;

    // Determine Network Topology

    case pHeader^.iNetworkPlus of
     0 : begin
              case pHeader^.iNetwork of
                0 : prot_LinkType := LINKTYPE_ETHERNET;
                1 : prot_LinkType := LINKTYPE_TOKEN_RING;
                5 : prot_LinkType := LINKTYPE_PPP_ETHER;                      // WAN (PPP), but shaped like Ethernet
               11 : prot_LinkType := LINKTYPE_IEEE802_11;
              else
                  begin
                       raise EInOutError.Create ('Error: Unknown network type in Sniffer CAP file (tCAPLoader.readTraceFileHeader).');

                       exit;
                  end;
              end
         end;
     2 : begin
              case pHeader^.iNetwork of
                0 : prot_LinkType := LINKTYPE_NULL;
                1 : prot_LinkType := LINKTYPE_ETHERNET;
                2 : prot_LinkType := LINKTYPE_TOKEN_RING;
                4 : prot_LinkType := LINKTYPE_PPP_ETHER;                      // WAN (PPP), but shaped like Ethernet
               10 : prot_LinkType := LINKTYPE_IEEE802_11;
              else
                  begin
                       raise EInOutError.Create ('Error: Unknown network type in Sniffer CAP file (tCAPLoader.readTraceFileHeader).');

                       exit;
                  end;
              end
         end
     else
         begin
              raise EInOutError.Create ('Error: Unknown network enhancement type in Sniffer CAP file (tCAPLoader.readTraceFileHeader).');

              exit;
         end;
    end;

    // Assign the capture module, which tells us on which hardware the trace was captured

    iCaptureModule := pHeader^.capType;

    // Check if this is a multichannel capture (a.k.a. full duplex, trunk)
    // Max should be 4 since that would be a FDX capture on a Adaptec 62044
    // Of course we have another special guest called S6000 series which
    // doesn't use the Channelcount byte for channels but for the capture
    // blade index. It wouldn't be fun to do something in a consistent way,
    // right?

    if iCaptureModule = ETH_CAPTYPE_GIGPOD2 then
       begin
            bBladeCount := pHeader^.ChannelCount;

            if bBladeCount = 1 then
               ReportMessage (msInfo, 'GE10/GE20 Device detected, capturing on blade ' + IntToStr (pHeader^.unknown8[0]), 'tCAPLoader.readTraceFileheader')
            else
                ReportMessage (msInfo, 'GE10/GE20 Device detected, capturing on ' + IntToStr (bBladeCount) + ' blades', 'tCAPLoader.readTraceFileheader');

            BChannelCount := 2; // TODO -oJasper: find out about the blade indexes used in a multiblade capture if we use only one port

            for i := 1 to bBladeCount * 2 do
                begin
                     InterfaceEntry := tCaptureInterface.create (i-1, LinkType, ioFromFile, true);

                     BladeIndex := ((i-1) div 2) + 1;

                     if (i mod 2) = 1 then
                        begin
                             InterfaceEntry.Name := IntToStr (BladeIndex) + '-A';
                             InterfaceEntry.Description := 'Network General S6000 GE10/GE20 Interface ' + IntToStr (BladeIndex) + '-A';
                        end
                     else
                         begin
                              InterfaceEntry.Name := IntToStr (BladeIndex) + '-B';
                              InterfaceEntry.Description := 'Network General S6000 GE10/GE20 Interface ' + IntToStr (BladeIndex) + '-B';
                         end;

                     InterfaceEntry.SpeedBPS := 1000000000;
                     InterfaceEntry.LinkType := LINKTYPE_ETHERNET;
                     InterfaceEntry.TimeStampResolution := 9;        // This interface has nanosecond resolution

                     InterfaceList.Add (InterfaceEntry);
                end;
       end
    else
        begin
             bChannelCount := pHeader^.ChannelCount;

             if bChannelCount = 0 then
                bChannelCount := 1;

             for i := 1 to bChannelCount do
                 begin
                      InterfaceEntry := tCaptureInterface.create (i-1, LinkType, ioFromFile, true);

                      InterfaceEntry.Name := 'Unknown';
                      InterfaceEntry.Description := 'Generic Interface from NAI CAP File';
                      InterfaceEntry.SpeedBPS := 1000000000;
                      InterfaceEntry.LinkType := LINKTYPE_ETHERNET;

                      InterfaceList.Add (InterfaceEntry);
                 end;
        end;

    // Appearantly traces taken from one NIC have a Channel Count of Zero, but we need it to be 1 for 1 Channel

    if bChannelCount = 0 then
       bChannelCount := 1;

    // We only do Ethernet and TokenRing right now, so we check for that

    if not ((LinkType = LINKTYPE_ETHERNET) or (LinkType = LINKTYPE_TOKEN_RING)) then
       begin
           raise EInOutError.Create ('Error: Unsupported network topology in Sniffer CAP file (tCAPLoader.readTraceFileHeader).');

           exit;
       end;

     // Time stuff - this is where it gets interesting or complicated/frustrating, depending on your point of view

     lwStartTime := pHeader^.iStartTime;
     StartTimeStamp := pHeader^.iTimeLow + pHeader^.iTimeHi * 4294967296;

     // Here we continue doing stuff like wireshark does

     case iFileFormat of
      CAP_FILE_NETXRAY_1_0 : begin
                                  dTraceTimeUnit := 1000.0;
                                  bPrecision := WTAP_FILE_TSPREC_MSEC;
                             end;
      CAP_FILE_NETXRAY_1_1 : begin
                                  dTraceTimeUnit := 1000000.0;
                                  bPrecision := WTAP_FILE_TSPREC_USEC;
                             end;

      CAP_FILE_NETXRAY_2_00x : begin
                                    // Here's the part where it gets nasty

                                    case LinkType of
                                     LINKTYPE_ETHERNET : begin
                                                              // Ethernet - the table to use depends on whether this is an NDIS or pod capture.

                                                              case pHeader^.capType of
                                                               CAPTYPE_NDIS          : begin
                                                                                            if pHeader^.iTimeUnit > NUM_NETXRAY_TIMEUNITS then
                                                                                               begin
                                                                                                    raise EInOutError.Create ('Error: Unknown timeunit ' + IntToStr(pHeader^.iTimeUnit) + ' for Ethernet/CAPTYPE_NDIS version ' + pHeader^.sVersion + ' Sniffer CAP file (CAPLoader.readTraceFileHeader).');

                                                                                                    exit;
                                                                                               end;

                                                                                            dTraceTimeUnit := TpS[pHeader^.iTimeUnit + 1];
                                                                                       end;
                                                               ETH_CAPTYPE_GIGPOD    : begin
                                                                                            if (pHeader^.iTimeUnit > NUM_NETXRAY_TIMEUNITS_GIGPOD) or (TpS_gigpod [pHeader^.iTimeUnit + 1] = 0) then
                                                                                               begin
                                                                                                    raise EInOutError.Create ('Error: Unknown timeunit ' + IntToStr(pHeader^.iTimeUnit) + ' for Ethernet/CAPTYPE_NDIS version ' + pHeader^.sVersion + ' Sniffer CAP file (CAPLoader.readTraceFileHeader).');

                                                                                                    exit;
                                                                                               end;

                                                                                            dTraceTimeUnit := TpS_gigpod[pHeader^.iTimeUnit + 1];
       {
                                                                                            // At least for 002.002 and 002.003 captures, the start time stamp is 0,
                                                                                            // not the value in the file.

                                                                                         if (iMinorVersion = 2) and (iMinorVersion = 3) then
                                                                                          StartTimeStamp := 0.0;
       }
                                                                                       end;
                                                               ETH_CAPTYPE_OTHERPOD  : begin
                                                                                            if (pHeader^.iTimeUnit > NUM_NETXRAY_TIMEUNITS_OTHERPOD) or (TpS_otherpod [pHeader^.iTimeUnit + 1] = 0) then
                                                                                               begin
                                                                                                    raise EInOutError.Create ('Error: Unknown timeunit ' + IntToStr(pHeader^.iTimeUnit) + ' for Ethernet/CAPTYPE_NDIS version ' + pHeader^.sVersion + ' Sniffer CAP file (CAPLoader.readTraceFileHeader).');

                                                                                                    exit;
                                                                                               end;

                                                                                            dTraceTimeUnit := TpS_otherpod[pHeader^.iTimeUnit + 1];

       {
                                                                                            // At least for 002.002 and 002.003 captures, the start time stamp is 0,
                                                                                            // not the value in the file.

                                                                                         if (iMinorVersion = 2) and (iMinorVersion = 3) then
                                                                                          StartTimeStamp := 0.0;
       }
                                                                                       end;

                                                               ETH_CAPTYPE_OTHERPOD2 : begin
                                                                                            if (pHeader^.iTimeUnit > NUM_NETXRAY_TIMEUNITS_OTHERPOD2) or (TpS_otherpod2 [pHeader^.iTimeUnit + 1] = 0) then
                                                                                               begin
                                                                                                    raise EInOutError.Create ('Error: Unknown timeunit ' + IntToStr(pHeader^.iTimeUnit) + ' for Ethernet/CAPTYPE_NDIS version ' + pHeader^.sVersion + ' Sniffer CAP file (CAPLoader.readTraceFileHeader).');

                                                                                                    exit;
                                                                                               end;

                                                                                            dTraceTimeUnit := TpS_otherpod2[pHeader^.iTimeUnit + 1];

                                                                                        // XXX: start time stamp in the one capture file examined of this type was 0;
                                                                                        //      We'll assume the start time handling is the same as for other pods.
                                                                                            //
       {
                                                                                            // At least for 002.002 and 002.003 captures, the start time stamp is 0,
                                                                                            // not the value in the file.

                                                                                         if (iMinorVersion = 2) and (iMinorVersion = 3) then
                                                                                          StartTimeStamp := 0.0;
       }
                                                                                       end;
                                                               ETH_CAPTYPE_GIGPOD2   : begin
                                                                                            // This should be a S6000 unit

                                                                                            if (pHeader^.iTimeUnit > NUM_NETXRAY_TIMEUNITS_GIGPOD2) or (TpS_gigpod2 [pHeader^.iTimeUnit + 1] = 0) then
                                                                                               begin
                                                                                                    raise EInOutError.Create ('Error: Unknown timeunit ' + IntToStr(pHeader^.iTimeUnit) + ' for Ethernet/CAPTYPE_NDIS version ' + pHeader^.sVersion + ' Sniffer CAP file (CAPLoader.readTraceFileHeader).');

                                                                                                    exit;
                                                                                               end;

                                                                                            dTraceTimeUnit := TpS_gigpod2[pHeader^.iTimeUnit + 1];

                                                                                        // XXX: start time stamp in the one capture file examined of this type was 0;
                                                                                        //      We'll assume the start time handling is the same as for other pods.
                                                                                            //
       {
                                                                                            // At least for 002.002 and 002.003 captures, the start time stamp is 0,
                                                                                            // not the value in the file.

                                                                                         if (iMinorVersion = 2) and (iMinorVersion = 3) then
                                                                                          StartTimeStamp := 0.0;
       }
                                                                                       end;
                                                              else
                                                                  begin
                                                                       raise EInOutError.Create ('Error: Unknown timeunit ' + IntToStr(pHeader^.iTimeUnit) + ' for Ethernet/CAPTYPE_NDIS version ' + pHeader^.sVersion + ' Sniffer CAP file (CAPLoader.readTraceFileHeader).');

                                                                       exit;
                                                                  end;
                                                              end;

                                                              //	If the number of ticks per second is greater than 1 million, make the precision be nanoseconds rather
                                                            // than microseconds.
                                                              //
                                                            // XXX - do values only slightly greater than one million correspond to a resolution sufficiently better than
                                                            // 1 microsecond to display more digits of precision?
                                                              // XXX - Seems reasonable to use nanosecs only if TPS >= 10M

                                                            if (dTraceTimeUnit >= 1e7) then
                                                              bPrecision := WTAP_FILE_TSPREC_NSEC
                                                              else
                                                               bPrecision := WTAP_FILE_TSPREC_USEC;
                                                         end;
                                    else
                                        begin
                                             dTraceTimeUnit := 0.0;
                                        end;
                                    end;
                               end;
     end;

     // Some Synerity time mojo

     TraceDateTime := UnixToDateTime (pHeader^.iStartTime + (iUTCDelta * 3600));       // we compensate for time zone difference
     i64StartTick := getStartTicksFromHeader (pHeader^.iStartTime);

     // Extract Line speed information

     lwLineSpeed := pHeader^.lwLinespeed;

     if (lwLinespeed < 0) or (lwLinespeed > lwMaxLineSpeed) then
        ReportMessage (msWarning, 'trace file has a suspicious linespeed value: ' + IntToStr (lwLineSpeed), 'tCAPLoader.readTraceFileheader');

     // Number of frames in File

     iTotalFramesInFile := pHeader^.iNumFrames;

     // Set Offset to first frame

     if iBatchFileOffset <> pHeader^.iStartOffset then
         begin
              // Uh oh... if this happens it means we got one of those trace files where some lunatic decided to
              // put in a few frames into the file out of order. That means the real first frame is behind other
              // frames that must be read later.
              // If I ever meet the guy who had this brain dead idea I'm going to choke him at least for a
              // minute!

              ReportMessage (msWarning, 'shifted CAP File start offset detected (' + IntToStr (pHeader^.iStartOffset) + '). Probably a ring buffer trace.', 'tCAPLoader.readTraceFileheader');

              // Apply the new offset

              iBatchFileOffset := pHeader^.iStartOffset;

              pBatchPosition := pByte (pData);
              inc (pBatchPosition, iBatchFileOffset);
         end;

     boHeaderRead := true;
end;

(* ------------------------------------------- *)

function tCAPLoader.ParseFrameFromOffset (const FrameIndex : integer; var FileOffset: integer; var DataPointer: pByte): tRawFrame;

var pFrameHeader   : pCAP_V2FrameHeader;
    pFrameBuffer   : pRawBuffer;

    boUseSliced    : boolean;

    helpBuf        : tRawBuffer;
    NewFrame       : tRawFrame;

    TimingDetails  : tTimingDetails;
    WireDetails    : tWireDetails;
    FilePosition   : tFileFrameLocation;
    TickValues     : tTickRec;

    Blade          : integer;
    Channel        : integer;

    ErrorFlags     : tErrorFlagsSet;

    InterfaceIndex : integer;
    InterfaceEntry : tCaptureInterface;

begin
     // Assign Header Values to Frame Object

     FilePosition.FileIndex := iFileNumber;                             // Taken from traceloader object - it knows what index the current file has
     FilePosition.FileFrameIndex := FrameIndex;
     FilePosition.FileOffset := FileOffset;

     WireDetails.initialize;
     WireDetails.LinkType := LinkType;                                  // Read from file header and stored in object variable

     // Check File Length if a record header is available

     if iFileSize < (FileOffset + sizeOf (tCAP_V2FrameHeader)) then
        begin
             raise EInOutError.Create ('Error: unexpected end of Sniffer CAP file when trying to read after frame #' + IntToStr (iTotalFramesRead) + ' (CAPLoader.parseframefromoffset).');
             exit;
        end;

     // Read Header

     pFrameHeader := pCAP_V2FrameHeader (DataPointer);

     // If we find the LocalCapture byte being one it means the capture card
     // had an active IP stack which transmitted while capture

     if pFrameHeader^.LocalCapture = 1 then
        boLocalFramesCaptured := true;

     // 1. Process the frame header structures for all the funny things CAP does

     // Here we have some moron avoidance code: usually the capturesize value tells us how many bytes were
     // actually captured. If the slicedsice value is greater than 0 it means that there were additional
     // random bytes captured which we have to ignore, but skip over.

     boUseSliced := false;

     if (pFrameHeader^.SlicedSize > 0) then
        begin
             if pFrameHeader^.SlicedSize < pFrameHeader^.iCaptureSize then
                begin
                     // Check if slicedsice is ridiculous, and if so use capturesize instead

                     if pFrameHeader^.SlicedSize >= 32 then
                        boUseSliced := true;
                end;
        end;

     if not (iCaptureModule = ETH_CAPTYPE_GIGPOD) then
        boUseSliced := false;

     // SlicedSice probably is only used for the whizzo "high speed" SysConnect gigabit modules.
     // For all other capture types we ignore SlicedSice

     if boUseSliced then
        WireDetails.CaptureSize := pFrameHeader^.SlicedSize - 4         // Looks like we captured more than we wanted, so use only the useful bits
     else
        WireDetails.CaptureSize := pFrameHeader^.iCaptureSize;          // Capturesize tells what we captured if slicing was okay

     WireDetails.WireSize := pFrameHeader^.iFrameSize;                  // Framesize tells the size on the wire

     ErrorFlags := CAPErrorFlagsToGenericFormat (pFrameHeader^.iErrorFlag);
     WireDetails.FrameFlags.SetErrorFlags(ErrorFlags);

     if WireDetails.FrameFlags.Value = 0 then
        WireDetails.FrameFlags.Present := false;

     WireDetails.FrameType := tftCAP;
     WireDetails.Dropped := 0;
     WireDetails.FrameFlags.Value := 0;

     TimingDetails.AbsoluteTime := TraceDateTime;
     TickValues := CAPTimetoTicks (FrameIndex, pFrameHeader^.rTime);
     TimingDetails.TraceTicks := TickValues.MicroSeconds;
     TimingDetails.NanoTicks := TickValues.NanoDeltaSeconds;

     // Determine Interface

     if iCaptureModule = ETH_CAPTYPE_GIGPOD2 then
        begin
             // Once again a special treatment for S6000 devices - map the blade and channel to an interface index

             Blade := pFrameHeader^.BladeIndex;
             Channel := pFrameHeader^.Channel;

             InterfaceIndex := ((Blade - 1) * 2) + Channel;

             WireDetails.InterfaceID := InterfaceIndex;
        end
     else
         begin
              WireDetails.InterfaceID := pFrameHeader^.Channel;
         end;

     // Determine LinkID

     WireDetails.LinkID := InterfaceList.getInterfaceByIndex (WireDetails.InterfaceID).PacketLinkID;

      // Determine if we have frame slicing active in this trace

     if (WireDetails.CaptureSize < pFrameHeader^.iFrameSize) then
         boFrameSlicing := true;

     // Move the pointer to the Frame Content Start

     inc (FileOffset, sizeOf (tCAP_V2FrameHeader));
     inc (DataPointer, sizeOf (tCAP_V2FrameHeader));

     // Check File Length if a frame is available

     if iFileSize < (FileOffset + WireDetails.CaptureSize) then
        begin
             raise EInOutError.Create ('Error: unexpected end of file (CAPLoader.parseframefromoffset).');
             exit;
        end;

     pFrameBuffer := pRawBuffer (DataPointer);

     if WireDetails.CaptureSize > maxFrameSize then
        raise EInOutError.Create ('Error: oversized frame encountered (CAPLoader.parseframefromoffset).');

     // We hand over a static buffer to the RawFrame.create which will move it to a dynamic buffer

     move (pFrameBuffer^, helpBuf, min (maxFrameSize, WireDetails.CaptureSize));

     // Create new Frame Object that will hold the data

     NewFrame := tRawFrame.create (TimingDetails, WireDetails, FilePosition, helpBuf);

     // Give it back

     result := NewFrame;
end;

(* ------------------------------------------- *)

procedure tCAPLoader.readBatchFromDisk;

var StartTime    : tDateTime;
    boReadDone   : boolean;
    newFrame     : tRawFrame;
    DeltaTime    : tDeltaTickTime;
    iDuration    : integer;

    pHelp        : pointer;
    helpBuf      : tRawBuffer;

begin
     // First we clear the frames we have in memory

     FrameList_ValidFrames.Clear;

     // Some initialisations

     StartTime := now;

     iFramesAvailable := 0;
     iFramesTossed := 0;

     boReadDone := false;

     if boEOF then
        exit;

     // Being here we should have read the file header and perhaps another batch of frames, so the iOffset needs to
     // be greater than 0

     if iBatchFileOffset = 0 then
        begin
             raise EInOutError.Create('Error: invalid file offset - Sniffer CAP file header not read? (CAPLoader.readBatchFromDisk).');

             exit;
        end;

     // Main Loop

     repeat
           inc (iFramesAvailable);
           inc (iTotalFramesRead);

           NewFrame := ParseFrameFromOffset (iTotalFramesRead, iBatchFileOffset, pBatchPosition);

           // We move by NewFrame.Framesize because that is how many bytes in the file we had for this frame (even
           // if we use less, f.e. if slicedsice is less that capture size)

           inc (iBatchFileOffset, NewFrame.FrameSize);
           inc (pBatchPosition, NewFrame.FrameSize);

           // 3. Add new Frame to our list of Frames

           self.addRawFrame(NewFrame);

           // 4. Check if we have as many frames as we have to read in one batch

           if (boFrameMemLimited and (iFramesAvailable >= iMaxFramesinMem)) then
              boReadDone := true;

           // 5. Check if the file is done

           if (iTotalFramesRead >= iTotalFramesInFile) then
              begin
                   boReadDone := true;
                   boEOF := true;
              end
           else
               // 6. Check if we need to wrap the offset because we have a stupid ring buffer trace

               if (iBatchFileOffset >= iFileSize)  then
                  begin
                       // we're at the end of the file and need to wrap to the first frame in the file
                       // because we started in the middle

                       ReportMessage (msInfo, 'End of File reached, wrapping to read remaining frames.', 'tCAPLoader.readBatchFromDisk');

                       iBatchFileOffset := 128;
                       pBatchPosition := pByte (pData);
                       inc (pBatchPosition, 128);
                  end;

     until boReadDone;

     iDuration := SecondsBetween (StartTime, now);

     if iDuration > 0 then
        iFPS := iFramesAvailable Div iDuration
     else
         iFPS := iFramesAvailable;
end;

(* ------------------------------------------- *)

function tCAPLoader.readSingleFrameByOffset(const Offset: integer; FrameIndex : integer = -1) : tRawFrame;

var pLocalPos    : pByte;
    NewFrame     : tRawFrame;
    HelpOffset   : integer;

begin
     HelpOffset := Offset; // Necessary because ParseFrameFromOffset will need to modify it, but it will be discarded afterwards anyway

     // Set the local pointer to the required Offset

     pLocalPos := pByte (pData);
     inc (pLocalPos, Offset);

     // Check File Length if a record header is available - this is NOT optional!

     if iFileSize < (Offset + sizeOf (tCAP_V2FrameHeader)) then
        begin
             raise EInOutError.Create ('Error: unexpected end of file (CAPLoader.readSingleFrameByOffset).');
             exit;
        end;

     // Create new Frame Object that will hold the data

     NewFrame := ParseFrameFromOffset (FrameIndex, HelpOffset, pLocalPos);

     // Add new Frame to our list of Frames

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tCAPLoader.readSingleFrameFromDisk (const FrameIndex: integer) : tRawFrame;

var pLocalPos    : pByte;
    iLocalOffset : integer;
    iCount       : integer;
    boReadDone   : boolean;
    boWrapped    : boolean;
    sHelp        : string;
    pHeader      : pCAP_Fileheader;
    pFrameHeader : pCAP_V2FrameHeader;

begin
     result := nil;

     if FrameIndex = 0 then
        raise EInOutError.Create ('Error: invalid frame index of 0 (CAPLoader.readSingleFrame).');

     boReadDone := false;
     boWrapped := false;

     iLocalOffset := 0;                  // This keeps track of where we are in the whole mapped file

     sHelp := '';
     pLocalPos := pByte (pData);

     // Header Stuff first... unfortunately we cannot skip this because the brain-dead ring buffer storing for CAP files :-(

     // Check File Length

     if iFileSize < length (CAP_MagicString) then
        raise EInOutError.Create ('Error: Sniffer CAP file too short (traceloader).');

     // 1. Check if the Magic String has the correct length and matches

     for iCount := 0 to (length (CAP_Magicstring) - 1) do
         begin
              sHelp := sHelp + chr (pData^[iCount]);

              inc (iLocalOffset);
              inc (pLocalPos);
         end;

     if sHelp <> CAP_MagicString then
        begin
             raise EInOutError.Create('Error: Not a Sniffer CAP file (traceloader).');

             exit;
        end;

     // Check File Length

     if iFileSize < (iLocalOffset + sizeOf (tCAP_FileHeader)) then
        begin
             raise EInOutError.Create ('Error: Sniffer CAP file too short (traceloader).');

             exit;
        end;

    // 2. Now we read the CAP file header

    pHeader := pCAP_Fileheader (pLocalPos);

    inc (iLocalOffset, sizeOf (tCAP_FileHeader));
    inc (pLocalPos, sizeOf (tCAP_FileHeader));

    // 3. Check if we have a ring buffer file

    if iLocalOffset <> pHeader^.iStartOffset then
        begin
             // Uh oh... if this happens it means we got one of those trace files where some lunatic decided to
             // put in a few frames into the file out of order. That means the real first frame is behind other
             // frames that must be read later.
             // If I ever meet the guy who had this brain dead idea I'm going to choke him at least for a
             // minute!

             ReportMessage (msWarning, 'shifted CAP File start offset detected (' + IntToStr (pHeader^.iStartOffset) + '). Probably a ring buffer trace.', 'tCAPLoader.readSingleFrameFromDisk');

             // Apply the new offset

             iLocalOffset := pHeader^.iStartOffset;

             pLocalPos := pByte (pData);
             inc (pLocalPos, iLocalOffset);
        end;

     iCount := 0;

     // Now pLocalPosition points to the first frame record header

     repeat
           inc (iCount);

           if iCount = FrameIndex then
              begin
                   result := readSingleFrameByOffset (iLocalOffset, FrameIndex);
                   boReadDone := true;
              end
           else
               begin
                    // Check if we need to wrap the offset because we have a stupid ring buffer trace

                    if not boWrapped and (iLocalOffset >= iFileSize)  then
                       begin
                            // we're at the end of the file and need to wrap to the first frame in the file
                            // because we started in the middle

                            ReportMessage (msInfo, 'End of File reached, wrapping to read remaining frames.', 'tCAPLoader.readSingleFrameFromDisk');

                            iLocalOffset := 128;
                            pLocalPos := pByte (pData);
                            inc (pLocalPos, 128);

                            boWrapped := true;
                       end;

                   // Check File Length if a record header is available

                    if iFileSize < (iLocalOffset + sizeOf (tCAP_V2FrameHeader)) then
                       begin
                            raise EInOutError.Create ('Error: unexpected end of file (CAPLoader.readSingleFrame).');

                            exit;
                       end;

                    // Read Header

                    pFrameHeader := pCAP_V2FrameHeader (pLocalPos);

                    inc (iLocalOffset, sizeOf (tCAP_V2FrameHeader));
                    inc (pLocalPos, sizeOf (tCAP_V2FrameHeader));

                    inc (iLocalOffset, pFrameHeader^.iCaptureSize);
                    inc (pLocalPos, pFrameHeader^.iCaptureSize);
               end;

     until boReadDone;
end;

(* ------------------------------------------- *)

end.
