{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_traceloader;

interface

uses Sysutils,
     Classes,
     Contnrs,
     base_structures,
     definition_simple_nameresolution,
     class_dupechecker,
     class_rawframe,
     class_interfacelist,
     class_timezonedetails;

type tLoaderStatus = (lsSuccess, lsUnknown, lsGeneralWarning, lsWrongHeader, lsGeneralError, lsIncompleteFile, lsReadFailed, lsAbort);

     tTraceLoaderMessage = record                                   // TODO: build a list structure with this to replace StrList_Messages
                                 Severity     : tMsgSeverity;
                                 FrameIndex   : int64;
                                 MessageText  : string;
                                 CodeLocation : string;

                                 procedure initialize;
                           end;

     tLoaderMessageList = array of tTraceLoaderMessage;

     tOnFrameLoaderMessage = procedure (Sender : tObject; const LoaderMessage : tTraceLoaderMessage) of object;

     ELoaderException = class (Exception) // alle Exceptions beginnen per Konvention mit einem E; Die Klasse Exception ist Vorfahr aller Exceptions

                              private
                                     priv_FrameIndex   : int64;
                                     priv_DiagHandle   : string;
                                     priv_ErrorMessage : string;
                              public
                                    constructor Create (DiagHandle, Msg: string; FrameIndex : int64);
                                    property FrameIndex : Int64 read priv_FrameIndex write priv_FrameIndex;
                                    property DiagnosticIndex : string read priv_DiagHandle write priv_DiagHandle;
                                    property Message : string read priv_ErrorMessage write priv_ErrorMessage;
                              end;

     tTraceLoader = class (tObject)

                          {$REGION 'Documentation'}
                          ///	<summary>
                          ///	  Constructor used to create a traceloader
                          ///	  instance
                          ///	</summary>
                          ///	<param name="Filename">
                          ///	  Filename of the tracefile that will be opened
                          ///	</param>
                          ///	<param name="FileIndex">
                          ///	  Index of the file if it is part of a filelist,
                          ///	  e.g. in a tTraceFileList object. Used to set
                          ///	  the according index in each frame read.
                          ///	</param>
                          ///	<param name="OwnsInterfacelist">
                          ///	  if true, the capture interface list will be
                          ///	  discarded when the loader is destroyed. This
                          ///	  must be set to false in cases where the
                          ///	  interfaces are stored and managed even after
                          ///	  the loader object is freed.
                          ///	</param>
                          {$ENDREGION}
                          constructor create (const Filename : string; const FileIndex : integer = 0; const OwnsInterfacelist : boolean = true);
                          procedure free;

                          protected
                                   EventLoaderMessage   : tOnFrameLoaderMessage;

                                   hFile                : tHandle;                // Filehandle to the trace
                                   hMap                 : tHandle;                // Memory Mapping handle
                                   pData                : pByteArray;             // This is a pointer that gives access to the memory mapped file
                                   pBatchPosition       : pByte;                  // This is the pointer that we "slide" throught the mapped file, as current read offset
                                                                                  // pPosition must always be pData + iFileOffset

                                   sTracefilename       : string;                 // Name of the trace file
                                   iFileNumber          : integer;                // Index of the tracefile; must be set by object that is creating the traceloader
                                   TraceDateTime        : tDateTime;              // Absolute DateTime of the Trace (Date & Time of the actual Capture)
                                   iFileSize            : int64;                  // Size of the File
                                   lwFramesInFile       : longword;               // Number of frames in the file if file header tells us, otherwise 0
                                   iBatchFileOffset     : integer;                // Offset in the file
                                   boEOF                : boolean;                // End of File reached
                                   boReadFailed         : boolean;                // Read error occured

                                   boFrameMemLimited    : boolean;                // Tells if the maxFramesInMem should be observed
                                   iMaxFramesInMem      : integer;                // Limits the ObjList_Frames to avoid spending all memory at once

                                   LoaderStatus         : tLoaderStatus;          // contains success or error status codes depending on what happened when loading a file

                                   iUTCDelta            : integer;                // Number of hours to compensate on loading
                                   lwLineSpeed          : longword;               // speed of network, in bits/second - needed for time calculations of every CAP frame
                                   iFramePos            : integer;                // counter for the current number of frames read
                                   iFramesAvailable     : integer;                // value that can be read to find out how many frames are ready
                                   iTotalFramesRead     : integer;                // value that tells how many total frames were yet by now
                                   iFramesSkipped       : integer;                // Frames that were not read (including Funny Frames)
                                   iFramesUnknown       : integer;                // Frames that we do not recognize
                                   iFPS                 : integer;                // Load Speed in Frames per Second

                                   // TODO : create an object to keep the FrameList to have more control about freeing frames on purpose

                                   FrameList_ValidFrames : tObjectList;           // Frames we accepted for processing (usually Dupe Frames don't make it in here, at least we try)
//                                 StrList_Messages      : tStringlist;           // Contains log messages if anything fishy turns up during processing a trace
                                   LoaderMessageList     : tLoaderMessageList;    // Contains log messages if anything fishy turns up during processing a trace
                                   LoaderMessageCount    : integer;

                                   dTraceTimeUnit        : double;
                                   boFrameSlicing        : boolean;               // true if frameslicing was active on capture
                                   boLocalFramesCaptured : boolean;               // true if Sniffer captured frames originating from the capture card
                                   bBladeCount           : byte;                  // Number of capture blades. One if its not a S6000 device with more than one
                                   bS6000BladeActive     : array [1..4] of byte;  // Contains the number of ports per Blade Index that were active in the capture. For S6000 this is usually 2.
                                   bChannelCount         : byte;                  // Number of trunked channels in trace
                                   bPrecision            : byte;
                                   iCaptureModule        : shortint;              // Tells with what capture hardware the file was captured (-1 for unknown)
                                   boHeaderRead          : boolean;               // Tells if the fileheader was read with "open" method, which is sometimes vital for reading single frames
                                   dtTimestamp           : tDateTime;             // File timestamp
                                   DupeCheckMode         : tDupeCheckMethod;      // The method to be applied for determining and disposing duplicate frames
                                   iFramesDropped        : integer;               // Number of frames dropped at capture time for the current batch
                                   iTotalFramesDropped   : integer;               // Number of frames dropped at capture time in total
                                   iFramesTossed         : integer;               // Number of frames dropped because they were duplicates for the current batch
                                   iTotalFramesTossed    : integer;               // Number of frames dropped because they were duplicates in total
                                   prot_LinkType         : longword;              // PCAP-NG based values

                                   DupeHandler           : tDupeChecker;          // Object that is used to run dupechecks on frames

                                   prot_FixLoadErrors    : boolean;               // if true, the loader will try to autocorrect problems

                                   // Additional Information Structures

                                   InterfaceCount        : integer;
                                   InterfaceList         : tCaptureInterfaceList;
                                   HeaderCommentList     : tStringList;
                                   HardwareInfoList      : tStringList;
                                   OperatingSystemList   : tStringList;
                                   UserApplicationList   : tStringList;

//                                 CaptureSystemGUID     : tGUIDBytes;            // Used by the Hone Project to associate programs/ports/packets to a system that took the capture
                                   CaptureSystemGUID     : tGUID;                 // Used by the Hone Project to associate programs/ports/packets to a system that took the capture

//                                 IPv4NameResolutionMgr : tAddressBookList;
//                                 IPv6NameResolutionMgr : tAddressBookList;
                                   IPv4NameResolutionMgr : tSimpleIPv4NameResolutionList;
                                   IPv6NameResolutionMgr : tSimpleIPv6NameResolutionList;

                                   TimeZoneHandler       : tTimeZoneHandler;

                                   procedure ReportMessage (const Severity: tMsgSeverity; const Messagetext: string; const CodeLocation : string = ''; const FrameIndex : int64 = 0);

                                   procedure setMaxFramesInMem (const iMax : integer);
                                   procedure setDupeCheckMode (const NewMode : tDupeCheckMethod);

                                   function getDupeTimeLimit : integer;
                                   procedure setDupeTimeLimit (newValue : integer);
                                   function getDupeFrameLimit : integer;
                                   procedure setDupeFrameLimit (newValue : integer);

                                   function getNumOfFramesAvailable : integer;

                                   function thisFileFormat : tCaptureFileType; virtual; abstract;
                                   function getCaptureModule : string;

                                   // This is a helper function that parses the tRawFrame from the file;

                                   function ParseFrameFromOffset(const FrameIndex : integer; var FileOffset: integer; var DataPointer: pByte): tRawFrame; virtual; abstract;
                          public
                                property onLoadMessage : tOnFrameLoaderMessage read EventLoaderMessage write EventLoaderMessage;

                                property Filename : string read sTracefilename;
                                property StartDateTime : tDateTime read TraceDateTime;
                                property Filesize : int64 read iFileSize;
                                property Position : integer read iBatchFileOffset;
                                property FramesInMemLimited : boolean read boFrameMemLimited write boFrameMemLimited;
                                property MaxFramesInMem : integer read iMaxFramesInMem write setMaxFramesInMem;
                                property FramesAvailable : integer read getNumOfFramesAvailable;
                                property FrameDuplicatesFound : integer read iFramesTossed;
                                property TotalFramesRead : integer read iTotalFramesRead;
                                property TotalFramesSkipped : integer read iFramesSkipped;
                                property TotalFramesUnknown : integer read iFramesUnknown;
                                property TotalFrameDuplicatesFound : integer read iTotalFramesTossed;
                                property LineSpeed : longword read lwLineSpeed;
                                property Channels : byte read bChannelCount;
                                property EOF : boolean read boEOF;
                                property ReadFailed : boolean read boReadFailed;
                                property FileFormat : tCaptureFileType read thisFileFormat;
                                property FPS : integer read iFPS;
                                property UTCDelta : integer read iUTCDelta write iUTCDelta;
                                property Status : tLoaderStatus read LoaderStatus;
                                property LinkType : longword read prot_LinkType;

                                property FrameSlicing    : boolean read boFrameSlicing;
                                property LocalFramesCaptured : boolean read boLocalFramesCaptured;
                                property CaptureModule : string read getCaptureModule;
//                              property Messages : tStringlist read StrList_Messages;
                                property MessageList : tLoaderMessageList read LoaderMessageList;
                                property MessageCount : integer read LoaderMessageCount;
                                property FileTimeStamp : tDateTime read dtTimestamp;

                                property DupeCheckMethod : tDupeCheckMethod read DupeCheckMode write SetDupeCheckMode;
                                property DupeHistoryTimeLimit : integer read getDupeTimeLimit write setDupeTimeLimit;
                                property DupeHistoryFrameLimit : integer read getDupeFrameLimit write setDupeFrameLimit;

                                property CaptureInterfaces : tCaptureInterfaceList read InterfaceList;
                                property NameResolutionManagerIPv4 : tSimpleIPv4NameResolutionList read IPv4NameResolutionMgr;
                                property NameResolutionManagerIPv6 : tSimpleIPv6NameResolutionList read IPv6NameResolutionMgr;

                                property HeaderComments : tStringlist read HeaderCommentList write HeaderCommentList;
                                property Hardware : tstringlist read HardwareInfoList;
                                property OperatingSystem : tstringlist read OperatingSystemList;
                                property Userapplication : tstringlist read UserApplicationList;

{$IFDEF playground}
                                property CAPHeader : tCAP_FileHeader read CAPFileHeader;
{$ENDIF}
                                property FixLoadErrors : boolean read prot_fixloaderrors write prot_fixloaderrors;

                                function openFile : boolean;
                                procedure closeFile;
                                procedure clearFramelist;

                                procedure addRawFrame (newFrame : tRawFrame);
                                function getRawFrame (const index : integer): tRawFrame;

                                function getInterfaceForFrameByFrameIndex (FrameIndex : integer) : tCaptureInterface;

                                // Read trace file header

                                procedure readTraceFileheader; virtual; abstract;

                                // Read a number of frames from disk

                                procedure readBatchFromDisk; virtual; abstract;

                                // Directly read a single frame from disk by frame index and adds it to the list. This does not modify the offset variable of the tracefile object!

                                function readSingleFrameFromDisk (const FrameIndex : integer) : tRawFrame; virtual; abstract;

                                // Directly read a single frame from disk by handle offset where the frame starts and add it to the list. This does not modify the offset variable of the tracefile object!

                                function readSingleFrameByOffset (const Offset : integer; FrameIndex : integer = -1) : tRawFrame; virtual; abstract;
                    end;


     procedure AddMessagesToList (const NewMessages : tLoaderMessageList; const NewMessageCount : integer; var ExistingMessages : tLoaderMessageList);


implementation

uses Math,
     Windows,
     SiAuto;

(* ------------------------------------------- *)

procedure AddMessagesToList (const NewMessages : tLoaderMessageList; const NewMessageCount : integer; var ExistingMessages : tLoaderMessageList);

var existLen : integer;
    i        : integer;

begin
     existLen := length (ExistingMessages);

     if NewMessageCount > 0 then
        begin
             setLength (ExistingMessages, existLen + NewMessageCount);

             for i := 0 to NewMessageCount - 1 do
                 ExistingMessages[existLen + i] := NewMessages[i];
        end;
end;

(* ------------------------------------------- *)
{ tTraceLoaderMessage }
(* ------------------------------------------- *)

procedure tTraceLoaderMessage.initialize;

begin
     Severity := msNone;
     FrameIndex := 0;
     MessageText := '';
     CodeLocation := '';
end;

(* ------------------------------------------- *)
{ ELoaderException }
(* ------------------------------------------- *)

constructor ELoaderException.Create (DiagHandle, Msg: string; FrameIndex: int64);

begin
     inherited Create (Msg);

     priv_FrameIndex := FrameIndex;
     priv_DiagHandle := DiagHandle;
end;

(* ------------------------------------------- *)
{ tTraceLoader }
(* ------------------------------------------- *)

constructor tTraceLoader.create (const Filename : string; const FileIndex : integer = 0; const OwnsInterfacelist : boolean = true);

var i : integer;

begin
     inherited create;

     sTracefilename := Filename;
     iFileNumber := FileIndex;

     iFramesAvailable := 0;
     iTotalFramesRead := 0;
     iTotalFramesTossed := 0;
     iTotalFramesDropped := 0;
     iFramesSkipped := 0;
     iFramesUnknown := 0;
     iFramePos := 0;
     iFPS := 0;
     iCaptureModule := CAPTYPE_UNKNOWN;
     bBladeCount := 1;

     DupeCheckMode := dcNone;
     DupeHandler := tDupeChecker.create (DupeCheckMode, 20, 1000);

     prot_fixloaderrors := true;
     prot_linktype := 0;

     for i := 1 to 4 do
         bS6000BladeActive[i] := 0;

     InterfaceList := tCaptureInterfaceList.create (OwnsInterfacelist);
     InterfaceCount := -1;

     boFrameSlicing := false;
     boLocalFramesCaptured := false;

     boHeaderRead := false;

     boFrameMemLimited := true;
     iMaxFramesInMem := 10000;

//   IPv4NameResolutionMgr := tAddressBookList.create;
//   IPv6NameResolutionMgr := tAddressBookList.create;

     TimeZoneHandler := tTimeZoneHandler.create;
//   FrameCommentList := tCommentList.create;
     HeaderCommentList := tStringList.create;
     HardwareInfoList := tStringList.create;
     OperatingSystemList := tStringList.create;
     UserApplicationList := tStringList.create;

     FrameList_ValidFrames := tObjectList.Create (false);         // Must be false, or else we accidentally discard frames when freeing the traceloader instance. Frame handling must be done at a higher structure!

//   StrList_Messages := tStringList.Create;
     LoaderMessageCount := 0;
     setlength (LoaderMessageList, 100);

     LoaderStatus := lsSuccess;

     CaptureSystemGUID := tGuid.Empty;
end;

(* ------------------------------------------- *)
{
function tTraceLoader.thisFileFormat: tCaptureFileType;

begin
     if self is tENCLoader then
        result := cfENC
     else
         if self is tCAPLoader then
            result := cfCAP
         else
             if self is tPCAPLoader then
                result := cfPCAP
             else
                 result := cfUnknown;
end;
}
(* ------------------------------------------- *)

function tTraceLoader.getCaptureModule: string;

begin
     case iCaptureModule of
     -1 : result := 'Unknown';
      0 : result := 'NDIS';
      2 : result := 'GigPod';
      3 : result := 'OtherPod';
      5 : result := 'OtherPod2';
      6 : result := 'GE10/GE20';                   // Wireshark calls it "GigPod2"
     else
         result := 'Unknown (' + IntToStr (iCaptureModule) + ')';
     end;
end;

(* ------------------------------------------- *)

// This routine adds a new frame to the list. Depending on the DupeCheck setting frames may be
// discarded before the actual addition.

procedure tTraceLoader.addRawFrame(newFrame: tRawFrame);

begin
     if (DupeCheckMode = dcNone) or (not DupeHandler.FrameIsDuplicate (newFrame)) then
        FrameList_ValidFrames.Add(newFrame)
     else
         begin
              inc (self.iFramesTossed);
              inc (self.iTotalFramesTossed);
         end;
end;

(* ------------------------------------------- *)

procedure tTraceLoader.clearFramelist;

begin
     FrameList_ValidFrames.Clear;
end;

(* ------------------------------------------- *)

procedure tTraceLoader.ReportMessage (const Severity: tMsgSeverity; const Messagetext : string; const CodeLocation : string = ''; const FrameIndex : int64 = 0);

var LoaderMessage : tTraceLoaderMessage;
    len           : integer;

begin
     LoaderMessage.Severity := Severity;
     LoaderMessage.MessageText := Messagetext;
     LoaderMessage.CodeLocation := CodeLocation;

     if Assigned (EventLoaderMessage) then
        EventLoaderMessage (self, LoaderMessage);

     inc (LoaderMessageCount);

     if LoaderMessageCount > length (LoaderMessageList) then
        setLength (LoaderMessageList, length (LoaderMessageList) + 100);

     LoaderMessageList[LoaderMessageCount - 1] := LoaderMessage;

//   StrList_Messages.Add(LoaderMessage.MessageText);
end;

(* ------------------------------------------- *)

procedure tTraceLoader.free;

begin
     freeAndNil (FrameList_ValidFrames);
//   freeAndNil (StrList_Messages);
     LoaderMessageCount := 0;
     setlength (LoaderMessageList, LoaderMessageCount);

//   IPv4NameResolutionMgr.free;
//   IPv4NameResolutionMgr := nil;

//   IPv6NameResolutionMgr.free;
//   IPv6NameResolutionMgr := nil;

     TimeZoneHandler.free;
     TimeZoneHandler := nil;

     DupeHandler.free;
     DupeHandler := nil;

//   FrameCommentList.free;
//   FrameCommentList := nil;

     HeaderCommentList.Free;
     HeaderCommentList := nil;

     HardwareInfoList.Free;
     HardwareInfoList := nil;

     OperatingSystemList.Free;
     OperatingSystemList := nil;

     UserApplicationList.Free;
     UserApplicationList := nil;

     InterfaceList.free;
     InterfaceList := nil;

     inherited;
end;

(* ------------------------------------------- *)

procedure tTraceLoader.setDupeCheckMode(const NewMode: tDupeCheckMethod);

begin
     DupeCheckMode := NewMode;

     if assigned (DupeHandler) then
        DupeHandler.DupeCheckMethod := NewMode;
end;

(* ------------------------------------------- *)

function tTraceLoader.getDupeFrameLimit: integer;

begin
     result := DupeHandler.MaxHistorySize;
end;

(* ------------------------------------------- *)

function tTraceLoader.getDupeTimeLimit: integer;

begin
     result := DupeHandler.MaxTimeDelta;
end;

(* ------------------------------------------- *)

procedure tTraceLoader.setDupeFrameLimit(newValue: integer);

begin
     DupeHandler.MaxHistorySize := newValue;
end;

(* ------------------------------------------- *)

procedure tTraceLoader.setDupeTimeLimit(newValue: integer);

begin
     DupeHandler.MaxTimeDelta := newValue;
end;

(* ------------------------------------------- *)

procedure tTraceLoader.setMaxFramesInMem(const iMax: integer);

begin
     iMaxFramesInMem := max (1, iMax);
     iMaxFramesInMem := min (iMaxFramesInMem, 10000000);
end;

(* ------------------------------------------- *)

function tTraceLoader.getNumOfFramesAvailable : integer;

begin
     result := iFramesAvailable;

     if iFramesAvailable <> FrameList_ValidFrames.Count then
        raise EInOutError.Create('Error: Frame count mismatch: ' + IntToStr (iFramesAvailable) + ' count <> ' + IntToStr (FrameList_ValidFrames.Count) + ' in the frame list (traceloader.getNumOfFrameAvailable).');
end;

(* ------------------------------------------- *)

// returns false if file was not found

function tTraceLoader.openFile : boolean;

var iHelp         : integer;

begin
     SiMain.EnterMethod ('tTraceLoader.openFile');

     if pData <> nil then
        SiMain.LogWarning ('Warning: file seems to be already mapped and reopened for a second time');

     result := true;
     boEOF := false;
     boReadFailed := false;

     boHeaderRead := false;

     // Clear old messages if any

//   StrList_Messages.Clear;
     LoaderMessageCount := 0;
     setlength (LoaderMessageList, LoaderMessageCount);

     if FileExists (sTraceFilename) then
        begin
             // Open File

             hFile := CreateFile (PChar (sTraceFileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);

             // Get details (Filesize and Timestamp)

             iFileSize := Windows.GetFileSize (hFile, nil);

             if iFilesize = 0 then
                begin
                     CloseHandle (hFile);
                     exit (false);
                end;

             iHelp := FileGetDate (hFile);
             dtTimeStamp := FileDateToDateTime (iHelp);

             // Create Memory Mapping

             hMap := CreateFileMapping (hFile, nil, PAGE_READONLY, 0, 0, nil);

             if hMap = 0 then
                begin
                     raise EInOutError.Create ('Error: File mapping failed (traceloader.openFile).');

                     exit;
                end;

             // Align the pointer to the mapping - in this case the offsets are 0/0, so we can only read files up to 2GB.
             // If bigger files need to be read, add more code here to slide the offset over the file

             pData := MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);

             if pData = nil then
                begin
                     SiMain.LogFatal ('Error: Memory file mapping failed, pData is NIL (traceloader.openFile).');
                     raise EInOutError.Create ('Error: File mapping failed (traceloader.openFile).');
                end;
        end
     else
         begin
              ReportMessage (msCritical, 'file not found', 'tTraceLoader.OpenFile');
              result := false;

              exit;
        end;

     SiMain.LeaveMethod ('tTraceLoader.openFile');
end;

(* ------------------------------------------- *)

procedure tTraceLoader.closeFile;

begin
     SiMain.EnterMethod ('tTraceLoader.closeFile');

     if iFramesAvailable > 0 then
        FrameList_ValidFrames.Clear;

     if pData <> nil then
        UnmapViewOfFile (pData);

     if hMap <> INVALID_HANDLE_VALUE then
        CloseHandle (hMap);

     if hFile <> INVALID_HANDLE_VALUE  then
        CloseHandle (hFile);

     pData := nil;

    SiMain.LeaveMethod ('tTraceLoader.closeFile');
end;

(* ------------------------------------------- *)

// Gets a frame from the loader, first frame starting with Index 1

function tTraceLoader.getRawFrame (const index : integer): tRawFrame;

begin
     if (index > 0) and (index <= FrameList_ValidFrames.Count) then
        result := (FrameList_ValidFrames.Items[index - 1] as tRawFrame)
     else
         raise EInOutError.Create ('Error: frame index out of bounds: ' + IntToStr (index) + ' (max: ' + IntToStr (FrameList_ValidFrames.Count) + ') (traceloader.getRawFrame).');
end;

(* ------------------------------------------- *)

function tTraceLoader.getInterfaceForFrameByFrameIndex (FrameIndex: integer): tCaptureInterface;

var frame : tRawFrame;
    entry : tCaptureInterface;

begin
     frame := (FrameList_ValidFrames.Items[FrameIndex - 1] as tRawFrame);
     entry := InterfaceList.getInterfaceByIndex(frame.InterfaceID);

     result := entry;
end;

(* ------------------------------------------- *)

end.

