{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parsemaster;

interface

uses base_structures,
     class_parser_baseclass,
     definition_ip_addresses,
     definition_5tuple,
     definition_tcp_structures,
     class_diagmessagehandler,
     common_parse_helpers,
     class_parser_tcp,
     class_parser_udp,
     class_rawframe;

// {$define debugparser}

type tParseEntry = record
                         Anchor        : tProtocolAnchor;
                         Parser        : tFrameParser;
                         CRCWasCorrect : boolean;
                         TunnelFlags   : tTunnelFlagSet;

                         procedure initialize;
                   end;

     tParseProcessMode = (ppmParseAll, ppmSanitize, ppmEdit, ppmUpToLayer2, ppmUpToLayer3, ppmUpToLayer4);
     tParsedProtocols = (ppEthernet, ppLinuxCooked, ppVLANTag, ppARP, ppMPLS, ppIPv4, ppIPv6, ppICMPv4, ppICMPv6, ppGRE, ppGTPv1, ppAYIYA, ppVXLAN, ppERSPAN, ppCAPWAP, ppGeneve, ppTZSP, ppTCP, ppUDP, ppDNS, ppDHCPv4, ppHTTP, ppRTPS);
     tParsedProtocolSet = set of tParsedProtocols;

     tDecodeScope = (psSource, psDestination, psBoth);

     {$REGION 'Documentation'}
     ///	<summary>
     ///	  A setting used in determineNextAnchorFromSocket routine to determine
     ///	  the next layer protocol
     ///	</summary>
     {$ENDREGION}
     tDecodeSetting = record
                            Port      : word;
                            PortScope : tDecodeScope;
                            Node4     : tIPv4Address;
                            Node6     : tIPv6Address;
                      end;

     tDecodeSettingsList = array of tDecodeSetting;

     tSocketLayerEntry = record
                               IPVersion            : byte;
                               EthernetAddrPair     : tMACAddressPair;
                               IPv4SocketPair       : tIPv4SocketPair;
                               IPv6SocketPair       : tIPv6SocketPair;
                               Layer4NextProtocol   : tLayer4Protocol;

                               procedure initialize;
                         end;

     tSocketLayerList = array of tSocketLayerEntry;

     tLayerOffset = record
                          StartOffset      : integer;
                          NextLayerOffset  : integer;

                          procedure initialize;
                    end;

     tLayerOffsetList = array [2..7] of tLayerOffset;


     tParseMaster = class (tObject)

                          {$REGION 'Documentation'}
                          ///	<param name="Frame">
                          ///	  RawFrame that will be parsed
                          ///	</param>
                          {$ENDREGION}
                          constructor create (Frame : tRawFrame; const ProcessMode : tParseProcessMode); overload;

                          {$REGION 'Documentation'}
                          ///	<param name="Frame">
                          ///	  RawFrame that will be parsed
                          ///	</param>
                          ///	<param name="DecodeSettings">
                          ///	  Decoding rules for TCP &amp; UDP payloads
                          ///	  (optional)
                          ///	</param>
                          {$ENDREGION}
                          constructor create (Frame : tRawFrame; const ProcessMode : tParseProcessMode; DecodeSettings : tDecodeSettingsList); overload;
                          procedure free;

                          private
                                 priv_frame              : tRawFrame;
                                 priv_ParseList          : array of tParseEntry;
                                 priv_ParseErrorCount    : integer;
                                 priv_parsemode          : tParseProcessMode;
                                 priv_DecodeSettings     : tDecodeSettingsList;
                                 priv_parsedprotocolset  : tParsedProtocolSet;
                                 priv_LayerOffsetList    : tLayerOffsetList;

                                 {$REGION 'Documentation'}
                                 ///	<summary>
                                 ///	  List of IP sockets that were found in
                                 ///	  the frame. If there is IP in IP it will
                                 ///	  contain more than one layer.
                                 ///	</summary>
                                 {$ENDREGION}
                                 priv_socketlayers       : tSocketLayerList;
                                 priv_tuples             : t5TupleSet;

                                 // Shortcut Infos - aggregates stuff for quick access without
                                 // going through all the single layers

                                 priv_EthernetPair        : tMACAddressPair;
                                 priv_VLANID              : word;
                                 priv_firstIPLayerVersion : byte;
                                 priv_lastIPLayerVersion  : byte;
                                 priv_IPv4SocketPair      : tIPv4SocketPair;
                                 priv_IPv6SocketPair      : tIPv6SocketPair;
                                 priv_IPID                : word;
                                 priv_Layer4Protocol      : byte;
                                 priv_TCPFlags            : tTCPFlagSet;
                                 priv_TCPSequenceNo       : tTCPSequence;
                                 priv_TCPAcknowledgeNo    : tTCPSequence;
                                 priv_TCPPayloadLength    : integer;
                                 priv_IsFragment          : boolean;

                                 priv_EtherLayerCount     : integer;
                                 priv_VLANCount           : integer;
                                 priv_IPLayerCount        : integer;

                                 // Remaining Payload

                                 priv_payloadchunk        : tFrameChunk;
                                 priv_paddingbytechunk    : tFrameChunk;
                                 priv_paddingbytecount    : integer;
                                 priv_containsfcs         : boolean;
                                 priv_ethernetfcs         : longword;
                                 priv_fcscorrect          : boolean;

                                 // Truncate settings

                                 priv_IsTruncated         : boolean;
                                 priv_truncatedby         : integer;

                                 // Diagnostics List

                                 priv_diagnosticmsglist   : tDiagnosticMsgList;

                                 {$REGION 'Documentation'}
                                 ///	<summary>
                                 ///	  Fills the parselist array with parsers
                                 ///	</summary>
                                 ///	<remarks>
                                 ///	  Is called after create, and should not
                                 ///	  be called from anywhere else, otherwise
                                 ///	  memory leak is probable because the
                                 ///	  parselist is not de-initialized if there
                                 ///	  are already parsers in it
                                 ///	</remarks>
                                 {$ENDREGION}
                                 procedure Initialize;
                                 procedure clear;
                                 procedure CreateParseList;
                                 function getParserCount: integer;
                                 function getHasRemainingPayload : boolean;
                                 function getRemaningPayloadLength : integer;

                                 procedure AddTupleToList (Tuple : t5Tuple);

                                 function NoAssemblerExists (HeaderType : tHeaderType) : boolean;

                                 function DetermineNextLayerOSI (const HeaderType : integer) : integer;
                                 {$REGION 'Documentation'}
                                 ///	<summary>
                                 ///	  Helper routine that tries to determine
                                 ///	  what the next layer protocol is by
                                 ///	  locking at the socket/ports
                                 ///	</summary>
                                 {$ENDREGION}
                                 function determineNextAnchorFromSocket (const Layer4Protocol : byte) : tHeaderType;
                                 /// <summary>
                                 ///   This routine is used to make sure that
                                 ///   when parsing is limited by ISO layer the
                                 ///   tunneling doesn't break the parsing
                                 ///   depth
                                 /// </summary>
                                 function NextProtocolIsTunnelLayer (const HeaderType : integer) : boolean;
                                 function getCopyOfRemainingPayload: tFrameChunk;
                                 function getCopyOfPaddingBytes : tFrameChunk;
                                 function getVLANIDsAsString : string;
                                 procedure SetFCSAndTrailerValues (TrailerLen : integer);

                                 /// <summary>
                                 ///   This function checks how complex a frame
                                 ///   is, mostly to help speed-filtering on
                                 ///   sockets later
                                 /// </summary>
                                 function DetermineComplexityRating : integer;

                                 procedure TransferDiagnosticMessages (const Parser : tFrameParser; const HeaderType : tHeaderType);

                          protected

                          public
                                property FrameToParse : tRawFrame read priv_frame write priv_frame;

                                {$REGION 'Documentation'}
                                ///	<summary>
                                ///	  The number of total layers decoded
                                ///	</summary>
                                {$ENDREGION}
                                property ParserCount : integer read getParserCount;
                                property ParserErrorCount : integer read priv_ParseErrorCount;

                                /// <summary>
                                ///   Returns a list of protocols
                                ///   found/identified during parsing
                                /// </summary>
                                property ParsedProtocols : tParsedProtocolSet read priv_parsedprotocolset;
                                /// <summary>
                                ///   Returns the lowest and highest offset for
                                ///   each ISO layer found. If a layer wasn't
                                ///   found the offsets are -1
                                /// </summary>
                                property LayerOffsetList : tLayerOffsetList read priv_LayerOffsetList;

                                {$REGION 'Documentation'}
                                ///	<summary>
                                ///	  This flag indicates if there are further
                                ///	  payload bytes that no parser exists for
                                ///	  yet. In that case, the RemainingPayload
                                ///	  property will contain the buffer of
                                ///	  unparsed bytes
                                ///	</summary>
                                ///	<value>
                                ///	  if true, unparsed payload is available
                                ///	</value>
                                {$ENDREGION}
                                property HasRemainingPayload : boolean read getHasRemainingPayload;
                                property RemainingPayloadLen : integer read getRemaningPayloadLength;
                                property RemainingPayloadCopy : tFrameChunk read getCopyOfRemainingPayload;
                                property PaddingByteCopy : tFrameChunk read getCopyOfPaddingBytes;
                                property PaddingByteCount : integer read priv_paddingbytecount;
                                property HasEthernetFCS : boolean read priv_containsfcs;
                                property EthernetFCS : longword read priv_ethernetfcs;
                                property EthernetFSCCorrect : boolean read priv_fcscorrect;

                                property EtherLayerCount : integer read priv_EtherLayerCount;
                                property VLANCount : integer read priv_VLANCount;
                                property IPLayerCount : integer read priv_IPLayerCount;
                                property IPLayers : tSocketLayerList read priv_socketlayers;
                                property Tuples : t5TupleSet read priv_tuples;

                                property MACAddressPair : tMACAddressPair read priv_EthernetPair;
                                property VLANID : word read priv_VLANID;
                                property VLANIDsAsString : string read getVLANIDsAsString;

                                {$REGION 'Documentation'}
                                ///	<summary>
                                ///	  IPv4 Socket pair found in the packet. In
                                ///	  case of multiple network/transport layers
                                ///	  the latest combination is kept
                                ///	</summary>
                                ///	<remarks>
                                ///	  Source and destination IPs and Ports are
                                ///	  just taken from the packet. They do NOT
                                ///	  indicate anything about client/server
                                ///	  yet. This must be done at higher levels
                                ///	  when having the full picture of what node
                                ///	  started the conversation!
                                ///	</remarks>
                                {$ENDREGION}
                                property IPv4SocketPair : tIPv4SocketPair read priv_IPv4SocketPair;
                                {$REGION 'Documentation'}
                                ///	<summary>
                                ///	  IPv6 Socket pair found in the packet. In
                                ///	  case of multiple network/transport layers
                                ///	  the latest combination is kept
                                ///	</summary>
                                ///	<remarks>
                                ///	  Source and destination IPs and Ports are
                                ///	  just taken from the packet. They do NOT
                                ///	  indicate anything about client/server
                                ///	  yet. This must be done at higher levels
                                ///	  when having the full picture of what node
                                ///	  started the conversation!
                                ///	</remarks>
                                {$ENDREGION}
                                property IPv6SocketPair : tIPv6SocketPair read priv_IPv6SocketPair;

                                property FirstIPLayerVersion : byte read priv_firstIPLayerVersion;
                                {$REGION 'Documentation'}
                                ///	<summary>
                                ///	  Tells what version the highest IP layer
                                ///	  was, which is important in case of
                                ///	  packets containing IPinIP
                                ///	</summary>
                                {$ENDREGION}
                                property LastIPLayerVersion : byte read priv_lastIPLayerVersion;
                                property Identification : word read priv_IPID;
                                property Layer4Protocol : byte read priv_Layer4Protocol;
                                property TCPSequenceNo : tTCPSequence read priv_TCPSequenceNo;
                                property TCPAcknowledgeNo : tTCPSequence read priv_TCPAcknowledgeNo;
                                property TCPPayloadLength : integer read priv_TCPPayloadLength;
                                property TCPFlags : tTCPFlagSet read priv_TCPFlags;

                                property IsFragment : boolean read priv_IsFragment;
                                property IsTruncated : boolean read priv_IsTruncated;
                                property TruncatedByteCount : integer read priv_truncatedby;
                                property ComplexityRating : integer read DetermineComplexityRating;

                                property DiagnosticMessages : tDiagnosticMsgList read priv_diagnosticmsglist;

                                {$REGION 'Documentation'}
                                ///	<summary>
                                ///	  Returns a frame parser type by layer index
                                ///	</summary>
                                ///	<param name="Index">
                                ///	  layer index; usually 1 for Ethernet, 2
                                ///	  for IP and so on
                                ///	</param>
                                ///	<returns>
                                ///	  tellswhat kind of parserwas used to
                                ///	  parse the values of the according layer
                                ///	</returns>
                                {$ENDREGION}
                                function getParserTypeByIndex (Index : integer) : tHeaderType;
                                function getParserParseInfoByIndex (Index : integer) : tParseEntry;
                                {$REGION 'Documentation'}
                                ///	<summary>
                                ///	  Returns a frameparser by layer index
                                ///	</summary>
                                ///	<param name="Index">
                                ///	  layer index; usually 1 for Ethernet, 2
                                ///	  for IP and so on
                                ///	</param>
                                ///	<returns>
                                ///	  Parser that was used to parse the values
                                ///	  of the according layer
                                ///	</returns>
                                {$ENDREGION}
                                function getParserByIndex (Index : integer) : tFrameParser;
                                function getParserByProtocol (const Protocol : tParsedProtocols; const Occurrence : byte = 1) : tFrameParser;
                                function getParserCRCOkayByIndex (Index : integer) : boolean;

                                function getDiagnosticMsgCount : integer;
                                function getDiagnosticMsg (const Index : integer) : tDiagnosticMsg;
                                procedure clearDiagnosticMsgList;

                                destructor Destroy; override;
                    end;


     function ParserTypeFromProtocol (const Protocol : tParsedProtocols) : integer;
     function ParsedProtocolsToString (Protocols : tParsedProtocolSet) : string;
     function ParseModeToString (Mode : tParseProcessMode) : string;

implementation

uses classes,
     SysUtils,
     toolkit,
     SiAuto,
     base_checksums,
     class_parser_ethernet,
     class_parser_linux_sll,
     class_parser_juniper,
     class_parser_tokenring,
     class_parser_802dot11_radiotap,
     class_parser_802dot11header,
     class_parser_802dot11management,
     class_parser_802dot1q,
     class_parser_llc,
     class_parser_mpls,
     class_parser_arp,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parser_ipv6_fragmentationheader,
     class_parser_icmpv4,
     class_parser_icmpv6,
     class_parser_dhcpv4,
     class_parser_dns,
     class_parser_http,
     class_parser_gre,
     class_parser_gtpv1,
     class_parser_ayiya,
     class_parser_vxlan,
     class_parser_netflow,
     class_parser_hsrp,
     class_parser_geneve,
     class_parser_capwap,
     class_parser_erspan,
     class_parser_tzsp,
     class_parser_rtps,
     class_parser_ntp;

(* -------------------------------------- *)

function ParserTypeFromProtocol (const Protocol : tParsedProtocols) : integer;

begin
     case Protocol of
      ppEthernet         : result := NEXT_HEADER_ETHERNET;
      ppLinuxCooked      : result := NEXT_HEADER_LINUX_COOKED;
      ppVLANTag          : result := NEXT_HEADER_8021_Q;
      ppARP              : result := NEXT_HEADER_ARP;
      ppIPv4             : result := NEXT_HEADER_IPv4;
      ppIPv6             : result := NEXT_HEADER_IPv6;
      ppICMPv4           : result := NEXT_HEADER_ICMP;
      ppICMPv6           : result := NEXT_HEADER_IPV6_ICMP;
      ppGRE              : result := NEXT_HEADER_GRE;
      ppGTPv1            : result := NEXT_HEADER_GTPv1;
      ppAYIYA            : result := NEXT_HEADER_AYIYA;
      ppVXLAN            : result := NEXT_HEADER_VXLAN;
      ppERSPAN           : result := NEXT_HEADER_ERSPAN;
      ppCAPWAP           : result := NEXT_HEADER_CAPWAP;
      ppGeneve           : result := NEXT_HEADER_Geneve;
      ppTZSP             : result := NEXT_HEADER_TZSP;
      ppTCP              : result := NEXT_HEADER_TCP;
      ppUDP              : result := NEXT_HEADER_UDP;
      ppDNS              : result := NEXT_HEADER_DNS;
      ppDHCPv4           : result := NEXT_HEADER_DHCPv4;
      ppHTTP             : result := NEXT_HEADER_HTTP;
      ppRTPS             : result := NEXT_HEADER_RTPS;
     else
         result := 0;
     end;
end;

(* -------------------------------------- *)

function ParsedProtocolsToString (Protocols : tParsedProtocolSet) : string;

var list : string;

begin
     // ppEthernet, ppLinuxCooked, ppVLANTag, ppARP, ppMPLS, ppIPv4, ppIPv6, ppICMPv4, ppICMPv6, ppTCP, ppUDP, ppDNS, ppDHCPv4, ppHTTP, ppRTPS
     // ppGRE, ppGTPv1, ppAYIYA, ppVXLAN, ppERSPAN, ppCAPWAP, ppGeneve, ppTZSP,

     if ppEthernet in Protocols then
        list := 'Ethernet '
     else
         list := '';

     if ppLinuxCooked in Protocols then
        list := list + 'LinuxCooked ';

     if ppVLANTag in Protocols then
        list := list + 'VLAN ';

     if ppARP in Protocols then
        list := list + 'ARP ';

     if ppMPLS in Protocols then
        list := list + 'MPLS ';

     if ppIPv4 in Protocols then
        list := list + 'IPv4 ';

     if ppIPv6 in Protocols then
        list := list + 'IPv6 ';

     if ppICMPv4 in Protocols then
        list := list + 'ICMPv4 ';

     if ppICMPv6 in Protocols then
        list := list + 'ICMPv6 ';

     if ppGRE in Protocols then
        list := list + ' GRE';

     if ppGTPv1 in Protocols then
        list := list + ' GTPv1';

     if ppAYIYA in Protocols then
        list := list + ' AYIYA';

     if ppVXLAN in Protocols then
        list := list + ' VXLAN';

     if ppERSPAN in Protocols then
        list := list + ' ERSPAN';

     if ppCAPWAP in Protocols then
        list := list + ' CAPWAP';

     if ppGeneve in Protocols then
        list := list + ' Geneve';

     if ppTZSP in Protocols then
        list := list + ' TZSP';

     if ppTCP in Protocols then
        list := list + 'TCP ';

     if ppUDP in Protocols then
        list := list + 'UDP ';

     if ppDNS in Protocols then
        list := list + 'DNS ';

     if ppDHCPv4 in Protocols then
        list := list + 'DHCPv4 ';

     if ppHTTP in Protocols then
        list := list + 'HTTP ';

     if ppRTPS in Protocols then
        list := list + 'RTPS ';

     list := Trim (list);
     list := StringReplace (list, ' ', ',', [rfReplaceAll]);

     result := list;
end;

(* -------------------------------------- *)

function ParseModeToString (Mode : tParseProcessMode) : string;

begin
     case Mode of
      ppmParseAll   : result := 'Parse All';
      ppmSanitize   : result := 'Sanitize';
      ppmEdit       : result := 'Edit';
      ppmUpToLayer2 : result := 'Layer2';
      ppmUpToLayer3 : result := 'Layer3';
      ppmUpToLayer4 : result := 'Layer4';
     end;
end;

(* -------------------------------------- *)
{ tSocketLayerEntry }
(* -------------------------------------- *)

procedure tSocketLayerEntry.initialize;

begin
     IPVersion := 0;
     EthernetAddrPair.initialize;
     IPv4SocketPair.Initialize;
     IPv6SocketPair.Initialize;
     Layer4NextProtocol := l4Any;
end;

(* -------------------------------------- *)
{ tLayerOffset }
(* -------------------------------------- *)

procedure tLayerOffset.initialize;

begin
     StartOffset := -1;
     NextLayerOffset := -1;
end;

(* -------------------------------------- *)
{ tParseEntry }
(* -------------------------------------- *)

procedure tParseEntry.initialize;

begin
     Anchor.HeaderType := 0;
     Anchor.Offset := 0;
     Parser := nil;
     CRCWasCorrect := false;
     TunnelFlags := [];
end;

(* -------------------------------------- *)
{ tParseMaster }
(* -------------------------------------- *)

constructor tParseMaster.create (Frame : tRawFrame; const ProcessMode : tParseProcessMode);

begin
     inherited create;

//   SiMain.LogDebug ('Creating ParseMaster, Mode ' + ParseModeToString (ProcessMode));

     priv_frame := Frame;

     initialize;

     // We have no decode settings, so set the list to zero

     setLength (priv_DecodeSettings, 0);

     priv_parsemode := ProcessMode;

     CreateParseList;

//   SiMain.LogDebug ('ParseMaster created, parsed Protocols: ' + ParsedProtocolsToString (ParsedProtocols));
//   SiMain.LogDebug ('ParseMaster Remaining Payload Length: ' + IntToStr (RemainingPayloadLen));
end;

(* -------------------------------------- *)

constructor tParseMaster.create (Frame: tRawFrame; const ProcessMode : tParseProcessMode; DecodeSettings: tDecodeSettingsList);

begin
     inherited create;

//   SiMain.LogDebug ('Creating ParseMaster, Mode ' + ParseModeToString (ProcessMode));

     priv_frame := Frame;

     initialize;

     priv_DecodeSettings := DecodeSettings;

     priv_parsemode := ProcessMode;
     priv_IsFragment := false;

     CreateParseList;

//   SiMain.LogDebug ('ParseMaster created, parsed Protocols: ' + ParsedProtocolsToString (ParsedProtocols));
//   SiMain.LogDebug ('ParseMaster Remaining Payload Length: ' + IntToStr (RemainingPayloadLen));
end;

(* -------------------------------------- *)

procedure tParseMaster.Initialize;
var
  i: Integer;

begin
     priv_ParseErrorCount := 0;

     priv_payloadchunk.initialize;
     priv_paddingbytechunk.initialize;

     priv_EtherLayerCount := 0;
     priv_VLANCount := 0;
     priv_IPLayerCount := 0;

     priv_paddingbytecount := 0;

     priv_containsfcs := false;
     priv_ethernetfcs := 0;
     priv_fcscorrect := true;

     priv_IPID := 0;

     priv_Layer4Protocol := 0;
     priv_TCPFlags := [];
     priv_TCPSequenceNo.initialize;
     priv_TCPAcknowledgeNo.initialize;
     priv_TCPPayloadLength := 0;
     priv_IsFragment := false;

     setLength (priv_ParseList, 0);
     setLength (priv_socketlayers, 0);
     setLength (priv_tuples, 0);

     priv_IPv4SocketPair.initialize;
     priv_IPv6SocketPair.initialize;

     priv_IsTruncated := priv_frame.WireSize > priv_frame.FrameSize;

     if priv_isTruncated then
        priv_truncatedby := priv_frame.WireSize - priv_frame.FrameSize
     else
         priv_truncatedby := 0;

     priv_parsedprotocolset := [];

     for i := 2 to 7 do
         priv_LayerOffsetList[i].initialize;
end;

(* -------------------------------------- *)

destructor tParseMaster.Destroy;

begin
//   SiMain.EnterMethod ('tParseMaster.Destroy');

     if priv_payloadchunk.Data <> nil then
        begin
//           SiMain.LogDebug ('Freeing PayloadChunk with length ' + IntToStr (priv_payloadchunk.Len));

             try
                priv_payloadchunk.free;
             except
                   SiMain.LogFatal ('Freeing PayloadChunk with length ' + IntToStr (priv_payloadchunk.Len) + ' failed!');
             end;
        end;


     if priv_paddingbytechunk.Data <> nil then
        begin
//           SiMain.LogDebug ('Freeing PaddingByteChunk with length ' + IntToStr (priv_paddingbytechunk.Len));

             try
                priv_paddingbytechunk.free;
             except
                   SiMain.LogFatal ('Freeing PaddingByteChunk with length ' + IntToStr (priv_paddingbytechunk.Len) + ' failed!');
             end;
        end;

     self.clear;

     inherited;

//   SiMain.LeaveMethod ('tParseMaster.Destroy');
end;

(* -------------------------------------- *)

procedure tParseMaster.free;

begin
     inherited;
end;

(* -------------------------------------- *)

// This function lets us know if we cannot reassemble a protocol

function tParseMaster.NoAssemblerExists (HeaderType: tHeaderType): boolean;

begin
     case HeaderType of
      NEXT_HEADER_JUNIPER,
      NEXT_HEADER_ERSPAN,
      NEXT_HEADER_GENEVE,
      NEXT_HEADER_CAPWAP,
      NEXT_HEADER_TZSP,
      NEXT_HEADER_NTP,
      NEXT_HEADER_HTTP,
      NEXT_HEADER_DNS     : result := true;
     else
         result := false;
     end
end;

(* -------------------------------------- *)

procedure tParseMaster.AddTupleToList (Tuple: t5Tuple);

var len : integer;

begin
     len := length (priv_tuples);
     setLength (priv_tuples, len + 1);       // yeah, I know, not efficient, but it works
     priv_tuples[len] := Tuple;
end;

(* -------------------------------------- *)

procedure tParseMaster.clear;

var anchor      : tProtocolAnchor;
    frameparser : tFrameParser;
    i           : integer;

begin
//   SiMain.EnterMethod ('tParseMaster.Clear');

     // Remove all parser objects

     for i := 1 to length (priv_ParseList) do
         begin
              frameparser := priv_ParseList[i-1].Parser;
              anchor := priv_ParseList[i-1].anchor;

              case anchor.HeaderType of
               NEXT_HEADER_ETHERNET,
               NEXT_HEADER_TRANS_ETHER_BRIDGE : (FrameParser as tEthernetParser).free;
               NEXT_HEADER_LINUX_COOKED       : (frameparser as tLinuxSLLParser).free;
               NEXT_HEADER_JUNIPER            : (frameparser as tJuniperParser).free;
               NEXT_HEADER_80211RADIO         : (frameparser as t802dot11RadioTapParser).free;
               NEXT_HEADER_802dot11Hdr        : (frameparser as t802dot11HeaderParser).free;
               NEXT_HEADER_802dot11Mgmt       : (frameparser as t802dot11ManagementParser).free;
//             NEXT_HEADER_TOKEN_RING         : (FrameParser as tTokenRingParser).free;
               NEXT_HEADER_LLC                : (frameparser as tLLCParser).free;
               NEXT_HEADER_8021_Q,
               NEXT_HEADER_802_QINQ_Outer,
               NEXT_HEADER_802_QINQ_Outer_OLD1,
               NEXT_HEADER_802_QINQ_Outer_OLD2,
               NEXT_HEADER_802_QINQ_Outer_OLD3: (FrameParser as t802dot1qParser).free;
               NEXT_HEADER_ARP,
               NEXT_HEADER_RARP               : (FrameParser as tARPParser).Free;
               NEXT_HEADER_MPLS_UNICAST,
               NEXT_HEADER_MPLS_MULTICAST     : (frameparser as tMPLSParser).free;
               NEXT_HEADER_IPv4_ENCAPSULATION,
               NEXT_HEADER_IPv4               : (FrameParser as tIPv4Parser).free;
               NEXT_HEADER_TYPE_IPV6,
               NEXT_HEADER_IPv6               : (FrameParser as tIPv6Parser).free;
               NEXT_HEADER_IPV6_FRAGMENTATION : (FrameParser as tIPv6FragmentationParser).free;
               NEXT_HEADER_ICMP               : (FrameParser as tICMPv4Parser).free;
               NEXT_HEADER_IPV6_ICMP          : (FrameParser as tICMPv6Parser).free;
               NEXT_HEADER_TCP                : (FrameParser as tTCPParser).free;
               NEXT_HEADER_UDP                : (FrameParser as tUDPParser).free;
               NEXT_HEADER_DHCPv4             : (FrameParser as tDHCPv4Parser).free;
               NEXT_HEADER_DNS                : (FrameParser as tDNSParser).free;
               NEXT_HEADER_HTTP               : (frameparser as tHTTPParser).free;
               NEXT_HEADER_GRE                : (FrameParser as tGREParser).free;
               NEXT_HEADER_GTPv1              : (frameparser as tGTPv1Parser).free;
               NEXT_HEADER_AYIYA              : (FrameParser as tAYIYAParser).free;
               NEXT_HEADER_NETFLOW            : (FrameParser as tNetflowParser).free;
               NEXT_HEADER_HSRP               : (FrameParser as tHSRPParser).free;
               NEXT_HEADER_VXLAN              : (FrameParser as tVXLANParser).free;
               NEXT_HEADER_ERSPAN             : (frameparser as tERSPANParser).free;
               NEXT_HEADER_RTPS               : (frameparser as tRTPSParser).free;
               NEXT_HEADER_NTP                : (frameparser as tNTPParser).free;
               NEXT_HEADER_TZSP               : (frameparser as tTZSPParser).free;
              else
                  frameparser.free;
              end;

              frameparser := nil;
         end;

     // Empty list

     setLength (priv_ParseList, 0);

     // Unlink (not free!) frame

     priv_frame := nil;

//   SiMain.LeaveMethod ('tParseMaster.Clear');
end;

(* -------------------------------------- *)

function tParseMaster.DetermineComplexityRating: integer;

var len : integer;
    i   : integer;

begin
     len := length (priv_parselist);

     if len > 2 then
        begin
             if (priv_parselist[0].Anchor.HeaderType = NEXT_HEADER_ETHERNET) and
                ((priv_parselist[1].Anchor.HeaderType = NEXT_HEADER_IPV4) or (priv_parselist[1].Anchor.HeaderType = NEXT_HEADER_IPV6)) and
                ((priv_parselist[2].Anchor.HeaderType = NEXT_HEADER_TCP) or (priv_parselist[2].Anchor.HeaderType = NEXT_HEADER_UDP)) then
                result := ComplexityL234
             else
                 if (len > 3) then
                    begin
                         if (priv_parselist[0].Anchor.HeaderType = NEXT_HEADER_ETHERNET) and
                            (priv_parselist[1].Anchor.HeaderType = NEXT_HEADER_8021_Q) and
                            ((priv_parselist[2].Anchor.HeaderType = NEXT_HEADER_IPV4) or (priv_parselist[1].Anchor.HeaderType = NEXT_HEADER_IPV6)) and
                            ((priv_parselist[3].Anchor.HeaderType = NEXT_HEADER_TCP) or (priv_parselist[2].Anchor.HeaderType = NEXT_HEADER_UDP)) then
                            result := ComplexityL2Q34
                         else
                             result := ComplexityHigh;
                    end
                 else
                     result := ComplexityUnknown;
        end
     else
         result := ComplexityUnknown;
end;

(* -------------------------------------- *)

function tParseMaster.determineNextAnchorFromSocket (const Layer4Protocol : byte) : tHeaderType;

var SourcePort : word;
    DestPort   : word;

begin
     result := NEXT_HEADER_PAYLOAD;

     // Determine what ports we need to look at

     if priv_lastIPLayerVersion = 4 then
        begin
             SourcePort := priv_IPv4SocketPair.Source.Port;
             DestPort := priv_IPv4SocketPair.Destination.Port;
        end
     else
         begin
             SourcePort := priv_IPv6SocketPair.Source.Port;
             DestPort := priv_IPv6SocketPair.Destination.Port;
         end;

     // DNS

     if (SourcePort = 53) or (DestPort = 53) then
        result := NEXT_HEADER_DNS;

     // DHCP

     if (SourcePort = 67) or (DestPort = 67) then
        result := NEXT_HEADER_DHCPv4;
{
     // HTTP

     if (SourcePort = 80) or (DestPort = 80) then
        result := NEXT_HEADER_HTTP;
}
     // HSRP

     if (SourcePort = 1985) and (DestPort = 1985) then
        result := NEXT_HEADER_HSRP;

     // GTP

     if (SourcePort = 2152) or (DestPort = 2152) then
        result := NEXT_HEADER_GTPv1;

     // AYIYA

     if (SourcePort = 5072) or (DestPort = 5072) then
        result := NEXT_HEADER_AYIYA;

     // VxLAN

     if (Layer4PRotocol = NEXT_HEADER_UDP) and ((SourcePort = 4789) or (DestPort = 4789)) then
        result := NEXT_HEADER_VXLAN;

     // Geneve

     if (Layer4Protocol = NEXT_HEADER_UDP) and ((SourcePort = 6081) or (DestPort = 6081)) then
        result := NEXT_HEADER_GENEVE;

     // NetFlow

     if (Layer4Protocol = NEXT_HEADER_UDP) and ((DestPort = 2055) or (DestPort = 9996)) then
        result := NEXT_HEADER_NETFLOW;

     // CAPWAP

     if (Layer4Protocol = NEXT_HEADER_UDP) and ((SourcePort = 5247) or (DestPort = 5247)) then
        result := NEXT_HEADER_CAPWAP;

     // NTP

     if (Layer4Protocol = NEXT_HEADER_UDP) and ((SourcePort = 123) and (DestPort = 123)) then
        result := NEXT_HEADER_NTP;

     // VxLAN

     if (Layer4PRotocol = NEXT_HEADER_UDP) and ((SourcePort = 37008) or (DestPort = 37008)) then
        result := NEXT_HEADER_TZSP;

     // More to come

     // TODO: if we have a port/IP decode settings list, check it for a matching rule
end;

(* -------------------------------------- *)

function tParseMaster.DetermineNextLayerOSI (const HeaderType: integer): integer;

begin
     result := 7;

     case HeaderType of
      NEXT_HEADER_80211RADIO,
      NEXT_HEADER_802dot11Hdr,
      NEXT_HEADER_802dot11Mgmt          : result := 1;
      NEXT_HEADER_ETHERNET,
      NEXT_HEADER_TRANS_ETHER_BRIDGE,
      NEXT_HEADER_LLC,
      NEXT_HEADER_8021_Q,
      NEXT_HEADER_802_QINQ_Outer,
      NEXT_HEADER_802_QINQ_Outer_OLD1,
      NEXT_HEADER_802_QINQ_Outer_OLD2,
      NEXT_HEADER_802_QINQ_Outer_OLD3,
      NEXT_HEADER_ARP,
      NEXT_HEADER_RARP,
      NEXT_HEADER_MPLS_UNICAST,
      NEXT_HEADER_MPLS_MULTICAST        : result := 2;
      NEXT_HEADER_IPV4,
      NEXT_HEADER_IPv4_ENCAPSULATION,
      NEXT_HEADER_IPv6,
      NEXT_HEADER_TYPE_IPV6,
      NEXT_HEADER_IPV6_FRAGMENTATION,
      NEXT_HEADER_GRE,
      NEXT_HEADER_GTPv1,
      NEXT_HEADER_AYIYA,
      NEXT_HEADER_GENEVE,
      NEXT_HEADER_VXLAN,
      NEXT_HEADER_ERSPAN,
      NEXT_HEADER_CAPWAP,
      NEXT_HEADER_TZSP,
      NEXT_HEADER_ICMP,
      NEXT_HEADER_IPV6_ICMP             : result := 3;
      NEXT_HEADER_TCP,
      NEXT_HEADER_UDP                   : result := 4;
     end;
end;

(* -------------------------------------- *)

function tParseMaster.NextProtocolIsTunnelLayer (const HeaderType : integer) : boolean;

begin
     case HeaderType of
         NEXT_HEADER_GRE,
         NEXT_HEADER_GTPv1,
         NEXT_HEADER_AYIYA,
         NEXT_HEADER_VXLAN,
         NEXT_HEADER_ERSPAN,
         NEXT_HEADER_CAPWAP,
         NEXT_HEADER_GENEVE : result := true;
     else
         result := false;
     end;
end;

(* -------------------------------------- *)

procedure tParseMaster.SetFCSAndTrailerValues (TrailerLen: integer);

var help     : pByte;
    LongHelp : pLongword;
    offset   : integer;
    crc      : Longword;
    crcchunk : tFrameChunk;
    isZero   : boolean;
    test     : string;
    i        : integer;

begin
     if TrailerLen = 0 then
        begin
             // neither padding nor FCS

             priv_containsfcs := false;
             priv_paddingbytecount := 0;
        end
     else
         if TrailerLen < 4 then
            begin
                 // too short for a FCS

                 priv_paddingbytecount := TrailerLen;
                 priv_containsfcs := false;
            end
         else
             if TrailerLen > 4 then
                begin
                     // could be padding and FCS, so let's find out. Usually we have padding bytes
                     // of 0, and and FCS > 0.
                     // Step one: check all bytes except the last 4 for being zero

                     isZero := true;
                     i := 0;

                     while (i < (Trailerlen - 4)) and isZero do
                           begin
                                help := pByte (priv_frame.Data) + (priv_frame.FrameSize - TrailerLen) + i;

                                if help^ > 0 then
                                   isZero := false;

                                inc (i);
                           end;

                     // Step two: if padding bytes are all zero, check if the last 4 aren't

                     if isZero then
                        begin
                             help := pByte (priv_frame.Data) + (priv_frame.FrameSize - 4);
                             LongHelp := pLongword (Help);

                             if LongHelp^ > 0 then
                                begin
                                     priv_containsfcs := true;
                                     priv_paddingbytecount := TrailerLen - 4;
                                     priv_ethernetfcs := swap32 (LongHelp^);
                                end
                             else
                                 begin
                                      priv_paddingbytecount := TrailerLen;
                                      priv_containsfcs := false;
                                 end;
                        end
                     else
                         begin
                              // not zero, so assume we have garbage padding and no FCS

                              priv_paddingbytecount := TrailerLen;
                              priv_containsfcs := false;
                         end;
                end
             else
                 begin
                      // we have 4 bytes of padding or FCS, so let's check if it's all zeroes

                      help := pByte (priv_frame.Data) + (priv_frame.FrameSize - 4);
                      LongHelp := pLongword (Help);
                      priv_ethernetfcs := swap32 (LongHelp^);

                      priv_containsfcs := true;
                 end;

     if priv_containsfcs then
        begin
             case priv_frame.LinkType of
              LINKTYPE_ETHERNET         : begin
                                               // Calculate frame CRC

                                               crc := priv_frame.CRC32;
                                          end;
              LINKTYPE_IEEE802_11_RADIO : begin
                                               // get the start offset of the 802.11 header (following the radio tap header)

                                               offset := priv_ParseList[0].Parser.NextAnchor.offset;

                                               crcchunk := priv_frame.FrameDataAsChunk (offset);

                                               // calculate CRC, minus the 4 bytes at the end (which are the existing CRC bytes)

                                               crc := getCRC32FromPointer (crcchunk.Data, crcchunk.Len - 4);
                                           end;
              end;

              priv_fcscorrect :=  (crc = priv_ethernetfcs);
        end;

     if priv_paddingbytecount > 0 then
        begin
             help := pByte (priv_frame.Data) + (priv_frame.FrameSize - TrailerLen);
             priv_paddingbytechunk.copyFrom (help, priv_paddingbytecount);
{$ifdef debug}
             test := priv_paddingbytechunk.getAsHEXString;
{$endif}
        end;
end;

(* -------------------------------------- *)

procedure tParseMaster.TransferDiagnosticMessages (const Parser : tFrameParser; const HeaderType : tHeaderType);

var RecCount  : integer;
    Count     : integer;
    MsgRecord : tDiagnosticMsg;
    i         : integer;

begin
     RecCount := Parser.getDiagnosticMsgCount;
     Count := length (priv_diagnosticmsglist);

     setlength (priv_diagnosticmsglist, Count + RecCount);

     for i := 1 to RecCount do
         begin
              MsgRecord := Parser.getDiagnosticMsg (i);
              MsgRecord.frameindex := priv_frame.FileFrameID;
              MsgRecord.parsertype := HeaderType;

              priv_diagnosticmsglist[Count + i - 1] := MsgRecord;
         end;
end;

(* -------------------------------------- *)

procedure tParseMaster.CreateParseList;

var FrameParser    : tFrameParser;
    ProtocolAnchor : tProtocolAnchor;
    ParentAnchor   : tProtocolAnchor;
    HelpAnchor     : tProtocolAnchor;
    ParseComplete  : boolean;
    ParseResult    : tParseResult;
    ChecksumTest   : word;
    ChecksumActual : word;
    IP4Pseudo      : tIPv4PseudoHeader;
    IP6Pseudo      : tIPv6PseudoHeader;
    IsICMPQuote    : boolean;
    WiFiSubType    : t802dot11FrameSubType;
    CRCBuffer      : tBufferWords;
    help           : pByte;
    HeaderLen      : integer;
    IPPayLoadLen   : integer;
    L4PayloadLen   : integer;
    TrailerLen     : integer;
    NextHdr        : byte;
//  len            : integer;
    layerlistlen   : integer;
    NextLayerOSI   : integer;
    i              : integer;
    Tuple          : t5Tuple;

begin
     // Init

     ParseComplete := false;
     WiFiSubType := mgmtftNone;
     IsICMPQuote := False;
     ParentAnchor.HeaderType := 0;
     ParentAnchor.Offset := 0;
     ProtocolAnchor.Offset := 0;
     i := 0;
     TrailerLen := 0;
     layerlistlen := 0;

     case priv_frame.LinkType of
      LINKTYPE_ETHERNET         : ProtocolAnchor.HeaderType := NEXT_HEADER_ETHERNET;
      LINKTYPE_LINUX_SLL        : ProtocolAnchor.HeaderType := NEXT_HEADER_LINUX_COOKED;
      LINKTYPE_IEEE802_11_RADIO : ProtocolAnchor.HeaderType := NEXT_HEADER_80211RADIO;
      LINKTYPE_JUNIPER          : ProtocolAnchor.HeaderType := NEXT_HEADER_JUNIPER;
      LINKTYPE_RAW,
      LINKTYPE_IPV4             : ProtocolAnchor.HeaderType := NEXT_HEADER_IPv4;
//    LINKTYPE_TOKEN_RING : ProtocolAnchor.HeaderType := NEXT_HEADER_TOKEN_RING;
     end;

     repeat
           // Create parser based on header type

            case ProtocolAnchor.HeaderType of
            NEXT_HEADER_NONE               : begin
                                                  TrailerLen := priv_frame.FrameSize - ProtocolAnchor.Offset;

                                                  if TrailerLen > 0 then
                                                     SetFCSAndTrailerValues (TrailerLen);

                                                  ParseComplete := true
                                             end;
            NEXT_HEADER_ETHERNET,
            NEXT_HEADER_TRANS_ETHER_BRIDGE : FrameParser := tEthernetParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_LINUX_COOKED       : FrameParser := tLinuxSLLParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_JUNIPER            : FrameParser := tJuniperParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_80211RADIO         : FrameParser := t802dot11RadioTapParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_802dot11Hdr        : FrameParser := t802dot11HeaderParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_802dot11Mgmt       : FrameParser := t802dot11ManagementParser.create (WiFiSubType, priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_LLC                : FrameParser := tLLCParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
//          NEXT_HEADER_TOKEN_RING         : FrameParser := tTokenRingParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_8021_Q,
            NEXT_HEADER_802_QINQ_Outer,
            NEXT_HEADER_802_QINQ_Outer_OLD1,
            NEXT_HEADER_802_QINQ_Outer_OLD2,
            NEXT_HEADER_802_QINQ_Outer_OLD3: FrameParser := t802dot1qParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_ARP,
            NEXT_HEADER_RARP               : FrameParser := tARPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_MPLS_UNICAST,
            NEXT_HEADER_MPLS_MULTICAST     : FrameParser := tMPLSParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);

            NEXT_HEADER_IPv4_ENCAPSULATION,
            NEXT_HEADER_IPv4               : FrameParser := tIPv4Parser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, IsICMPQuote, priv_Frame.FileFrameID);
            NEXT_HEADER_TYPE_IPV6,
            NEXT_HEADER_IPv6               : FrameParser := tIPv6Parser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_IPV6_FRAGMENTATION : begin
                                                  FrameParser := tIPv6FragmentationParser.create (priv_frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
                                                  priv_IsFragment := true;
                                             end;
            NEXT_HEADER_ICMP               : FrameParser := tICMPv4Parser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_IPV6_ICMP          : FrameParser := tICMPv6Parser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, IPPayloadLen, priv_Frame.FileFrameID);
            // NEXT_HEADER ICMPv6
            NEXT_HEADER_GRE                : FrameParser := tGREParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_TCP                : FrameParser := tTCPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, IPPayloadLen, priv_Frame.FileFrameID);
            NEXT_HEADER_UDP                : FrameParser := tUDPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, IPPayloadLen, priv_Frame.FileFrameID);
            NEXT_HEADER_GTPv1              : FrameParser := tGTPv1Parser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_DHCPv4             : FrameParser := tDHCPv4Parser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, L4PayloadLen, priv_Frame.FileFrameID);
            NEXT_HEADER_DNS                : FrameParser := tDNSParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, L4PayloadLen, priv_Frame.FileFrameID);
            NEXT_HEADER_HTTP               : FrameParser := tHTTPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_AYIYA              : FrameParser := tAYIYAParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_NETFLOW            : FrameParser := tNetFlowParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_HSRP               : FrameParser := tHSRPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_GENEVE             : FrameParser := tGeneveParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_VXLAN              : FrameParser := tVXLANParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_ERSPAN             : FrameParser := tERSPANParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_CAPWAP             : FrameParser := tCAPWAPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_TZSP               : FrameParser := tTZSPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_RTPS               : FrameParser := tRTPSParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_NTP                : FrameParser := tNTPParser.create (priv_Frame.Data, ProtocolAnchor.Offset, priv_Frame.FrameSize, priv_frame.WireSize, priv_Frame.FileFrameID);
            NEXT_HEADER_PAYLOAD            : begin
                                                  SetFCSAndTrailerValues (TrailerLen);

                                                  priv_payloadchunk.Len := priv_frame.FrameSize - (ProtocolAnchor.Offset + TrailerLen);

                                                  if priv_payloadchunk.Len > 0 then
                                                     begin
                                                          getmem (priv_payloadchunk.Data, priv_payloadchunk.Len);

                                                          help := pByte (priv_frame.Data) + ProtocolAnchor.Offset;
                                                          move (help^, priv_payloadchunk.Data^, priv_payloadchunk.Len);
                                                     end;

                                                  ParseComplete := true;
                                             end;
           else
               SetFCSAndTrailerValues (TrailerLen);

               priv_payloadchunk.Len := priv_frame.FrameSize - (ProtocolAnchor.Offset + TrailerLen);

               if priv_payloadchunk.Len > 0 then
                  begin
                       getmem (priv_payloadchunk.Data, priv_payloadchunk.Len);

                       help := pByte (priv_frame.Data) + ProtocolAnchor.Offset;
                       try
                          move (help^, priv_payloadchunk.Data^, priv_payloadchunk.Len);
                       except
                             SiMain.LogFatal ('Error: could not copy payload data while parsing frame #' + priv_frame.fileframeid.ToString + ' (tParseMaster.CreateParseList)');
                       end;
                  end;

               ParseComplete := true;
           end;

           // Make sure Frameparser is nil if there is nothing more to parse

           if ParseComplete then
              FrameParser := nil
           else
               begin
                    // Record current protocol type in the protocol chain array

                    inc (i);
                    setLength (priv_ParseList, i);

                    // Assign parser and anchor

                    priv_ParseList[i-1].anchor := ProtocolAnchor;
                    priv_ParseList[i-1].Parser := FrameParser;
                    priv_ParseList[i-1].CRCWasCorrect := false;  // Init
                    priv_ParseList[i-1].TunnelFlags := [];

                    FrameParser.ParentAnchor := ParentAnchor;

                    // Note the current anchor for later

                    ParentAnchor := ProtocolAnchor;

                    try
                       // First, check if the parser we created is plausible; this will only be checked "for real" in those
                       // parsers that are based on tcp/udp ports higher than 1024.

                       if FrameParser.checkIsPlausible then
                          begin
                               // --- Do the actual parsing ---

                               ParseResult := FrameParser.parse;

                               if FrameParser.getDiagnosticMsgCount > 0  then
                                  TransferDiagnosticMessages (FrameParser, ProtocolAnchor.HeaderType);
                          end
                       else
                           begin
                                // Bypass parsing and treat the rest as unparsable payload

                                ProtocolAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                                ParseResult := prFailPlausibleCheck;
                           end;

                       // Continue if parsing was okay, including if the further payload of the current layer needs to be ignored

                       if (ParseResult = prSuccess) or (ParseResult = prFailPayloadTruncated) then
                          begin
                               // Do specific things depending on layer

                               case ProtocolAnchor.HeaderType of
                                NEXT_HEADER_ETHERNET,
                                NEXT_HEADER_TRANS_ETHER_BRIDGE : begin
                                                                      layerlistlen := length (priv_socketlayers) + 1;
                                                                      setLength (priv_socketlayers, layerlistlen);
                                                                      priv_socketlayers[layerlistlen-1].initialize;

                                                                      // Record in the layer list

                                                                      priv_EthernetPair.Source := (FrameParser as tEthernetParser).SourceAddress;
                                                                      priv_EthernetPair.Destination := (FrameParser as tEthernetParser).DestinationAddress;

                                                                      priv_socketlayers[layerlistlen-1].EthernetAddrPair.Source := priv_EthernetPair.Source;
                                                                      priv_socketlayers[layerlistlen-1].EthernetAddrPair.Destination := priv_EthernetPair.Destination;

                                                                      inc (priv_EtherLayerCount);

                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppEthernet];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_LINUX_COOKED       : begin
                                                                      layerlistlen := length (priv_socketlayers) + 1;
                                                                      setLength (priv_socketlayers, layerlistlen);
                                                                      priv_socketlayers[layerlistlen-1].initialize;

                                                                      // Create pseudo ethernet values if possible

                                                                      if (FrameParser as tLinuxSLLParser).MACAddress <> VALUE_ZERO_MAC then
                                                                         begin
                                                                              case (FrameParser as tLinuxSLLParser).PacketType of
                                                                               sptUnicastToUs : begin
                                                                                                     priv_EthernetPair.Source := VALUE_ZERO_MAC;
                                                                                                     priv_EthernetPair.Destination := (FrameParser as tLinuxSLLParser).MACAddress;
                                                                                                end;
                                                                               sptSentByUs    : begin
                                                                                                     priv_EthernetPair.Source := (FrameParser as tLinuxSLLParser).MACAddress;
                                                                                                     priv_EthernetPair.Destination := VALUE_ZERO_MAC;
                                                                                                end;
                                                                              else
                                                                                  priv_EthernetPair.Source := VALUE_ZERO_MAC;
                                                                                  priv_EthernetPair.Destination := VALUE_ZERO_MAC;
                                                                              end;
                                                                         end;

                                                                      priv_socketlayers[layerlistlen-1].EthernetAddrPair.Source := priv_EthernetPair.Source;
                                                                      priv_socketlayers[layerlistlen-1].EthernetAddrPair.Destination := priv_EthernetPair.Destination;

                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppLinuxCooked];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_802dot11Hdr        : begin
                                                                      // Get the frame subtype in case we need to parse a management frame

                                                                      WiFiSubType := (FrameParser as t802dot11HeaderParser).FrameSubType;
                                                                 end;
                                NEXT_HEADER_8021_Q,
                                NEXT_HEADER_802_QINQ_Outer,
                                NEXT_HEADER_802_QINQ_Outer_OLD1,
                                NEXT_HEADER_802_QINQ_Outer_OLD2,
                                NEXT_HEADER_802_QINQ_Outer_OLD3: begin

                                                                      priv_VLANID := (FrameParser as t802dot1qParser).VLANID;

                                                                      inc (priv_VLANCount);

                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppVLANTag];
                                                                 end;
                                NEXT_HEADER_ARP,
                                NEXT_HEADER_RARP               : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppARP];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_MPLS_UNICAST,
                                NEXT_HEADER_MPLS_MULTICAST     : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppMPLS];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelMPLS];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelMPLS];
                                                                 end;
                                NEXT_HEADER_GRE                : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppGRE];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelGRE];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelGRE];
                                                                 end;
                                NEXT_HEADER_GTPv1              : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppGTPv1];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelGTPv1];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelGTPv1];
                                                                 end;
                                NEXT_HEADER_AYIYA              : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppAYIYA];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelAYIYA];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelAYIYA];
                                                                 end;
                                NEXT_HEADER_VXLAN              : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppVXLAN];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelVXLAN];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelVXLAN];
                                                                 end;
                                NEXT_HEADER_ERSPAN             : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppERSPAN];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelERSPAN];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelERSPAN];
                                                                 end;
                                NEXT_HEADER_CAPWAP             : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppCAPWAP];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelCAPWAP];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelCAPWAP];
                                                                 end;
                                NEXT_HEADER_GENEVE             : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppGeneve];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelGeneve];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelGeneve];
                                                                 end;
                                NEXT_HEADER_TZSP               : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppTZSP];
                                                                      priv_ParseList[i-1].TunnelFlags := [tfTunnelLayer, tfTunnelTZSP];

                                                                      if i > 1 then
                                                                         priv_ParseList[i-2].TunnelFlags := [tfTunnelCarrier, tfTunnelTZSP];
                                                                 end;
                                NEXT_HEADER_IPv4_ENCAPSULATION,
                                NEXT_HEADER_IPv4               : begin
                                                                      layerlistlen := length (priv_socketlayers);

                                                                      // If this is the first IP layer following an ethernet layer we're good, otherwise
                                                                      // we need to add a layer first

                                                                      if (layerlistlen = 0) or (priv_socketlayers[layerlistlen-1].IPVersion <> 0)  then
                                                                         begin
                                                                              layerlistlen := layerlistlen + 1;
                                                                              setLength (priv_socketlayers, layerlistlen);
                                                                              priv_socketlayers[layerlistlen-1].initialize;
                                                                         end;

                                                                      // Store the IP version

                                                                      if (priv_IPLayerCount = 0) then
                                                                         priv_firstIPLayerVersion := 4;

                                                                      priv_lastIPLayerVersion := 4;

                                                                      // Transfer other interesting fields

                                                                      priv_IPID := (FrameParser as tIPv4Parser).Identification;

                                                                      priv_IPv4SocketPair.Source.IPAddress := (FrameParser as tIPv4Parser).Source;
                                                                      priv_IPv4SocketPair.Destination.IPAddress := (FrameParser as tIPv4Parser).Destination;

                                                                      priv_Layer4Protocol := (FrameParser as tIPv4Parser).Protocol;

                                                                      // TODO: check if socketlayers use is correct; I'm not sure if a new layer needs to be added here

                                                                      // priv_socketlayers[layerlistlen].EthernetAddrPair.initialize;
                                                                      priv_socketlayers[layerlistlen-1].IPVersion := 4;
                                                                      priv_socketlayers[layerlistlen-1].IPv4SocketPair := priv_IPv4SocketPair;
                                                                      priv_socketlayers[layerlistlen-1].IPv6SocketPair.Initialize;

                                                                      // Record the IP payload length; needed for parsing higher layers
                                                                      // also, find out if there's a trailer

                                                                      TrailerLen := (FrameParser as tIPv4Parser).TotalLength;
                                                                      IPPayloadLen := (FrameParser as tIPv4Parser).IPPayloadLength;

                                                                      TrailerLen := priv_frame.WireSize - (ProtocolAnchor.Offset + TrailerLen);

                                                                      // For truncated ICMP payloads we can have negative Trailerlengths, which will cause havoc on assembly,
                                                                      // so make sure we do not have those

                                                                      if TrailerLen < 0 then
                                                                         TrailerLen := 0;

                                                                      inc (priv_IPLayerCount);

                                                                      // Do CRC Stuff

                                                                      IP4Pseudo := (FrameParser as tIPv4Parser).PseudoHeader;

                                                                      ChecksumActual := (FrameParser as tIPv4Parser).Checksum;
                                                                      ChecksumTest := getIPv4Checksum ((FrameParser as tIPv4Parser).HeaderAsWordBuffer, (FrameParser as tIPv4Parser).IHL);
                                                                      priv_ParseList[i-1].CRCWasCorrect := (ChecksumActual = ChecksumTest);
{
                                                                      // Mark as fragment if it is, unless it's an ICMP quoted header

                                                                      -> Moved to IPv4Parser

                                                                      if not isICMPQuote then
                                                                         priv_IsFragment := (FrameParser as tIPv4Parser).IsFragment;
}
                                                                      priv_isFragment := (FrameParser as tIPv4Parser).IsFragment;
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppIPv4];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_IPV6,
                                NEXT_HEADER_TYPE_IPV6          : begin
                                                                      layerlistlen := length (priv_socketlayers);

                                                                      // If this is the first IP layer following an ethernet layer we're good, otherwise
                                                                      // we need to add a layer first

                                                                      if (layerlistlen = 0) or (priv_socketlayers[layerlistlen-1].IPVersion <> 0)  then
                                                                         begin
                                                                              layerlistlen := layerlistlen + 1;
                                                                              setLength (priv_socketlayers, layerlistlen);
                                                                              priv_socketlayers[layerlistlen-1].initialize;
                                                                         end;

                                                                      // Store the IP version

                                                                      if (priv_IPLayerCount = 0) then
                                                                         priv_firstIPLayerVersion := 6;

                                                                      priv_lastIPLayerVersion := 6;

                                                                      priv_IPv6SocketPair.Source.IPAddress := (FrameParser as tIPv6Parser).Source;
                                                                      priv_IPv6SocketPair.Destination.IPAddress := (FrameParser as tIPv6Parser).Destination;

                                                                      NextHDR := (FrameParser as tIPv6Parser).NextHeader;

                                                                      case NextHdr of
                                                                       6, 17, 58 : priv_Layer4Protocol := NextHdr;
                                                                      else
                                                                          priv_Layer4Protocol := 0;   // TODO: need to parse IPv6 Extension headers to find the correct value
                                                                      end;

                                                                      // priv_socketlayers[layerlistlen].EthernetAddrPair.initialize;
                                                                      priv_socketlayers[layerlistlen-1].IPVersion := 6;
                                                                      priv_socketlayers[layerlistlen-1].IPv6SocketPair := priv_IPv6SocketPair;
                                                                      priv_socketlayers[layerlistlen-1].IPv4SocketPair.Initialize;

                                                                      // Record the IP payload length; needed for parsing higher layers

                                                                      IPPayloadLen := (FrameParser as tIPv6Parser).PayloadLength;

                                                                      inc (priv_IPLayerCount);

                                                                      // See if there is a trailing FCS or other stuff

                                                                      TrailerLen := (FrameParser as tIPv6Parser).PayloadLength + 40;
                                                                      TrailerLen := priv_frame.WireSize - (ProtocolAnchor.Offset + TrailerLen);

                                                                      // For truncated ICMP payloads we can have negative Trailerlengths, which will cause havoc on assembly,
                                                                      // so make sure we do not have those

                                                                      if TrailerLen < 0 then
                                                                         TrailerLen := 0;

                                                                      // Do CRC stuff - can't calculate CRC because IPv6 doesn't have any

                                                                      IP6Pseudo := (FrameParser as tIPv6Parser).PseudoHeader;

                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppIPv6];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_ICMP               : begin
                                                                      ChecksumActual := (FrameParser as tICMPv4Parser).Checksum;
                                                                      ChecksumTest := getICMPv4Checksum ((FrameParser as tICMPv4Parser).HeaderAsWordBuffer, (FrameParser as tICMPv4Parser).PayloadAsWordBuffer);
                                                                      priv_ParseList[i-1].CRCWasCorrect := (ChecksumActual = ChecksumTest);

                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppICMPv4];

                                                                      priv_socketlayers[layerlistlen-1].Layer4NextProtocol := l4ICMP;

                                                                      // Note that following headers are quotes

                                                                      IsICMPQuote := True;

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_IPV6_ICMP          : begin
                                                                      ChecksumActual := (FrameParser as tICMPv6Parser).Checksum;
                                                                      ChecksumTest := getICMPv6Checksum (IP6Pseudo, (FrameParser as tICMPv6Parser).HeaderAsWordBuffer, (FrameParser as tICMPv6Parser).PayloadAsWordBuffer);
                                                                      priv_ParseList[i-1].CRCWasCorrect := (ChecksumActual = ChecksumTest);

                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppICMPv6];

                                                                      priv_socketlayers[layerlistlen-1].Layer4NextProtocol := l4ICMP;

                                                                      // Note that following headers are quotes

                                                                      IsICMPQuote := True;

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_TCP                : begin
                                                                      L4PayloadLen := IPPayloadLen - Frameparser.LayerSize;

                                                                      CRCBuffer :=  (FrameParser as tTCPParser).TCPSectionAsWordBuffer;

                                                                      if priv_lastIPLayerVersion = 4 then
                                                                         begin
                                                                              priv_IPv4SocketPair.Source.Port := (FrameParser as tTCPParser).SourcePort;
                                                                              priv_IPv4SocketPair.Destination.Port := (FrameParser as tTCPParser).DestinationPort;

                                                                              priv_socketlayers[layerlistlen-1].IPv4SocketPair.Source.Port := priv_IPv4SocketPair.Source.Port;
                                                                              priv_socketlayers[layerlistlen-1].IPv4SocketPair.Destination.Port := priv_IPv4SocketPair.Destination.Port;

                                                                              Tuple.initialize (priv_socketlayers[layerlistlen-1].IPv4SocketPair, l4TCP);

                                                                              // Verify CRC

                                                                              ChecksumActual := (FrameParser as tTCPParser).Checksum;
                                                                              ChecksumTest := getTCPChecksum (IP4Pseudo, CRCBuffer);

                                                                              priv_ParseList[i-1].CRCWasCorrect := (ChecksumActual = ChecksumTest);
                                                                         end
                                                                      else
                                                                          begin
                                                                              priv_IPv6SocketPair.Source.Port := (FrameParser as tTCPParser).SourcePort;
                                                                              priv_IPv6SocketPair.Destination.Port := (FrameParser as tTCPParser).DestinationPort;

                                                                              priv_socketlayers[layerlistlen-1].IPv6SocketPair.Source.Port := priv_IPv6SocketPair.Source.Port;
                                                                              priv_socketlayers[layerlistlen-1].IPv6SocketPair.Destination.Port := priv_IPv6SocketPair.Destination.Port;

                                                                              Tuple.initialize (priv_socketlayers[layerlistlen-1].IPv6SocketPair, l4TCP);

                                                                              // Verify CRC

                                                                              ChecksumActual := (FrameParser as tTCPParser).Checksum;
                                                                              ChecksumTest := getTCPChecksum (IP6Pseudo, CRCBuffer);

                                                                              priv_ParseList[i-1].CRCWasCorrect := (ChecksumActual = ChecksumTest);
                                                                          end;

                                                                      priv_TCPSequenceNo := (FrameParser as tTCPParser).SequenceNumber;
                                                                      priv_TCPAcknowledgeNo := (FrameParser as tTCPParser).AcknowledgeNumber;
                                                                      priv_TCPFlags := (FrameParser as tTCPParser).Flags;
                                                                      priv_TCPPayloadLength := (FrameParser as tTCPParser).PayloadLength;

                                                                      // Determine next header in case it is dynamic (meaning, payload len > 0)

                                                                      if FrameParser.NextAnchor.HeaderType = NEXT_HEADER_DYNAMIC then
                                                                         begin
                                                                              HelpAnchor.HeaderType := determineNextAnchorFromSocket (NEXT_HEADER_TCP);
                                                                              HelpAnchor.Offset := FrameParser.NextAnchor.Offset;

                                                                              // We need the HelpAnchor because objects do not allow to assign to record property sub elements

                                                                              FrameParser.NextAnchor := HelpAnchor;
                                                                         end;

                                                                       // Record in layer list and protocol set

                                                                       priv_socketlayers[layerlistlen-1].Layer4NextProtocol := l4TCP;

                                                                       priv_parsedprotocolset := priv_parsedprotocolset + [ppTCP];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_UDP                : begin
                                                                      L4PayloadLen := IPPayloadLen - Frameparser.LayerSize;

                                                                      CRCBuffer := (FrameParser as tUDPParser).UDPSectionAsWordBuffer;

                                                                      if priv_lastIPLayerVersion = 4 then
                                                                         begin
                                                                              priv_IPv4SocketPair.Source.Port := (FrameParser as tUDPParser).SourcePort;
                                                                              priv_IPv4SocketPair.Destination.Port := (FrameParser as tUDPParser).DestinationPort;

                                                                              priv_socketlayers[layerlistlen-1].IPv4SocketPair.Source.Port := priv_IPv4SocketPair.Source.Port;
                                                                              priv_socketlayers[layerlistlen-1].IPv4SocketPair.Destination.Port := priv_IPv4SocketPair.Destination.Port;

                                                                              Tuple.initialize (priv_socketlayers[layerlistlen-1].IPv4SocketPair, l4UDP);

                                                                              ChecksumActual := (FrameParser as tUDPParser).Checksum;
                                                                              ChecksumTest := getUDPChecksum (IP4Pseudo, CRCBuffer);

                                                                              priv_ParseList[i-1].CRCWasCorrect := (ChecksumActual = ChecksumTest);
                                                                         end
                                                                      else
                                                                          begin
                                                                              priv_IPv6SocketPair.Source.Port := (FrameParser as tUDPParser).SourcePort;
                                                                              priv_IPv6SocketPair.Destination.Port := (FrameParser as tUDPParser).DestinationPort;

                                                                              priv_socketlayers[layerlistlen-1].IPv6SocketPair.Source.Port := priv_IPv6SocketPair.Source.Port;
                                                                              priv_socketlayers[layerlistlen-1].IPv6SocketPair.Destination.Port := priv_IPv6SocketPair.Destination.Port;

                                                                              Tuple.initialize (priv_socketlayers[layerlistlen-1].IPv6SocketPair, l4UDP);

                                                                              ChecksumActual := (FrameParser as tUDPParser).Checksum;
                                                                              ChecksumTest := getUDPChecksum (IP6Pseudo, CRCBuffer);

                                                                              priv_ParseList[i-1].CRCWasCorrect := (ChecksumActual = ChecksumTest);
                                                                          end;

                                                                      if FrameParser.NextAnchor.HeaderType = NEXT_HEADER_DYNAMIC then
                                                                         begin
                                                                              HelpAnchor.HeaderType := determineNextAnchorFromSocket (NEXT_HEADER_UDP);
                                                                              HelpAnchor.Offset := FrameParser.NextAnchor.Offset;

                                                                              // We need the HelpAnchor because objects do not allow to assign to record property sub elements

                                                                              FrameParser.NextAnchor := HelpAnchor;
                                                                         end;

                                                                      // Record in layer list and protocol set

                                                                      priv_socketlayers[layerlistlen-1].Layer4NextProtocol := l4UDP;

                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppUDP];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_DNS                : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppDNS];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_DHCPv4             : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppDHCPv4];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_HTTP               : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppHTTP];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                                NEXT_HEADER_RTPS               : begin
                                                                      priv_parsedprotocolset := priv_parsedprotocolset + [ppRTPS];

                                                                      // Update offset table

                                                                      priv_LayerOffsetList[FrameParser.OSILayer].StartOffset := ProtocolAnchor.Offset;
                                                                      priv_LayerOffsetList[FrameParser.OSILayer].NextLayerOffset := FrameParser.NextAnchor.Offset;
                                                                 end;
                               end;

                               NextLayerOSI := DetermineNextLayerOSI (FrameParser.NextAnchor.HeaderType);

                               // Check tunnel stuff

                               if i > 1 then
                                  begin
                                       // if the previous layer is a tunnel mechanism, mark the current as tunnel payload

                                       if (tfTunnelLayer in priv_ParseList[i-2].TunnelFlags) or (tfTunnelPayload in priv_ParseList[i-2].TunnelFlags) then
                                          priv_ParseList[i-1].TunnelFlags := priv_ParseList[i-1].TunnelFlags + [tfTunnelPayload];
                                  end;

                               // Determine next parser setting depending on parse mode

                               case priv_parsemode of
                                ppmSanitize   : begin
                                                     // If we do not have an assembler for that protocol, set the next anchor to payload
                                                     // but keep the correct anchor for the parser itself

                                                     if NoAssemblerExists (FrameParser.NextAnchor.HeaderType) then
                                                        begin
                                                             ProtocolAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                                                             ProtocolAnchor.Offset := FrameParser.NextAnchor.Offset;
                                                        end
                                                     else
                                                         ProtocolAnchor := FrameParser.NextAnchor;
                                                end;
                                ppmUpToLayer2 : begin
                                                     // If the parse mode is set to stop after layer 2 (Ethernet etc.), mark the rest as payload

                                                     if (FrameParser.OSILayer = 2) and (NextLayerOSI > 2) then
                                                        begin
                                                             ProtocolAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                                                             ProtocolAnchor.Offset := FrameParser.NextAnchor.Offset;
                                                        end
                                                     else
                                                         ProtocolAnchor := FrameParser.NextAnchor;
                                                end;
                                ppmUpToLayer3 : begin
                                                     // If the parse mode is set to stop after layer 3 (IPv4/IPv6), mark the rest as payload

                                                     if (FrameParser.OSILayer = 3) and (NextLayerOSI > 3) then
                                                        begin
                                                             ProtocolAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                                                             ProtocolAnchor.Offset := FrameParser.NextAnchor.Offset;
                                                        end
                                                     else
                                                         ProtocolAnchor := FrameParser.NextAnchor;
                                                end;
                                ppmUpToLayer4 : begin
                                                     // If the parse mode is set to stop after layer 4 (TCP/UDP), mark the rest as payload

                                                     if (FrameParser.OSILayer = 4) and (NextLayerOSI > 4) and not (NextProtocolIsTunnelLayer (FrameParser.NextAnchor.HeaderType)) then
                                                        begin
                                                             ProtocolAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                                                             ProtocolAnchor.Offset := FrameParser.NextAnchor.Offset;
                                                        end
                                                     else
                                                         ProtocolAnchor := FrameParser.NextAnchor;
                                                end;
                               else
                                   ProtocolAnchor := FrameParser.NextAnchor;
                               end;
                          end
                       else
                           begin
                                if ParseResult = prFailInvalidValues then
                                   SiMain.LogError ('Invalid values while parsing frame #' + priv_frame.FileFrameID.ToString + ' of file ' + priv_frame.TracefileID.ToString);

                                // Bypass parsing and treat the rest as unparsable payload

                                ProtocolAnchor.HeaderType := NEXT_HEADER_PAYLOAD;

                                // We also need to void the last parser we put in the list

                                setLength (priv_ParseList, length (priv_ParseList) - 1);
                           end;

                    except
                          SiMain.LogError ('Exception triggered while parsing frame #' + priv_frame.FileFrameID.ToString + ' of file ' + priv_frame.TracefileID.ToString);

                          inc (priv_ParseErrorCount);
                    end;
              end;

     until ParseComplete;

{$ifdef debugparser}

     SiMain.LogVerbose ('Parsed frame #' + intTostr (priv_frame.FileFrameID));

{$endif}
end;

(* -------------------------------------- *)

function tParseMaster.getParserCount: integer;

begin
     result := length (priv_ParseList);
end;

(* -------------------------------------- *)

function tParseMaster.getParserTypeByIndex(Index: integer): tHeaderType;

begin
     if Index > length (priv_ParseList) then
        result := NEXT_HEADER_UNKNOWN
     else
         result := priv_ParseList[Index - 1].Anchor.HeaderType;
end;

(* -------------------------------------- *)

function tParseMaster.getParserByIndex (Index: integer): tFrameParser;

begin
     if Index > length (priv_ParseList) then
        result := nil
     else
         result := priv_ParseList[Index - 1].Parser;
end;

(* -------------------------------------- *)

function tParseMaster.getParserByProtocol (const Protocol: tParsedProtocols; const Occurrence : byte = 1): tFrameParser;

var ParserID  : integer;
    count     : integer;
    findcount : byte;
    found     : boolean;
    i         : integer;

begin
     Count := getParserCount;
     findcount := 0;
     found := False;

     if (count > 0) and (Protocol in priv_parsedprotocolset) then
        begin
             ParserID := ParserTypeFromProtocol (Protocol);

             if ParserID > 0 then
                begin
                     i := 0;

                     repeat
                           repeat
                                 inc (i)
                           until (getParserTypeByIndex (i) = ParserID) or (i >= Count);

                           // Check if we have the correct protocol and the right occurrence

                           if getParserTypeByIndex (i) = ParserID then
                              begin
                                   inc (findcount);

                                   if findcount >= Occurrence then
                                      found := True;
                              end;

                     until found or (i >= Count);

                     // Check if the index is still correct and a valid occurrence

                     if found and (getParserTypeByIndex (i) = ParserID) then
                        result := getParserByIndex (i)
                     else
                         result := nil;
                end
             else
                 raise Exception.Create('Error: Can''t find parser ID (tParseMaster.getParserByProtocol)');
        end
     else
         result := nil;
end;

(* -------------------------------------- *)

function tParseMaster.getParserCRCOkayByIndex(Index: integer): boolean;

begin
     if Index > length (priv_ParseList) then
        result := false
     else
         result := priv_ParseList[Index - 1].CRCWasCorrect;
end;

(* -------------------------------------- *)

function tParseMaster.getParserParseInfoByIndex (Index: integer): tParseEntry;

begin
     if Index > length (priv_ParseList) then
        result.initialize
     else
         result := priv_ParseList[Index - 1];
end;

(* -------------------------------------- *)

function tParseMaster.getHasRemainingPayload: boolean;

begin
     result := priv_payloadchunk.Len > 0;
end;

(* -------------------------------------- *)

function tParseMaster.getRemaningPayloadLength: integer;

begin
     result := priv_payloadchunk.Len;
end;

(* -------------------------------------- *)

function tParseMaster.getCopyOfRemainingPayload : tFrameChunk;

var chunk : tFrameChunk;

begin
     chunk.Len := priv_payloadchunk.Len;

     getmem (chunk.Data, chunk.Len);

     move (priv_payloadchunk.Data^, chunk.data^, chunk.Len);

     result := chunk;
end;

(* ------------------------------------------- *)

function tParseMaster.getCopyOfPaddingBytes : tFrameChunk;

var chunk : tFrameChunk;

begin
     chunk.Len := priv_payloadchunk.Len;

     getmem (chunk.Data, chunk.Len);
     chunk.copyFrom (priv_paddingbytechunk);

     result := chunk;
end;

(* ------------------------------------------- *)

function tParseMaster.getDiagnosticMsgCount : integer;

begin
     result := length (priv_diagnosticmsglist);
end;

(* ------------------------------------------- *)

function tParseMaster.getDiagnosticMsg (const Index : integer) : tDiagnosticMsg;

begin
     result := priv_diagnosticmsglist[Index - 1];
end;

(* ------------------------------------------- *)

procedure tParseMaster.clearDiagnosticMsgList;

begin
     setLength (priv_diagnosticmsglist, 0);
end;

(* -------------------------------------- *)

function tParseMaster.getVLANIDsAsString: string;

var value  : string;
    parser : t802dot1QParser;
    i      : integer;

begin
     if priv_VLANCount > 0 then
        begin
             if priv_VLANCount = 1 then
                value := IntToStr (priv_VLANID)
             else
                 begin
                      value := '';

                      // We need to get all VLAN IDs from the parsers

                      for i := 1 to priv_VLANCount do
                          begin
                               parser := getParserByProtocol (ppVLANTag, i) as t802dot1qParser;

                               if assigned (parser) then
                                  value := value + ' ' + IntToStr (parser.VLANID);
                          end;

                      // Reformat the string

                      value := Trim (value);
                      value := StringReplace (value, ' ', '-', [rfReplaceAll]);
                 end;
        end
     else
         value := '';

     result := value;
end;

(* -------------------------------------- *)


end.

