{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_802dot11_radiotap;

interface

uses base_structures,
     base_checksums,
     common_parse_helpers,
     class_parser_baseclass;

type t802dot11HeaderFlag = (rtTSFT, rtFlags, rtRate, rtChannel, rtFHSS, rtdBmAntennaSignal, rtdBmAntennaNoise, rtLockQuality, rtTXAttenuation, rtdbTXAttenuation, rtdBmTXPower, rtAntenna, rtdbAntennaSignal,
                             rtdbAntennaNoise, rtRXflags, rtChannelPlus, rtHTinformation, rtAMPDUStatus, rtVHTinformation, rtRadioNSTapNext, rtVendorNSNext, rtExt);
     t802dot11HeaderFlagSet = set of t802dot11HeaderFlag;

     t802dot11OptionalFlag = (rtofCFP, rtofPreamble, rtofWEP, rtofFragmentation, rtofFCSAtEnd, rtofDataPad, rtofBadFCS, rtofShortGI);
     t802dot11OptionalFlagSet = set of t802dot11OptionalFlag;

     t802dot11OptionalChannelTypeFlag = (rtcfTurbo, rtcfCCK, rtcfOFDM, rtcf2GHZ, rtcf5GHz, rtcfPassive, rtcfDynCCKOFDM, rtcfGFSK, rtcfGSM, rtcfStaticTurbo, rtcfHalfRate, rtcfQuarterRate);
     t802dot11OptionalChannelTypeFlagSet = set of t802dot11OptionalChannelTypeFlag;

     t802dot11RadioTapHeader = packed record
                                            HeaderVersion    : byte;
                                            HeaderPadding    : byte;
                                            HeaderLength     : word;
                                            HeaderFlags      : longword;
                               end;
     p802dot11RadioTapHeader = ^t802dot11RadioTapHeader;

     t802dot11RadioTapParser = class (tFrameParser)

                         private
                                priv_Version               : byte;
                                priv_HeaderLen             : word;
                                priv_HeaderFlags           : t802dot11HeaderFlagSet;

                                priv_optTSFT               : uint64;
                                priv_optFlags              : t802dot11OptionalFlagSet;
                                priv_optDataRateMbps       : double;
                                priv_optChannelFrequency   : word;
                                priv_optChannelFlags       : t802dot11OptionalChannelTypeFlagSet;
                                priv_FHSS                  : byte;
                                priv_optdbmAntennaSignal   : shortint;
                                priv_optdbmAntennaNoise    : shortint;
                                priv_optSignalQuality      : word;
                                priv_optAntennaIndex       : byte;
                                priv_optdbAntennaSignal    : shortint;
                                priv_optdbAntennaNoise     : shortint;

                                procedure clearOptionalValues;

                                function getHeaderFlagsFromLong (const FlagValues : longword): t802dot11HeaderFlagSet;
                                function getOptionalFlagsFromByte (const FlagValues : byte): t802dot11OptionalFlagSet;
                                function getOptionalChannelFlagsFroMWord (const Flagvalues : word): t802dot11OptionalChannelTypeFlagSet;

                                function parseOptionalValues (const Offset : integer) : integer;

                         public
                               property Version : byte read priv_version;
                               property HeaderLength : word read priv_headerlen;
                               property HeaderFlags : t802dot11HeaderFlagSet read priv_headerflags;

                               property TSFT : uint64 read priv_optTSFT;
                               property OptionalFlags : t802dot11OptionalFlagSet read priv_optflags;
                               property DataRateMBps : double read priv_optdatarateMbps;
                               property ChannelFrequency : word read priv_optChannelFrequency;
                               property ChannelFlags : t802dot11OptionalChannelTypeFlagSet read priv_optChannelFlags;
                               property FHSS : byte read priv_FHSS;
                               property dbmAntennaSignal : shortint read priv_optdbmAntennaSignal;
                               property dbmAntennaNoise : shortint read priv_optdbmAntennaNoise;
                               property SignalQuality : word read priv_optSignalQuality;
                               property Antenna : byte read priv_optAntennaIndex;
                               property dbAntennaSignal : shortint read priv_optdbAntennaSignal;
                               property dbAntennaNoise : shortint read priv_optdbAntennaNoise;

                               function checkIsPlausible : boolean; override;
                               function parse : tParseResult; override;
                   end;


implementation

uses Toolkit;

(* ------------------------------------------- *)
{ t802dot11RadioTapParser }
(* ------------------------------------------- *)

function t802dot11RadioTapParser.checkIsPlausible: boolean;

begin
     // if we get here, it's by link layer type decision, so its okay

     result := true;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapParser.getHeaderFlagsFromLong (const FlagValues: longword): t802dot11HeaderFlagSet;

var isSet   : boolean;
    flagset : t802dot11HeaderFlagSet;
    flag    : t802dot11HeaderFlag;
    i       : integer;

begin
     flagset := [];

     for i := 0 to 31 do
         begin
              isSet := BitIsSet (FlagValues, i);

              if isSet then
                 begin
                      case i of
                        0 : flag := rtTSFT;
                        1 : flag := rtFlags;
                        2 : flag := rtRate;
                        3 : flag := rtChannel;
                        4 : flag := rtFHSS;
                        5 : flag := rtdBmAntennaSignal;
                        6 : flag := rtdBmAntennaNoise;
                        7 : flag := rtLockQuality;
                        8 : flag := rtdbTXAttenuation;
                        9 : flag := rtTXAttenuation;
                       10 : flag := rtdBmTXPower;
                       11 : flag := rtAntenna;
                       12 : flag := rtdbAntennaSignal;
                       13 : flag := rtdbAntennaNoise;
                       14 : flag := rtRXflags;
                       18 : flag := rtChannelPlus;
                       19 : flag := rtHTinformation;
                       20 : flag := rtAMPDUStatus;
                       21 : flag := rtVHTinformation;
                       29 : flag := rtRadioNSTapNext;
                       30 : flag := rtVendorNSNext;
                       31 : flag := rtExt;
                      end;

                      flagset := flagset + [flag];
                 end;
         end;

     result := flagset;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapParser.getOptionalFlagsFromByte (const FlagValues: byte): t802dot11OptionalFlagSet;

var isSet   : boolean;
    flagset : t802dot11OptionalFlagSet;
    flag    : t802dot11OptionalFlag;
    i       : integer;

begin
     flagset := [];

     for i := 0 to 7 do
         begin
              isSet := BitIsSet (FlagValues, i);

              if isSet then
                 begin
                      case i of
                        0 : flag := rtofCFP;
                        1 : flag := rtofPreamble;
                        2 : flag := rtofWEP;
                        3 : flag := rtofFragmentation;
                        4 : flag := rtofFCSAtEnd;
                        5 : flag := rtofDataPad;
                        6 : flag := rtofBadFCS;
                        7 : flag := rtofShortGI;
                      end;

                      flagset := flagset + [flag];
                 end;
         end;

     result := flagset;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapParser.getOptionalChannelFlagsFroMWord (const Flagvalues: word): t802dot11OptionalChannelTypeFlagSet;

var isSet   : boolean;
    flagset : t802dot11OptionalChannelTypeFlagSet;
    flag    : t802dot11OptionalChannelTypeFlag;
    i       : integer;

begin
     flagset := [];

     for i := 4 to 15 do
         begin
              isSet := BitIsSet (FlagValues, i);

              if isSet then
                 begin
                      case i of
                        4 : flag := rtcfTurbo;
                        5 : flag := rtcfCCK;
                        6 : flag := rtcfOFDM;
                        7 : flag := rtcf2GHZ;
                        8 : flag := rtcf5GHz;
                        9 : flag := rtcfPassive;
                       10 : flag := rtcfDynCCKOFDM;
                       11 : flag := rtcfGFSK;
                       12 : flag := rtcfGSM;
                       13 : flag := rtcfStaticTurbo;
                       14 : flag := rtcfHalfRate;
                       15 : flag := rtcfQuarterRate;
                      end;

                      flagset := flagset + [flag];
                 end;
         end;

     result := flagset;
end;

(* ------------------------------------------- *)

procedure t802dot11RadioTapParser.clearOptionalValues;

begin
     priv_optflags := [];
     priv_optdatarateMbps := 0;
     priv_optChannelFrequency := 0;
     priv_optChannelFlags := [];
end;

(* ------------------------------------------- *)

function t802dot11RadioTapParser.parseOptionalValues (const Offset: integer) : integer;

var opttotallen : integer;
    flag        : t802dot11HeaderFlag;
    bHelp       : byte;
    sHelp       : shortint;
    wHelp       : word;
    u64Help     : uint64;
    localoffset : integer;

begin
     opttotallen := 0;
     localoffset := Offset;

     clearOptionalValues;

     // So far I only implement those flags I have traces for

     if rtTSFT in priv_headerflags then
        begin
             // Value in microseconds of the MACs 64-bit 802.11 Time Synchronization Function timer

             u64Help := ParseIn64FromOffset (LocalOffset);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             priv_optTSFT := u64Help;
        end;

     if rtFlags in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_FLAGS

             bHelp := ParseByteFromOffset (LocalOffset);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             // Got the byte, now get the flags

             priv_optflags := getOptionalFlagsFromByte (bHelp);
        end;

     if rtRate in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_RATE

             bHelp := ParseByteFromOffset (LocalOffset);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             // Got the byte, now calculate the MBps (the byte tells the blocks of 500kb)

             priv_optdatarateMbps := (500 * bHelp) / 1000;
        end;

     if rtChannel in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_CHANNEL

             wHelp := ParseWordFromOffset (LocalOffset, false, true);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 2);

             priv_optChannelFrequency := wHelp;

             // IEEE80211_RADIOTAP_RX_FLAGS (or TX, not sure how to tell yet)

             wHelp := ParseWordFromOffset (LocalOffset, false, true);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 2);

             priv_optChannelFlags := getOptionalChannelFlagsFroMWord (wHelp);
        end;

     if rtFHSS in priv_headerflags then
        begin
             bHelp := ParseByteFromOffset (LocalOffset);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             priv_FHSS := bHelp;
        end;

     if rtdBmAntennaSignal in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_DBM_ANTSIGNAL (decibels from one milliwatt (dBm))

             sHelp := shortint (ParseByteFromOffset (LocalOffset));

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             priv_optdbmAntennaSignal := sHelp;
        end;

     if rtdBmAntennaNoise in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_DBM_ANTSIGNAL (decibels from one milliwatt (dBm))

             sHelp := shortint (ParseByteFromOffset (LocalOffset));

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             priv_optdbmAntennaNoise := sHelp;
        end;

     if rtLockQuality in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_LOCK_QUALITY

             wHelp := ParseWordFromOffset (LocalOffset, false, true);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 2);

             priv_optSignalQuality := wHelp;
        end;

     if rtAntenna in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_ANTENNA

             bHelp := ParseByteFromOffset (LocalOffset);

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             priv_optAntennaIndex := bHelp;
        end;

     if rtdbAntennaSignal in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_DB_ANTSIGNAL (decibels from one milliwatt (dBm))

             sHelp := shortint (ParseByteFromOffset (LocalOffset));

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             priv_optdbAntennaSignal := sHelp;
        end;

     if rtdbAntennaNoise in priv_headerflags then
        begin
             // IEEE80211_RADIOTAP_DB_ANTNOISE (decibels from one milliwatt (dBm))

             sHelp := shortint (ParseByteFromOffset (LocalOffset));

             if prot_lasterrorstatus = prFailUndersize then
                exit (LocalOffset - 1);

             priv_optdbAntennaNoise := sHelp;
        end;


     // More options later



     // Calculate Option Length

     result := localoffset - Offset;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapParser.parse: tParseResult;

var status    : tParseResult;
    pData     : p802dot11RadioTapHeader;
    hdrlen    : integer;
    optlen    : integer;

begin
     priv_osilayer := 1;     // Set Layer this parser is for
     prot_layersize := sizeOf (t802dot11RadioTapHeader);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (t802dot11RadioTapHeader) then
        status := prFailUndersize
     else
         try
            hdrlen := sizeOf (t802dot11RadioTapHeader);

            pData := p802dot11RadioTapHeader (priv_framedata + priv_frameoffset);

            priv_version := pData^.HeaderVersion;
            priv_headerlen := pData^.HeaderLength;

            priv_headerflags := getHeaderFlagsFromLong (pData^.HeaderFlags);

            // If the extended flag is set there is another 32 bit flag set present.
            // Lets skip it because we don't have info how to parse it

            if rtExt in priv_headerflags then
               inc (hdrlen, 4);

            // Parse options based on what the flags tell us is present

            optlen := parseOptionalValues (priv_frameoffset + hdrlen);
            prot_layersize := hdrlen + optlen;

            // Determine next anchor

            priv_nextAnchor.HeaderType := NEXT_HEADER_802dot11Hdr;
            priv_nextAnchor.Offset := priv_frameoffset + prot_layersize;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
