{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_802dot11header;

interface

uses base_structures,
     base_checksums,
     common_parse_helpers,
     definition_ip_addresses,
     class_parser_baseclass;

type t802dot11HeaderFCFFlag = (fcfIBSS, fcfToAP, fcfFromAP, fcfWDS, fcfMoreFragments, fcfRetry, fcfPwrMgmt, fcfMoreData, fcfProtected, fcfOrdered);
     t802dot11HeaderFCFFlagSet = set of t802dot11HeaderFCFFlag;

     t802dot11FrameType = (ft802None, ft802Management, ft802Control, ft802Data);

     t802dot11MgmtDataHeaderStruct = packed record
                                                  HeaderFrameControl    : byte;
                                                  HeaderFCFFlags        : byte;
                                                  HeaderDuration        : word;
                                                  HeaderAddress1        : tMACBytes;
                                                  HeaderAddress2        : tMACBytes;
                                                  HeaderAddress3        : tMACBytes;
                                                  HeaderFragmentSeq     : word;
                                     end;
     p802dot11MgmtDataHeaderStruct = ^t802dot11MgmtDataHeaderStruct;

     t802dot11CtrlHeaderStruct = packed record
                                              HeaderFrameControl    : byte;
                                              HeaderFCFFlags        : byte;
                                              HeaderDuration        : word;
                                              ReceiverAddress1      : tMACBytes;
                                 end;
     p802dot11CtrlHeaderStruct = ^t802dot11CtrlHeaderStruct;

     t802dot11HeaderParser = class (tFrameParser)

                                   private
                                          priv_isOverCAPWAP     : boolean;

                                          priv_Version          : byte;
                                          priv_FrameType        : t802dot11FrameType;
                                          priv_FrameSubType     : t802dot11FrameSubType;

                                          priv_HeaderFlags      : t802dot11HeaderFCFFlagSet;
                                          priv_Duration         : word;

                                          priv_receiveraddr     : tMACAddress;
                                          priv_destinationaddr  : tMACAddress;
                                          priv_transmitteraddr  : tMACAddress;
                                          priv_BSSId            : tMACAddress;
                                          priv_sourceaddr       : tMACAddress;
                                          priv_FragmentNumber   : word;
                                          priv_SequenceNumber   : word;
                                          priv_HTControl        : longword;   // this is simple an unparsed 32 bit value (mostly flags) as of now

                                          function getFrameSubTypeFromByte (const Value : byte) : t802dot11FrameSubType;
                                          function getMgmtHeaderFlagsFromByte (const FlagValues : byte): t802dot11HeaderFCFFlagSet;

                                          function ParseManagementHeader (const Offset : integer) : tParseResult;
                                          function ParseControlHeader (const Offset : integer) : tParseResult;
                                          function ParseDataHeader (const Offset : integer) : tParseResult;

                                   public
                                         property Version : byte read priv_version;
                                         property FrameType : t802dot11FrameType read priv_FrameType;
                                         property FrameSubType : t802dot11FrameSubType read priv_FrameSubType;
                                         property HeaderFlags : t802dot11HeaderFCFFlagSet read priv_HeaderFlags;
                                         property Duration : word read priv_Duration write priv_Duration;

                                         property ReceiverAddress : tMACAddress read priv_receiveraddr;
                                         property DestinationAddress : tMACAddress read priv_destinationaddr;
                                         property TransmitterAddress : tMACAddress read priv_transmitteraddr;
                                         property SourceAddress : tMACAddress read priv_sourceaddr;
                                         property BSSId : tMACAddress read priv_BSSId;
                                         property FragmentNumber : word read priv_fragmentnumber;
                                         property SequenceNumber : word read priv_SequenceNumber;
                                         property HTControl : longword read priv_HTControl;

                                         property isOverCAPWAP : boolean read priv_isOverCAPWAP;

                                         function checkIsPlausible : boolean; override;
                                         function parse : tParseResult; override;
                             end;


implementation

uses Toolkit;

(* ------------------------------------------- *)
{ t802dot11HeaderParser }
(* ------------------------------------------- *)

function t802dot11HeaderParser.checkIsPlausible: boolean;

begin
     // if we get here, it's by link layer type decision (well, RadioTap, actually), so its okay

     result := true;
end;

(* ------------------------------------------- *)

function t802dot11HeaderParser.getFrameSubTypeFromByte (const Value: byte): t802dot11FrameSubType;

begin
     case priv_FrameType of
      ft802Management : begin
                             case Value of
                                0 : result := mgmtftAssocReq;
                                1 : result := mgmtftAssocResp;
                                2 : result := mgmtftReassocReq;
                                3 : result := mgmtftReassocResp;
                                4 : result := mgmtftProbeReq;
                                5 : result := mgmtftProbeResp;
                                8 : result := mgmtftBeacon;
                                9 : result := mgmtftATIM;
                               10 : result := mgmtftDisassoc;
                               11 : result := mgmtftAuth;
                               12 : result := mgmtftDeAuth;
                               13 : result := mgmtftAction;
                               14 : result := mgmtftActionNoAck;
                             end;
                        end;
      ft802Control    : begin
                             case Value of
                              0..6 : result := ctrlftNone;
                                 7 : result := ctrlWrapper;
                                 8 : result := ctrlftBlockAckReq;
                                 9 : result := ctrlftBlockAck;
                                10 : result := ctrlftPSPoll;
                                11 : result := ctrlftRTS;
                                12 : result := ctrlftCTS;
                                13 : result := ctrlftACK;
                                14 : result := ctrlftCFend;
                                15 : result := ctrlftCFEndAck;
                              end;
                        end;
      ft802Data       : begin
                             case Value of
                                0 : result := dftData;
                                1 : result := dftDataCFAck;
                                2 : result := dftDataCFPoll;
                                3 : result := dftDataCFAckCFPoll;
                                4 : result := dftNull;
                                5 : result := dftCFAck;
                                6 : result := dftCFPoll;
                                7 : result := dftCFAckCFPoll;
                                8 : result := dftQoSData;
                                9 : result := dftQoSDataCFAck;
                               10 : result := dftQoSDataCFPoll;
                               11 : result := dftQoSCFAckCFPoll;
                               12 : result := dftQoSNull;
                               13 : result := dftNone;
                               14 : result := dftQoSCFPollNoData;
                               15 : result := dftQoSCFAckNoData;
                             end;
                        end;
     end;
end;

(* ------------------------------------------- *)

function t802dot11HeaderParser.getMgmtHeaderFlagsFromByte (const FlagValues: byte) : t802dot11HeaderFCFFlagSet;

var isSet   : boolean;
    flagset : t802dot11HeaderFCFFlagSet;
    flag    : t802dot11HeaderFCFFlag;
    help    : byte;
    i       : integer;

begin
     // Check the directional bits

     help := FlagValues and $3;

     case help of
      0 : flagset := [fcfIBSS];
      1 : flagset := [fcfToAP];
      2 : flagset := [fcfFromAP];
      3 : flagset := [fcfWDS];
     end;

     // Parse the rest

     for i := 2 to 7 do
         begin
              isSet := BitIsSet (FlagValues, i);

              if isSet then
                 begin
                      case i of
                        2 : flag := fcfMoreFragments;
                        3 : flag := fcfRetry;
                        4 : flag := fcfPwrMgmt;
                        5 : flag := fcfMoreData;
                        6 : flag := fcfProtected;
                        7 : flag := fcfOrdered;
                      end;

                      flagset := flagset + [flag];
                 end;
         end;

     result := flagset;
end;

(* ------------------------------------------- *)

function t802dot11HeaderParser.ParseManagementHeader (const Offset: integer) : tParseResult;

var status      : tParseResult;
    pData       : p802dot11MgmtDataHeaderStruct;
    help        : word;
    localoffset : integer;

begin
     prot_layersize := sizeOf (t802dot11MgmtDataHeaderStruct);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (t802dot11MgmtDataHeaderStruct) then
        status := prFailUndersize
     else
         try
            pData := p802dot11MgmtDataHeaderStruct (priv_framedata + priv_frameoffset);

            // Headers following a CAPWAP layer have FCF and Flags swapped

            priv_isOverCAPWAP := priv_parentanchor.HeaderType = NEXT_HEADER_CAPWAP;

            if priv_isOverCAPWAP then
               priv_HeaderFlags := getMgmtHeaderFlagsFromByte (pData^.HeaderFrameControl)
            else
                priv_HeaderFlags := getMgmtHeaderFlagsFromByte (pData^.HeaderFCFFlags);

            priv_Duration := pData^.HeaderDuration;

            priv_receiveraddr.RangeBytes := pData^.HeaderAddress1;
            priv_destinationaddr.RangeBytes := pData^.HeaderAddress1;
            priv_transmitteraddr.RangeBytes := pData^.HeaderAddress2;
            priv_BSSId.RangeBytes := pData^.HeaderAddress2;
            priv_sourceaddr.RangeBytes := pData^.HeaderAddress3;

            help := pData^.HeaderFragmentSeq;

            priv_SequenceNumber := (help and $fff0) shr 4;
            priv_FragmentNumber := (help and $000f);

            localoffset := sizeOf (t802dot11MgmtDataHeaderStruct);

            if fcfOrdered in priv_HeaderFlags then
               begin
                    // this frame has an HT Control Field

                    priv_HTControl := pLongword (priv_framedata + priv_frameoffset + localoffset)^;
               end;

            // Determine next anchor

            if fcfProtected in priv_HeaderFlags then
               begin
                    // there is some 4 or 8 byte crypto stuff here sometimes, but I have no
                    // idea yet how to determine that. So let's all call it payload ;-)

                    priv_nextAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                    priv_nextAnchor.Offset := priv_frameoffset + localoffset;
               end
            else
                begin
                     priv_nextAnchor.HeaderType := NEXT_HEADER_802dot11Mgmt;
                     priv_nextAnchor.Offset := priv_frameoffset + localoffset;
                end;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

function t802dot11HeaderParser.ParseControlHeader (const Offset: integer) : tParseResult;

var status    : tParseResult;
    pData     : p802dot11CtrlHeaderStruct;
    help      : word;

begin
     prot_layersize := sizeOf (t802dot11CtrlHeaderStruct);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (t802dot11CtrlHeaderStruct) then
        status := prFailUndersize
     else
         try
            pData := p802dot11CtrlHeaderStruct (priv_framedata + priv_frameoffset);

            // Headers following a CAPWAP layer have FCF and Flags swapped

            priv_isOverCAPWAP := priv_parentanchor.HeaderType = NEXT_HEADER_CAPWAP;

            if priv_isOverCAPWAP then
               priv_HeaderFlags := getMgmtHeaderFlagsFromByte (pData^.HeaderFrameControl)
            else
                priv_HeaderFlags := getMgmtHeaderFlagsFromByte (pData^.HeaderFCFFlags);

            priv_Duration := pData^.HeaderDuration;
            priv_receiveraddr.RangeBytes := pData^.ReceiverAddress1;

            // Determine next anchor

            priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (t802dot11CtrlHeaderStruct);

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

function t802dot11HeaderParser.ParseDataHeader (const Offset: integer) : tParseResult;

var status    : tParseResult;
    pData     : p802dot11MgmtDataHeaderStruct;
    help      : word;

begin
     prot_layersize := sizeOf (t802dot11MgmtDataHeaderStruct);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (t802dot11MgmtDataHeaderStruct) then
        status := prFailUndersize
     else
         try
            pData := p802dot11MgmtDataHeaderStruct (priv_framedata + priv_frameoffset);

            // Headers following a CAPWAP layer have FCF and Flags swapped

            if priv_parentanchor.HeaderType = NEXT_HEADER_CAPWAP then
               priv_HeaderFlags := getMgmtHeaderFlagsFromByte (pData^.HeaderFrameControl)
            else
                priv_HeaderFlags := getMgmtHeaderFlagsFromByte (pData^.HeaderFCFFlags);

            priv_Duration := pData^.HeaderDuration;

            priv_receiveraddr.RangeBytes := pData^.HeaderAddress1;
            priv_destinationaddr.RangeBytes := pData^.HeaderAddress1;
            priv_transmitteraddr.RangeBytes := pData^.HeaderAddress2;
            priv_BSSId.RangeBytes := pData^.HeaderAddress2;
            priv_sourceaddr.RangeBytes := pData^.HeaderAddress3;

            help := pData^.HeaderFragmentSeq;

            priv_FragmentNumber := (help and $000f);
            priv_SequenceNumber := (help and $fff0) shr 4;

            // Determine next anchor

            if fcfProtected in priv_HeaderFlags then
               begin
                    // there is some 4 or 8 byte crypto stuff here sometimes, but I have no
                    // idea yet how to determine that. So let's all call it payload ;-)

                    priv_nextAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                    priv_nextAnchor.Offset := priv_frameoffset + sizeOf (t802dot11MgmtDataHeaderStruct);
               end
            else
                begin
                     priv_nextAnchor.HeaderType := NEXT_HEADER_LLC;
                     priv_nextAnchor.Offset := priv_frameoffset + sizeOf (t802dot11MgmtDataHeaderStruct);
                end;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

function t802dot11HeaderParser.parse: tParseResult;

var status    : tParseResult;
    pHelpByte : pByte;
    bHelp     : byte;

begin
     priv_osilayer := 1;     // Set Layer this parser is for

     status := prSuccess;

     if priv_framelength < priv_frameoffset + 1 then
        status := prFailUndersize
     else
         begin
              // Determine Type and Subtype - for headers following a CAPWAP layer there seems
              // to be a mistake that forces us to swap the type/subtybe byte position, meaning it
              // is the second byte in the layer, not the first

              priv_isOverCAPWAP := priv_parentanchor.HeaderType = NEXT_HEADER_CAPWAP;

              if priv_isOverCAPWAP then
                 pHelpByte := (priv_framedata + priv_frameoffset + 1)
              else
                  pHelpByte := (priv_framedata + priv_frameoffset);

              // Version

              priv_Version := (pHelpByte^ and 3);

              // Frame Type

              bHelp := (pHelpByte^ and $C) shr 2;

              case bHelp of
               0 : priv_FrameType := ft802Management;
               1 : priv_FrameType := ft802Control;
               2 : priv_FrameType := ft802Data;
              end;

              // Subtype

              bHelp := (pHelpByte^ and $f0) shr 4;
              priv_FrameSubType := getFrameSubTypeFromByte (bHelp);

              // Now parse the rest based on the frame type

              case priv_FrameType of
               ft802Management : Status := ParseManagementHeader (priv_frameoffset);
               ft802Control    : Status := ParseControlHeader (priv_frameoffset);
               ft802Data       : Status := ParseDataHeader (priv_frameoffset);
              end;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
