{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_ayiya;

interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;


type tAYIYAIDType = (aitNone, aitInteger, aitASCII);
     tAYIYAHashMethod = (ahmNoHash, ahmMD5, ahmSHA1);
     tAYIYAAuthenticationMethod = (aamNoAuth, aamSharedSecret, aamPubKey);
     tAYIYAOperationCode = (aocNOP, aocForward, aocEchoReq, aocEchoReqAndFwd, aocEchoResp, aocMOTD, aocQueryReq, aocQueryResp);

     pAYIYAStruct = ^tAYIYAStruct;
     tAYIYAStruct = packed record
                                 IDLenType            : byte;
                                 SigLenHshMeth        : byte;
                                 AuthMethOpCode       : byte;
                                 NextHeader           : byte;
                                 EpochTime            : longword;
                    end;

     tAYIYAParser  = class (tFrameParser)

                           private
                                  priv_IDLen     : byte;
                                  priv_IDType    : tAYIYAIDType;
                                  priv_SigLen    : word;
                                  priv_HshMeth   : tAYIYAHashMethod;
                                  priv_AuthMeth  : tAYIYAAuthenticationMethod;
                                  priv_OpCode    : tAYIYAOperationCode;
                                  priv_EpochTime : longword;
                                  priv_Identity  : tBufferBytes;
                                  priv_signature : tBufferBytes;

                           public
                                 property IDLength : byte read priv_IDLen;
                                 property IDType : tAYIYAIDType read priv_IDType;
                                 property SignatureLength : word read priv_SigLen;
                                 property HashMethod : tAYIYAHashMethod read priv_HshMeth;
                                 property AuthMethod : tAYIYAAuthenticationMethod read priv_AuthMeth;
                                 property OpCode : tAYIYAOperationCode read priv_OpCode;
                                 property EpochTime : longword read priv_EpochTime;
                                 property Identity : tBufferBytes read priv_Identity;
                                 property Signature : tBufferBytes read priv_signature;

                                 function checkIsPlausible : boolean; override;
                                 function parse : tParseResult; override;
                     end;


implementation

uses Math,
     Toolkit;

(* ------------------------------------------- *)
{ tAYIYAParser }
(* ------------------------------------------- *)

function tAYIYAParser.checkIsPlausible: boolean;

var pData  : pAYIYAStruct;
    help   : byte;
    value  : boolean;

begin
     priv_osilayer := 3;     // Set Layer this parser is for

     value := true;

     if priv_framelength < priv_frameoffset + sizeOf (tAYIYAStruct) then
        value := false
     else
         begin
              pData := pAYIYAStruct (priv_framedata + priv_frameoffset);

              help := pData^.IDLenType;
              help := help and $f;

              if help > 2 then    // IDType must be 0,1 or 2.
                 value := false;

              help := pData^.SigLenHshMeth;
              help := help and $f;

              if help > 2 then    // Hash Method must be 0,1 or 2.
                 value := false;

              help := pData^.AuthMethOpCode;
              help := (help and $f0) shr 4;

              if help > 2 then    // Auth Method must be 0,1 or 2.
                 value := false;

              help := pData^.AuthMethOpCode;
              help := help and $f;

              if help > 7 then    // OpCode is 0 to 7 (at the moment).
                 value := false;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function tAYIYAParser.parse: tParseResult;

var pData    : pAYIYAStruct;
    status   : tParseResult;
    help     : byte;
    help2    : byte;
    pHelp    : pByte;
    dHelp    : double;
    modifier : integer;
    i        : integer;

begin
     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tAYIYAStruct) then
        status := prFailUndersize
     else
         try
            pData := pAYIYAStruct (priv_framedata + priv_frameoffset);

            // ID Length

            help := pData^.IDLenType;

            help2 := (help and $f0) shr 4;
            dHelp := power (2, help2);
            priv_IDLen := round (dHelp);

            // IDType

            help := help and $f;

            case help of
             0 : priv_IDType := aitNone;
             1 : priv_IDType := aitInteger;
             2 : priv_IDType := aitASCII;
            end;

            // Sig Length

            help := pData^.SigLenHshMeth;
            help2 := (help and $f0) shr 4;

            priv_SigLen := help2 * 4;

            // Hash Method

            help := help and $f;

            case help of
             0 : priv_HshMeth := ahmNoHash;
             1 : priv_HshMeth := ahmMD5;
             2 : priv_HshMeth := ahmSHA1;
            end;

            // Auth Method

            help := pData^.AuthMethOpCode;
            help2 := (help and $f0) shr 4;

            case help2 of
             0 : priv_AuthMeth := aamNoAuth;
             1 : priv_AuthMeth := aamSharedSecret;
             2 : priv_AuthMeth := aamPubKey;
            end;

            help := help and $f;

            // OpCode

            case help of
             0 : priv_OpCode := aocNOP;
             1 : priv_OpCode := aocForward;
             2 : priv_OpCode := aocEchoReq;
             3 : priv_OpCode := aocEchoReqAndFwd;
             4 : priv_OpCode := aocEchoResp;
             5 : priv_OpCode := aocMOTD;
             6 : priv_OpCode := aocQueryReq;
             7 : priv_OpCode := aocQueryResp;
            end;

            priv_EpochTime := pData^.EpochTime;
            priv_EpochTime := Swap32 (priv_EpochTime);

            // Identity

            setlength (priv_Identity, priv_IDLen);
            modifier := 8;                           // Length of static header

            for i := 0 to priv_IDLen - 1 do
                begin
                     pHelp := pByte (priv_framedata + priv_frameoffset + modifier);
                     priv_Identity[i] := pHelp^;

                     inc (modifier);
                end;

            // Signature

            setLength (priv_signature, priv_SigLen);

            for i := 0 to priv_SigLen - 1 do
                begin
                     pHelp := pByte (priv_framedata + priv_frameoffset + modifier);
                     priv_signature[i] := pHelp^;

                     inc (modifier);
                end;

            // Determine Layersize

            prot_layersize := sizeOf (tAYIYAStruct) + priv_IDLen + priv_SigLen;

            // Prepare next anchor

            priv_nextAnchor.HeaderType := pData^.NextHeader;
            priv_nextAnchor.Offset := priv_frameoffset + modifier;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)


end.
