{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_baseclass;

interface

uses SysUtils,
     base_structures,
     common_parse_helpers,
     class_diagmessagehandler;

const maxProcolsHeaders = 32;

type tProtocolChainEntry = record
                                 index     : integer;               // Counter
                                 anchor    : tProtocolAnchor;
                           end;

     // Static Protocol Chain List

     tProtocolChain = array [1..maxProcolsHeaders] of tProtocolChainEntry;
{
     EParsingException = class (Exception)

                               private
                                      priv_FrameIndex   : int64;
                                      priv_DiagHandle   : string;
                                      priv_ErrorMessage : string;
                               public
                                     constructor Create (DiagHandle, Msg: string; FrameIndex : int64);
                                     property FrameIndex : Int64 read priv_FrameIndex write priv_FrameIndex;
                                     property DiagnosticIndex : string read priv_DiagHandle write priv_DiagHandle;
                                     property Message : string read priv_ErrorMessage write priv_ErrorMessage;
                               end;
}
     tFrameParser  = class (tObject)

                           constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength : integer; const FrameIndex : integer = 0);
                           procedure free;

                           private

                           strict protected

                                  priv_osilayer          : byte;             // Set to the according value in the actual parser code

                                  priv_framedata         : pByte;
                                  priv_frameoffset       : integer;
                                  priv_framelength       : integer;
                                  priv_wirelength        : integer;
                                  priv_frameindex        : int64;
                                  prot_layersize         : integer;

                                  priv_parentAnchor      : tProtocolAnchor;
                                  priv_nextAnchor        : tProtocolAnchor;

                                  priv_diagnosticmsglist : tDiagnosticMsgList;
                                  prot_lasterrorstatus   : tParseResult;

                                  function ParseByteFromOffset (var Offset : integer; const DoNotAdvanceOffset : boolean = false): byte;
                                  function ParseWordFromOffset (var Offset : integer; const DoNotAdvanceOffset : boolean = false; const DoNotByteswap : boolean = false): word;
                                  function ParseLongwordFromOffset (var Offset : integer; const DoNotAdvanceOffset : boolean = false; const DoNotByteswap : boolean = false): longword;
                                  function ParseIn64FromOffset (var Offset : integer; const DoNotAdvanceOffset : boolean = false; const DoNotByteswap : boolean = false): uint64;

                           public
                                 /// <summary>
                                 ///   Tells the OSI layer this parser is
                                 ///   working on, which is sometimes a bit
                                 ///   hard to determine. This is not read from
                                 ///   frame but hardcoded for each parser.
                                 ///   It's used for situations where parsing
                                 ///   should stop at a certain layer.
                                 /// </summary>
                                 property OSILayer : byte read priv_osilayer;
                                 /// <summary>
                                 ///   Index of the frame within the current
                                 ///   trace file
                                 /// </summary>
                                 property FrameIndex : int64 read priv_frameindex write priv_frameindex;
                                 property LayerSize : integer read prot_layersize;
                                 /// <summary>
                                 ///   Specifies the offset within the frame
                                 ///   bytes where the parser will start
                                 ///   parsing
                                 /// </summary>
                                 property FrameOffset : integer read priv_frameoffset;

                                 property ParentAnchor : tProtocolAnchor read priv_parentAnchor write priv_parentAnchor;
                                 property NextAnchor : tProtocolAnchor read priv_nextAnchor write priv_nextanchor;

                                 procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgHandle : integer); overload;
                                 procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgHandle : integer; const Param1 : string); overload;
                                 procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgHandle : integer; const Param1, Param2 : string); overload;
                                 procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgHandle : integer; const Param1, Param2, Param3 : string); overload;
                                 procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgHandle : integer; const Param1, Param2, Param3, Param4 : string); overload;
                                 procedure addDiagnosticMsg (const Msg : tDiagnosticMsg); overload;
                                 function getDiagnosticMsgCount : integer;
                                 function getDiagnosticMsg (const Index : integer) : tDiagnosticMsg;
                                 procedure clearDiagnosticMsgList;

                                 function checkIsPlausible : boolean; virtual; abstract;
                                 function parse : tParseResult; virtual; abstract;
                     end;


implementation

uses Toolkit;

(* ------------------------------------------- *)
{ EParsingException }
(* ------------------------------------------- *)
{
constructor EParsingException.Create (MsgID : integer; Msg: string; FrameIndex : int64);

begin
     inherited Create (Msg);

     priv_FrameIndex := FrameIndex;
     priv_DiagHandle := DiagHandle;
end;
}
(* ------------------------------------------- *)
{ tFrameParser }
(* ------------------------------------------- *)

constructor tFrameParser.create (FrameData : pointer; const FrameOffset, DataLength, WireLength : integer; const FrameIndex : integer = 0);

begin
     inherited create;

     priv_osilayer := 0;

     priv_framedata := FrameData;
     priv_frameoffset := FrameOffset;
     priv_framelength := DataLength;
     priv_wirelength := WireLength;
     priv_frameindex := FrameIndex;

     prot_layersize := 0;
     prot_lasterrorstatus := prSuccess;

     clearDiagnosticMsgList;
end;

(* ------------------------------------------- *)

procedure tFrameParser.free;

begin
     setlength (priv_diagnosticmsglist, 0);

     inherited;
end;

(* ------------------------------------------- *)

function tFrameParser.ParseByteFromOffset (var Offset: integer; const DoNotAdvanceOffset : boolean = false): byte;

var value : byte;
    help  : pByte;

begin
     value := 0;

     if (Offset + 1) > priv_framelength then
        prot_lasterrorstatus := prFailUndersize
     else
         begin
              help := pByte (priv_framedata + offset);

              value := help^;
         end;

     if not DoNotAdvanceOffset then
        inc (Offset);

     result := value;
end;

(* ------------------------------------------- *)

function tFrameParser.ParseWordFromOffset (var Offset: integer; const DoNotAdvanceOffset : boolean = false; const DoNotByteswap : boolean = false): word;

var value : word;
    help  : pWord;

begin
     value := 0;

     if (Offset + 1) > priv_framelength then
        prot_lasterrorstatus := prFailUndersize
     else
         begin
              help := pWord (priv_framedata + offset);

              if DoNotByteSwap then
                 value := help^
              else
                  value := swap16 (help^);
         end;

     if not DoNotAdvanceOffset then
        inc (Offset, 2);

     result := value;
end;

(* ------------------------------------------- *)

function tFrameParser.ParseLongwordFromOffset (var Offset: integer; const DoNotAdvanceOffset : boolean = false; const DoNotByteswap : boolean = false): longword;

var value : longword;
    help  : pLongword;

begin
     value := 0;

     if (Offset + 1) > priv_framelength then
        prot_lasterrorstatus := prFailUndersize
     else
         begin
              help := pLongword (priv_framedata + offset);

              if DoNotByteSwap then
                 value := help^
              else
                  value := swap32 (help^);
         end;

     if not DoNotAdvanceOffset then
        inc (Offset, 4);

     result := value;
end;

(* ------------------------------------------- *)

function tFrameParser.ParseIn64FromOffset (var Offset: integer; const DoNotAdvanceOffset : boolean = false; const DoNotByteswap : boolean = false): uint64;

var value : Uint64;
    help  : pUint64;

begin
     value := 0;

     if (Offset + 1) > priv_framelength then
        prot_lasterrorstatus := prFailUndersize
     else
         begin
              help := pUint64 (priv_framedata + offset);

              if DoNotByteSwap then
                 value := help^
              else
                  value := swap64 (help^);
         end;

     if not DoNotAdvanceOffset then
        inc (Offset, 8);

     result := value;
end;

(* ------------------------------------------- *)

procedure tFrameParser.addDiagnosticMsg (const Severity : tMsgSeverity; const MsgHandle : integer);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize (Severity, MsgHandle);

     RecCount := length (priv_diagnosticmsglist) + 1;
     setLength (priv_diagnosticmsglist, RecCount);
     priv_diagnosticmsglist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tFrameParser.addDiagnosticMsg (const Severity: tMsgSeverity; const MsgHandle : integer; const Param1: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize(Severity, MsgHandle, Param1);

     RecCount := length (priv_diagnosticmsglist) + 1;
     setLength (priv_diagnosticmsglist, RecCount);
     priv_diagnosticmsglist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tFrameParser.addDiagnosticMsg (const Severity: tMsgSeverity; const MsgHandle : integer; const Param1, Param2: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize (Severity, MsgHandle, Param1, Param2);

     RecCount := length (priv_diagnosticmsglist) + 1;
     setLength (priv_diagnosticmsglist, RecCount);
     priv_diagnosticmsglist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tFrameParser.addDiagnosticMsg (const Severity: tMsgSeverity; const MsgHandle : integer; const Param1, Param2, Param3: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize (Severity, MsgHandle, Param1, Param2, Param3);

     RecCount := length (priv_diagnosticmsglist) + 1;
     setLength (priv_diagnosticmsglist, RecCount);
     priv_diagnosticmsglist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tFrameParser.addDiagnosticMsg (const Severity: tMsgSeverity; const MsgHandle : integer; const Param1, Param2, Param3, Param4: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize (Severity, MsgHandle, Param1, Param2, Param3, Param4);

     RecCount := length (priv_diagnosticmsglist) + 1;
     setLength (priv_diagnosticmsglist, RecCount);
     priv_diagnosticmsglist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tFrameParser.addDiagnosticMsg (const Msg: tDiagnosticMsg);

var RecCount  : integer;

begin
     RecCount := length (priv_diagnosticmsglist) + 1;
     setLength (priv_diagnosticmsglist, RecCount);
     priv_diagnosticmsglist[RecCount-1] := Msg;
end;

(* ------------------------------------------- *)

function tFrameParser.getDiagnosticMsgCount : integer;

begin
     result := length (priv_diagnosticmsglist);
end;

(* ------------------------------------------- *)

function tFrameParser.getDiagnosticMsg (const Index : integer) : tDiagnosticMsg;

begin
     result := priv_diagnosticmsglist[Index - 1];
end;

(* ------------------------------------------- *)

procedure tFrameParser.clearDiagnosticMsgList;

begin
     setLength (priv_diagnosticmsglist,0);
end;

(* ------------------------------------------- *)

end.
