{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_dhcpv4;

interface

uses base_structures,
     definition_ip_addresses,
     common_parse_helpers,
     class_parser_baseclass;


const MaxIPListAddresses = 10;
      MaxFQDNs = 10;

type tDHCPv4MessageOpCode = (docBootRequest, docBootReply);
     tDHCPv4MessageType = (dmtNone, dmtDiscover, dmtOffer, dmtRequest, dmtAcknowledge, dmtNACK, dmtDecline, dmtRelease, dmtInform);
     tDHCPv4Flag = (dfUnicast, dfBroadcast);
     tDHCPOptionOverload = (dooNone, dooBootFile, dooServername, dooBoth);
     tDHCPOptionSIPServerEncoding =(dssFQDNList, dssIPList);

     tDHCPv4MACPAdding = array [1..10] of byte;

     pStruct_DHCPv4 = ^tStruct_DHCPv4;
     tStruct_DHCPv4 = packed record
                             MessageType         : byte;
                             HardwareAddrType    : byte;
                             HardwareAddrLength  : byte;
                             Hops                : byte;
                             TransactionID       : longword;
                             Seconds             : word;
                             Flags               : word;
                             ClientIPAddress     : longword;
                             YourClientIPAddress : longword;
                             ServerIPAddress     : longword;
                             RelayIPAddress      : longword;
                             ClientHWAddress     : tMACBytes;
                             ClientHWPadding     : tDHCPv4MACPAdding;
                             ServerHostName      : string[63];         // according to specs it is 64 byte zero terminated
                             BootFileName        : string[127];        // according to specs it is 128 byte zero terminated
                             MagicCookie         : longword;
                    end;

     tDHCPv4AddressList = array [1..MaxIPListAddresses] of tIPv4Address;
     tDHCPParamReqList = array of byte;
     tDHCPv4FQDNList = array [1..MaxFQDNs] of tFQDN;

     tDHCPv4OptionHdr = record
                               Code           : byte;
                               Length         : byte;
                               addresscount   : byte;
                               addresslist    : tDHCPv4AddressList;
                               MACAddress     : tMACAddress;
                               number64       : uint64;
                               numberword     : word;
                               numberlongword : longword;
                               numberint      : integer;
                               numberbyte     : byte;
                               text           : string;
//                             FQDN           : tFQDN;
                               FQDNCount      : byte;
                               FQDNList       : tDHCPv4FQDNList;
                               Value          : array of byte;
                               ParamReqList   : tDHCPParamReqList;
                               SIPSrvEncoding : tDHCPOptionSIPServerEncoding;

                               procedure initialize;
                        end;

     tDHCPv4OptionList = array of tDHCPv4OptionHdr;

     tDHCPv4Parser  = class (tFrameParser)

                            constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength, DHCPSectionLength : integer; const FrameIndex : integer = 0); reintroduce;

                            private
                                   priv_sectionlen          : integer;
                                   priv_lastoffset          : integer;

                                   priv_Messagetype         : tDHCPv4MessageOpCode;
                                   priv_HardwareAddrType    : byte;
                                   priv_HardwareAddrLength  : byte;
                                   priv_Hops                : byte;
                                   priv_TransactionID       : longword;
                                   priv_Seconds             : word;
                                   priv_Flags               : tDHCPv4Flag;
                                   priv_ClientIPAddress     : tIPv4Address;
                                   priv_YourClientIPAddress : tIPv4Address;
                                   priv_ServerIPAddress     : tIPv4Address;
                                   priv_RelayIPAddress      : tIPv4Address;
                                   priv_ClientHWAddress     : tMACAddress;
                                   priv_ClientHWPadding     : tDHCPv4MACPAdding;
                                   priv_ServerHostName      : string;
                                   priv_BootFileName        : string;
                                   priv_MagicCookie         : longword;

                                   // Values taken from options for quicker access

                                   priv_DHCPMessageType     : tDHCPv4MessageType;
                                   priv_ClientIdentifier    : tMACAddress;
                                   priv_ServerIdentifier    : tIPv4Address;
                                   priv_SubnetMask          : tIPv4Address;
                                   priv_Hostname            : tFQDN;
                                   priv_Domainname          : tFQDN;
                                   priv_RequestedIPAddress  : tIPv4Address;
                                   priv_LeaseTime           : Longword;
                                   priv_Router              : tIPv4Address;
                                   priv_DNSServers          : tDHCPv4AddressList;
                                   priv_NTPServers          : tDHCPv4AddressList;
                                   priv_OptionOverload      : tDHCPOptionOverload;
                                   priv_OptionalMessage     : string;
                                   priv_MaxMessageSize      : word;

                                   priv_dhcpoptions         : tDHCPv4OptionList;

                                   priv_padsize             : integer;

                                   function parseIPv4AddressListOption (const OptionLen : byte; LocalOffset : integer) : tDHCPv4AddressList;
                                   function ParseMACAddressOption (LocalOffset : integer) : tMACAddress;
                                   function ParseStringOption (Length : integer; LocalOffset: integer): string;
                                   function parseFQDNListOption (Length, Localoffset : integer) : tDHCPv4FQDNList;
                                   function ParseDHCPv4Options (Offset : integer) : integer;

                                   procedure InitOptionalValues;
                            public
                                  property MessageOpCode : tDHCPv4MessageOpCode read priv_Messagetype;
                                  property HardwareAddressType : byte read priv_HardwareAddrType;
                                  property HardwareAddressLength : byte read priv_HardwareAddrLength;
                                  property Hops : byte read priv_Hops;
                                  property TransactionID : longword read priv_TransactionID;
                                  property Seconds : word read priv_Seconds;
                                  property Flags : tDHCPv4Flag read priv_Flags;
                                  property ClientIPAddress : tIPv4Address read priv_ClientIPAddress;
                                  property YourClientIPAddress : tIPv4Address read priv_YourClientIPAddress;
                                  property ServerIPAddress : tIPv4Address read priv_ServerIPAddress;
                                  property RelayIPAddress : tIPv4Address read priv_RelayIPAddress;
                                  property ClientHWAddress : tMACAddress read priv_ClientHWAddress;
                                  property ClientHWPadding : tDHCPv4MACPAdding read priv_ClientHWPadding;
                                  property ServerHostName : string read priv_ServerHostName;
                                  property BootFileName : string read priv_BootFileName;
                                  property MagicCookie : longword read priv_MagicCookie;

                                  property DHCPMessageType : tDHCPv4MessageType read priv_DHCPMessageType;
                                  property DHCPClientIdentifier : tMACAddress read priv_ClientIdentifier;
                                  property DHCPServerIdentifier : tIPv4Address read priv_ServerIdentifier;
                                  property DHCPSubnetMask : tIPv4Address read priv_SubnetMask;
                                  property DHCPHostName : tFQDN read priv_Hostname;
                                  property DHCPDomainName : tFQDN read priv_Domainname;
                                  property DHCPRequestedIPAddress : tIPv4Address read priv_RequestedIPAddress;
                                  property DHCPLeaseTime : Longword read priv_LeaseTime;
                                  property DHCPRouter : tIPv4Address read priv_router;
                                  property DHCPNameServers : tDHCPv4AddressList read priv_DNSServers;
                                  property DHCPNTPServers : tDHCPv4AddressList read priv_NTPServers;
                                  property DHCPOptionOverload : tDHCPOptionOverload read priv_OptionOverload;
                                  property DHCPOptionalMessage : string read priv_OptionalMessage;
                                  property DHCPMaxMessageSize : word read priv_MaxMessageSize;

                                  property OptionList : tDHCPv4OptionList read priv_dhcpoptions;

                                  property PaddingSize : integer read priv_padsize;

                                  function checkIsPlausible : boolean; override;
                                  function parse : tParseResult; override;
                      end;


const DHCP_OPT_Pad =	0;
      DHCP_OPT_Subnet_Mask =	1;
      DHCP_OPT_Time_Offset =	2;
      DHCP_OPT_Router =	3;
      DHCP_OPT_Time_Server =	4;
      DHCP_OPT_Name_Server =	5;
      DHCP_OPT_Domain_Name_Server =	6;
      DHCP_OPT_Log_Server =	7;
      DHCP_OPT_Quote_Server =	8;
      DHCP_OPT_LPR_Server =	9;
      DHCP_OPT_Impress_Server =	10;
      DHCP_OPT_Resource_Location_Server =	11;
      DHCP_OPT_Host_Name =	12;
      DHCP_OPT_Boot_File_Size =	13;
      DHCP_OPT_Merit_Dump_File =	14;
      DHCP_OPT_Domain_Name =	15;
      DHCP_OPT_Swap_Server =	16;
      DHCP_OPT_Root_Path =	17;
      DHCP_OPT_Extensions_Path =	18;
      DHCP_OPT_IP_Forwarding_enable_disable =	19;
      DHCP_OPT_NonLocal_Source_Routing_enable_disable =	20;
      DHCP_OPT_Policy_Filter =	21;
      DHCP_OPT_Maximum_Datagram_Reassembly_Size =	22;
      DHCP_OPT_Default_IP_TimeToLive =	23;
      DHCP_OPT_Path_MTU_Aging_Timeout =	24;
      DHCP_OPT_Path_MTU_Plateau_Table =	25;
      DHCP_OPT_Interface_MTU =	26;
      DHCP_OPT_All_Subnets_Local =	27;
      DHCP_OPT_Broadcast_Address =	28;
      DHCP_OPT_Perform_Mask_Discovery =	29;
      DHCP_OPT_Mask_supplier =	30;
      DHCP_OPT_Perform_router_discovery =	31;
      DHCP_OPT_Router_solicitation_address =	32;
      DHCP_OPT_Static_routing_table =	33;
      DHCP_OPT_Trailer_encapsulation =	34;
      DHCP_OPT_ARP_cache_timeout =	35;
      DHCP_OPT_Ethernet_encapsulation =	36;
      DHCP_OPT_Default_TCP_TTL =	37;
      DHCP_OPT_TCP_keepalive_interval =	38;
      DHCP_OPT_TCP_keepalive_garbage =	39;
      DHCP_OPT_Network_Information_Service_Domain =	40;
      DHCP_OPT_Network_Information_Servers =	41;
      DHCP_OPT_NTP_servers =	42;
      DHCP_OPT_Vendor_specific_information =	43;
      DHCP_OPT_NetBIOS_over_TCPIP_nameserver =	44;
      DHCP_OPT_NetBIOS_over_TCPIP_Datagram_Distribution_Server =	45;
      DHCP_OPT_NetBIOS_over_TCPIP_Node_Type =	46;
      DHCP_OPT_NetBIOS_over_TCPIP_Scope =	47;
      DHCP_OPT_X_Window_System_Font_Server =	48;
      DHCP_OPT_X_Window_System_Display_Manager =	49;
      DHCP_OPT_Requested_IP_Address =	50;
      DHCP_OPT_IP_address_lease_time =	51;
      DHCP_OPT_Option_overload =	52;
      DHCP_OPT_DHCP_message_type =	53;
      DHCP_OPT_Server_identifier =	54;
      DHCP_OPT_Parameter_request_list =	55;
      DHCP_OPT_Message =	56;
      DHCP_OPT_Maximum_DHCP_message_size =	57;
      DHCP_OPT_Renew_time_value =	58;
      DHCP_OPT_Rebinding_time_value =	59;
      DHCP_OPT_ClassIdentifier =	60;
      DHCP_OPT_ClientIdentifier =	61;
      DHCP_OPT_NetWareIP_Domain_Name =	62;
      DHCP_OPT_NetWareIP_information =	63;
      DHCP_OPT_Network_Information_ServiceDomain =	64;
      DHCP_OPT_Network_Information_ServiceServers =	65;
      DHCP_OPT_TFTP_server_name =	66;
      DHCP_OPT_Bootfile_name =	67;
      DHCP_OPT_Mobile_IP_Home_Agent =	68;
      DHCP_OPT_Simple_Mail_Transport_Protocol_Server =	69;
      DHCP_OPT_Post_Office_Protocol_Server =	70;
      DHCP_OPT_Network_News_Transport_Protocol_Server =	71;
      DHCP_OPT_Default_World_Wide_Web_Server =	72;
      DHCP_OPT_Default_Finger_Server =	73;
      DHCP_OPT_Default_Internet_Relay_Chat_Server =	74;
      DHCP_OPT_StreetTalk_Server =	75;
      DHCP_OPT_StreetTalk_Directory_Assistance_Server =	76;
      DHCP_OPT_User_Class_Information =	77;
      DHCP_OPT_SLP_Directory_Agent =	78;
      DHCP_OPT_SLP_Service_Scope =	79;
      DHCP_OPT_Rapid_Commit =	80;
      DHCP_OPT_FQDN_Fully_Qualified_Domain_Name =	81;
      DHCP_OPT_Relay_Agent_Information =	82;
      DHCP_OPT_Internet_Storage_Name_Service =	83;
      DHCP_OPT_NDS_servers =	85;
      DHCP_OPT_NDS_tree_name =	86;
      DHCP_OPT_NDS_context =	87;
      DHCP_OPT_BCMCS_Controller_Domain_Name_list =	88;
      DHCP_OPT_BCMCS_Controller_IPv4_address_list =	89;
      DHCP_OPT_Authentication =	90;
      DHCP_OPT_ClientLastTransactionTime =	91;
      DHCP_OPT_associatedIP =	92;
      DHCP_OPT_Client_System_Architecture_Type =	93;
      DHCP_OPT_Client_Network_Interface_Identifier =	94;
      DHCP_OPT_LDAP_Lightweight_Directory_Access_Protocol =	95;
      DHCP_OPT_Client_Machine_Identifier =	97;
      DHCP_OPT_Open_Groups_User_Authentication =	98;
      DHCP_OPT_GEOCONF_CIVIC =	99;
      DHCP_OPT_IEEE_10031_TZ_String =	100;
      DHCP_OPT_Reference_to_the_TZ_Database =	101;
      DHCP_OPT_NetInfo_Parent_Server_Address =	112;
      DHCP_OPT_NetInfo_Parent_Server_Tag =	113;
      DHCP_OPT_URL =	114;
      DHCP_OPT_AutoConfigure =	116;
      DHCP_OPT_Name_Service_Search =	117;
      DHCP_OPT_Subnet_Selection =	118;
      DHCP_OPT_DNS_domain_search_list =	119;
      DHCP_OPT_SIP_Servers_DHCP_Option =	120;
      DHCP_OPT_Classless_Static_Route_Option =	121;
      DHCP_OPT_CableLabs_Client_Configuration =	122;
      DHCP_OPT_GeoConf =	123;
      DHCP_OPT_VendorIdentifying_Vendor_Class =	124;
      DHCP_OPT_VendorIdentifying_VendorSpecific =	125;
      DHCP_OPT_TFPT_Server_IP_address =	128;
      DHCP_OPT_Call_Server_IP_address =	129;
      DHCP_OPT_Discrimination_string =	130;
      DHCP_OPT_Remote_statistics_server_IP_address =	131;
      DHCP_OPT_8021P_VLAN_ID =	132;
      DHCP_OPT_8021Q_L2_Priority =	133;
      DHCP_OPT_Diffserv_Code_Point =	134;
      DHCP_OPT_HTTP_Proxy_for_PhoneSpecific_Applications =	135;
      DHCP_OPT_PANA_Authentication_Agent =	136;
      DHCP_OPT_LoST_Server =	137;
      DHCP_OPT_CAPWAP_Access_Controller_addresses =	138;
      DHCP_OPT_OPTIONIPv4_Address_MoS =	139;
      DHCP_OPT_OPTIONIPv4_FQDN_MoS =	140;
      DHCP_OPT_SIP_UA_Configuration_Service_Domains =	141;
      DHCP_OPT_OPTIONIPv4_Address_ANDSF =	142;
      DHCP_OPT_OPTIONIPv6_Address_ANDSF =	143;
      DHCP_OPT_TFTP_server_address =	150;
      DHCP_OPT_Etherboot =	150;
      DHCP_OPT_IP_Telephone =	176;
      DHCP_OPT_pxelinuxMagicString =	208;
      DHCP_OPT_pxelinuxConfigfileText =	209;
      DHCP_OPT_pxelinuxpathprefixText =	210;
      DHCP_OPT_pxelinuxreboottime_unsigned_integer_32_bits =	211;
      DHCP_OPT_OPTION_6RD =	212;
      DHCP_OPT_OPTION_V4_ACCESS_DOMAIN =	213;
      DHCP_OPT_Subnet_Allocation =	220;
      DHCP_OPT_Private_use =	254;
      DHCP_OPT_End =	255;


implementation

uses SysUtils,
     toolkit,
     base_messages;

(* ------------------------------------------- *)
{ tDHCPv4OptionHdr }
(* ------------------------------------------- *)

procedure tDHCPv4OptionHdr.initialize;

var i : integer;

begin
     Code := 0;
     Length := 0;

     addresscount := 0;

     for i := 1 to MaxIPListAddresses do
         addresslist[i].initialize;

     FQDNCount := 0;

     for i := 1 to MaxFQDNs do
         FQDNList[i].initialize;

     MACAddress := VALUE_ZERO_MAC;
     number64 := 0;
     numberword := 0;
     numberlongword := 0;
     numberint := 0;
     numberbyte := 0;
     text := '';
//   FQDN.initialize;
     setLength (Value, 0);
     setLength (ParamReqList, 0);
end;

(* ------------------------------------------- *)
{ tDHCPv4Parser }
(* ------------------------------------------- *)

constructor tDHCPv4Parser.create (FrameData: pointer; const FrameOffset, DataLength, WireLength, DHCPSectionLength, FrameIndex: integer);

begin
     inherited create (FrameData, FrameOffset, DataLength, WireLength, FrameIndex);

     priv_sectionlen := DHCPSectionLength;

     // calculate the end offset we can parse up to

     if (priv_frameoffset + priv_sectionlen) > priv_framelength  then
        priv_lastoffset := priv_framelength
     else
         priv_lastoffset := priv_frameoffset + priv_sectionlen;
end;

(* ------------------------------------------- *)

function tDHCPv4Parser.checkIsPlausible: boolean;

begin
     result := true;                     // Not much else it can be, port is pretty specific
end;

(* ------------------------------------------- *)

function tDHCPv4Parser.parseIPv4AddressListOption (const OptionLen : byte; LocalOffset : integer) : tDHCPv4AddressList;

var Help32  : pLongword;
    addlist : tDHCPv4AddressList;
    i       : integer;

begin
     for i := 1 to (OptionLen DIV 4) do
         begin
              Help32 := pLongword (priv_framedata + localoffset);
              addlist[i].ValueAsLong := swap32 (Help32^);
              addlist[i].MaskBitCount := 32;

              inc (localoffset, 4);
         end;

     result := addlist;
end;

(* ------------------------------------------- *)

function tDHCPv4Parser.parseFQDNListOption (Length, Localoffset: integer): tDHCPv4FQDNList;

var parsedone  : boolean;
    fqdnresult : tFQDNResult;
    fqdnlist   : tDHCPv4FQDNList;
    count      : integer;
    ParsedLen  : integer;
    i          : integer;

begin
     parsedone := false;
     count := 0;
     ParsedLen := 0;

     for i := 1 to maxFQDNs do
         fqdnlist[i].initialize;

     repeat
//         fqdnresult := readFQDNFromOffset (priv_framedata, priv_framelength, Localoffset, Localoffset);
           fqdnresult := readFQDNFromOffset (priv_framedata, priv_sectionlen, Localoffset, Localoffset);

           inc (count);
           fqdnlist[count] := fqdnresult.Name;

           inc (Localoffset, fqdnresult.OffsetIncrement);
           inc (ParsedLen, fqdnresult.OffsetIncrement);

           parsedone := ParsedLen >= Length;

     until parsedone;

     result := fqdnlist;
end;

(* ------------------------------------------- *)

function tDHCPv4Parser.ParseMACAddressOption (LocalOffset: integer): tMACAddress;

var value : tMACAddress;

begin
     value.RangeBytes := pMACBytes (priv_framedata + localoffset)^;

     result := value;
end;

(* ------------------------------------------- *)

function tDHCPv4Parser.ParseStringOption (Length : integer; LocalOffset: integer): string;

var help  : pANSIChar;
    value : string;

begin
     Help := pANSIChar (priv_framedata + LocalOffset);
     value := string (Help);

     result := copy (value, 1, Length);
end;

(* ------------------------------------------- *)

// Returns the length of the option block

function  tDHCPv4Parser.ParseDHCPv4Options (Offset: integer) : integer;

var localoffset    : integer;
    optioncount    : integer;
    optiontotallen : integer;
    code           : Byte;
    len            : Byte;
    i              : integer;
    Help8          : pByte;
    Help16         : pWord;
    Help32         : pLongword;
    HelpString     : string;
    OptionValue    : tDHCPv4OptionHdr;
    SkipOption     : boolean;
    boAbort        : boolean;

begin
     priv_osilayer := 5;     // Set Layer this parser is for

     optioncount := 0;
     optiontotallen := 0;
     localoffset := Offset;

     boAbort := false;

     repeat
           OptionValue.initialize;
           SkipOption := false;

           if localoffset < priv_lastoffset + 1 then
              begin
                   // Read code, length, value

                   code := pByte (priv_framedata + localoffset)^;

                   if not ((code = DHCP_OPT_PAD) or (code = DHCP_OPT_END)) then
                      begin
                           if localoffset + 1 < priv_lastoffset + 1 then
                              begin
                                   len := pByte (priv_framedata + localoffset + 1)^;

                                   // Check if we have enough bytes left for the option content

                                   if localoffset + 2 + len >= priv_lastoffset then
                                      begin
                                           boAbort := true;
                                           addDiagnosticMsg (msWarning, MSG_PARSER_OUT_OF_DATA_FOR_LENGTH, (localoffset + 1).toString, len.toString, priv_lastoffset.ToString);
                                      end;
                              end
                           else
                               begin
                                    boAbort := true;
                                    addDiagnosticMsg (msWarning, MSG_PARSER_OUT_OF_DATA, (localoffset + 1).toString, priv_lastoffset.ToString);
                               end;
                      end;
              end
           else
               begin
                    boAbort := true;
                    addDiagnosticMsg (msWarning, MSG_PARSER_OUT_OF_DATA, localoffset.toString, priv_lastoffset.ToString);
               end;

           if not boAbort then
              begin
                   if (code > DHCP_OPT_PAD) and (code < DHCP_OPT_End) then
                      begin
                           inc (localoffset, 2);
                           inc (optiontotallen, 2);

                           OptionValue.initialize;
                           OptionValue.Code := code;
                           OptionValue.Length := len;

                           case OptionValue.Code of
                            DHCP_OPT_Subnet_Mask                                 : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;

                                                                                        priv_SubnetMask := OptionValue.addresslist[1];
                                                                                   end;
                            DHCP_OPT_Time_Offset                                 : begin
                                                                                        Help32 := pLongword (priv_framedata + localoffset);
                                                                                        OptionValue.numberlongword := Help32^;
                                                                                   end;
                            DHCP_OPT_Router                                      : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;

                                                                                        priv_Router := OptionValue.addresslist[1];
                                                                                   end;
                            DHCP_OPT_Time_Server                                 : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;
                                                                                   end;
                            DHCP_OPT_Name_Server                                 : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;

                                                                                        priv_DNSServers := OptionValue.addresslist;
                                                                                   end;
                            DHCP_OPT_Domain_Name_Server                          : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;

                                                                                        priv_DNSServers := OptionValue.addresslist;
                                                                                   end;
                            DHCP_OPT_Host_Name                                   : begin
                                                                                        HelpString := ParseStringOption (OptionValue.Length, localoffset);
        //                                                                              OptionValue.FQDN.setFQDN (HelpString);
        //                                                                              priv_Hostname := OptionValue.FQDN;

                                                                                        OptionValue.FQDNCount := 1;
                                                                                        OptionValue.FQDNList[1].setFQDN(Helpstring);
                                                                                        priv_Hostname := OptionValue.FQDNList[1];
                                                                                   end;
                            DHCP_OPT_Domain_Name                                 : begin
                                                                                        HelpString := ParseStringOption (OptionValue.Length, localoffset);
        //                                                                              OptionValue.FQDN.setFQDN (HelpString);
        //                                                                              priv_Domainname := OptionValue.FQDN;

                                                                                        OptionValue.FQDNCount := 1;
                                                                                        OptionValue.FQDNList[1].setFQDN(Helpstring);
                                                                                        priv_Hostname := OptionValue.FQDNList[1];
                                                                                   end;
                            DHCP_OPT_Broadcast_Address                           : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;
                                                                                   end;
                            DHCP_OPT_TCP_keepalive_interval                      : begin
                                                                                        Help32 := pLongword (priv_framedata + localoffset);
                                                                                        OptionValue.numberlongword := swap32 (Help32^);
                                                                                   end;
                            DHCP_OPT_NTP_servers                                 : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;

                                                                                        priv_NTPServers := OptionValue.addresslist;
                                                                                   end;
                            DHCP_OPT_NetBIOS_over_TCPIP_nameserver               : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;
                                                                                   end;
                            DHCP_OPT_NetBIOS_over_TCPIP_Node_Type                : begin
                                                                                        Help8 := pByte (priv_framedata + localoffset);
                                                                                        OptionValue.numberbyte := Help8^;
                                                                                   end;
                            DHCP_OPT_Requested_IP_Address                        : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;

                                                                                        priv_RequestedIPAddress := OptionValue.addresslist[1];
                                                                                   end;
                            DHCP_OPT_IP_address_lease_time                       : begin
                                                                                        Help32 := pLongword (priv_framedata + localoffset);
                                                                                        OptionValue.numberlongword := swap32 (Help32^);

                                                                                        priv_LeaseTime := OptionValue.numberlongword;
                                                                                   end;
                            DHCP_OPT_Option_overload                             : begin
                                                                                        Help8 := pByte (priv_framedata + localoffset);
                                                                                        OptionValue.numberbyte := Help8^;

                                                                                        case OptionValue.numberbyte of
                                                                                         1 : begin
                                                                                                  priv_OptionOverload := dooBootFile;
                                                                                                  priv_BootFileName := '';
                                                                                             end;
                                                                                         2 : begin
                                                                                                  priv_OptionOverload := dooServername;
                                                                                                  priv_ServerHostName := '';
                                                                                             end;
                                                                                         3 : begin
                                                                                                  priv_OptionOverload := dooBoth;

                                                                                                  priv_BootFileName := '';
                                                                                                  priv_ServerHostName := '';
                                                                                             end;
                                                                                        end;

                                                                                        // TODO: parse the additional stuff in the header when we're done with the rest of the normal options
                                                                                   end;
                            DHCP_OPT_DHCP_message_type                           : begin
                                                                                        Help8 := pByte (priv_framedata + localoffset);
                                                                                        OptionValue.numberbyte := Help8^;

                                                                                        case OptionValue.numberbyte of
                                                                                         1 : priv_DHCPMessageType := dmtDiscover;
                                                                                         2 : priv_DHCPMessageType := dmtOffer;
                                                                                         3 : priv_DHCPMessageType := dmtRequest;
                                                                                         4 : priv_DHCPMessageType := dmtDecline;
                                                                                         5 : priv_DHCPMessageType := dmtAcknowledge;
                                                                                         6 : priv_DHCPMessageType := dmtNACK;
                                                                                         7 : priv_DHCPMessageType := dmtRelease;
                                                                                         8 : priv_DHCPMessageType := dmtInform;
                                                                                        end;
                                                                                   end;
                            DHCP_OPT_Server_identifier                           : begin
                                                                                        OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length, localoffset);
                                                                                        OptionValue.addresscount := len div 4;

                                                                                        priv_ServerIdentifier := OptionValue.addresslist[1];
                                                                                   end;
                            DHCP_OPT_Parameter_request_list                      : begin
                                                                                        setLength (OptionValue.ParamReqList, OptionValue.Length);

                                                                                        for i := 1 to OptionValue.Length do
                                                                                            begin
                                                                                                 Help8 := pByte (priv_framedata + localoffset + (i - 1));

                                                                                                 OptionValue.ParamReqList[i-1] := Help8^;
                                                                                            end;
                                                                                   end;
                            DHCP_OPT_Message                                     : begin
                                                                                        OptionValue.text := ParseStringOption (OptionValue.Length, localoffset);

                                                                                        priv_OptionalMessage := OptionValue.text;
                                                                                   end;
                            DHCP_OPT_Maximum_DHCP_message_size                   : begin
                                                                                        Help16 := pWord (priv_framedata + localoffset);
                                                                                        OptionValue.numberword := swap16 (Help16^);

                                                                                        priv_MaxMessageSize := OptionValue.numberword;
                                                                                   end;
                            DHCP_OPT_Renew_time_value                            : begin
                                                                                        Help32 := pLongword (priv_framedata + localoffset);
                                                                                        OptionValue.numberlongword := swap32 (Help32^);

                                                                                        priv_LeaseTime := OptionValue.numberlongword;
                                                                                   end;
                            DHCP_OPT_Rebinding_time_value                        : begin
                                                                                        Help32 := pLongword (priv_framedata + localoffset);
                                                                                        OptionValue.numberlongword := swap32 (Help32^);

                                                                                        priv_LeaseTime := OptionValue.numberlongword;
                                                                                   end;
                            DHCP_OPT_ClassIdentifier                             : begin
                                                                                        OptionValue.text := ParseStringOption (OptionValue.Length, localoffset);
                                                                                   end;
                            DHCP_OPT_ClientIdentifier                            : begin
                                                                                        Help8 := pByte (priv_framedata + localoffset);

                                                                                        OptionValue.numberbyte := Help8^;
                                                                                        OptionValue.MACAddress := ParseMACAddressOption (localoffset + 1);

                                                                                        priv_ClientIdentifier := OptionValue.MACAddress;
                                                                                   end;
                            DHCP_OPT_FQDN_Fully_Qualified_Domain_Name            : begin
                                                                                        setlength (OptionValue.Value, 3);

                                                                                        OptionValue.Value[0] := pByte (priv_framedata + localoffset)^;
                                                                                        OptionValue.Value[1] := pByte (priv_framedata + localoffset + 1)^;
                                                                                        OptionValue.Value[2] := pByte (priv_framedata + localoffset + 2)^;

                                                                                        HelpString := ParseStringOption (OptionValue.Length - 3, localoffset + 3);

                                                                                        OptionValue.FQDNCount := 1;
                                                                                        OptionValue.FQDNList[1].setFQDN (HelpString);
                                                                                   end;
                            DHCP_OPT_AutoConfigure                               : begin
                                                                                        Help8 := pByte (priv_framedata + localoffset);
                                                                                        OptionValue.numberbyte := Help8^;
                                                                                   end;

                            DHCP_OPT_SIP_Servers_DHCP_Option                     : begin
                                                                                        Help8 := pByte (priv_framedata + localoffset);

                                                                                        case Help8^ of
                                                                                         0 : begin
                                                                                                  OptionValue.SIPSrvEncoding := dssFQDNList;

                                                                                                  // Parse FQDNs

                                                                                                  OptionValue.FQDNList := parseFQDNListOption (OptionValue.Length - 1, localoffset + 1);

                                                                                                  // Determine number of FQDNs returned by checking if the FQDN is not empty

                                                                                                  for i := 1 to 10 do
                                                                                                      if OptionValue.FQDNList[i].FullFQDN <> '' then
                                                                                                         OptionValue.FQDNCount := i;
                                                                                             end;
                                                                                         1 : begin
                                                                                                  OptionValue.SIPSrvEncoding := dssIPList;

                                                                                                  // Parse IPs

                                                                                                  OptionValue.addresslist := parseIPv4AddressListOption (OptionValue.Length - 1, localoffset + 1);
                                                                                                  OptionValue.addresscount := len div 4;
                                                                                             end;
                                                                                        end;


                                                                                   end
                           else
                               // Can't parse this option

                               SkipOption := true;
                           end;

                           inc (localoffset, len);
                           inc (optiontotallen, len);

                           if not SkipOption then
                              begin
                                   // Add new record

                                   setLength (priv_dhcpoptions, optioncount + 1);

                                   priv_dhcpoptions[optioncount] := OptionValue;
                                   inc (optioncount);
                              end;
                      end
                   else
                       begin
                            // Add new record

                            setLength (priv_dhcpoptions, optioncount + 1);

                            priv_dhcpoptions[optioncount].code := code;
                            inc (optioncount);

                            inc (localoffset);
                            inc (optiontotallen);
                       end;
              end

     until (code = DHCP_OPT_End) or boAbort;

     // Determine length. Add 2 for the initial byte positioned on and the EndOfOpts

//   result := (localoffset + 2) - Offset;
     result := optiontotallen;
end;

(* ------------------------------------------- *)

procedure tDHCPv4Parser.InitOptionalValues;

var i : integer;

begin
     priv_DHCPMessageType := dmtNone;

     priv_ServerIdentifier.ValueAsLong := 0;
     priv_SubnetMask.ValueAsLong := 0;
     priv_Hostname.initialize;
     priv_RequestedIPAddress .ValueAsLong := 0;
     priv_LeaseTime := 0;
     priv_Router .ValueAsLong := 0;
     priv_OptionOverload := dooNone;

     for i := 1 to MaxIPListAddresses do
         priv_DNSServers[i].ValueAsLong := 0;
end;

(* ------------------------------------------- *)

function tDHCPv4Parser.parse: tParseResult;

var pData      : pStruct_DHCPv4;
    OptSize    : integer;
    status     : tParseResult;
    Help8      : Byte;
    Help16     : word;
    Help32     : Longword;
    pHelp32    : pLongword;
    HelpChar   : pANSIChar;
    HelpString : string;


begin
     status := prSuccess;
     OptSize := 0;
     priv_padsize := 0;

     InitOptionalValues;

     if priv_framelength < priv_frameoffset + sizeOf (tStruct_DHCPv4) then
        status := prFailUndersize
     else
         try
            pData := pStruct_DHCPv4 (priv_framedata + priv_frameoffset);

            Help8 := pData^.MessageType;

            case Help8 of
             1 : priv_Messagetype := docBootRequest;
             2 : priv_Messagetype := docBootReply;
            end;

            priv_HardwareAddrType := pData^.HardwareAddrType;
            priv_HardwareAddrLength := pData^.HardwareAddrLength;
            priv_Hops := pData^.Hops;

            priv_TransactionID := pData^.TransactionID;
            priv_TransactionID := Swap32 (priv_TransactionID);

            priv_Seconds := pData^.Seconds;

            if priv_Seconds > 255 then
               priv_Seconds := Swap16 (priv_Seconds);

            Help16 := pData^.Flags;
            Help16 := Swap16 (Help16);

            if (Help16 and $8000) = $8000 then
               priv_Flags := dfBroadcast
            else
                priv_Flags := dfUnicast;

            Help32 := pData^.ClientIPAddress;
            priv_ClientIPAddress.ValueAsLong := Swap32 (Help32);
            priv_ClientIPAddress.MaskBitCount := 32;

            Help32 := pData^.YourClientIPAddress;
            priv_YourClientIPAddress.ValueAsLong := Swap32 (Help32);
            priv_YourClientIPAddress.MaskBitCount := 32;

            Help32 := pData^.ServerIPAddress;
            priv_ServerIPAddress.ValueAsLong := Swap32 (Help32);
            priv_ServerIPAddress.MaskBitCount := 32;

            Help32 := pData^.RelayIPAddress;
            priv_RelayIPAddress.ValueAsLong := Swap32 (Help32);
            priv_RelayIPAddress.MaskBitCount := 32;

            priv_ClientHWAddress.RangeBytes := pData^.ClientHWAddress;
            priv_ClientHWPadding := pData^.ClientHWPadding;

            // Get the Hostname and Bootfile strings. The pData strings are just used as placeholders
            // because of the zero terminated strings in the payload

            HelpChar := pANSIChar (priv_framedata + priv_frameoffset + 44);
            HelpString := string (HelpChar);

            priv_ServerHostName := copy (HelpString, 1, 63);

            HelpChar := pANSIChar (priv_framedata + priv_frameoffset + 108);
            HelpString := string (HelpChar);

            priv_BootFileName := copy (HelpString, 1, 127);

            priv_MagicCookie := pData^.MagicCookie;
            priv_MagicCookie := Swap32 (priv_MagicCookie);

            // Parse DHCP options, if any

            if (priv_frameoffset + sizeOf (tStruct_DHCPv4)) < priv_lastoffset then
               OptSize := ParseDHCPv4Options (priv_frameoffset + sizeOf (tStruct_DHCPv4));

            prot_layersize := sizeOf (tStruct_DHCPv4);
            prot_layersize := prot_layersize + OptSize;

            // Determine Padding

            priv_padsize := (priv_framelength - priv_frameoffset) - prot_layersize;

            if priv_padsize > 3 then
               begin
                    // Check for FCS at the end

                    pHelp32 := pLongWord (priv_framedata + priv_framelength - 4);

                    if pHelp32^ <> 0 then
                       dec (priv_padsize, 4);
               end;

            prot_layersize := prot_layersize + priv_padsize;

            priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
            priv_nextAnchor.Offset := priv_frameoffset + prot_layersize;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
