{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_gtpv1;   // GPRS Tunneling Protocol V1

interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;


// http://www.tcpdump.org/linktypes/LINKTYPE_LINUX_SLL.html

type tGTPFlags = (gtpReserved, gtpExtensionPresent, gtpSequencePresent, gtpNPDUPresent);
     TGTPFlagSet = set of tGTPFlags;
     tGTPProtocolType = (gtpptGTP, gtpptGTP0);

     pGTPv1Struct = ^tGTPv1Struct;
     tGTPv1Struct = packed record
                                 Flags       : byte;
                                 MessageType : byte;
                                 Length      : word;
                                 TEID        : longword;
{                                SequenceNo  : word;        Variable, cannot assume they always exist
                                 NPDUNumber  : byte;
                                 NextExtHdr  : byte;
}
                    end;

     tGTPv1Parser  = class (tFrameParser)

                              private
                                     priv_Version      : byte;
                                     priv_ProtocolType : tGTPProtocolType;
                                     priv_flags        : TGTPFlagSet;
                                     priv_MessageType  : byte;
                                     priv_TotalLength  : word;
                                     priv_TEID         : longword;
                                     priv_SequenceNo   : word;
                                     priv_NPDUNumber   : byte;
                                     priv_NextExtHdr   : byte;

                              public
                                    property Version : byte read priv_Version;
                                    property ProtocolType : tGTPProtocolType read priv_ProtocolType;
                                    property Flags : TGTPFlagSet read priv_flags;
                                    property MessageType : byte read priv_MessageType;
                                    property TotalLength : word read priv_TotalLength;
                                    property TEID : longword read priv_TEID;
                                    property SequenceNo : word read priv_SequenceNo;
                                    property NPDUNumber : byte read priv_NPDUNumber;
                                    property NextExtHdr : byte read priv_NextExtHdr;

                                    function checkIsPlausible : boolean; override;
                                    function parse : tParseResult; override;
                        end;


const GTP_MSG_END_MARKER = 254;
      GTP_MSG_TPDU = 255;

implementation

uses Math,
     Toolkit;

(* ------------------------------------------- *)
{ tGTPv1Parser }
(* ------------------------------------------- *)

function tGTPv1Parser.checkIsPlausible: boolean;

var pData    : pGTPv1Struct;

begin
     result := true;

     // Since we can only work with T-PDU messages, we check for it, otherwise refuse

     if priv_framelength < priv_frameoffset + sizeOf (tGTPv1Struct) then
        result := false
     else
         begin
              pData := pGTPv1Struct (priv_framedata + priv_frameoffset);

              // $ff is the T-PDU message type

              result := (pData^.MessageType = $ff);
         end;
end;

(* ------------------------------------------- *)

function tGTPv1Parser.parse: tParseResult;

var pData     : pGTPv1Struct;
    status    : tParseResult;
    pHelpWord : pWord;
    HelpWord  : Word;
    pHelpByte : pByte;
    helpByte  : byte;
    optlen    : integer;
    offset    : integer;

begin
     priv_osilayer := 2;     // Set Layer this parser is for
     prot_layersize := sizeOf (tGTPv1Struct);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tGTPv1Struct) then
        status := prFailUndersize
     else
         try
            pData := pGTPv1Struct (priv_framedata + priv_frameoffset);

            // Parse Flags byte

            helpByte := pData^.Flags;
            priv_Version := (helpByte AND $e0) shr 5;

            if (helpByte and $10) = $10 then
               priv_ProtocolType := gtpptGTP
            else
                priv_ProtocolType := gtpptGTP0;

            if (helpByte and $8) = $8 then
               priv_flags := [gtpReserved]
            else
                priv_flags := [];

            if (helpByte and $4) = $4 then
               priv_flags := priv_flags + [gtpExtensionPresent];

            if (helpByte and $2) = $2 then
               priv_flags := priv_flags + [gtpSequencePresent];

            if (helpByte and $1) = $1 then
               priv_flags := priv_flags + [gtpNPDUPresent];

            // Parse Rest

            priv_MessageType := pData^.MessageType;
            priv_TotalLength := swap16 (pData^.Length);
            priv_TEID := Swap32 (pData^.TEID);

            // Parse the optional values, if present

            optlen := 0;
            offset := priv_frameoffset + sizeOf (tGTPv1Struct);

            if gtpSequencePresent in priv_flags then
               begin
                    pHelpWord := pWord (priv_framedata + offset);
                    HelpWord := pHelpWord^;
                    priv_SequenceNo := Swap16 (HelpWord);

                    inc (optlen, 2);
                    inc (offset, 2);
               end
            else
                priv_SequenceNo := 0;

            if gtpNPDUPresent in priv_flags then
               begin
                    pHelpByte := pByte (priv_framedata + offset);
                    priv_NPDUNumber := pHelpByte^;

                    inc (offset);
               end
            else
                priv_NPDUNumber := 0;

            if gtpExtensionPresent in priv_flags then
               begin
                    pHelpByte := pByte (priv_framedata + offset);
                    priv_NextExtHdr := pHelpByte^;

                    inc (offset);
               end
            else
                priv_NextExtHdr := 0;

            // Prepare Anchor. If optional values were present they are always 32 bits wide

            if priv_flags = [] then
               offset := priv_frameoffset + sizeOf (tGTPv1Struct)
            else
                offset := priv_frameoffset + sizeOf (tGTPv1Struct) + 4;

            pHelpByte := pByte (priv_framedata + offset);
            helpByte := (pHelpByte^ and $f0) shr 4;

            // Check if there is IP4

            case helpByte of
             4 : priv_nextAnchor.HeaderType := NEXT_HEADER_IPv4;
             6 : priv_nextAnchor.HeaderType := NEXT_HEADER_IPv6;
            else
                priv_nextAnchor.HeaderType := NEXT_HEADER_UNKNOWN;
            end;

            priv_nextAnchor.Offset := offset;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)


end.
