{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_http;

interface

uses Classes,
     base_structures,
     definition_http_structures,
     common_parse_helpers,
     class_parser_baseclass;

type tHTTPParser  = class (tFrameParser)

                          constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength : integer; const FrameIndex : integer = 0); reintroduce;

                           private
                                  priv_IsRequest         : boolean;

                                  priv_HTTPMethod        : tHTTPMethod;
                                  priv_HTTPVersion       : string;
                                  priv_Statuscode        : integer;
                                  priv_ResponsePhrase    : string;
                                  priv_URI               : string;
                                  priv_Host              : string;
                                  priv_UserAgent         : string;
                                  priv_Connection        : tHTTPConnectionState;
                                  priv_ContentLength     : integer;
                                  priv_ContentType       : string;
                                  priv_XForwardedFor     : string;
                                  priv_HTTPHeaderList    : tHTTPHeaderList;

                                  procedure initialize;
                                  procedure determineRequestResponse (value : string);
                                  procedure AddHTTPHeaderToList (const Entry : tHTTPHeaderEntry);

                           public
                                 destructor Destroy; override;

                                 property IsRequest : boolean read priv_IsRequest;
                                 property Method : tHTTPMethod read priv_HTTPMethod;
                                 property Version : string read priv_HTTPVersion;
                                 property StatusCode : integer read priv_StatusCode;
                                 property ResponsePhrase : string read priv_ResponsePhrase;
                                 property URI : string read priv_URI;
                                 property Host : string read priv_Host;
                                 property UserAgent : string read priv_UserAgent;
                                 property ContentLength : integer read priv_ContentLength;
                                 property XForwardedFor : string read priv_XForwardedFor;
                                 property Headers : tHTTPHeaderList read priv_HTTPHeaderList;

                                 function checkIsPlausible : boolean; override;
                                 function parse : tParseResult; override;
                       end;

implementation

uses SysUtils,
     Toolkit;

(* ------------------------------------------- *)
{ tHTTPParser }
(* ------------------------------------------- *)

constructor tHTTPParser.create (FrameData : pointer; const FrameOffset, DataLength, WireLength : integer; const FrameIndex : integer = 0);

begin
     inherited create (FrameData, FrameOffset, DataLength, Wirelength, FrameIndex);

     initialize;
end;

(* ------------------------------------------- *)

destructor tHTTPParser.Destroy;

begin

     inherited;
end;

(* ------------------------------------------- *)

procedure tHTTPParser.initialize;

begin
     priv_IsRequest := True;
     priv_HTTPMethod := HTTPMethodUnknown;
     priv_HTTPVersion := '';
     priv_Statuscode := 0;
     priv_ResponsePhrase := '';
     priv_URI := '';
     priv_Host := '';
     priv_UserAgent := '';
     priv_Connection := HTTPConnectionKeepAlive;
     priv_ContentLength := 0;
     priv_ContentType := '';
     priv_XForwardedFor := '';

     setLength (priv_HTTPHeaderList, 0);
end;

(* ------------------------------------------- *)

function tHTTPParser.checkIsPlausible: boolean;

var value : boolean;
    test  : string;

begin
     // Check if "HTTP/" is found in the first line

     test := pANSIChar (priv_framedata + priv_frameoffset);
     test := leftstring (test, CRLF);

     value := (pos ('HTTP/', test) > 0);

     result := value;
end;

(* ------------------------------------------- *)

procedure tHTTPParser.determineRequestResponse (value : string);

var index : integer;
    help  : string;

begin
     value := trim (value);
     index := pos ('HTTP/', value);

     if index = 1 then
        begin
             // Response, e.g. "HTTP/1.1 200 OK"

             priv_IsRequest := False;

             priv_HTTPVersion := copy (value, 1, 8);
             help := copy (value, 10, 3);
             TryStrToInt (help, priv_statuscode);

             priv_ResponsePhrase := copy (value, 14, length (value) - 13);
        end
     else
         begin
              // Request, e.g. "GET / HTTP/1.1"

              priv_IsRequest := true;

              help := leftstring (value, ' ');
              priv_HTTPMethod := StrToHTTPMethod (help);

              help := rightstring (value, ' ');
              priv_URI := leftstring (help, ' ');

              priv_HTTPVersion := rightstring (help, ' ');
         end;
end;

(* ------------------------------------------- *)

procedure tHTTPParser.AddHTTPHeaderToList (const Entry: tHTTPHeaderEntry);

var len  : integer;
    help : string;

begin
     len := length (priv_HTTPHeaderList);
     setLength (priv_HTTPHeaderList, len + 1);
     priv_HTTPHeaderList[len] := Entry;

     // Transfer some specific header values to their object fields

     help := trim (lowercase (Entry.Name));

     if help = 'host' then
        priv_Host := Entry.Value
     else
     if help = 'user-agent' then
        priv_UserAgent := Entry.Value
     else
     if help = 'x-forwarded-for' then
        priv_XForwardedFor := Entry.Value
     else
     if help = 'content-lengt' then
        priv_ContentLength := Entry.ValueAsInteger
     else
     if help = 'content-type' then
        priv_ContentType := Entry.Value
     else
     if help = 'connection' then
        begin
             if lowercase (Entry.Value) = 'close' then
                priv_Connection := HTTPConnectionClose
             else
                 priv_Connection := HTTPConnectionKeepAlive;
        end;
end;

(* ------------------------------------------- *)

function tHTTPParser.parse: tParseResult;

var status      : tParseResult;
    HTTPChunk   : tFrameChunk;
    HTTPStrings : TStringList;
    entry       : tHTTPHeaderEntry;
    EoLFound    : boolean;
    i           : integer;

begin
     priv_osilayer := 7;     // Set Layer this parser is for

     status := prSuccess;

     prot_layersize := priv_framelength - priv_frameoffset;

     if prot_layersize > 0 then
        begin
             HTTPStrings := TStringList.Create;

             // Transfer Payload to a FrameChunk

             HTTPChunk.copyFrom (priv_framedata + priv_frameoffset, prot_layersize);

             // And now to a stringlist

             HTTPStrings.Text := string (HTTPChunk.getAsANSIString);

             EoLFound := False;

             for i := 1 to HTTPStrings.Count do
                 begin
                      if pos ('HTTP/', HTTPStrings[i-1]) > 0 then
                         determineRequestResponse (HTTPStrings[i-1])
                      else
                          if pos (HTTPHeaderDelimiter, HTTPStrings[i-1]) > 0 then
                             begin
                                  if HTTPStrings[i-1] <> '' then
                                     entry.setfromString (HTTPStrings[i-1])
                                  else
                                      EoLFound := True;

                                  AddHTTPHeaderToList (entry);
                             end;
                 end;

             HTTPStrings.Free;
             HTTPChunk.free;

             if not EoLFound then
                status := prFailSegmentedPayload;

             // Prepare next anchor

             priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
             priv_nextAnchor.Offset := priv_framelength;
        end
     else
         status := prFailUndersize;

      result := status;
end;

(* ------------------------------------------- *)


end.
