{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_icmpv4;

interface

uses base_structures,
     definition_ip_addresses,
     common_parse_helpers,
     class_parser_baseclass;


type pICMPv4Struct = ^tICMPv4Struct;
     tICMPv4Struct = packed record
                                  icmp_type     : byte;
                                  icmp_code     : byte;
                                  checksum      : word;
                            end;

     pEchoStruct = ^tEchoStruct;
     tEchoStruct = packed record
                                  echo_identifier : word;
                                  echo_sequence   : word;
                            end;

     tICMPv4Parser = class (tFrameParser)

                           private
                                  priv_type          : byte;
                                  priv_code          : byte;
                                  priv_checksum      : word;
                                  priv_payload       : tRawBuffer;
                                  priv_payloadlength : integer;
                                  priv_payloadoffset : integer;

                                  // Optional values

                                  priv_echo_identifier        : word;                   // stored in Big Endian
                                  priv_echo_sequence          : word;                   // stored in Big Endian
                                  priv_next_hop_mtu           : word;
                                  priv_redirect_gwy           : tIPv4Address;
                                  priv_OriginalDatagramLength : word;
                                  priv_icmpsectionwords       : tBufferWords;

                                  function getPayloadString : string;
                                  function getPayLoadWordBuffer : tBufferWords;
                                  function determineNextAnchor (Offset : integer) : tProtocolAnchor;

                           public
                                 property ICMPType : byte read priv_type;
                                 property ICMPCode : byte read priv_code;
                                 property Checksum : word read priv_checksum;
                                 property EchoIdentifier : word read priv_echo_identifier;
                                 property EchoSequence : word read priv_echo_sequence;
                                 property NextHopMTU : word read priv_next_hop_mtu;
                                 property RedirectGwy : tIPv4Address read priv_redirect_gwy;
                                 property OriginalDatagramLength : word read priv_OriginalDatagramLength;
                                 property Payload  : tRawBuffer read priv_payload;
                                 property PayloadAsString : string read getPayloadString;
                                 property PayloadAsWordBuffer : tBufferWords read getPayLoadWordBuffer;
                                 property PayloadLength : integer read priv_payloadlength;
                                 property HeaderAsWordBuffer : tBufferWords read priv_icmpsectionwords;

                                 function checkIsPlausible : boolean; override;
                                 function parse : tParseResult; override;
                     end;


implementation

uses Toolkit;


(* ------------------------------------------- *)
{ ICMPv4Parser }
(* ------------------------------------------- *)

function tICMPv4Parser.checkIsPlausible: boolean;

begin
     result := true;                     // Obviously
end;

(* ------------------------------------------- *)

function tICMPv4Parser.determineNextAnchor (Offset : integer) : tProtocolAnchor;

var anchor     : tProtocolAnchor;
    EchoStruct : pEchoStruct;
    pbHelp     : pByte;
    bHelp      : Byte;
    pwHelp     : pWord;
    wHelp      : word;
    plHelp     : pLongword;
    lHelp      : longword;

begin
//   priv_payloadlength := 0;

     // Since there are always 4 bytes following the ICMP header, the next anchor starts at +4

     anchor.Offset := offset + 4;
     priv_payloadoffset := anchor.Offset;

     priv_payloadlength := priv_framelength - priv_payloadoffset;  // treat everytthing as payload, even if parsed in other parsers later

     // Determine values depending on type

     case priv_type of
       0,
       8 : begin
                // IPv4 Echo Request/Reply

                EchoStruct := pEchoStruct (priv_framedata + Offset);

                wHelp := EchoStruct^.echo_identifier;
                priv_echo_identifier := swap16 (wHelp);

                wHelp := EchoStruct^.echo_sequence;
                priv_echo_sequence := swap16 (wHelp);

                // Set correct next anchor offset and payload length

//              priv_payloadlength := priv_framelength - priv_payloadoffset;

                anchor.Offset := Offset + 4 + priv_payloadlength;
                anchor.HeaderType := NEXT_HEADER_NONE;
           end;
       3 : begin
                // Destination unreachable

                case priv_code of
                 4 : begin
                          // Fragmentation needed

                          pwHelp := pWord (priv_framedata + Offset + 2);
                          wHelp := pwHelp^;

                          priv_next_hop_mtu := Swap16 (wHelp);
                     end;
                else

                end;

                // ICMP messages that have IPv4 quoted packets as payload

                anchor.HeaderType := NEXT_HEADER_IPv4;
           end;
       5 : begin
                // Redirect

                plHelp := pLongword (priv_framedata + Offset);
                lHelp := plHelp^;

                priv_redirect_gwy.ValueAsLong := Swap32 (lHelp);
                priv_redirect_gwy.MaskBitCount := 32;

                // ICMP messages that have IPv4 quoted packets as payload

                anchor.HeaderType := NEXT_HEADER_IPv4;
           end;
      11 : begin
                // TTL exceeded, as extended in rfc4884

                pbHelp := pByte (priv_framedata + Offset + 1);
                bHelp := pbHelp^;

                priv_OriginalDatagramLength := 4 * bHelp;

                // ICMP messages that have IPv4 quoted packets as payload

                anchor.HeaderType := NEXT_HEADER_IPv4;
           end;
     else
         anchor.HeaderType := NEXT_HEADER_UNKNOWN;

         // determine remaining payload length

//       priv_payloadlength := priv_framelength - priv_payloadoffset;
     end;

     result := anchor;
end;

(* ------------------------------------------- *)

function tICMPv4Parser.getPayLoadWordBuffer: tBufferWords;

begin
     result := GetRawBufferAsWordBuffer (priv_payload, priv_payloadlength);
end;

(* ------------------------------------------- *)

function tICMPv4Parser.getPayloadString: string;

var i     : integer;
    value : string;

begin
     value := '';

     for i := 1 to priv_payloadlength do
         value := value + chr (priv_payload[i]);

     result := value;
end;

(* ------------------------------------------- *)

function tICMPv4Parser.parse: tParseResult;

var pData    : pICMPv4Struct;
    pPayload : pByte;
    status   : tParseResult;
    help     : word;
//  CRCsize  : integer;
    calcCRC  : boolean;

begin
     priv_osilayer := 3;     // Set Layer this parser is for
     priv_echo_identifier := 0;
     priv_echo_sequence := 0;
     priv_next_hop_mtu := 0;
     priv_redirect_gwy.initialize;
     priv_OriginalDatagramLength := 0;

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tICMPv4Struct) then
        status := prFailUndersize
     else
         try
            // Get basic ICMP structure

            pData := pICMPv4Struct (priv_framedata + priv_frameoffset);

            priv_type := pData^.icmp_type;
            priv_code := pData^.icmp_code;

            help := pData^.checksum;

            priv_checksum := swap16 (help);

            // Determine substructures and then the next anchor and payload length

            priv_nextAnchor := determineNextAnchor (priv_frameoffset + sizeOf (tICMPv4Struct));

            // Get the payload if any

            if priv_payloadlength > 0 then
               begin
                    pPayload := pByte (priv_framedata + priv_payloadoffset);

                    move (pPayload^, priv_payload, priv_payloadlength);
               end;

            // Determine LayerSize

            case priv_type of
             0,
             8  : prot_layersize := sizeOf (tICMPv4Struct) + 4 + priv_payloadlength;
            else
                prot_layersize := sizeOf (tICMPv4Struct) + 4;
            end;

            // Prepare the word array for crc calculation if possible - frame must be long enough

            CalcCRC := (priv_framelength = priv_wirelength);

            // But not too long - segmentation offloading seems to be the death to getWordsStructFromOffset, so we skip it in that case

            if priv_framelength > 9018 then
               CalcCRC := false;

            if CalcCRC then
               begin
//                  CRCsize := sizeOf (tICMPv4Struct) + 4 + priv_payloadlength;

                    // transfer the ICMP header info to a word buffer, so that the parsemaster can use it to calculate the CRC.
                    // it will pull the payload values for that via the PayloadAsWordBuffer property, so no need for adding them here

                    priv_icmpsectionwords := getWordsStructFromOffset (priv_FrameData, priv_frameOffset, sizeOf (tICMPv4Struct) + 4);
               end
            else
                begin
                     status := prFailPayloadTruncated;
                     setlength (priv_icmpsectionwords, 0);
                end;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
