{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_rtps;

// This parser is rudimentary only; it doesn't even try to understand the SubMessages, it just copies them as byte arrays
// This is for very basic sanitization by MessageID type only, so there is no Defensive Transformation for RTPS!

interface

uses base_structures,
     base_checksums,
     common_parse_helpers,
     class_parser_baseclass;


type tRTPSSubMessageFlag = (rtpsfEndianess, rtpsfTimestamp);
     tRTPSSubMessageFlagSet = set of tRTPSSubMessageFlag;

     tRTPSStruct  = packed record
                                 Protocol          : array[0..3] of byte;
                                 Version           : word;
                                 VendorID          : word;
                    end;
     pRTPSStruct = ^tRTPSStruct;

     tRTPSGUIDStruct = packed record
                                 GUIDPrefixHostID  : longword;
                                 GUIDPrefixAppID   : longword;
                                 GUIDPrefixCounter : longword;
                       end;
     pRTPSGUIDStruct = ^tRTPSGUIDStruct;

     tRTPSSubMessageStruct = packed record
                                          MessageID    : byte;
                                          Flags        : byte;
                                          Len          : word;
                             end;
     pRTPSSubMessageStruct = ^tRTPSSubMessageStruct;

     tRTPSDataMessageStruct = packed record
                                           ExtraFlags        : word;
                                           LenQoS            : word;
                                           ReaderEntity      : longword;
                                           WriterEntity      : longword;
                                           WriterSeqHi       : longword;
                                           WriterSeqLo       : longword;
                                     end;
     pRTPSDataMessageStruct = ^tRTPSDataMessageStruct;

     tRTPSDataMessageEncapStruct = packed record
                                                 EncapsulationKind : word;
                                                 EncapsulationOpts : word;
                                    end;
     pRTPSDataMessageEncapStruct = ^tRTPSDataMessageEncapStruct;

     tRTPSSerializedDataEntry = record
                                      DataType        : word;
                                      DataLength      : word;
                                      Available       : word;             // number of bytes we really got (when frame is truncated or fragmented)
                                      Data            : array of byte;
                                end;
     pRTPSSerializedDataEntry = ^tRTPSSerializedDataEntry;

     tRTPSSerializedDataList = array of tRTPSSerializedDataEntry;

     tRTPSDataMessage = record
                              ExtraFlags        : word;
                              LenQoS            : word;
                              ReaderEntity      : longword;
                              WriterEntity      : longword;
                              WriterSeq         : uint64;
                              EncapsulationKind : word;
                              EncapsulationOpts : word;
                              InlineQoSList     : tRTPSSerializedDataList;
                              Data              : array of byte;
                              DataList          : tRTPSSerializedDataList;

                              procedure initialize;
                        end;

     tRTPSDataFragMessage = record
                                  ExtraFlags        : word;
                                  LenQoS            : word;
                                  ReaderEntity      : longword;
                                  WriterEntity      : longword;
                                  WriterSeq         : uint64;
                                  FragmentStartNum  : longword;
                                  FragmentInSubMsg  : word;
                                  FragmentSize      : word;
                                  SampleSize        : longword;
                                  InlineQoSList     : tRTPSSerializedDataList;
                                  Data              : array of byte;

                                  procedure initialize;
                            end;

     tRTPSDataBatchMessage = record
                                   ExtraFlags        : word;
                                   LenQoS            : word;
                                   ReaderEntity      : longword;
                                   WriterEntity      : longword;
                                   BatchSeq          : uint64;
                                   FirstSample       : uint64;
                                   OffsetLastSample  : longword;
                                   BatchSampleCount  : longword;
                                   OctetsToEncapID   : longword;
                                   procedure initialize;
                             end;

     tRTPSSubMessage = record
                             MessageID        : byte;
                             Flags            : tRTPSSubMessageFlagSet;
                             Len              : word;
                             Available        : word;             // number of bytes we really got (when frame is truncated or fragmented)
                             data             : array of byte;
                             DataMessage      : tRTPSDataMessage;
                             DataFragMessage  : tRTPSDataFragMessage;
                             DataBatchMessage : tRTPSDataBatchMessage;
                       end;

     tRTPSSubMessageList = array of tRTPSSubMessage;

     tRTPSParser   = class (tFrameParser)

                           constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength, SectionLength : integer; const FrameIndex : integer = 0); reintroduce;

                           private
                                  priv_sectionlen    : integer;
                                  priv_lastoffset    : integer;

                                  priv_MajorVersion  : byte;
                                  priv_MinorVersion  : byte;
                                  priv_VendorID      : word;
                                  priv_GUIDHostID    : longword;
                                  priv_GUIDAppID     : longword;
                                  priv_GUIDCounter   : longword;
                                  priv_submessages   : tRTPSSubMessageList;
                                  priv_isPing        : boolean;

                                  function ParseSubMessageFlags (const value : byte) : tRTPSSubMessageFlagSet;

                                  function ParseSerializedDataEntryFromOffset (Offset: integer) : tRTPSSerializedDataEntry;
                                  function ParseDATASubmessage (Offset : integer; DataLen : integer;Flags : tRTPSSubMessageFlagSet) : tRTPSDataMessage;
                                  procedure ParseRTPSSubMessages (Offset : integer);

                           public
                                 property MajorVersion : byte read priv_MajorVersion;
                                 property MinorVersion : byte read priv_MinorVersion;
                                 property VendorID : word read priv_VendorID;
                                 property GUIDHostID : longword read priv_GUIDHostID;
                                 property GUIDAppID : longword read priv_GUIDAppID;
                                 property GUIDCounter : longword read priv_GUIDCounter;
                                 property SubMessages : tRTPSSubMessageList read priv_submessages;
                                 property IsPing : boolean read priv_isPing;

                                 function checkIsPlausible : boolean; override;
                                 function parse : tParseResult; override;
                     end;


CONST RTPS_ID_DATA       = $15;
      RTPS_ID_DATA_FRAG  = $16;
      RTPS_ID_DATA_BATCH = $18;

      RTPS_PID_SENTINEL                      = $0001;
      RTPS_PID_TYPECODE                      = $8004;
      RTPS_PID_OBJECT                        = $0072;
      RTPS_PID_PROPERTY_LIST                 = $0059;
      RTPS_PID_UNICAST_LOCATOR               = $002f;
      RTPS_PID_MULTICAST_LOCATOR             = $0030;
      RTPS_PID_DEFAULT_UNICAST_LOCATOR       = $0031;
      RTPS_PID_METATRAFFIC_UNICAST_LOCATOR   = $0032;
      RTPS_PID_METATRAFFIC_MULTICAST_LOCATOR = $0033;
      RTPS_PID_DEFAULT_MULTICAST_LOCATOR     = $0048;

      RTPS_PING = 'NDDSPING';

implementation

uses Math,
     Toolkit;

(* ------------------------------------------- *)
{ tRTPSDataMessage }
(* ------------------------------------------- *)

procedure tRTPSDataMessage.initialize;

begin
     ExtraFlags := 0;
     LenQoS := 0;
     ReaderEntity := 0;
     WriterEntity := 0;
     WriterSeq := 0;
     setLength (InlineQoSList, 0);
     EncapsulationKind := 0;
     EncapsulationOpts := 0;
     setLength (Data, 0);
     setLength (DataList, 0);
end;

(* ------------------------------------------- *)
{ tRTPSDataFragMessage }
(* ------------------------------------------- *)

procedure tRTPSDataFragMessage.initialize;

begin
     ExtraFlags := 0;
     LenQoS := 0;
     ReaderEntity := 0;
     WriterEntity := 0;
     WriterSeq := 0;
     setLength (InlineQoSList, 0);
     FragmentStartNum := 0;
     FragmentInSubMsg := 0;
     FragmentSize := 0;
     SampleSize := 0;
     setLength (Data, 0);
end;

(* ------------------------------------------- *)
{ tRTPSDataBatchMessage }
(* ------------------------------------------- *)

procedure tRTPSDataBatchMessage.initialize;

begin
     ExtraFlags := 0;
     LenQoS := 0;
     ReaderEntity := 0;
     WriterEntity := 0;
     BatchSeq := 0;
     FirstSample := 0;
     OffsetLastSample := 0;
     BatchSampleCount := 0;
     OctetsToEncapID := 0;




end;

(* ------------------------------------------- *)
{ tRTPSParser }
(* ------------------------------------------- *)

constructor tRTPSParser.create (FrameData: pointer; const FrameOffset, DataLength, WireLength, SectionLength, FrameIndex: integer);

begin
     inherited create (FrameData, FrameOffset, DataLength, WireLength, FrameIndex);

     priv_sectionlen := SectionLength;

     // calculate the end offset we can parse up to

     if (priv_frameoffset + priv_sectionlen) > priv_framelength  then
        priv_lastoffset := priv_framelength
     else
         priv_lastoffset := priv_frameoffset + priv_sectionlen - 1;
end;

(* ------------------------------------------- *)

function tRTPSParser.checkIsPlausible: boolean;

begin
     // We only got here when UDP did a look-ahead at the RTPS magic string

     result := true;
end;

(* ------------------------------------------- *)

function tRTPSParser.ParseSubMessageFlags (const value: byte): tRTPSSubMessageFlagSet;

begin
     result := [];

     if (value and 1) = 1  then
        result := result + [rtpsfEndianess];

     if (value and 2) = 2  then
        result := result + [rtpsfTimestamp];
end;

(* ------------------------------------------- *)

function tRTPSParser.ParseSerializedDataEntryFromOffset (Offset: integer) : tRTPSSerializedDataEntry;

var TLVHdr    : pTLVWordWordHeader;
    Entry     : tRTPSSerializedDataEntry;
    pHelpByte : pByte;
    i         : integer;
    len       : integer;

begin
     if Offset + 4 <= priv_lastoffset then
        begin
             TLVHdr := pTLVWordWordHeader (priv_framedata + Offset);

             Entry.DataType := TLVHdr.OptionType;
             Entry.DataLength := TLVHdr.OPtionLength;

             inc (Offset, 4);

             len := Min (Entry.DataLength, priv_sectionlen - Offset);

             Entry.Available := len;
             setLength (Entry.Data, len);

             for i := 0 to len - 1 do
                 begin
                      // The TLV Hdr is 4 bytes, so we skip it and then get each byte

                      pHelpByte := pByte (priv_framedata + Offset + i);

                      Entry.data[i] := pHelpByte^;
                 end;
        end
     else
         begin
              Entry.DataType := 0;
              Entry.DataLength := 0;
              Entry.Available := 0;
              setLength (Entry.Data, 0);
         end;

     result := Entry;
end;

(* ------------------------------------------- *)

function tRTPSParser.ParseDATASubmessage (Offset: integer; DataLen : integer; Flags: tRTPSSubMessageFlagSet): tRTPSDataMessage;

var pDataMessageHdr : pRTPSDataMessageStruct;
    pEncapHeader    : pRTPSDataMessageEncapStruct;
    DataMessage     : tRTPSDataMessage;
    TLVHdr          : pTLVWordWordHeader;
    pHelpByte       : pByte;
    len             : integer;
    ParsedLen       : integer;
    Entry           : tRTPSSerializedDataEntry;
    i               : integer;

begin
     pDataMessageHdr := pRTPSDataMessageStruct (priv_framedata + Offset);

     DataMessage.ExtraFlags := pDataMessageHdr^.ExtraFlags;
     DataMessage.LenQoS := pDataMessageHdr^.LenQoS;
     DataMessage.ReaderEntity := swap32 (pDataMessageHdr^.ReaderEntity);
     DataMessage.WriterEntity := swap32 (pDataMessageHdr^.WriterEntity);

     // Unfortuntely the 64 bit sequence is stored as hi and lo 32 bit values

     DataMessage.WriterSeq := pDataMessageHdr^.WriterSeqHi;
     DataMessage.WriterSeq := DataMessage.WriterSeq shl 32;
     DataMessage.WriterSeq := DataMessage.WriterSeq + pDataMessageHdr^.WriterSeqLo;

     Offset := Offset + sizeOf (tRTPSDataMessageStruct);

     ParsedLen := sizeOf (tRTPSDataMessageStruct);

     // Deal with QoS if it's in there

     if rtpsfTimestamp in flags then
        begin
             len := 0;

             repeat
                   // Parse QoS entries

                   Entry := ParseSerializedDataEntryFromOffset (Offset);

                   inc (len);
                   setLength (DataMessage.InlineQoSList, len);
                   DataMessage.InlineQoSList[len-1] := Entry;

                   if Entry.DataType = RTPS_PID_SENTINEL then
                      begin
                           // Adjust the offset

                           Offset := Offset + 4;
                           ParsedLen := ParsedLen + 4;
                      end
                   else
                       begin
                            // Adjust the offset

                            Offset := Offset + 4 + Entry.DataLength;
                            ParsedLen := ParsedLen + 4 + Entry.DataLength;
                       end;

             until Entry.DataType = RTPS_PID_SENTINEL;
        end;

     // Parse encapsulation stuff

     pEncapHeader := pRTPSDataMessageEncapStruct (priv_FrameData + Offset);

     if rtpsfEndianess in Flags then
        DataMessage.EncapsulationKind := swap16 (pEncapHeader^.EncapsulationKind)
     else
         DataMessage.EncapsulationKind := pEncapHeader^.EncapsulationKind;

     DataMessage.EncapsulationOpts := pEncapHeader^.EncapsulationOpts;

     Offset := Offset + sizeOf (tRTPSDataMessageEncapStruct);

     ParsedLen := ParsedLen + sizeOf (tRTPSDataMessageEncapStruct);

     // Read serialized data, either specificially for given data types, or just into a byte array

     if (DataMessage.WriterEntity = $3c2) or (DataMessage.WriterEntity = $4c2) or (DataMessage.WriterEntity = $100c2) then
        begin
             // Parse the single entries for this DATA submessage

             len := 0;
             setLength (DataMessage.DataList, len);

             while (ParsedLen < DataLen) and (Offset < priv_lastoffset) do
                   begin
                        Entry := ParseSerializedDataEntryFromOffset (Offset);

                        // Add new submessage to the list

                        inc (len);
                        setLength (DataMessage.DataList, len);
                        DataMessage.DataList[len-1] := Entry;

                        // Adjust the offset

                        Offset := Offset + 4 + Entry.DataLength;

                        ParsedLen := ParsedLen + 4 + Entry.DataLength;
                   end;
        end
     else
         begin
              // Just copy the serialized data to a byte array

              len := DataLen - ParsedLen;
              setLength (DataMessage.Data, len);

              for i := 0 to len - 1 do
                  begin
                       pHelpByte := pByte (priv_framedata + Offset + i);

                       DataMessage.Data[i] := pHelpByte^;
                  end;
         end;

     result := DataMessage;
end;

(* ------------------------------------------- *)

procedure tRTPSParser.ParseRTPSSubMessages (Offset: integer);

var pSubMessageHdr  : pRTPSSubMessageStruct;
    SubMessage      : tRTPSSubMessage;
    pDataMessageHdr : pRTPSDataMessageStruct;
    WriterType      : longword;
    pHelpByte       : pByte;
    len             : word;
    i               : integer;

begin
     while Offset < priv_lastoffset do
           begin
                SubMessage.DataMessage.initialize;

                pSubMessageHdr := pRTPSSubMessageStruct (priv_framedata + Offset);

                SubMessage.MessageID := pSubMessageHdr^.MessageID;
                SubMessage.Flags := ParseSubMessageFlags (pSubMessageHdr^.Flags);

                if rtpsfEndianess in SubMessage.Flags then
                   SubMessage.Len := pSubMessageHdr^.Len                  // Len is not byte swapped
                else
                    SubMessage.Len := swap16 (pSubMessageHdr^.Len);

                Inc (Offset, sizeOf (tRTPSSubMessageStruct));

                // Set the data length, either as specified or less (if the frame is truncated)

                len := min (pSubMessageHdr.Len, priv_sectionlen - offset);

                SubMessage.Available := len;
                setLength (SubMessage.data, len);

                // Copy the whole submessage byte by byte

                for i := 0 to len - 1 do
                    begin
                         // The Submessage Hdr is 4 bytes, so we skip it and then get each byte

                         pHelpByte := pByte (priv_framedata + Offset + i);

                         SubMessage.data[i] := pHelpByte^;
                    end;

                // Now the special stuff in case we have one of the writerEntityIDs that we need to keep
                // and if so, add it as a data message to the current submessage

                case SubMessage.MessageID of
                 RTPS_ID_DATA       : begin
{
                                           pDataMessageHdr := pRTPSDataMessageStruct (priv_framedata + Offset + 4);
                                           WriterType := swap32 (pDataMessageHdr^.WriterEntity);

//                                         if (WriterType = $3c2) or (WriterType = $4c2) then
}
                                           SubMessage.DataMessage := ParseDATASubmessage (Offset, SubMessage.Len, SubMessage.Flags);
                                      end;
                 RTPS_ID_DATA_FRAG  : begin

                                      end;
                 RTPS_ID_DATA_BATCH : begin

                                      end;
                end;

                // Add new submessage to the list, but only if the message is complete

                if SubMessage.Available = SubMessage.Len then
                   begin
                        len := length (priv_submessages);

                        setLength (priv_submessages, len + 1);
                        priv_submessages[len] := SubMessage;
                   end;

                // Adjust the offset

                Offset := Offset + SubMessage.Len;
                prot_layersize := prot_layersize + 4 + SubMessage.Len;
           end;
end;

(* ------------------------------------------- *)

function tRTPSParser.parse: tParseResult;

var pData    : pRTPSStruct;
    pGUID    : pRTPSGUIDStruct;
    status   : tParseResult;
    hWord    : word;
    Len      : integer;
    help     : pANSIChar;
    sHelp    : string;

begin
     priv_osilayer := 7;     // Set Layer this parser is for
     prot_layersize := sizeOf (tRTPSStruct) + sizeOf (tRTPSGUIDStruct);

     status := prSuccess;

     priv_isPing := false;

     if priv_lastoffset < priv_frameoffset + sizeOf (tRTPSStruct) then
        status := prFailUndersize
     else
         try
            pData := pRTPSStruct (priv_framedata + priv_frameoffset);

            hWord := swap16 (pData^.Version);

            priv_MajorVersion  := (hWord and $ff00) shr 8;
            priv_MinorVersion  := (hWord and $00ff);
            priv_VendorID := swap16 (pData^.VendorID);
{
            priv_GUIDHostID := swap32 (pData^.GUIDPrefixHostID);
            priv_GUIDAppID := swap32 (pData^.GUIDPrefixAppID);
            priv_GUIDCounter := swap32 (pData^.GUIDPrefixCounter);
}
            Len := priv_lastoffset - (priv_frameoffset + sizeOf (tRTPSStruct));

            if Len = 8 then
               begin
                    // This is probably a NDDSPING

                    Help := pANSIChar (priv_framedata + priv_frameoffset + sizeOf (tRTPSStruct));
                    sHelp := string (Help);
                    sHelp := copy (sHelp, 1, 8);

                    priv_isPing := (sHelp = RTPS_PING);
               end;

            if not priv_isPing and (priv_lastoffset > priv_frameoffset + sizeOf (tRTPSStruct) + sizeOf (tRTPSGUIDStruct)) then
               begin
                    pGUID := pRTPSGUIDStruct (priv_framedata + priv_frameoffset + sizeOf (tRTPSStruct));

                    priv_GUIDHostID := swap32 (pGUID^.GUIDPrefixHostID);
                    priv_GUIDAppID := swap32 (pGUID^.GUIDPrefixAppID);
                    priv_GUIDCounter := swap32 (pGUID^.GUIDPrefixCounter);

                    // Now parse SubMessages

                    ParseRTPSSubMessages (priv_frameOffset + sizeOf (tRTPSStruct) + sizeOf (tRTPSGUIDStruct));
               end;

             // Prepare next anchor

             priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
             priv_nextAnchor.Offset := priv_lastoffset + 1;

         except
               status := prFailInvalidValues;
         end;

     // Fix layersize in case the frame is a fragment or sliced

//   prot_layersize := Min (prot_layersize, priv_framelength - priv_frameoffset);
     prot_layersize := priv_sectionlen;

     result := status;
end;

(* ------------------------------------------- *)

end.
