{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_udp;

interface

uses base_structures,
     base_checksums,
     common_parse_helpers,
     definition_ip_addresses,
     class_parser_baseclass;


type pUDPStruct = ^tUDPStruct;
     tUDPStruct = packed record
                               sourceport      : word;
                               destinationport : word;
                               length          : word;
                               crc             : word;
                          end;

     tUDPParser   = class (tFrameParser)

                          constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength, UDPSectionLength : integer; const FrameIndex : integer = 0); reintroduce;

                          private
                                 priv_sourceport      : word;
                                 priv_destinationport : word;
                                 priv_length          : word;
                                 priv_crc             : word;

                                 priv_sectionlen      : integer;
                                 priv_lastoffset      : integer;

                                 priv_udpsectionwords : tBufferWords;

                                 function DetermineHeaderTypeFromPayloadMagic : tHeaderType;

                          public
                                property SourcePort : word read priv_sourceport;
                                property DestinationPort : word read priv_destinationport;
                                property Length : word read priv_length;
                                property Checksum : word read priv_crc;

                                property UDPSectionAsWordBuffer : tBufferWords read priv_udpsectionwords;

                                function checkIsPlausible : boolean; override;
                                function parse : tParseResult; override;

                                function PortPair : tPortPair;
                    end;


implementation

uses Toolkit;

(* ------------------------------------------- *)

constructor tUDPParser.create (FrameData: pointer; const FrameOffset, DataLength, WireLength, UDPSectionLength, FrameIndex: integer);

begin
     inherited create (FrameData, FrameOffset, DataLength, WireLength, FrameIndex);

     priv_sectionlen := UDPSectionLength;

     // calculate the end offset we can parse up to

     if (priv_frameoffset + priv_sectionlen) > priv_framelength  then
        priv_lastoffset := priv_framelength
     else
         priv_lastoffset := priv_frameoffset + priv_sectionlen - 1;
end;

(* ------------------------------------------- *)

function tUDPParser.checkIsPlausible: boolean;

begin
     result := true;                // Not much we can do anyway
end;

(* ------------------------------------------- *)

function tUDPParser.DetermineHeaderTypeFromPayloadMagic : tHeaderType;

var value : tHeaderType;
    help  : pANSIChar;
    sHelp : string;

begin
     value := NEXT_HEADER_DYNAMIC;

     // Check if we have RTPS next

     if (priv_nextAnchor.Offset + 4 < priv_lastoffset) then
        begin
             Help := pANSIChar (priv_framedata + priv_nextAnchor.Offset);
             sHelp := string (Help);
             sHelp := copy (sHelp, 1, 4);

             if sHelp = 'RTPS' then
                value := NEXT_HEADER_RTPS;
        end;

     result := value;
end;

(* ------------------------------------------- *)

function tUDPParser.parse: tParseResult;

var pData   : pUDPStruct;
    status  : tParseResult;
    CalcCRC : boolean;
    help    : word;
    iCount  : integer;


begin
     priv_osilayer := 4;     // Set Layer this parser is for
     prot_layersize := 8;

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tUDPStruct) then
        status := prFailUndersize
     else
         try
            pData := pUDPStruct (priv_framedata + priv_frameoffset);

            help := pData^.sourceport;
            priv_sourceport := Swap16 (help);

            help := pData^.destinationport;
            priv_destinationport := Swap16 (help);

            help := pData^.length;
            priv_length := Swap16 (help);

            help := pData^.crc;
            priv_crc := Swap16 (help);

            // Prepare the word array for crc calculation if possible - frame must be long enough

            CalcCRC := (priv_framelength = priv_wirelength);

            if CalcCRC then
               priv_udpsectionwords := getWordsStructFromOffset (priv_FrameData, priv_frameOffset, priv_length)
            else
                begin
                     status := prFailPayloadTruncated;
                     setlength (priv_udpsectionwords, 0);
                end;

            // Prepare next anchor

            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tUDPStruct);

            // Check if there is UDP payload at all, otherwise set it to none

            if (priv_length > 8) and (priv_nextAnchor.Offset < priv_framelength) then
               priv_nextAnchor.HeaderType := DetermineHeaderTypeFromPayloadMagic
            else
                priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

function tUDPParser.PortPair: tPortPair;

begin
     result.Value1 := priv_sourceport;
     result.Value2 := priv_destinationport;
end;

(* ------------------------------------------- *)

end.

