{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

program pcaptouch;

{$APPTYPE CONSOLE}

{$R *.res}

uses
  System.SysUtils,
  DateUtils,
  {$i base\loaderincludes.inc}
  class_dirparser in 'base\class_dirparser.pas';

(* ------------------------------------------- *)

procedure showhelp;

begin
     writeln ('pcaptouch [options] <filename>');
     writeln;
     writeln ('Options:');
     writeln ('/r         recursive');
     writeln ('/v         verbose mode');
     writeln ('/e         fix extension');
     writeln ('/h         show help');
end;

(* ------------------------------------------- *)

function getFileTimestamp (const Filename : string) : tDateTime;

var sr    : TSearchRec;

begin
     if FindFirst(Filename, faAnyFile, sr ) = 0 then
        begin
             result := sr.TimeStamp;

             FindClose(sr) ;
        end
     else
         result := 0;
end;

(* ------------------------------------------- *)

procedure touchfile (const filename : string);

var loader     : tTraceLoader;
    frame      : tRawFrame;
    timestamp  : tDateTime;
    current    : tDateTime;
    help       : integer;
    rcode      : integer;
    attrs      : integer;
    verbose    : boolean;
    fixext     : boolean;
    ext        : string;
    correctext : string;

begin
     attrs := filegetattr (fileName);

     if (attrs and faReadOnly > 0) or (attrs and faSysFile > 0) then
        begin
             writeln ('Skipping "' + filename + '": file is write protected');
             exit;
        end;

     verbose := FindCmdLineSwitch ('v');
     fixext := FindCmdLineSwitch ('e');

     current := getFileTimestamp (filename);
     ext := ExtractFileExt (filename);

     loader := createTraceLoader (filename);

     case loader.FileFormat of
      cfENC    : correctext := '.enc';
      cfCAP    : correctext := '.cap';
      cfPCAP   : correctext := '.pcap';
      cfPCAPng : correctext := '.pcapng';
     end;

     if not assigned (loader) and verbose then
        writeln ('Skipping "' + filename + '": unknown file format');

     if assigned (loader) and loader.openFile then
        begin
             try
                loader.readTraceFileheader;
             except
                   writeln ('Skipping "' + filename + '": cannot read file header');
                   exit;
             end;

             frame := loader.readSingleFrameFromDisk(1);
             loader.closeFile;

             if assigned (frame) then
                begin
                     timestamp := frame.Timestamp.datetime;

                     // Convert to filedate and back (otherwise it seems that the filetime will often be sligthly off every time when comparing them)

                     help := DateTimeToFileDate (timestamp);
                     timestamp := FileDateToDateTime (help);

                     frame.free;

                     if (timestamp <> current) then
                        begin
                             if (timestamp <> 0) then
                                begin
                                     try
                                        rcode := FileSetDate (filename, DateTimeToFileDate (timestamp));
                                     finally
                                            case rcode of
                                             0 : begin
                                                      if verbose then
                                                         writeln ('Successfully adjusted timestamp for file "' + filename + '" to ' + DateTimeToStr (timestamp));
                                                 end;
                                             1 : writeln ('Setting time for file "' + filename + '" failed: INVALID_FUNCTION');
                                             2 : writeln ('Setting time for file "' + filename + '" failed: FILE_NOT_FOUND');
                                             3 : writeln ('Setting time for file "' + filename + '" failed: PATH_NOT_FOUND');
                                             4 : writeln ('Setting time for file "' + filename + '" failed: TOO_MANY_OPEN_FILES');
                                             5 : writeln ('Setting time for file "' + filename + '" failed: ACCESS_DENIED');
                                             6 : writeln ('Setting time for file "' + filename + '" failed: INVALID_HANDLE');
                                             else
                                                 writeln ('Setting time for file "' + filename + '" failed');
                                            end;
                                     end;
                                end
                             else
                                 writeln ('Could not set time for "' + filename + '": invalid timestamp of first frame');
                        end
                     else
                         if verbose then
                            writeln ('Skipped "' + filename + '": file timestamp already correct');
                end;
        end
     else
         writeln ('Setting time for "' + filename + '" failed: could not read file');

     if fixext and (ext <> correctext) then
        begin
             if verbose then
                writeln ('Fixed extension for "' + filename + '" from ' + ext + ' to ' + correctext);

             RenameFile (filename, ChangeFileExt (filename, correctext));
        end;
end;

(* ------------------------------------------- *)

procedure main;

var scanner  : tDirParser;
    filename : string;
    param    : string;
    path     : string;
    i        : integer;

begin
     path := '';
     filename := '';

     if (ParamCount = 0) or FindCmdLineSwitch ('h') then
        showhelp
     else
         begin
              for i  := 1 to ParamCount do
                  begin
                       param := ParamStr (i);

                       if pos ('/', param) = 0 then
                          begin
                               path := ExtractFilePath (param);
                               filename := ExtractFileName (param);
                          end;
                  end;

              if filename <> '' then
                 begin
                      if FileExists (path + filename) then
                         touchfile (path + filename)
                      else
                          begin
                               // Directory processing mode

                               scanner := tDirParser.create;
                               scanner.BaseDirectory := path;
                               scanner.FileMask := filename;
                               scanner.BulkSetValidExtensions ('pcap,pcapng,cap,enc,dmp', ',', false);
                               scanner.IncludeSubDirs := FindCmdLineSwitch ('r');

                               scanner.execute;

                               for i := 1 to scanner.FileCount do
                                   begin
                                        filename := scanner.getFileName (i);

                                        touchfile (filename);
                                   end;
                          end;
                 end
              else
                  showhelp;
         end;
end;

(* ------------------------------------------- *)

begin
     try
        main;

     except
           on E: Exception do
              Writeln(E.ClassName, ': ', E.Message);
     end;
end.
