{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_fqdnmanager;

interface

uses Generics.Collections,
     Classes,
     VirtualTrees,
     definition_ip_addresses;

type tFQDNLabelCounter = record
                               FQDNLabel : tFQDNLabel;
                               UseCount  : integer;
                         end;
     tFQDNLabelCountList = array of tFQDNLabelCounter;

     tFQDNDetailNode = class (tObject)

                             constructor create (const DomainName : string);

                             private
                                    priv_Name          : tFQDNLabel;
                                    priv_Parent        : tFQDNDetailNode;
                                    priv_SubDomains    : tObjectDictionary<string, tFQDNDetailNode>;
                                    priv_Answers       : tDictionary<string, tNameResolutionAnswer>;
                                    priv_FullFQDN      : string;
                                    priv_VSTNode       : pVirtualNode;

                             public
                                   property Name : tFQDNLabel read priv_name;
                                   property VirtualTreeNode : pVirtualNode read priv_VSTNOde;

                                   procedure AddAnswer (const NewAnswer : tNameResolutionAnswer);

                                   destructor Destroy; override;
                       end;

     tFQDNManager = class (tObject)

                          constructor create (const IgnoreCase : boolean = false);

                          private
                                 priv_FQDNRootNode        : tFQDNDetailNode;
                                 priv_dictFullFQDNList    : tObjectDictionary<string, tFQDNDetailNode>;
                                 priv_dictFQDNLabelCount  : tDictionary<string, integer>;
                                 priv_FullFQDNList        : tStringList;
                                 priv_DuplicatesAdded     : integer;
                                 priv_IgnoreCase          : boolean;

                                 function getKeyForFQDNString (const value : string) : string;
                                 function getKeyForFQDN (const FQDN : tFQDN) : string;
                                 function getFQDNCount : integer;

                                 procedure AddSubNodesToTree (var Tree : TVirtualStringTree; ParentNode : pVirtualNode; entry : tFQDNDetailNode);

                          public
                                property count : integer read getFQDNCount;
                                /// <summary>
                                ///   Total Number of FQDNs that were added
                                ///   more than once
                                /// </summary>
                                property DuplicatesAdded : integer read priv_DuplicatesAdded;
                                /// <summary>
                                ///   if set, the storage will ignore character
                                ///   case of FQDNs when adding/comparing
                                ///   entries
                                /// </summary>
                                property IgnoreCase : boolean read priv_IgnoreCase write priv_IgnoreCase;
                                /// <summary>
                                ///   Adds a FQDN to the FQDN manager and
                                ///   returns the final subelement
                                /// </summary>
                                /// <param name="value">
                                ///   New FQDN to be added
                                /// </param>
                                /// <returns>
                                ///   Last element of the FQDN that was added
                                ///   (last node in the tree added for the new
                                ///   FQDN)
                                /// </returns>
                                function AddFQDN (const value : tFQDN) : tFQDNDetailNode;
                                /// <summary>
                                ///   Returns true if a FQDN was already added
                                ///   earlier.
                                /// </summary>
                                /// <remarks>
                                ///   This function honors the case
                                ///   sensitiveness setting, so <see href="http://www.packet-foo.com">
                                ///   www.packet-foo.com</see> is not the same
                                ///   as <see href="http://www.packet-FOO.com">
                                ///   WWW.packet-FOO.com</see> if set
                                /// </remarks>
                                function FQDNExists (const Value : tFQDN) : boolean;
                                /// <summary>
                                ///   Returns the internal detail node for the
                                ///   FQDN, which contains the last subelement
                                ///   of the FQDN (plus pointers to all parent
                                ///   elements up to the root node)
                                /// </summary>
                                /// <param name="Value">
                                ///   FQDN for which the detail node should be
                                ///   returned
                                /// </param>
                                /// <returns>
                                ///   detail node for the FQDN, or NIL if the
                                ///   FQDN does not exist
                                /// </returns>
                                function GetFDQNDetailNode (const Value : tFQDN) : tFQDNDetailNode;
                                /// <summary>
                                ///   Returns the virtualtree node with the
                                ///   last element of the FQDN, e.g. for adding
                                ///   more subnodes
                                /// </summary>
                                function GetFQDNVirtualTreeNode (const Value : tFQDN) : pVirtualNode;
                                /// <summary>
                                ///   Fills a VirtualStringTree with nodes from
                                ///   the internal FQDN tree
                                /// </summary>
                                procedure DisplayTree (var Tree : TVirtualStringTree; const AggregatedMode : boolean);
                                /// <summary>
                                ///   Returns the number of times each sub
                                ///   element of an FQDN (a "label") was seen
                                ///   in all FQDNs added to the list
                                /// </summary>
                                function GetLabelCounters : tFQDNLabelCountList;

                                destructor Destroy; override;
                    end;

     tFQDNTreeNodeType = (tntFQDN, tntDetailQuery, tntDetailAnswer);

     tFQDNTreeNode = record
                           nodetype           : tFQDNTreeNodeType;
                           name               : string;
                           framecount         : integer;
                           querycount         : integer;
                           answercount        : integer;
                           SuccessfullQueries : integer;
                           FailedQueries      : integer;
                           ServerFailures     : integer;
                           FormatErrors       : integer;
                           NXDomain           : integer;
                           NotImplemented     : integer;
                           QueryRefused       : integer;
                           NotAuthorized      : integer;

                           entry : tFQDNDetailNode;

                           procedure initialize (TreeNodeType : tFQDNTreeNodeType);
                     end;
     pFQDNTreeNode = ^tFQDNTreeNode;

implementation

uses SysUtils;

(* ----------------------------------- *)
{ tFQDNTreeNode }
(* ----------------------------------- *)

procedure tFQDNTreeNode.initialize (TreeNodeType: tFQDNTreeNodeType);

begin
     nodetype := TreeNodeType;
     name := '';
     framecount := 0;
     querycount := 0;
     answercount := 0;
     SuccessfullQueries := 0;
     FailedQueries := 0;
     ServerFailures := 0;
     FormatErrors := 0;
     NXDomain := 0;
     NotImplemented := 0;
     QueryRefused := 0;
     NotAuthorized := 0;
end;

(* ----------------------------------- *)
{ tFQDNNode }
(* ----------------------------------- *)

constructor tFQDNDetailNode.create (const DomainName: string);

begin
     inherited create;

     priv_name := DomainName;
     priv_subdomains := tObjectDictionary<string, tFQDNDetailNode>.create ([doOwnsValues]);
     priv_Answers := tDictionary<string, tNameResolutionAnswer>.create;
     priv_FullFQDN := '';
     priv_VSTNode := nil;
end;

(* ----------------------------------- *)

destructor tFQDNDetailNode.Destroy;

var key : string;

begin
     // recursive cleanup

     for key in priv_subdomains.keys do
         priv_subdomains.Remove(key);

     priv_Answers.Free;

     inherited;
end;

(* ----------------------------------- *)

procedure tFQDNDetailNode.AddAnswer (const NewAnswer: tNameResolutionAnswer);

var key : string;

begin
     key := NewAnswer.DictionaryKey;

     if not priv_Answers.ContainsKey (key) then
        priv_Answers.Add (key, NewAnswer);
end;

(* ----------------------------------- *)
{ tFQDNManager }
(* ----------------------------------- *)

constructor tFQDNManager.create (const IgnoreCase : boolean = false);

begin
     inherited create;

     priv_FQDNRootNode := tFQDNDetailNode.create ('.');
     priv_dictFullFQDNList := tObjectDictionary<string, tFQDNDetailNode>.create;
     priv_dictFQDNLabelCount  := tDictionary<string, integer>.create;
     priv_FullFQDNList := tStringList.Create;
     priv_DuplicatesAdded := 0;
     priv_IgnoreCase := IgnoreCase;
end;

(* ----------------------------------- *)

destructor tFQDNManager.Destroy;

begin
     priv_dictFullFQDNList.Free;
     priv_dictFQDNLabelCount.Free;
     priv_FullFQDNList.Free;
     priv_FQDNRootNode.Free;

     inherited;
end;

(* ----------------------------------- *)

procedure tFQDNManager.AddSubNodesToTree (var Tree: TVirtualStringTree; ParentNode : pVirtualNode; entry: tFQDNDetailNode);

var node       : pVirtualNode;
    data       : pFQDNTreeNode;
    key        : string;

begin
     if assigned (entry) then
        begin
             node := Tree.AddChild (ParentNode);
             Entry.priv_VSTNode := node;

             data := Tree.GetNodeData (node);

             data^.initialize (tntFQDN);

             if entry.priv_SubDomains.Count > 0 then
                data^.name := entry.priv_name + ' (' + entry.priv_SubDomains.Count.ToString + ')'
             else
                 data^.name := entry.priv_name;

             data^.entry := entry;

             for key in entry.priv_subdomains.Keys do
                 AddSubNodesToTree (Tree, node, entry.priv_subdomains.Items[key]);
        end;
end;

(* ----------------------------------- *)

procedure tFQDNManager.DisplayTree (var Tree: TVirtualStringTree; const AggregatedMode : boolean);

var entry    : tFQDNDetailNode;
    subentry : tFQDNDetailNode;
    node     : pVirtualNode;
    data     : pFQDNTreeNode;
    key      : string;
    i        : integer;

begin
     Tree.BeginUpdate;
     Tree.Clear;

     if AggregatedMode then
        begin
             // Build a reverse FQDN tree with subnodes

             entry := priv_FQDNRootNode;

             for key in entry.priv_subdomains.Keys do
                 begin
                      subentry := entry.priv_subdomains.Items[key];

                      AddSubNodesToTree (Tree, nil, subentry)
                 end;
        end
     else
         begin
              // Build a plain list without subnodes

              priv_FullFQDNList.Sort;

              for i := 1 to priv_FullFQDNList.Count do
                  begin
                       key := priv_FullFQDNList[i-1];

                       entry := priv_dictFullFQDNList.Items[key];

                       node := Tree.AddChild (nil);
                       data := Tree.GetNodeData (node);

                       data^.initialize (tntFQDN);
                       data^.name := entry.priv_FullFQDN;
                       data^.entry := entry;

                       entry.priv_VSTNode := node;
                  end;
         end;

     Tree.EndUpdate;
end;

(* ----------------------------------- *)

function tFQDNManager.getKeyForFQDNString (const value: string): string;

begin
     if priv_IgnoreCase then
        result := lowercase (trim (value))
     else
         result := trim (value);
end;

(* ----------------------------------- *)

function tFQDNManager.getKeyForFQDN (const FQDN: tFQDN): string;

begin
     result := getKeyForFQDNString (FQDN.FullFQDN);
end;

(* ----------------------------------- *)

function tFQDNManager.getFQDNCount: integer;

begin
     result := priv_dictFullFQDNList.Count;
end;

(* ----------------------------------- *)

function tFQDNManager.GetLabelCounters: tFQDNLabelCountList;

var key   : string;
    count : integer;
    list  : tFQDNLabelCountList;

begin
     count := priv_dictFQDNLabelCount.Count;
     setLength (list, count);

     for key in priv_dictFQDNLabelCount.Keys do
         begin
              list[count].FQDNLabel := key;
              list[count].UseCount := priv_dictFQDNLabelCount.Items[key];
         end;

     result := list;
end;

(* ----------------------------------- *)

function tFQDNManager.AddFQDN (const value : tFQDN) : tFQDNDetailNode;

var entry    : tFQDNDetailNode;
    newentry : tFQDNDetailNode;
    part     : string;
    fqdnkey  : string;
    subkey   : string;
    count    : integer;
    i        : integer;

begin
     result := nil;

     // Check if the new FQDN is not empty and not yet stored

     fqdnkey := getKeyForFQDN (value);

     if (trim (value.FullFQDN) <> '') and not priv_dictFullFQDNList.ContainsKey (fqdnkey) then
        begin
             entry := priv_FQDNRootNode;

             // Process all sub elements of the new value

             for i := value.ElementCount downto 1 do
                 begin
                      part := value.Elements[i-1];

                      // Get key for subelement, case sensitive or not

                      subKey := getKeyForFQDNString (part);

                      // Check if the element already exists, and add it if it doesn't
                      // this way the FQDN is added from the root node down

                      if entry.priv_subdomains.ContainsKey (subkey) then
                         begin
                              entry := entry.priv_subdomains.Items[subkey];

                               if i = 1 then
                                  begin
                                       // Update FullFQDN, because now this is not just a subelement of a longer type anymore

                                       assert (entry.priv_FullFQDN = '', 'Entry has existing FQDN (tFQDNManager.AddFQDN)');

                                       entry.priv_FullFQDN := value.FullFQDN;

                                       // this can happen if a shorter FQDN is a final node being part of a longer existing FQDN

                                       if priv_IgnoreCase then
                                          priv_FullFQDNList.Add (lowercase (trim (value.FullFQDN)))
                                       else
                                           priv_FullFQDNList.Add (trim (value.FullFQDN));

                                       priv_dictFullFQDNList.Add (fqdnkey, entry);
                                  end;
                         end
                      else
                          begin
                               newEntry := tFQDNDetailNode.create (part);

                               // Add parent/child links

                               newEntry.priv_parent := entry;
                               entry.priv_subdomains.Add (subkey, newEntry);

                               // if we're at the end, add the final node to the full FQDN list

                               if i = 1 then
                                  begin
                                       newEntry.priv_FullFQDN := value.FullFQDN;

                                       if priv_IgnoreCase then
                                          priv_FullFQDNList.Add (lowercase (trim (value.FullFQDN)))
                                       else
                                           priv_FullFQDNList.Add (trim (value.FullFQDN));

                                       priv_dictFullFQDNList.Add (fqdnkey, newEntry);
                                  end;

                               // Count the occurrences

                               if priv_dictFQDNLabelCount.ContainsKey (subkey) then
                                  priv_dictFQDNLabelCount.Items[subkey] := priv_dictFQDNLabelCount.Items[subkey] + 1
                               else
                                   priv_dictFQDNLabelCount.Add (subkey, 1);

                               entry := newEntry;
                          end;
                 end;

             result := entry;
        end
     else
         inc (priv_DuplicatesAdded);
end;

(* ----------------------------------- *)

function tFQDNManager.FQDNExists (const Value: tFQDN): boolean;

var key : string;

begin
     key := getKeyForFQDN (value);

     result := priv_dictFullFQDNList.ContainsKey (key);
end;

(* ----------------------------------- *)

function tFQDNManager.GetFDQNDetailNode (const Value: tFQDN): tFQDNDetailNode;

var key : string;

begin
     key := getKeyForFQDN (value);

     if priv_dictFullFQDNList.ContainsKey (key) then
        result := priv_dictFullFQDNList.Items[key]
     else
         result := nil;
end;

(* ----------------------------------- *)

function tFQDNManager.GetFQDNVirtualTreeNode (const Value: tFQDN): pVirtualNode;

var entry : tFQDNDetailNode;
    key   : string;

begin
     key := getKeyForFQDN (value);

     if priv_dictFullFQDNList.ContainsKey (key) then
        result := priv_dictFullFQDNList.Items[key].VirtualTreeNode
     else
         result := nil;
end;

(* ----------------------------------- *)


end.
