{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_inteldb_manager;

interface

uses SysUtils,
     base_structures,
     definition_ip_addresses,

     // FireDAC

     FireDAC.Stan.Intf, FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf,
     FireDAC.Phys.Intf, FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.dAPT,
     FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.ExprFuncs,
     FireDAC.VCLUI.Wait, FireDAC.Comp.UI, FireDAC.Phys.SQLite,
     Data.DB, FireDAC.Comp.Client,

     pcapng_structures,
     IntelDB_Tools;


type // Intel DB Manager Object

     tOnDBVersionWrong = procedure (Sender : tObject; const IsVersion, ShouldVersion : string) of object;

     tIntelDBManager = class (tObject)

                             {$REGION 'Documentation'}
                             ///	<summary>
                             ///	  Creates an instance of the trace IntelDB
                             ///	  manager, which can store and retrieve
                             ///	  details about the trace file that was
                             ///	  specified in the Filename variable. This
                             ///	  means that for each file a instance of the
                             ///	  IntelDB manager must be created; you cannot
                             ///	  reuse it for another file.
                             ///	</summary>
                             ///	<param name="DBFileName">
                             ///	  Name of the trace intel database file
                             ///	</param>
                             ///	<param name="Filename">
                             ///	  Name of the trace file to retrieve and store
                             ///	  intel about
                             ///	</param>
                             ///	<param name="FileSize">
                             ///	  Size of the trace file
                             ///	</param>
                             ///	<param name="FileDate">
                             ///	  Date/Time of the trace file
                             ///	</param>
                             ///	<param name="UseDiskDatabase">
                             ///	  If true, the SQLite DB will be written to
                             ///	  disk right away, instead of at disconnect
                             ///	</param>
                             ///	<param name="SharedDBMode">
                             ///	  If true, the SQLite DB can be opened
                             ///	  concurrently to check results while running
                             ///	  TW
                             ///	</param>
                             {$ENDREGION}
                             constructor create (const DBFileName, Filename: string; const FileSize : int64; FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp; const UseDiskDatabase : boolean=false; const SharedDBMode : boolean = false); overload;
                             constructor create; overload;

                             procedure free;

                             private
                                    priv_db                   : TFDConnection;
                                    priv_backupdb             : TFDSQLiteBackup;
                                    priv_queryaddress         : TFDQuery;
                                    priv_queryconversation    : TFDQuery;
                                    priv_queryaddresslink     : TFDQuery;
                                    priv_dbdirty              : boolean;

                                    priv_usediskdatabase      : boolean;
                                    priv_sharedmode           : boolean;

                                    priv_inteldbfilename      : string;
                                    priv_inteldbfileindex     : integer;

                                    priv_tracefilename        : string;
                                    priv_generalstatistics    : tGeneralTraceStats;
                                    priv_pcapngstatistics     : tPCAPNGStatistic;

                                    priv_fullfilescanned      : boolean;
                                    priv_pcapngstructscanned  : boolean;
                                    priv_conversationsscanned : boolean;
                                    priv_fileframeparsedcount : integer;
                                    priv_filefirstframetime   : tDateTime;
                                    priv_filelastframetime    : tDateTime;

//                                  priv_dbversionwrong       : boolean;
                                    priv_dbversion            : string;

                                    // Events

                                    EventOnDBVersionWrong     : tOnDBVersionWrong;

                                    // Internal helper routines

                                    function getFileName : string;
                                    function getFilePath : string;
                                    function getFileStatusFromDB (const Filename: string; const FileSize : int64; const FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp) : tFileIntelStatus;

                                    {$REGION 'Documentation'}
                                    ///	<summary>
                                    ///	  Helper routine to assign the
                                    ///	  standard values all chain block
                                    ///	  nodes have to the query in one
                                    ///	  single call
                                    ///	</summary>
                                    {$ENDREGION}
                                    procedure AddStandardBlockParamsToQuery (Query : TFDQUery; ChainNode: pBlockChainEntry; EntryIndex : integer);
                                    procedure StoreGeneralPCAPngStatisticsToDB (Query : TFDQUery);
                                    procedure ReadGeneralPCAPngStatisticsFromDB (Query : TFDQUery);
                                    procedure StoreSHBToIntelDB (Query : TFDQUery; ChainNode : pBlockChainEntry; EntryIndex : integer);
                                    procedure StoreNRBToIntelDB (Query : TFDQUery; ChainNode : pBlockChainEntry; EntryIndex : integer);
                                    procedure StoreIDBToIntelDB (Query : TFDQUery; ChainNode : pBlockChainEntry; EntryIndex : integer);
                                    procedure StoreISBToIntelDB (Query : TFDQUery; ChainNode : pBlockChainEntry; EntryIndex : integer);
                                    procedure StorePacketBlockToIntelDB (Query : TFDQUery; ChainNode : pBlockChainEntry; EntryIndex : integer);
                                    procedure StoreBasicBlockToIndelDB (Query : TFDQUery; ChainNode : pBlockChainEntry; EntryIndex : integer);
                                    procedure AssignQueryDetailsToChainNode (Query : TFDQUery; out NewEntry : pBlockChainEntry);
                                    procedure ReadNRBEntriesFromDB (var BlockNodeEntry : pBlockChainEntry);

                                    // File scan Status

                                    function getFileProcessingStatus: tProcessingStatus;

                                    // General Statistics

                                    procedure writeGeneralStatistics (Statistics : tGeneralTraceStats);
                                    function readGeneralStatistics : tGeneralTraceStats;

                             public
                                   property Database : TFDConnection read priv_db write priv_db;
                                   property FileProcessingStatus : tProcessingStatus read getFileProcessingStatus;
                                   property FullFileScanned : boolean read priv_fullfilescanned;
                                   property PCAPNGStructureScanned : boolean read priv_pcapngstructscanned;
                                   property ConversationsScanned : boolean read priv_conversationsscanned;
                                   property FirstFrameTimestamp : tDateTime read priv_filefirstframetime;
                                   property LastFrameTimestamp : tDateTime read priv_filelastframetime;
                                   property PreparsedTotalFrameCount : integer read priv_fileframeparsedcount;
                                   property GeneralStatistics : tGeneralTraceStats read readGeneralStatistics write writeGeneralStatistics;
                                   property CurrentFileIndex : integer read priv_inteldbfileindex;
                                   property CurrentTraceFilename : string read priv_tracefilename;
                                   property DatabaseDirty : boolean read priv_dbdirty write priv_dbdirty;

                                   property OnDBVersionWrong : tOnDBVersionWrong read EventOnDBVersionWrong write EventOnDBVersionWrong;

                                   procedure Connect;
                                   procedure Disconnect;
                                   /// <summary>
                                   ///   Copies the memory based SQLite
                                   ///   database to disk. This is usually
                                   ///   called while scanning files to enforce
                                   ///   writing results of each file to disk
                                   ///   before scanning the next (which might
                                   ///   crash) . This also clears the DBDirty
                                   ///   flag.
                                   /// </summary>
                                   procedure WriteMemDBtoDisk;

                                   {$REGION 'Documentation'}
                                   ///	<summary>
                                   ///	  This tells the intel DB manager which
                                   ///	  trace file name it is supposed to work
                                   ///	  with
                                   ///	</summary>
                                   ///	<param name="Filename">
                                   ///	  Name of the trace file values are
                                   ///	  stored and retrieved for
                                   ///	</param>
                                   {$ENDREGION}
                                   procedure AssignTraceFile (const Filename: string; const FileSize : int64; const FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp);

                                   function DBVersionIsWrong : boolean;

                                   // PCAPng Statistics

                                   function PCAPngStatisticsRecordsExistInDB : boolean;
                                   procedure writePCAPngStatisticsToDB (Statistics : tPCAPNGStatistic);
                                   function readPCAPngStatisticsFromDB : tPCAPNGStatistic;

                                   // Comments

                                   procedure WriteFrameComments (CommentList : tPCAPNGPacketCommentList);
                                   function ReadFrameComments : tPCAPNGPacketCommentList;

                                   {$REGION 'Documentation'}
                                   ///	<summary>
                                   ///	  Writes additional details to the
                                   ///	  filelist table, like first/last frame
                                   ///	  timestamp and frame count
                                   ///	</summary>
                                   {$ENDREGION}
                                   procedure writeAdditionalFileDetails;

                                   // File scan Status

                                   procedure setFileProcessingStatus (const ScanStatus : tProcessingStatus);

                                   function rename (const NewFilename : string) : boolean;
                       end;

     tIntelDBParameters = record
                                IntelDB               : tIntelDBManager;

                                UsePersistentIntelDB  : boolean;
                                DoNotUseIntelDB       : boolean;
                                HideIntelDBFile       : boolean;
                                UseLocalDBFiles       : boolean;        // store one db per directory
                                GlobalDBPath          : string;        // Path for storing dbs if not using locals

                                procedure initialize;
                          end;

const inteldbversion = '1.7';

implementation

uses classes,   // for TResourceStream
     types,     // for rt_rcdata;
     DateUtils,
     SiAuto,
     toolkit,
     databasetools,
     definition_simple_nameresolution;
//     class_parser_baseclass;


const standardblockfields = 'fileid,blocksequence,blocktypeid,sectionstartoffset,sectionendoffset,firstblockstartoffset,lastblockendoffset,blocklength,count';
      standardblockparams = ':fileid,:blocksequence,:blocktypeid,:sectionstartoffset,:sectionendoffset,:firstblockstartoffset,:lastblockendoffset,:blocklength,:count';


(* ------------------------------------------- *)
{ tIntelDBManager }
(* ------------------------------------------- *)

constructor tIntelDBManager.create (const DBFileName, Filename: string; const FileSize : int64; FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp; const UseDiskDatabase : boolean=false; const SharedDBMode : boolean = false);

begin
     SiMain.EnterMethod ('tIntelDBManager.create');

     SiMain.LogDebug ('DBFilename: ' + DBFilename);
     SiMain.LogDebug ('Filename: ' + Filename);
     SiMain.LogDebug ('Filesize: ' + IntToStr (FileSize));
     SiMain.LogDebug ('First Frametime: ' + DateTimeToStr (FirstFrameTime.datetime));

     if UseDiskDatabase then
        SiMain.LogDebug ('Using disk based SQLite database')
     else
         SiMain.LogDebug ('Using memory based SQLite database');

     if SharedDBMode then
        SiMain.LogDebug ('Database is in shared mode')
     else
         SiMain.LogDebug ('Database in in exclusive mode');

     inherited create;

     priv_inteldbfileindex := 0;
     priv_dbdirty := false;

//   priv_dbversionwrong := false;

     priv_inteldbfilename := DBFileName;
     priv_usediskdatabase := UseDiskDatabase;
     priv_sharedmode := SharedDBMode;

     // Init statistics

     priv_fileframeparsedcount := 0;
     priv_filefirstframetime := 0;
     priv_filelastframetime := 0;

     // Connect to database and add/retrieve the file details

     Connect;
     AssignTraceFile (Filename, FileSize, FileDate, FirstFrameTime, LastFrameTime);

     priv_queryaddress := TFDQUery.Create (nil);
     priv_queryaddress.Connection := priv_db;

     priv_queryconversation := TFDQUery.Create(nil);
     priv_queryconversation.Connection := priv_db;

     priv_queryaddresslink := TFDQUery.Create(nil);
     priv_queryaddresslink.Connection := priv_db;

     SiMain.LeaveMethod ('tIntelDBManager.create');
end;

(* ------------------------------------------- *)

constructor tIntelDBManager.create;

begin
     SiMain.EnterMethod ('tIntelDBManager.create');
     SiMain.LogDebug ('No parameters given');

     inherited create;

     priv_usediskdatabase := true;
     priv_sharedmode := true;

     Connect;

     priv_queryaddress := TFDQUery.Create (nil);
     priv_queryaddress.Connection := priv_db;

     priv_queryconversation := TFDQUery.Create(nil);
     priv_queryconversation.Connection := priv_db;

     priv_queryaddresslink := TFDQUery.Create(nil);
     priv_queryaddresslink.Connection := priv_db;

     SiMain.LeaveMethod ('tIntelDBManager.create');
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.free;

begin
     SiMain.EnterMethod ('tIntelDBManager.free');

     if priv_db.Connected then
        Disconnect;

     // Free stuff

     priv_db.Free;

     priv_queryaddress.Free;
     priv_queryaddress := nil;

     priv_queryconversation.Free;
     priv_queryconversation := nil;

     priv_queryaddresslink.Free;
     priv_queryaddresslink := nil;

     inherited;

     SiMain.LeaveMethod ('tIntelDBManager.free');
end;

(* ------------------------------------------- *)

function tIntelDBManager.getFileName: string;

begin
     result := extractFilename (priv_tracefilename);
end;

(* ------------------------------------------- *)

function tIntelDBManager.getFilePath: string;

begin
     result := extractFilePath (priv_tracefilename);
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.Connect;

var CreateDB    : boolean;
    DiskDB      : TFDConnection;
    Query       : TFDQUery;
    DBStructure : TResourceStream;
    DBFileSize  : int64;
    help        : AnsiString;
    sHelp       : string;

begin
     SiMain.LogVerbose ('Connecting to IntelDB');

     if not (assigned (priv_db) and priv_db.Connected) then
        begin
             priv_db := TFDConnection.Create(nil);
             priv_db.DriverName := 'SQLite';

             // Check if the intel db file exists

             if FileExists (priv_inteldbfilename) then
                begin
                     priv_db.Params.Add ('SQLite');
                     priv_db.Params.Add ('Database=' + priv_inteldbfilename);

                     if priv_sharedmode then
                        begin
                             priv_db.Params.Add('LockingMode=Normal');
                             priv_db.Params.Add('Synchronous=Normal');
                        end;

                     priv_db.Open;

                     // Check the database size and version

                     DBFileSize := getFileSize (priv_inteldbfilename);
                     CreateDB := (DBFileSize = 0);

                     if CreateDB or DBVersionIsWrong then
                        begin
                             // If it is empty or too old, we simply delete it and force recreation

                             priv_db.Close;

                             SiMain.LogWarning ('Deleting empty or outdated IntelDB file: "' + priv_inteldbfilename + '"');
                             deletefile (priv_inteldbfilename);

                             CreateDB := true;
                        end
                     else
                         priv_db.Close;
                end
             else
                 CreateDB := true;

             priv_db.Params.Clear;

             // Check

             if priv_usediskdatabase then
                begin
                     priv_db.DriverName := 'SQLite';
                     priv_db.Params.Add ('Database=' + priv_inteldbfilename);
                     priv_db.Open;
                end
             else
                 begin
                      priv_db.DriverName := 'SQLite';
                      priv_db.Params.Add ('Database=:memory:');
                      priv_db.Open;

                      if not CreateDB then
                         begin
                              DiskDB := TFDConnection.Create(nil);
                              DiskDB.DriverName := 'SQLite';
                              DiskDB.Params.Values['DriverID'] := 'SQLite';
                              DiskDB.Params.Values['Database'] := priv_inteldbfilename;
                              DiskDB.Params.Values['OpenMode'] := 'ReadOnly';
                              DiskDB.Open;

                              // Read database into memory if it is not a disk based run

                              SQliteBackup (DiskDB, priv_db);

                              DiskDB.Free;
                         end;
                 end;

             // Create Query

             Query := TFDQUery.Create(nil);
             Query.Connection := priv_db;

             // Create Database if not there yet

             if CreateDB then
                begin
                     // Get structure from Ressource (configured at Project -> Resources and Images)

                     DBStructure := TResourceStream.Create (HInstance, 'traceintel_dbstruct', RT_RCDATA);
                     setLength (help, DBStructure.Size);
                     DBStructure.ReadBuffer(help[1], DBStructure.Size);

                     Query.SQL.Text := help;
                     Query.ExecSQL;
                     DBStructure.Free;

                     // Write Meta, e.g. DB Version

                     Query.SQL.Text := 'INSERT INTO meta (name,value) VALUES ("version",:version);';
                     Query.ParamByName ('version').AsString := inteldbversion;
                     Query.ExecSQL;

                     priv_dbdirty := true;
                end;

             Query.Free;
        end;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.WriteMemDBtoDisk;

var DiskDB : TFDConnection;
    boOkay : boolean;

begin
     SiMain.LogVerbose ('Writing memory IntelDB to disk');

     boOkay := true;

     if FileExists (priv_inteldbfilename) then
        try
           DeleteFile (priv_inteldbfilename);
        except
              boOkay := false;
        end;

     if boOkay then
        begin
             DiskDB := TFDConnection.Create(nil);
             DiskDB.DriverName := 'SQLite';
             DiskDB.Params.Values['DriverID'] := 'SQLite';
             DiskDB.Params.Values['Database'] := priv_inteldbfilename;
             DiskDB.Open;

             // Write memory database to disk

             SQliteBackup (priv_db, DiskDB);

             DiskDB.Free;
        end;

     priv_dbdirty := False;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.Disconnect;

begin
     if priv_dbdirty then
        begin
             SiMain.LogVerbose ('Disconnecting from IntelDB, mem DB status is dirty');

             if not priv_usediskdatabase then
                WriteMemDBtoDisk;
        end
     else
         SiMain.LogVerbose ('Disconnecting from IntelDB, mem DB status is clean');

     priv_db.Close;
end;

(* ------------------------------------------- *)

function tIntelDBManager.DBVersionIsWrong : boolean;

var Query   : tFDQuery;
    invalid : boolean;

begin
     SiMain.EnterMethod ('tIntelDBManager.DBVersionIsWrong');

     Query := tFDQuery.Create (nil);
     Query.Connection := priv_db;

     Query.Open ('SELECT * FROM meta WHERE name="version";');

     if Query.RecordCount > 0 then
        begin
             priv_dbversion := Query.FieldByName ('value').AsString;

             if priv_dbversion <> inteldbversion then
                begin
                     SiMain.LogWarning ('Database version is ' + priv_dbversion + ', but should be ' + inteldbversion);

                     if assigned (EventOnDBVersionWrong) then
                        EventOnDBVersionWrong (self, priv_dbversion, inteldbversion);

                     invalid := true;
                end
             else
                 begin
                      SiMain.LogVerbose ('Database version is ' + priv_dbversion);

                      invalid := false;
                 end;
         end;

     Query.Free;

     result := invalid;

     SiMain.LeaveMethod ('tIntelDBManager.DBVersionIsWrong');
end;

(* ------------------------------------------- *)

function tIntelDBManager.getFileStatusFromDB (const Filename: string; const FileSize : int64; const FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp) : tFileIntelStatus;

begin
     // Call the tool function in inteldb_tools

     result := getIntelDBFileStatus (priv_db, Filename, FileSize, FileDate, FirstFrameTime, LastFrameTime);
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.AssignTraceFile (const Filename: string; const FileSize : int64; const FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp);

var Query         : TFDQUery;
    Status        : tFileIntelStatus;
    boForceAdd    : boolean;
    boDeleteEntry : boolean;

begin
     SiMain.EnterMethod('tIntelDBManager.AssignTraceFile');

     // Create Query

     Query := TFDQUery.Create(nil);
     Query.Connection := priv_db;

     priv_tracefilename := Filename;
     priv_generalstatistics.initialize;

     // Add file to database if not in there yet

     Status := getFileStatusFromDB (Filename, FileSize, FileDate, FirstFrameTime, LastFrameTime);

     if Status.DBIndex = 0 then
        begin
             boDeleteEntry := false;
             boForceAdd := true
        end
     else
         begin
              boDeleteEntry := not DBFileEntryMatchFound (Status);
              boForceAdd := boDeleteEntry;
         end;

     if boDeleteEntry then
         begin
              SiMain.LogVerbose('Removing old file entry for "' + filename + '"from database');

              // remove old entry

              Query.SQL.Text := 'DELETE FROM filelist WHERE id=:id;';
              Query.SQL.Add('DELETE FROM filestatistics WHERE fileid=:id;');
              Query.SQL.Add('DELETE FROM blockchainentries WHERE fileid=:id;');
              Query.SQL.Add('DELETE FROM pcapngfilestatistics WHERE fileid=:id;');
              Query.ParamByName ('id').AsInteger := Status.DBIndex;
              Query.ExecSQL;

              priv_dbdirty := true;
         end;

     if boForceAdd then
        begin
             SiMain.LogDebug ('Adding file "' + Filename + '" to database');

             // Put file into database

             Query.SQL.Text := 'INSERT INTO filelist (filename,filepath,filesize,filedate,firstframetimestamp,firstframenanoseconds,lastframetimestamp,lastframenanoseconds,LayerComplexity) ' +
                               'VALUES (:filename,:filepath,:filesize,:filedate,:firstframetimestamp,:firstframenanoseconds,:lastframetimestamp,:lastframenanoseconds,:LayerComplexity);';

             Query.ParamByName ('filename').AsString := lowercase (getFileName);
             Query.ParamByName ('filepath').AsString := lowercase (getFilePath);
             Query.ParamByName ('filesize').AsLargeInt := Filesize;
             Query.ParamByName ('filedate').AsDateTime := FileDate;
             Query.ParamByName ('firstframetimestamp').AsDateTime := FirstFrameTime.datetime;
             Query.ParamByName ('firstframenanoseconds').AsInteger := FirstFrameTime.nanoseconds;
             Query.ParamByName ('lastframetimestamp').AsDateTime := LastFrameTime.datetime;
             Query.ParamByName ('lastframenanoseconds').AsInteger := LastFrameTime.nanoseconds;
             Query.ParamByName ('layercomplexity').AsInteger := ComplexityUnknown;

             Query.ExecSQL;

             priv_dbdirty := true;

             // Get the status from the DB, including the file index

             Status := getFileStatusFromDB (Filename, FileSize, FileDate, FirstFrametime, LastFrameTime);
             priv_inteldbfileindex := Status.DBIndex;

             SiMain.LogDebug ('TraceIntel DB index for new file "' + Filename + '" is ' + IntToStr (priv_inteldbfileindex));
        end
     else
         begin
              SiMain.LogDebug ('Found file "' + Filename + '" in database, reading details');

              // Transfer file index from status record

              priv_inteldbfileindex := Status.DBIndex;

              SiMain.LogDebug ('TraceIntel DB index for file "' + Filename + '" is ' + IntToStr (priv_inteldbfileindex));

              // Read AdditionalFileDetails

              Query := TFDQUery.Create(nil);
              Query.Connection := priv_db;

              Query.SQL.Text := 'SELECT * from filelist WHERE id=:id;';
              Query.ParamByName ('id').AsInteger := priv_inteldbfileindex;
              Query.Open;

              if Query.RecordCount > 0 then
                 begin
                      priv_fullfilescanned := Query.FieldByName ('fullfilescanned').AsBoolean;
                      priv_pcapngstructscanned  := Query.FieldByName ('pcapngstructscanned').AsBoolean;
                      priv_conversationsscanned := Query.FieldByName ('conversationsscanned').AsBoolean;
                 end
              else
                  begin
                       SiMain.LogError ('Error: could not retrieve file details');

                       raise Exception.Create('Error: could not retrieve file details in tIntelDBManager.AssignTraceFile');
                  end;

              // Read general statistics

              if priv_fullfilescanned then
                 readGeneralStatistics;
         end;

     Query.Free;

     SiMain.LeaveMethod('tIntelDBManager.AssignTraceFile');
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.WriteGeneralStatistics (Statistics: tGeneralTraceStats);

var Query : TFDQUery;

begin
     Query := TFDQUery.Create (nil);
     Query.Connection := priv_db;

     Query.SQL.Text := 'SELECT id FROM filestatistics WHERE fileid=:fileid;';
     Query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
     Query.Open;

     if Query.RecordCount = 0 then
        begin
             Query.SQL.Text := 'INSERT INTO filestatistics (fileid,firstframetime,firstframetimenanosecs,firstframeoffset,lastframetime,lastframetimenanosecs,lastframeoffset,minframesize,maxframesize,averageframesize,framecount,totalframebytes,totalwirebytes,' +
                                                           'isFrameSliced,SliceIsConstant,SliceSize,TimeOrderIsCorrect,OutOfTimeFrames) ' +
                               'VALUES (:fileid,:firstframetime,:firstframetimenanosecs,:firstframeoffset,:lastframetime,:lastframetimenanosecs,:lastframeoffset,:minframesize,:maxframesize,:averageframesize,:framecount,:totalframebytes,:totalwirebytes,' +
                                       ':isFrameSliced,:SliceIsConstant,:SliceSize,:TimeOrderIsCorrect,:OutOfTimeFrames);';
        end
     else
         begin
              Query.SQL.Text := 'UPDATE filestatistics SET firstframetime=:firstframetime,firstframetimenanosecs=:firstframetimenanosecs,firstframeoffset=:firstframeoffset,lastframetime=:lastframetime,lastframetimenanosecs=:lastframetimenanosecs,lastframeoffset=:lastframeoffset,' +
                                'minframesize=:minframesize,maxframesize=:maxframesize,averageframesize=:averageframesize,framecount=:framecount,totalframebytes=:totalframebytes,totalwirebytes=:totalwirebytes,' +
                                'isFrameSliced=:isFrameSliced,SliceIsConstant=:SliceIsConstant,SliceSize=:SliceSize,TimeOrderIsCorrect=:TimeOrderIsCorrect,OutOfTimeFrames=:OutOfTimeFrames WHERE fileid=:fileid;';
         end;

     Query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
     Query.ParamByName ('firstframetime').AsDateTime := Statistics.FirstFrameTime.datetime;
     Query.ParamByName ('firstframetimenanosecs').AsInteger := Statistics.FirstFrameTime.nanoseconds;
     Query.ParamByName ('firstframeoffset').AsLargeInt := Statistics.FirstFrameOffset;
     Query.ParamByName ('lastframetime').AsDateTime := Statistics.LastFrameTime.datetime;
     Query.ParamByName ('lastframetimenanosecs').AsInteger := Statistics.LastFrameTime.nanoseconds;
     Query.ParamByName ('lastframeoffset').AsLargeInt := Statistics.LastFrameOffset;
     Query.ParamByName ('minframesize').AsInteger := Statistics.MinFrameSize;
     Query.ParamByName ('maxframesize').AsInteger := Statistics.MaxFrameSize;
     Query.ParamByName ('averageframesize').AsInteger := Statistics.AverageFrameSize;
     Query.ParamByName ('framecount').AsInteger := Statistics.FrameCount;
     Query.ParamByName ('totalframebytes').AsLargeInt := Statistics.TotalFrameBytes;
     Query.ParamByName ('totalwirebytes').AsLargeInt := Statistics.TotalWireBytes;
     Query.ParamByName ('isFrameSliced').AsBoolean := Statistics.isSliced;
     Query.ParamByName ('SliceIsConstant').AsBoolean := Statistics.SliceSizeConstant;
     Query.ParamByName ('SliceSize').AsInteger := Statistics.SliceSize;
     Query.ParamByName ('TimeOrderIsCorrect').AsBoolean := Statistics.TimeOrderCorrect;
     Query.ParamByName ('OutOfTimeFrames').AsInteger := Statistics.OutOfTimeFrames;

     Query.ExecSQL;

     priv_dbdirty := true;

     // Set file to fullfilescanned

     Query.SQL.Text := 'UPDATE filelist SET fullfilescanned=:fullfilescanned WHERE id=:id;';
     Query.ParamByName ('id').AsInteger := priv_inteldbfileindex;
     Query.ParamByName ('fullfilescanned').AsBoolean := true;
     Query.ExecSQL;

     Query.Free;
end;

(* ------------------------------------------- *)

function tIntelDBManager.readGeneralStatistics: tGeneralTraceStats;

var Query  : TFDQUery;
    values : tGeneralTraceStats;

begin
     if priv_generalstatistics.FrameCount = 0 then
        priv_generalstatistics := readGeneralStatisticsFromDB (priv_db, priv_inteldbfileindex);
{
        begin
             Query := TFDQUery.Create (nil);
             Query.Connection := priv_db;

             Query.SQL.Text := 'SELECT * FROM filestatistics WHERE fileid=:fileid;';
             Query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
             Query.Open;

             priv_generalstatistics.FirstFrameTime.datetime := Query.FieldByName ('firstframetime').AsDateTime;
             priv_generalstatistics.FirstFrameTime.nanoseconds := Query.FieldByName ('firstframetimenanosecs').AsInteger;
             priv_generalstatistics.FirstFrameOffset := Query.FieldByName ('firstframeoffset').AsLargeInt;
             priv_generalstatistics.LastFrameTime.datetime := Query.FieldByName ('lastframetime').AsDateTime;
             priv_generalstatistics.LastFrameTime.nanoseconds := Query.FieldByName ('lastframetimenanosecs').AsInteger;
             priv_generalstatistics.LastFrameOffset := Query.FieldByName ('lastframeoffset').AsLargeInt;
             priv_generalstatistics.MinFrameSize := Query.FieldByName ('minframesize').AsInteger;
             priv_generalstatistics.MaxFrameSize := Query.FieldByName ('maxframesize').AsInteger;
             priv_generalstatistics.AverageFrameSize := Query.FieldByName ('averageframesize').AsInteger;
             priv_generalstatistics.FrameCount := Query.FieldByName ('framecount').AsInteger;
             priv_generalstatistics.TotalFrameBytes := Query.FieldByName ('totalframebytes').AsLargeInt;
             priv_generalstatistics.TotalWireBytes := Query.FieldByName ('totalwirebytes').AsLargeInt;
             priv_generalstatistics.isSliced := Query.FieldByName ('isFrameSliced').AsBoolean;
             priv_generalstatistics.SliceSizeConstant := Query.FieldByName ('SliceIsConstant').AsBoolean;
             priv_generalstatistics.SliceSize := Query.FieldByName ('slicesize').AsInteger;
             priv_generalstatistics.TimeOrderCorrect := Query.FieldByName ('TimeOrderIsCorrect').AsBoolean;
             priv_generalstatistics.OutOfTimeFrames := Query.FieldByName ('OutOfTimeFrames').AsInteger;

             Query.Free;
        end;
}
     result := priv_generalstatistics;
end;

(* ------------------------------------------- *)

function tIntelDBManager.PCAPngStatisticsRecordsExistInDB: boolean;

var Query : TFDQUery;

begin
     Query := TFDQUery.Create (nil);
     Query.Connection := priv_db;

     result := false;

     // Check if we have a index for the filelist in the DB, and if so, wether there are block chain entries stored

     if priv_inteldbfileindex > 0 then
        begin
             Query.Open ('SELECT id FROM blockchainentries WHERE fileid=' + IntToStr (priv_inteldbfileindex));

             result := Query.RecordCount > 0;
        end;

     Query.Free;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.WriteFrameComments (CommentList: tPCAPNGPacketCommentList);

var Query : TFDQUery;
    i     : integer;

begin
     Query := TFDQUery.Create (nil);
     Query.Connection := priv_db;

     for i := 1 to length (CommentList) do
         begin
              Query.SQL.Text := 'INSERT INTO framecomments (fileid,frameid,comment) VALUES (:fileid,:frameid,:comment);';
              Query.ParamByName('fileid').AsInteger := priv_inteldbfileindex;
              Query.ParamByName('frameid').AsInteger := CommentList[i-1].FrameNo;
              Query.ParamByName('comment').AsString := CommentList[i-1].Comment;

              Query.ExecSQL;

              priv_dbdirty := true;
         end;

     Query.Free;
end;

(* ------------------------------------------- *)

function tIntelDBManager.ReadFrameComments: tPCAPNGPacketCommentList;

var Query : TFDQUery;
    list  : tPCAPNGPacketCommentList;
    i     : integer;

begin
     Query := TFDQUery.Create (nil);
     Query.Connection := priv_db;

     Query.SQL.Text := 'SELECT * FROM framecomments WHERE fileid=:fileid;';
     Query.ParamByName('fileid').AsInteger := priv_inteldbfileindex;
     Query.Open;

     setLength (list, Query.RecordCount);

     for i := 1 to Query.RecordCount do
         begin
              list[i-1].FrameNo := Query.FieldByName('frameid').AsInteger;
              list[i-1].Comment := Query.FieldByName('comment').AsString;
         end;

     Query.Free;

     result := list;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.AddStandardBlockParamsToQuery (Query: TFDQUery; ChainNode: pBlockChainEntry; EntryIndex: integer);

begin
     Query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
     Query.ParamByName ('blocksequence').AsInteger := EntryIndex;
     Query.ParamByName ('blocktypeid').AsInteger := ChainNode^.BlockTypeID;
     Query.ParamByName ('sectionstartoffset').AsLargeInt := ChainNode^.SectionStartOffset;
     Query.ParamByName ('sectionendoffset').AsLargeInt := ChainNode^.SectionEndOffset;
     Query.ParamByName ('firstblockstartoffset').AsLargeInt := ChainNode^.FirstBlockStartOffset;
     Query.ParamByName ('lastblockendoffset').AsLargeInt := ChainNode^.LastBlockEndOffset;
     Query.ParamByName ('blocklength').AsInteger := ChainNode^.BlockLength;
     Query.ParamByName ('count').AsLargeInt := ChainNode^.Count;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.StoreSHBToIntelDB (Query: TFDQUery; ChainNode: pBlockChainEntry; EntryIndex: integer);

begin
     Query.Params.Clear;
     Query.SQL.Text := 'INSERT INTO blockchainentries (' + standardblockfields + ',shb_majorversion,shb_minorversion,shb_byteordermagic,shb_sectionlength,shb_comment,shb_hardware,shb_os,shb_userappl,shb_guid)' +
                                                       ' VALUES (' + standardblockparams + ',:shb_majorversion,:shb_minorversion,:shb_byteordermagic,:shb_sectionlength,:shb_comment,:shb_hardware,:shb_os,:shb_userappl,:shb_guid);';

     AddStandardBlockParamsToQuery (Query, ChainNode, EntryIndex);

     Query.ParamByName('shb_majorversion').AsInteger := ChainNode^.SHBMajorVersion;
     Query.ParamByName('shb_minorversion').AsInteger := ChainNode^.SHBMinorVersion;
     Query.ParamByName('shb_byteordermagic').AsInteger := ChainNode^.SHBByteOrderMagic;
     Query.ParamByName('shb_sectionlength').AsString := InttoHex (ChainNode^.SHBSectionLength, 8);
     Query.ParamByName('shb_comment').AsString := ChainNode^.SHBComment;
     Query.ParamByName('shb_hardware').AsString := ChainNode^.SHBHardware;
     Query.ParamByName('shb_os').AsString := ChainNode^.SHBOS;
     Query.ParamByName('shb_userappl').AsString := ChainNode^.SHBUserAppl;
     Query.ParamByName('shb_guid').AsString := GUIDToString (ChainNode^.SHBGUID);

     Query.ExecSQL;

     priv_dbdirty := true;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.StoreIDBToIntelDB (Query: TFDQUery; ChainNode: pBlockChainEntry; EntryIndex: integer);

begin
     Query.Params.Clear;
     Query.SQL.Text := 'INSERT INTO blockchainentries (' + standardblockfields + ',idb_index,idb_linktype,idb_snaplen,idb_name,idb_description,idb_tsresolution,idb_filter)' +
                                                       ' VALUES (' + standardblockparams + ',:idb_index,:idb_linktype,:idb_snaplen,:idb_name,:idb_description,:idb_tsresolution,:idb_filter);';

     AddStandardBlockParamsToQuery (Query, ChainNode, EntryIndex);

     Query.ParamByName('idb_index').AsInteger := ChainNode^.IDBIndex;
     Query.ParamByName('idb_linktype').AsInteger := ChainNode^.IDBLinkType;
     Query.ParamByName('idb_snaplen').AsLongword := ChainNode^.IDBSnapLen;
     Query.ParamByName('idb_name').AsString := ChainNode^.IDBName;
     Query.ParamByName('idb_description').AsString := ChainNode^.IDBDescription;
     Query.ParamByName('idb_tsresolution').AsInteger := ChainNode^.IDBTSResolution;
     Query.ParamByName('idb_filter').AsString := ChainNode^.IDBFilter.FilterString;

     Query.ExecSQL;

     priv_dbdirty := true;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.StoreISBToIntelDB (Query: TFDQUery; ChainNode: pBlockChainEntry; EntryIndex: integer);

begin
     Query.Params.Clear;
     Query.SQL.Text := 'INSERT INTO blockchainentries (' + standardblockfields + ',isb_interfaceid,isb_timestamp,isb_starttime,isb_endtime,isb_received,isb_dropped)' +
                                                       ' VALUES (' + standardblockparams + ',:isb_interfaceid,:isb_timestamp,:isb_starttime,:isb_endtime,:isb_received,:isb_dropped);';

     AddStandardBlockParamsToQuery (Query, ChainNode, EntryIndex);

     Query.ParamByName('isb_interfaceid').AsInteger := ChainNode^.ISBInterfaceID;
     Query.ParamByName('isb_timestamp').AsDateTime := ChainNode^.ISBTimestamp;
     Query.ParamByName('isb_starttime').AsDateTime := ChainNode^.ISBStartTime;
     Query.ParamByName('isb_endtime').AsDateTime := ChainNode^.ISBEndTime;
     Query.ParamByName('isb_received').AsInteger := ChainNode^.ISBReceived;
     Query.ParamByName('isb_dropped').AsInteger := ChainNode^.ISBDropped;

     Query.ExecSQL;

     priv_dbdirty := true;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.StoreNRBToIntelDB (Query: TFDQUery; ChainNode: pBlockChainEntry; EntryIndex: integer);

var i       : integer;
    entry4  : tSimpleIPv4NameResolutionEntry;
    entry6  : tSimpleIPv6NameResolutionEntry;

begin
     Query.Params.Clear;
     Query.SQL.Text := 'INSERT INTO blockchainentries (' + standardblockfields + ',nrb_ipv4recordcount,nrb_ipv6recordcount)' +
                                                       ' VALUES (' + standardblockparams + ',:nrb_ipv4recordcount,:nrb_ipv6recordcount);';

     AddStandardBlockParamsToQuery (Query, ChainNode, EntryIndex);

     Query.ParamByName('nrb_ipv4recordcount').AsInteger := ChainNode^.NRBIPv4RecordCount;
     Query.ParamByName('nrb_ipv6recordcount').AsInteger := ChainNode^.NRBIPv6RecordCount;

     Query.ExecSQL;

     priv_dbdirty := true;

     for i := 1 to ChainNode^.NRBIPv4AddressList.count do
         begin
              entry4 := ChainNode^.NRBIPv4AddressList.GetEntry(i);

              Query.Params.Clear;

              Query.SQL.Text := 'INSERT INTO nrbentries (fileid,ipversion,fqdn,address) VALUES (:fileid,4,:fqdn,:address);';
              Query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
              Query.ParamByName ('fqdn').AsString := entry4.FQDN;
              Query.ParamByName ('address').AsString := entry4.IPv4Address.ValueAsString;

              Query.ExecSQL;
         end;

     for i := 1 to ChainNode^.NRBIPv6AddressList.count do
         begin
              entry6 := ChainNode^.NRBIPv6AddressList.GetEntry(i);

              Query.Params.Clear;

              Query.SQL.Text := 'INSERT INTO nrbentries (fileid,ipversion,fqdn,address) VALUES (:fileid,6,:fqdn,:address);';
              Query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
              Query.ParamByName ('fqdn').AsString := entry6.FQDN;
              Query.ParamByName ('address').AsString := entry6.IPv6Address.ValueAsString;

              Query.ExecSQL;
         end;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.StorePacketBlockToIntelDB (Query: TFDQUery; ChainNode: pBlockChainEntry; EntryIndex: integer);

begin
     Query.Params.Clear;
     Query.SQL.Text := 'INSERT INTO blockchainentries (' + standardblockfields + ') VALUES (' + standardblockparams + ');';

     AddStandardBlockParamsToQuery (Query, ChainNode, EntryIndex);

     Query.ExecSQL;

     priv_dbdirty := true;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.StoreBasicBlockToIndelDB (Query: TFDQUery; ChainNode: pBlockChainEntry; EntryIndex: integer);

begin
     Query.Params.Clear;
     Query.SQL.Text := 'INSERT INTO blockchainentries (' + standardblockfields + ') VALUES (' + standardblockparams + ');';

     AddStandardBlockParamsToQuery (Query, ChainNode, EntryIndex);

     Query.ExecSQL;

     priv_dbdirty := true;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.AssignQueryDetailsToChainNode (Query: TFDQUery; out NewEntry: pBlockChainEntry);

var sHelp : string;

begin
     NewEntry^.BlockTypeID := Query.FieldByName('blocktypeid').AsInteger;
     NewEntry^.BlockTypeName := GetBlockTypeName (NewEntry^.BlockTypeID);
     NewEntry^.SectionStartOffset := Query.FieldByName('sectionstartoffset').AsLargeInt;
     NewEntry^.SectionEndOffset := Query.FieldByName('sectionendoffset').AsLargeInt;
     NewEntry^.FirstBlockStartOffset := Query.FieldByName('firstblockstartoffset').AsLargeInt;
     NewEntry^.LastBlockEndOffset := Query.FieldByName('lastblockendoffset').AsLargeInt;
     NewEntry^.BlockLength := Query.FieldByName('blocklength').AsInteger;
     NewEntry^.Count := Query.FieldByName('count').AsLargeInt;

     case NewEntry^.BlockTypeID of
      BLOCKTYPE_SECTION_HEADER        : begin
                                             NewEntry^.SHBMajorVersion := Query.FieldByName('shb_majorversion').AsInteger;
                                             NewEntry^.SHBMinorVersion := Query.FieldByName('shb_minorversion').AsInteger;
                                             NewEntry^.SHBByteOrderMagic := Query.FieldByName('shb_byteordermagic').AsInteger;

                                             sHelp := '$' + Query.FieldByName('shb_sectionlength').AsString;
                                             NewEntry^.SHBSectionLength := StrToUInt64 (sHelp);
                                             NewEntry^.SHBComment := Query.FieldByName('shb_comment').AsString;
                                             NewEntry^.SHBHardware := Query.FieldByName('shb_hardware').AsString;
                                             NewEntry^.SHBOS := Query.FieldByName('shb_os').AsString;
                                             NewEntry^.SHBUserAppl := Query.FieldByName('shb_userappl').AsString;
                                             sHelp := Query.FieldByName('shb_guid').AsString;
                                             NewEntry^.SHBGUID := StringToGUID (sHelp);
                                        end;
      BLOCKTYPE_INTERFACE_DESCRIPTION : begin
                                             NewEntry^.IDBIndex := Query.FieldByName('idb_index').AsInteger;
                                             NewEntry^.IDBLinkType := Query.FieldByName('idb_linktype').AsInteger;
                                             NewEntry^.IDBSnapLen := Query.FieldByName('idb_snaplen').AsLongWord;
                                             NewEntry^.IDBName := Query.FieldByName('idb_name').AsString;
                                             NewEntry^.IDBDescription := Query.FieldByName('idb_description').AsString;
                                             NewEntry^.IDBTSResolution := Query.FieldByName('idb_tsresolution').AsInteger;
                                             NewEntry^.IDBFilter.FilterType := cftLibPCAP;                                    // TODO: store and retrieve the filter type
                                             NewEntry^.IDBFilter.FilterString := Query.FieldByName('idb_filter').AsString;
                                        end;
      BLOCKTYPE_INTERFACE_STATISTICS  : begin
                                             NewEntry^.ISBInterfaceID := Query.FieldByName('isb_interfaceid').AsInteger;
                                             NewEntry^.ISBTimestamp := Query.FieldByName('isb_timestamp').AsDateTime;
                                             NewEntry^.ISBStartTime := Query.FieldByName('isb_starttime').AsDateTime;
                                             NewEntry^.ISBEndTime := Query.FieldByName('isb_endtime').AsDateTime;
                                             NewEntry^.ISBReceived := Query.FieldByName('isb_received').AsInteger;
                                             NewEntry^.ISBDropped := Query.FieldByName('isb_dropped').AsInteger;
                                        end;
      BLOCKTYPE_NAME_RESOLUTION       : begin
                                             NewEntry^.NRBIPv4RecordCount := Query.FieldByName('nrb_ipv4recordcount').AsInteger;
                                             NewEntry^.NRBIPv6RecordCount := Query.FieldByName('nrb_ipv6recordcount').AsInteger;
                                        end;
     end;

     NewEntry^.NextBlockType := nil;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.StoreGeneralPCAPngStatisticsToDB (Query: TFDQUery);

begin
     Query.Params.Clear;
     Query.SQL.Text := 'INSERT INTO pcapngfilestatistics (fileid,sectioncount,filepacketcount,fileinterfacecount,statisticblockcount,statpacketcount,statfilteracceptcount,statosdropcount,statusdeliveredcount,errorcode,errortext)' +
                       ' VALUES (:fileid,:sectioncount,:filepacketcount,:fileinterfacecount,:statisticblockcount,:statpacketcount,:statfilteracceptcount,:statosdropcount,:statusdeliveredcount,:errorcode,:errortext);';

     Query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
     Query.ParamByName ('sectioncount').AsInteger := priv_pcapngstatistics.SectionCount;
     Query.ParamByName ('filepacketcount').AsInteger := priv_pcapngstatistics.FilePacketCount;
     Query.ParamByName ('fileinterfacecount').AsInteger := priv_pcapngstatistics.FileInterfaceCount;
     Query.ParamByName ('statisticblockcount').AsInteger := priv_pcapngstatistics.StatisticBlocKCount;
     Query.ParamByName ('statpacketcount').AsInteger := priv_pcapngstatistics.StatPacketCount;
     Query.ParamByName ('statfilteracceptcount').AsInteger := priv_pcapngstatistics.StatFilterAcceptCount;
     Query.ParamByName ('statosdropcount').AsInteger := priv_pcapngstatistics.StatOsDropCount;
     Query.ParamByName ('statusdeliveredcount').AsInteger := priv_pcapngstatistics.StatUsrDeliveredCount;
     Query.ParamByName ('errorcode').AsInteger := priv_pcapngstatistics.ErrorCode;
     Query.ParamByName ('errortext').AsString := priv_pcapngstatistics.ErrorText;

     Query.ExecSQL;

     priv_dbdirty := true;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.ReadGeneralPCAPngStatisticsFromDB (Query: TFDQUery);

begin
     Query.Open('SELECT * FROM pcapngfilestatistics WHERE fileid=' + IntToStr (priv_inteldbfileindex) + ';');

     priv_pcapngstatistics.SectionCount := Query.FieldByName ('sectioncount').AsInteger;
     priv_pcapngstatistics.FilePacketCount := Query.FieldByName ('filepacketcount').AsInteger;
     priv_pcapngstatistics.FileInterfaceCount := Query.FieldByName ('fileinterfacecount').AsInteger;
     priv_pcapngstatistics.StatisticBlocKCount := Query.FieldByName ('statisticblockcount').AsInteger;
     priv_pcapngstatistics.StatPacketCount := Query.FieldByName ('statpacketcount').AsInteger;
     priv_pcapngstatistics.StatFilterAcceptCount := Query.FieldByName ('statfilteracceptcount').AsInteger;
     priv_pcapngstatistics.StatOsDropCount := Query.FieldByName ('statosdropcount').AsInteger;
     priv_pcapngstatistics.StatUsrDeliveredCount := Query.FieldByName ('statusdeliveredcount').AsInteger;
     priv_pcapngstatistics.ErrorCode := Query.FieldByName ('errorcode').AsInteger;
     priv_pcapngstatistics.ErrorText := Query.FieldByName ('errortext').AsString;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.writePCAPngStatisticsToDB (Statistics: tPCAPNGStatistic);

var CurrentChainNode : pBlockChainEntry;
    count            : integer;
    Query            : TFDQUery;

begin
     Query := TFDQUery.Create (nil);
     Query.Connection := priv_db;

     // Assign statistics

     priv_pcapngstatistics := Statistics;
     CurrentChainNode := priv_pcapngstatistics.BlockChain;

     StoreGeneralPCAPngStatisticsToDB (Query);

     count := 0;

     // Write to DB

     while CurrentChainNode <> nil do
           begin
                inc (count);

                case CurrentChainNode^.BlockTypeID of
                 BLOCKTYPE_SECTION_HEADER        : StoreSHBToIntelDB (Query, CurrentChainNode, Count);
                 BLOCKTYPE_INTERFACE_DESCRIPTION : StoreIDBToIntelDB (Query, CurrentChainNode, Count);
                 BLOCKTYPE_NAME_RESOLUTION       : StoreNRBToIntelDB (Query, CurrentChainNode, Count);
                 BLOCKTYPE_PACKET,
                 BLOCKTYPE_SIMPLE_PACKET,
                 BLOCKTYPE_ENHANCED_PACKET       : StorePacketBlockToIntelDB (Query, CurrentChainNode, Count);
                 BLOCKTYPE_INTERFACE_STATISTICS  : StoreISBToIntelDB (Query, CurrentChainNode, Count);
                 BLOCKTYPE_IRIG_TIMESTAMP,
                 BLOCKTYPE_ARINC_429_IN_AFDX,
                 BLOCKTYPE_HONE_PROCESS_EVENT,
                 BLOCKTYPE_HONE_CONNECTION_EVENT,
                 BLOCKTYPE_SYSDIG_MACHINE_INFO,
                 BLOCKTYPE_SYSDIG_PROCESS_LIST,
                 BLOCKTYPE_SYSDIG_FD_LIST,
                 BLOCKTYPE_SYSDIG_EVENT,
                 BLOCKTYPE_SYSDIG_INTERFACE_LIST,
                 BLOCKTYPE_SYSDIG_USER_LIST       : StoreBasicBlockToIndelDB (Query, CurrentChainNode, Count);
                end;

                CurrentChainNode := CurrentChainNode.NextBlockType;
           end;

     query.SQL.Text := 'UPDATE filelist SET pcapngstructscanned=:pcapngstructscanned WHERE id=:fileid';
     query.ParamByName ('pcapngstructscanned').AsBoolean := true;
     Query.ExecSQL;

     priv_dbdirty := true;

     Query.Free;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.ReadNRBEntriesFromDB (var BlockNodeEntry: pBlockChainEntry);

var IPv4Address   : tIPv4Address;
    IPv6Address   : tIPv6Address;
    Query         : TFDQUery;
    i             : integer;

begin
     Query := TFDQUery.Create (nil);
     Query.Connection := priv_db;

     // Read NRB entries

     Query.Open('SELECT * FROM nrbentries WHERE fileid=' + IntToStr (priv_inteldbfileindex) + ' ORDER BY ipversion;');
     Query.FetchAll;

     for i := 1 to Query.RecordCount do
         begin
              if Query.FieldByName('ipversion').AsInteger = 4 then
                 begin
                      IPv4Address.setValueFromString(Query.FieldByName('address').AsString);
                      BlockNodeEntry^.NRBIPv4AddressList.AddIPandFQDN (IPv4Address, Query.FieldByName('fqdn').AsString);
                 end
              else
                  begin
                      IPv6Address.setValueFromString(Query.FieldByName('address').AsString);
                      BlockNodeEntry^.NRBIPv6AddressList.AddIPandFQDN (IPv6Address, Query.FieldByName('fqdn').AsString);
                  end;

              Query.Next;
         end;

     Query.Free;
end;

(* ------------------------------------------- *)

function tIntelDBManager.readPCAPngStatisticsFromDB: tPCAPNGStatistic;

var NewEntry      : pBlockChainEntry;
    PrevEntry     : pBlockChainEntry;
    Query         : TFDQUery;
    i             : integer;

begin
     Query := TFDQUery.Create (nil);
     Query.Connection := priv_db;

     // Read general statistics

     ReadGeneralPCAPngStatisticsFromDB (Query);

     // DB is already prepared in getPCAPngStatistics

     Query.Open('SELECT * FROM blockchainentries WHERE fileid=' + IntToStr (priv_inteldbfileindex) + ' ORDER BY blocksequence;');
     Query.FetchAll;

     for i := 1 to Query.RecordCount do
         begin
              if i > 1 then
                 PrevEntry := NewEntry;

              // Create a new entry and assign the data

              new (NewEntry);
              AssignQueryDetailsToChainNode (Query, NewEntry);

              // Link blocks

              if i > 1 then
                 begin
                      NewEntry^.PreviousBlockType := PrevEntry;
                      PrevEntry^.NextBlockType := NewEntry;
                 end
              else
                  begin
                       NewEntry^.PreviousBlockType := nil;
                       priv_pcapngstatistics.BlockChain := NewEntry;
                  end;

              if NewEntry^.BlockTypeID = BLOCKTYPE_NAME_RESOLUTION then
                 ReadNRBEntriesFromDB (NewEntry);

              Query.Next;
         end;

     result := priv_pcapngstatistics;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.writeAdditionalFileDetails;

var Query : TFDQUery;

begin
     Query := TFDQUery.Create(nil);
     Query.Connection := priv_db;
     Query.SQL.Text := 'UPDATE filelist SET firstframetimestamp=:firstframetimestamp,lastframetimestamp=:lastframetimestamp,framecount=:framecount WHERE id=:id;';
     Query.ParamByName ('id').AsInteger := priv_inteldbfileindex;
     Query.ParamByName ('firstframetimestamp').AsDateTime := priv_filefirstframetime;
     Query.ParamByName ('lastframetimestamp').AsDateTime := priv_filelastframetime;
     Query.ParamByName ('framecount').AsInteger := priv_fileframeparsedcount;

     Query.ExecSQL;

     priv_dbdirty := true;

     Query.Free;
end;

(* ------------------------------------------- *)
{ Read and set Filescan Status }
(* ------------------------------------------- *)

function tIntelDBManager.getFileProcessingStatus: tProcessingStatus;

var query       : TFDQUery;
    value       : integer;
    pcapscanned : boolean;

begin
     query := TFDQUery.Create(nil);
     query.Connection := priv_db;

     query.SQL.Text := 'SELECT pcapngstructscanned,processingstatus FROM filelist WHERE id=:fileid';
     query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
     query.Open;

     if query.RecordCount > 0 then
        begin
             pcapscanned := query.FieldByName ('pcapngstructscanned').AsBoolean;
             value := query.FieldByName ('processingstatus').AsInteger;

             case value of
               1 : result := psNewFile;
               2 : result := psScanningBasics;
               3 : result := psScanningPCAPngStruct;
               4 : result := psScanningConversations;
              10 : result := psTaskPending;
              11 : result := psTaskProcessing;
              12 : result := psTaskComplete;
              13 : result := psTaskAborted;
              14 : result := psTaskFailed;
             end;
        end
     else
         begin
              // Its not in the database, so its new

              result := psNewFile;
         end;

     query.Free;
end;

(* ------------------------------------------- *)

procedure tIntelDBManager.setFileProcessingStatus (const ScanStatus: tProcessingStatus);

var query : TFDQUery;

begin
     query := TFDQUery.Create(nil);
     query.Connection := priv_db;

     query.SQL.Text := 'UPDATE filelist SET processingstatus=:processingstatus WHERE id=:fileid';
     query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;

     case ScanStatus of
      psNewFile               : query.ParamByName ('processingstatus').AsInteger := 1;
      psScanningBasics        : query.ParamByName ('processingstatus').AsInteger := 2;
      psScanningPCAPngStruct  : query.ParamByName ('processingstatus').AsInteger := 3;
      psScanningConversations : query.ParamByName ('processingstatus').AsInteger := 4;
      psTaskPending           : query.ParamByName ('processingstatus').AsInteger := 10;
      psTaskProcessing        : query.ParamByName ('processingstatus').AsInteger := 11;
      psTaskComplete          : query.ParamByName ('processingstatus').AsInteger := 12;
      psTaskAborted           : query.ParamByName ('processingstatus').AsInteger := 13;
      psTaskFailed            : query.ParamByName ('processingstatus').AsInteger := 14;
     end;

     query.ExecSQL;

     priv_dbdirty := true;

     query.Free;
end;

(* ------------------------------------------- *)

function tIntelDBManager.rename (const NewFilename: string): boolean;

var Query : TFDQUery;

begin
     Query := TFDQUery.Create(nil);
     Query.Connection := priv_db;

     query.SQL.Text := 'UPDATE filelist SET filename=:newfilename WHERE id=:fileid';
     query.ParamByName ('fileid').AsInteger := priv_inteldbfileindex;
     Query.ParamByName ('newfilename').AsString := NewFilename;

     Query.ExecSQL;
     priv_dbdirty := True;

     Query.Free;
end;

(* ------------------------------------------- *)
{ tIntelDBParameters }
(* ------------------------------------------- *)

procedure tIntelDBParameters.initialize;

begin
     IntelDB := nil;

     DoNotUseIntelDB       := false;
     UsePersistentIntelDB  := false;
     HideIntelDBFile       := false;
     UseLocalDBFiles       := true;
     GlobalDBPath          := '';
end;

(* ------------------------------------------- *)

end.


