{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_replacement_manager;

interface

// {$define monitordb}

uses Generics.Collections,

     // FireDAC

     FireDAC.Stan.Intf, FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf,
     FireDAC.Phys.Intf, FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.dAPT,
     FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.ExprFuncs,
     FireDAC.VCLUI.Wait, FireDAC.Comp.UI, FireDAC.Phys.SQLite, FireDAC.Phys.SQLiteWrapper,
     Data.DB, FireDAC.Comp.Client, FireDAC.Stan.Param,

     // structures

     base_structures,
     definition_anonymization_struct,
     definition_ip_addresses,
     definition_5tuple,
     definition_tcp_structures,
     class_ipv4subnethandler,
     class_ipv6subnethandler;

type tStringOriginType = (qtAny, qtSHBComment,qtSHBHardware,qtSHBOS, qtSHBApplication, qt802dot11);
     tIntegerOriginType = (ioAny, ioMPLSLabel);

     tReplacementScope = (rsGenerated, rsFromSettings);

     tStringQueryEntry = record
                               StringOrigin        : tStringOriginType;
                               Original            : string;
                               Replacement         : string;
                         end;

     tStringQueryList = record
                              EntryList : array of tStringQueryEntry;

                              procedure initialize;
                              procedure add (Entry : tStringQueryEntry); overload;
                              procedure add (Origin : tStringOriginType; OriginalString : string; ReplacementString : string = ''); overload;
                              function getEntry (Index : integer) : tStringQueryEntry;
                              function count : integer;
                        end;

     tTCPSequenceAdjustmentDelta = record
                                         TCPSequence     : tTCPSequence;
                                         Delta           : integer;

                                         class operator equal (a, b : tTCPSequenceAdjustmentDelta) : boolean;
                                         class operator notequal (a, b : tTCPSequenceAdjustmentDelta) : boolean;
                                         class operator greaterthan (a, b : tTCPSequenceAdjustmentDelta) : boolean;
                                         class operator lessthan (a, b : tTCPSequenceAdjustmentDelta) : boolean;
                                         class operator greaterthanorequal (a, b : tTCPSequenceAdjustmentDelta) : boolean;
                                         class operator lessthanorequal (a, b : tTCPSequenceAdjustmentDelta) : boolean;
                                   end;

     tTCPSequenceAdjustmentMarker = record
                                          Tuple        : t5Tuple;
                                          Adjustments  : array of tTCPSequenceAdjustmentDelta;

                                          procedure initizalize (const value : t5Tuple);
                                          procedure addAdjustment (Entry : tTCPSequenceAdjustmentDelta);
                                    end;

     tReplacementManager = class (tObject)

                                 constructor create (const DBFile : string; const AnonymizationParameters : tAnonymizeParameters);
                                 procedure free;

                                 private
                                        priv_database                        : tFDConnection;
                                        priv_anonparams                      : tAnonymizeParameters;

                                        priv_SSIDIndex                       : integer;

                                        priv_IPv4Subnethandler               : tIPv4SubnetHandler;
                                        priv_IPv6SubnetHandler               : tIPv6SubnetHandler;
                                        priv_TCPSequenceDeltaMarkers         : tDictionary<string, tTCPSequenceAdjustmentMarker>;

                                        priv_LookupMAC                       : tDictionary<tMACAddress, tMACAddress>;
                                        priv_LookupIPv4                      : tDictionary<tIPv4Address, tIPv4Address>;
                                        priv_LookupIPv6                      : tDictionary<tIPv6Address, tIPv6Address>;
                                        priv_LookupStrings                   : tDictionary<string, string>;

                                        priv_MPLSLabelReplacementCount       : integer;    // used to prevent trouble if maximum label range is lower than different labels exist

                                        priv_dbfile                          : string;

                                        {$REGION 'Documentation'}
                                        ///	<summary>
                                        ///	  This routine creates the SQLite
                                        ///	  database that is used to keep
                                        ///	  replacement records. In debug
                                        ///	  mode the database will be placed
                                        ///	  into a file for easier debugging
                                        ///	  which in release mode the
                                        ///	  database will be in memory.
                                        ///	</summary>
                                        {$ENDREGION}
                                        procedure PrepareDatabase;
                                        procedure TransferAddressesFromParameters;

                                        procedure storeCounter (const Name : string; const count : integer);
                                        function getCounter (const Name : string) : integer;

                                        {$REGION 'Documentation'}
                                        /// <summary>
                                        ///   Main routine that randomizes an
                                        ///   IPv6 address. It checks a couple
                                        ///   of ranges to keep the replacement
                                        ///   address in the same boundaries if
                                        ///   required.
                                        /// </summary>
                                        /// <param name="OriginalAddress">
                                        ///   Original IPv6 address that a
                                        ///   random replacement is required for
                                        /// </param>
                                        /// <returns>
                                        ///   New, more or less random IPv6
                                        ///   address
                                        /// </returns>
                                        {$ENDREGION}
                                        function getRandomIPv6Address (const OriginalAddress : tIPv6Address) : tIPv6Address;
                                        {$REGION 'Documentation'}
                                        /// <summary>
                                        ///   Creates a matching solicited node
                                        ///   address for the replacement
                                        ///   address if it is not already in
                                        ///   the database. When a new
                                        ///   solicited node replacement
                                        ///   address is created, it is based
                                        ///   on the MCast MAC that needs to
                                        ///   match it. If that MAC does not
                                        ///   yet exists, it is created as well.
                                        /// </summary>
                                        /// <param name="OriginalAddress">
                                        ///   Original IPv6 address
                                        /// </param>
                                        /// <param name="Replacement">
                                        ///   Replacement IPv6 address
                                        /// </param>
                                        {$ENDREGION}
                                        procedure CreateSolicitedNodeReplacement (const OriginalAddress, Replacement : tIPv6Address);

                                        function getVLANCount: integer;
                                        function getIPv4Count: integer;
                                        function getIPv6Count: integer;
                                        function getMACCount: integer;

                                        function getVLANReplacementFromDB (const VLAN : tVLANID) : tVLANID;
                                        function getIPv4AddressReplacementFromDB (const IPAddress : tIPv4Address) : tIPv4Address;
                                        function getIPv4AddressReplacementFromSubnet (const IPAddress : tIPv4Address) : tIPv4Address;
                                        function getIPv6AddressReplacementFromDB (const IPAddress : tIPv6Address) : tIPv6Address;
                                        {$REGION 'Documentation'}
                                        /// <summary>
                                        ///   Determines if there is a prefix
                                        ///   pair where the original IP
                                        ///   address belongs to the original
                                        ///   subnet. If so, a new IP address
                                        ///   with the according replacement
                                        ///   prefix is created.
                                        /// </summary>
                                        {$ENDREGION}
                                        function getIPv6AddressReplacementFromPrefix (const IPAddress : tIPv6Address) : tIPv6Address;
                                        function getIPv6PrefixReplacement (const IPAddress : tIPv6Address) : tIPv6Address;
                                        function getReplacementForFQDNFromDB (const OldFQDN : tFQDN) : tFQDNPair;

                                        function OriginalAlreadyExists (const VLAN : tVLANID) : boolean; overload;
                                        function OriginalAlreadyExists (const MACAddress : tMACAddress) : boolean; overload;
                                        function OriginalAlreadyExists (const IPAddress: tIPv4Address): boolean; overload;
                                        function OriginalAlreadyExists (const IPAddress : tIPv6Address) : boolean; overload;
                                        function OriginalAlreadyExists (const Value : String; const Origin : tStringOriginType = qtAny) : boolean; overload;
                                        function OriginalAlreadyExists (const Value : tGUID) : boolean; overload;
                                        function OriginalAlreadyExists (const Protocol : byte; Socket : tIPv4Socket) : boolean; overload;
                                        function OriginalAlreadyExists (const Protocol : byte; Socket : tIPv6Socket) : boolean; overload;
                                        function OriginalAlreadyExists (const FQDN : tFQDN) : boolean; overload;
                                        function OriginalAlreadyExists (const Value : int64; const Origin : tIntegerOriginType = ioAny) : boolean; overload;

                                        function ReplacementAlreadyExists (const VLAN : tVLANID) : boolean; overload;
                                        function ReplacementAlreadyExists (const MACAddress : tMACAddress) : boolean; overload;
                                        function ReplacementAlreadyExists (const IPAddress : tIPv4Address) : boolean; overload;
                                        function ReplacementAlreadyExists (const IPAddress : tIPv6Address) : boolean; overload;
                                        function ReplacementAlreadyExists (const Value : String; const Origin : tStringOriginType = qtAny) : boolean; overload;
                                        function ReplacementAlreadyExists (const Value : tGUID) : boolean; overload;
                                        function ReplacementAlreadyExists (const Protocol : byte; Socket : tIPv4Socket) : boolean; overload;
                                        function ReplacementAlreadyExists (const Protocol : byte; Socket : tIPv6Socket) : boolean; overload;
                                        function ReplacementAlreadyExists (const Protocol : byte; Port : word) : boolean; overload;
                                        function ReplacementAlreadyExists (const FQDN : tFQDN) : boolean; overload;
                                        function ReplacementAlreadyExists (const Value : int64; const Origin : tIntegerOriginType = ioAny) : boolean; overload;

                                 public
                                       // TODO: add properties/code to set the type of replacements according to the anonymization parameters

                                       property VLANCount : integer read getVLANCount;
                                       property MACCount : integer read getMACCount;
                                       property IPv4Count : integer read getIPv4Count;
                                       property IPv6Count : integer read getIPv6Count;
                                       property IPv4SubnetManager : tIPv4SubnetHandler read priv_IPv4Subnethandler write priv_IPv4Subnethandler;
                                       property IPv6SubnetManager : tIPv6SubnetHandler read priv_IPv6Subnethandler write priv_IPv6Subnethandler;

                                       procedure AddReplacementPair (const OldVLAN, NewVLAN : tVLANID); overload;
                                       procedure AddReplacementPair (const OldMAC, NewMAC : tMACAddress); overload;
                                       procedure AddReplacementPair (const OldIP, NewIP : tIPv4Address); overload;
                                       procedure AddReplacementPair (const OldIP, NewIP : tIPv6Address); overload;
                                       procedure AddReplacementPair (const OldString, NewString : string; const Origin : tStringOriginType); overload;
                                       procedure AddReplacementPair (const OldGUID, NewGUID : tGUID); overload;
                                       procedure AddReplacementPair (const Protocol : byte; const OldSocket : tIPv4Socket; const NewSocket : tIPv4Socket); overload;
                                       procedure AddReplacementPair (const Protocol : byte; const OldSocket : tIPv6Socket; const NewSocket : tIPv6Socket); overload;
                                       procedure AddReplacementPair (const SocketPair : tIPv4SocketPair; const OldIPID, NewIPID : tIPID); overload;
                                       procedure AddReplacementPair (const SocketPair : tIPv6SocketPair; const OldIPID, NewIPID : tIPID); overload;
                                       procedure AddReplacementPair (const Protocol : byte; const OldPort, NewPort : word); overload;
                                       procedure AddReplacementPair (const OldFQDN, NewFQDN : tFQDN); overload;
                                       procedure AddReplacementPair (const OldInteger, NewInteger : int64; const Origin : tIntegerOriginType); overload;
                                       procedure AddTCPSequenceDelta (const Tuple : t5Tuple; const TCPSequence : tTCPSequence; const delta : integer); overload;

                                       function getMACReplacementFromDB (const MAC : tMACAddress) : tMACAddress;

                                       function getReplacementForVLAN (const OldVLAN : tVLANID) : tVLANID;
                                       /// <summary>
                                       ///   Determines a replacement MAC
                                       ///   address based on replacement
                                       ///   parameters. An optional parameter
                                       ///   allows forcing randomization for
                                       ///   IPv6 EUI64 addresses.
                                       /// </summary>
                                       /// <param name="MAC">
                                       ///   Original MAC
                                       /// </param>
                                       /// <param name="Force">
                                       ///   if true, randomization will occur
                                       ///   even if not enabled in the
                                       ///   replacement parameters
                                       /// </param>
                                       function getReplacementForAddress (const MAC : tMACAddress; const Force : boolean = false) : tMACAddress; overload;
                                       function getReplacementForAddress (const IPAddress : tIPv4Address) : tIPv4Address; overload;
                                       function getReplacementForAddress (const IPAddress : tIPv6Address) : tIPv6Address; overload;
                                       function getReplacementForString (const OldString : string; const Origin : tStringOriginType) : string;
                                       function getReplacementForGUID (const OldGuid : tGUID) : tGUID;
                                       function getReplacementForPort (const Protocol : byte; const OldPort : word) : word; overload;
                                       function getReplacementForSocket (const Protocol : byte; const OldSocket : tIPv4Socket) : tIPv4Socket; overload;
                                       function getReplacementForSocket (const Protocol : byte; const OldSocket : tIPv6Socket) : tIPv6Socket; overload;
                                       function getReplacementForIPID (const OldIPID : tIPID; const OldSocketPair: tIPv4SocketPair) : tIPID; overload;
                                       function getReplacementForIPID (const OldIPID : tIPID; const OldSocketPair: tIPv6SocketPair) : tIPID; overload;
                                       function getReplacementForFQDN (const OldFQDN : tFQDN) : tFQDN;
                                       function getReplacementForInteger (const OldInteger : int64; const Origin : tIntegerOriginType) : int64;
                                       function getReplacementForMPLSLabel (const OldLabel : longword) : longword;

                                       /// <summary>
                                       ///   Returns the number of bytes that
                                       ///   the sequence number needs to be
                                       ///   adjusted. This is caused by TCP
                                       ///   payload replacements that are
                                       ///   longer or shorter than the
                                       ///   original, and are accumulated over
                                       ///   all packets that were changed with
                                       ///   a sequence number lower than the
                                       ///   current one.
                                       /// </summary>
                                       /// <param name="Tuple">
                                       ///   Conversation tuple to check for
                                       ///   required adjustments
                                       /// </param>
                                       /// <param name="SequenceNo">
                                       ///   Current sequence number processed
                                       /// </param>
                                       /// <param name="WrapCount">
                                       ///   usually zero, unless the sequence
                                       ///   number wrapped
                                       /// </param>
                                       /// <returns>
                                       ///   delta bytes
                                       /// </returns>
                                       function getTCPSequenceDelta (const Tuple : t5Tuple; const TCPSequence : tTCPSequence) : integer;
                           end;


implementation

uses classes,   // for TResourceStream
     types,     // for rt_rcdata
     sysutils,
     toolkit,
     SiAuto;

const dbversion = '1.0';
      wipedbcommand = 'PRAGMA writable_schema = 1;' + CRLF +
                      'delete from sqlite_master where type in (''table'', ''index'', ''trigger'');' + CRLF +
                      'PRAGMA writable_schema = 0;';

(* ------------------------------------------- *)

function getStringOriginIndexFromType (Origin : tStringOriginType) : integer;

begin
     // Yeah, I know there's Ord () but this is safer when changing order of items later

     case Origin of
         qtAny            : result := 0;
         qtSHBComment     : result := 1;
         qtSHBHardware    : result := 2;
         qtSHBOS          : result := 3;
         qtSHBApplication : result := 4;
     end;
end;

(* ------------------------------------------- *)

function getStringOriginTypeFromIndex (Index : integer) : tStringOriginType;

begin
     // Yeah, I know there's Ord () but this is safer when changing order of items later

     case Index of
         0 : result := qtAny;
         1 : result := qtSHBComment;
         2 : result := qtSHBHardware;
         3 : result := qtSHBOS;
         4 : result := qtSHBApplication;
     end;
end;

(* ------------------------------------------- *)

function getIntegerOriginIndexFromType (Origin : tIntegerOriginType) : integer;

begin
     // Yeah, I know there's Ord () but this is safer when changing order of items later

     case Origin of
         ioAny       : result := 0;
         ioMPLSLabel : result := 1;
     end;
end;

(* ------------------------------------------- *)

function getIntegerOriginTypeFromIndex (Index : integer) : tIntegerOriginType;

begin
     // Yeah, I know there's Ord () but this is safer when changing order of items later

     case Index of
         0 : result := ioAny;
         1 : result := ioMPLSLabel;
     end;
end;

(* ------------------------------------------- *)
{ tStringQueryList }
(* ------------------------------------------- *)

procedure tStringQueryList.initialize;

begin
     setlength (EntryList, 0);
end;

(* ------------------------------------------- *)

procedure tStringQueryList.add (Entry: tStringQueryEntry);

var count : integer;

begin
     count := length (EntryList);

     setlength (EntryList, count);
     EntryList[count-1] := Entry;
end;

(* ------------------------------------------- *)

procedure tStringQueryList.add(Origin: tStringOriginType; OriginalString: string; ReplacementString : string = '');

var Entry : tStringQueryEntry;

begin
     Entry.StringOrigin := Origin;
     Entry.Original := OriginalString;
     Entry.Replacement := ReplacementString;
end;

(* ------------------------------------------- *)

function tStringQueryList.count: integer;

begin
     result := length (EntryList);
end;

(* ------------------------------------------- *)

function tStringQueryList.getEntry(Index: integer): tStringQueryEntry;

begin
     result := EntryList[Index-1];
end;

(* ------------------------------------------- *)
{ tTCPSequenceAdjustmentDelta }
(* ------------------------------------------- *)

class operator tTCPSequenceAdjustmentDelta.equal (a, b: tTCPSequenceAdjustmentDelta): boolean;

begin
     result := (a.TCPSequence = b.TCPSequence) and (a.Delta = b.Delta);
end;

(* ------------------------------------------- *)

class operator tTCPSequenceAdjustmentDelta.notequal (a, b: tTCPSequenceAdjustmentDelta): boolean;

begin
     result := not (a=b);
end;

(* ------------------------------------------- *)

// >, <, <=, >= are not looking at the delta as it doesn't help

class operator tTCPSequenceAdjustmentDelta.greaterthan (a, b: tTCPSequenceAdjustmentDelta): boolean;

begin
     result := a.TCPSequence > b.TCPSequence;
end;

(* ------------------------------------------- *)

class operator tTCPSequenceAdjustmentDelta.greaterthanorequal (a, b: tTCPSequenceAdjustmentDelta): boolean;

begin
     result := a.TCPSequence >= b.TCPSequence;
end;

(* ------------------------------------------- *)

class operator tTCPSequenceAdjustmentDelta.lessthan (a, b: tTCPSequenceAdjustmentDelta): boolean;

begin
     result := a.TCPSequence < b.TCPSequence;
end;

(* ------------------------------------------- *)

class operator tTCPSequenceAdjustmentDelta.lessthanorequal (a, b: tTCPSequenceAdjustmentDelta): boolean;

begin
     result := a.TCPSequence <= b.TCPSequence;
end;

(* ------------------------------------------- *)
{ tTCPSequenceAdjustmentMarker }
(* ------------------------------------------- *)

procedure tTCPSequenceAdjustmentMarker.initizalize (const value : t5Tuple);

begin
     Tuple := value;
     setLength (Adjustments, 0);
end;

(* ------------------------------------------- *)

procedure tTCPSequenceAdjustmentMarker.addAdjustment (Entry: tTCPSequenceAdjustmentDelta);

var len   : integer;
    index : integer;

begin
     len := length (Adjustments);

     if len = 0 then
        begin
             setLength (Adjustments, 1);
             Adjustments[0] := Entry;
        end
     else
         begin
              index := 0;

              // Find the correct spot to insert the new adjustment
              // as they need to be sorted by sequence number

              while (index < len) and (Entry > Adjustments[index]) do
                    inc (index);

              insert (Entry, Adjustments, index);
         end;
end;

(* ------------------------------------------- *)
{ tReplacementManager }
(* ------------------------------------------- *)

constructor tReplacementManager.create (const DBFile : string; const AnonymizationParameters : tAnonymizeParameters);

begin
     inherited create;
     randomize;

     priv_SSIDIndex := 0;
     priv_MPLSLabelReplacementCount := 0;

     priv_anonparams := AnonymizationParameters;

     if priv_anonparams.ReplacementDBForcePath then
        priv_dbfile := IncludeTrailingPathDelimiter (priv_anonparams.ReplacementDBFilePath) + ExtractFileName (DBFile)
     else
         priv_dbfile := DBFile;

     priv_LookupMAC := tDictionary<tMACAddress, tMACAddress>.create (1000);
     priv_LookupIPv4 := tDictionary<tIPv4Address, tIPv4Address>.create (1000);
     priv_LookupIPv6 := tDictionary<tIPv6Address, tIPv6Address>.create (1000);
     priv_LookupStrings := tDictionary<string, string>.create (1000);

//   priv_keepdbafterfinishing := KeepFileAfterFinish;

     priv_IPv4Subnethandler := tIPv4SubnetHandler.create;
     priv_IPv6SubnetHandler := tIPv6SubnetHandler.create;

     priv_TCPSequenceDeltaMarkers := tDictionary<string, tTCPSequenceAdjustmentMarker>.create (256);

//   priv_DBinMem := CreateDBInMem;

     PrepareDatabase;
     TransferAddressesFromParameters;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.free;

begin
     try
        priv_database.Close;
     finally
            priv_database.Free;
     end;

     priv_LookupMAC.Free;
     priv_LookupIPv4.Free;
     priv_LookupIPv6.Free;
     priv_LookupStrings.Free;

     priv_IPv4Subnethandler.free;
     priv_IPv6SubnetHandler.free;
     priv_TCPSequenceDeltaMarkers.Free;

     if not priv_anonparams.ReplacementDBKeepDiskFile then
        begin
             try
                if FileExists (priv_dbfile) then
                   DeleteFile (priv_dbfile);
             finally
                    inherited free;
             end;
        end;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.PrepareDatabase;

var Query        : tFDQuery;
    CreateDB     : boolean;
    DBStructure  : TResourceStream;
    help         : AnsiString;
    sVersion     : string;
    sHelp        : string;
    boWipeTables : boolean;

begin
     siMain.EnterMethod ('PrepareDatabase');

     priv_database := tFDConnection.Create(nil);
     priv_database.DriverName := 'SQLite';
     priv_database.Params.Add ('SQLite');

     if FileExists (priv_dbfile) and priv_anonparams.ReplacementDBWipeOnStart then
        begin
             DeleteFile (priv_dbfile);

             // check if the delete failed, and if so, mark tables for wipe

             boWipeTables := fileexists (priv_dbfile);
        end;

     priv_database.Params.Add('LockingMode=Normal');
     priv_database.Params.Add('Synchronous=Normal');

{$ifdef monitordb}

     priv_database.Params.Add('MonitorBy=Remote');

{$endif}

     CreateDB := not fileexists (priv_dbfile);

     if not priv_anonparams.ReplacementDBUseDiskFile then
        begin
             priv_database.Params.Add ('Database=:memory:');
             SiMain.LogVerbose ('Opening in-memory replacement database');
        end
     else
         begin
              priv_database.Params.Add ('Database=' + priv_dbfile);
              SiMain.LogVerbose ('Opening disk based replacement database at "' + priv_dbfile + '"');
         end;

     priv_database.Open;

{$ifdef monitordb}

     priv_database.ConnectionIntf.Tracing := true;

{$endif}

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     if priv_database.Connected then
        begin
             SiMain.LogVerbose ('Database successfully connected');

             if boWipeTables then
                begin
                     SiMain.LogVerbose ('Wiping replacement database tables');

                     Query.SQL.Text := WipeDBCommand;
                     Query.ExecSQL;

                     // Reopen database to complete the wipe

                     priv_database.Close;
                     priv_database.Open;

                     CreateDB := true;
                end;
        end;

     if CreateDB then
        begin
             SiMain.LogDebug ('Creating database table structure');

             // Get structure from Ressource (configured at Project -> Resources and Images)

             DBStructure := TResourceStream.Create (HInstance, 'replacementmgr_dbstruct', RT_RCDATA);
             setLength (help, DBStructure.Size);
             DBStructure.ReadBuffer(help[1], DBStructure.Size);

             Query.SQL.Text := help;
             Query.ExecSQL;
             DBStructure.Free;

             // Write Meta

             Query.SQL.Text := 'INSERT INTO rdb_meta (name,value) VALUES ("version",:version);';
             Query.ParamByName ('version').AsString := dbversion;
             Query.ExecSQL;
        end;

     Query.Free;

     siMain.LeaveMethod ('PrepareDatabase');
end;

(* ------------------------------------------- *)

procedure tReplacementManager.TransferAddressesFromParameters;

var VLAN1    : tVLAN;
    VLAN2    : tVLAN;
    Port     : word;
    Socket1  : tIPv4Socket;
    Socket2  : tIPv6Socket;
    Socket3  : tIPv4Socket;
    Socket4  : tIPv6Socket;
    Subnet1  : tIPv4SubnetDefinition;
    Subnet2  : tIPv6SubnetDefinition;
    len      : integer;
    i        : integer;
{$ifdef debug}
    help     : string;
{$endif}

begin
     SiMain.EnterMethod ('TransferAddressesFromParameters');

     // Hint: Payload replacement strings aren't copied since they're matched in a loop against originals anyway

     // Transfer 802.11 parameters

     if priv_anonparams.WiFiDoReplace then
        begin
             if priv_anonparams.WiFiReplaceBySSIDList then
                begin
                     len := length (priv_anonparams.WiFiSSIDReplacementStrings);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' SSID replacement sets to database');

                     for i := 1 to len do
                         begin
                              if not OriginalAlreadyExists (priv_anonparams.WiFiSSIDReplacementStrings[i-1].original, qt802dot11) then
                                 AddReplacementPair (priv_anonparams.WiFiSSIDReplacementStrings[i-1].original, priv_anonparams.WiFiSSIDReplacementStrings[i-1].replacement, qt802dot11);
                         end;
                end;
        end;

     // Transfer Ethernet parameters

     if priv_anonparams.EthernetDoReplace then
        begin
             // MAC address List

             if priv_anonparams.EthernetReplaceByMACList then
                begin
                     len := length (priv_anonparams.EthernetMACList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' MAC address replacement sets to database');

                     for i := 1 to len do
                         begin
                              // Make sure to add stuff only if it does not collide with existing stuff

                              if not priv_anonparams.ReplacementDBWipeOnStart then
                                 begin
                                      if not OriginalAlreadyExists (priv_anonparams.EthernetMACList[i-1].Source) then
                                         AddReplacementPair (priv_anonparams.EthernetMACList[i-1].Source, priv_anonparams.EthernetMACList[i-1].Destination)
                                 end
                              else
                                  AddReplacementPair (priv_anonparams.EthernetMACList[i-1].Source, priv_anonparams.EthernetMACList[i-1].Destination)
                         end;
                end;
        end;

     // Transfer VLAN parameters

     if priv_anonparams.VLANDoReplace then
        begin
             // VLAN ID List

             if priv_anonparams.VLANDoReplaceVLANByIDList then
                begin
                     len := length (priv_anonparams.VLANReplaceByVLANIDList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' VLAN ID replacement sets to database');

                     for i := 1 to len do
                         begin
                              VLAN1 := priv_anonparams.VLANReplaceByVLANIDList[i-1].Value1;
                              VLAN2 := priv_anonparams.VLANReplaceByVLANIDList[i-1].Value2;

                              if not priv_anonparams.ReplacementDBWipeOnStart then
                                 begin
                                      // Make sure to add stuff only if it does not collide with existing stuff

                                      if not priv_anonparams.ReplacementDBWipeOnStart then
                                         begin
                                              if not OriginalAlreadyExists (VLAN1) then
                                                 AddReplacementPair (VLAN1, VLAN2);
                                         end;
                                 end
                              else
                                  AddReplacementPair (VLAN1, VLAN2);
                         end;
                end;
        end;

     // Transfer IPv4 parameters

     if priv_anonparams.IPv4DoReplace then
        begin
             // IP Address List

             if priv_anonparams.IPv4DoReplaceByIPList then
                begin
                     len := length (priv_anonparams.IPv4AddressList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' IPv4 address replacement sets to database');

                     for i := 1 to len do
                         begin
                              if not priv_anonparams.ReplacementDBWipeOnStart then
                                 begin
                                      // Make sure to add stuff only if it does not collide with existing stuff

                                      if not OriginalAlreadyExists (priv_anonparams.IPv4AddressList[i-1].Source) then
                                         AddReplacementPair (priv_anonparams.IPv4AddressList[i-1].Source, priv_anonparams.IPv4AddressList[i-1].Destination);
                                 end
                              else
                                  AddReplacementPair (priv_anonparams.IPv4AddressList[i-1].Source, priv_anonparams.IPv4AddressList[i-1].Destination);
                         end;
                end;

             // IP Network List

             if priv_anonparams.IPv4ReplaceBySubnetlist then
                begin
                     len := length (priv_anonparams.IPv4SubnetList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' IPv4 subnet replacement sets to database');

                     for i := 1 to len do
                         begin
                              Subnet1 := tIPv4SubnetDefinition.create (priv_anonparams.IPv4SubnetList[i-1].Source);
                              Subnet1.ReplacementNetAddress := priv_anonparams.IPv4SubnetList[i-1].Destination;

                              priv_IPv4Subnethandler.AddSubnet (Subnet1);
                         end;
                end;
        end;

     // Transfer IPv6 parameters

     if priv_anonparams.IPv6DoReplace then
        begin
             // IP Address List

             if priv_anonparams.IPv6DoReplaceByIPList then
                begin
                     len := length (priv_anonparams.IPv6AddressList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' IPv6 address replacement sets to database');

                     for i := 1 to len do
                         begin
                              if not priv_anonparams.ReplacementDBWipeOnStart then
                                 begin
                                      // Make sure to add stuff only if it does not collide with existing stuff

                                      if not OriginalAlreadyExists (priv_anonparams.IPv6AddressList[i-1].Source) then
                                         AddReplacementPair (priv_anonparams.IPv6AddressList[i-1].Source, priv_anonparams.IPv6AddressList[i-1].Destination);
                                 end
                              else
                                  AddReplacementPair (priv_anonparams.IPv6AddressList[i-1].Source, priv_anonparams.IPv6AddressList[i-1].Destination);
                         end;
                end;

             // IP Network List

             if priv_anonparams.IPv6ReplaceBySubnetlist then
                begin
                     len := length (priv_anonparams.IPv6SubnetList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' IPv6 subnet replacement sets to database');

                     for i := 1 to len do
                         begin
                              Subnet2 := tIPv6SubnetDefinition.create (priv_anonparams.IPv6SubnetList[i-1].Source);
                              Subnet2.ReplacementNetAddress := priv_anonparams.IPv6SubnetList[i-1].Destination;

                              priv_IPv6Subnethandler.AddSubnet (Subnet2);
                         end;
                end;
        end;

     // TCP port parameters

     if priv_anonparams.TCPDoReplace then
        begin
             // TCP Socket list

             if priv_anonparams.TCPReplaceSocketsByList then
                begin
                     len := length (priv_anonparams.TCPSocketPortList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' TCP socket replacement sets to database');

                     for i := 1 to len do
                         begin
                              if priv_anonparams.TCPSocketPortList[i-1].Source.IPAddress.IPVersion = 4 then
                                 begin
                                      Socket1 := priv_anonparams.TCPSocketPortList[i-1].Source.getIPv4Socket;
                                      Socket3 := priv_anonparams.TCPSocketPortList[i-1].Destination.getIPv4Socket;

                                      if priv_anonparams.ReplacementDBWipeOnStart or (not OriginalAlreadyExists (protTCP, Socket1)) then
                                         AddReplacementPair (protTCP, Socket1, Socket3);
                                 end
                              else
                                  begin
                                      Socket2 := priv_anonparams.TCPSocketPortList[i-1].Source.getIPv6Socket;
                                      Socket4 := priv_anonparams.TCPSocketPortList[i-1].Destination.getIPv6Socket;

                                      if priv_anonparams.ReplacementDBWipeOnStart or (not OriginalAlreadyExists (protTCP, Socket2)) then
                                         AddReplacementPair (protTCP, Socket2, Socket4);
                                  end;
                         end;
                end;

             // TCP Port list (it's like a socket list, but the IPs are zero)

             if priv_anonparams.TCPReplacePortsByList then
                begin
                     len := length (priv_anonparams.TCPPortList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' TCP port replacement sets to database');

                     for i := 1 to length (priv_anonparams.TCPPortList) do
                         begin
                              Socket1.VLANID := 0;
                              Socket1.IPAddress.ValueAsLong := 0;
                              Socket1.Port := priv_anonparams.TCPPortList[i-1].Value1;

                              Socket2.VLANID := 0;
                              Socket2.IPAddress.initialize(false);
                              Socket2.Port := priv_anonparams.TCPPortList[i-1].Value1;

                              Socket3.VLANID := 0;
                              Socket3.IPAddress.ValueAsLong := 0;
                              Socket3.Port := priv_anonparams.TCPPortList[i-1].Value2;

                              Socket4.VLANID := 0;
                              Socket4.IPAddress.initialize(false);
                              Socket4.Port := priv_anonparams.TCPPortList[i-1].Value2;

                              // Add both IPv4 and IPv6 ports

                              if not priv_anonparams.ReplacementDBWipeOnStart then
                                 begin
                                      if not OriginalAlreadyExists (protTCP, Socket1) then
                                         AddReplacementPair (protTCP, Socket1, Socket3);

                                      if not OriginalAlreadyExists (protTCP, Socket2) then
                                         AddReplacementPair (protTCP, Socket2, Socket4);
                                 end
                              else
                                  begin
                                       AddReplacementPair (protTCP, Socket1, Socket3);
                                       AddReplacementPair (protTCP, Socket2, Socket4);
                                  end;
                         end;
                end;
        end;

     // UDP port parameters

     if priv_anonparams.UDPDoReplace then
        begin
             // UDP Socket list

             if priv_anonparams.UDPReplaceSocketsByList then
                begin
                     len := length (priv_anonparams.UDPSocketPortList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' UDP socket replacement sets to database');

                     for i := 1 to len do
                         begin
                              if priv_anonparams.UDPSocketPortList[i-1].Source.IPAddress.IPVersion = 4 then
                                 begin
                                      Socket1 := priv_anonparams.UDPSocketPortList[i-1].Source.getIPv4Socket;
                                      Socket3 := priv_anonparams.UDPSocketPortList[i-1].Destination.getIPv4Socket;

                                      if priv_anonparams.ReplacementDBWipeOnStart or (not OriginalAlreadyExists (protUDP, Socket1)) then
                                         AddReplacementPair (protUDP, Socket1, Socket3);
                                 end
                              else
                                  begin
                                      Socket2 := priv_anonparams.UDPSocketPortList[i-1].Source.getIPv6Socket;
                                      Socket4 := priv_anonparams.UDPSocketPortList[i-1].Destination.getIPv6Socket;

                                      if priv_anonparams.ReplacementDBWipeOnStart or (not OriginalAlreadyExists (protUDP, Socket2)) then
                                         AddReplacementPair (protUDP, Socket2, Socket4);
                                  end;
                         end;
                end;

             // UDP Port list

             if priv_anonparams.UDPReplacePortsByList then
                begin
                     len := length (priv_anonparams.UDPPortList);

                     SiMain.LogDebug ('Transfering ' + len.ToString + ' UDP port replacement sets to database');

                     for i := 1 to len do
                         begin
                              Socket1.IPAddress.ValueAsLong := 0;
                              Socket1.Port := priv_anonparams.UDPPortList[i-1].Value1;

                              Socket2.IPAddress.initialize (false);
                              Socket2.Port := priv_anonparams.UDPPortList[i-1].Value1;

                              Socket3.IPAddress.ValueAsLong := 0;
                              Socket3.Port := priv_anonparams.UDPPortList[i-1].Value2;

                              Socket4.IPAddress.initialize(false);
                              Socket4.Port := priv_anonparams.UDPPortList[i-1].Value2;

                              // Add both IPv4 and IPv6 ports

                              if not priv_anonparams.ReplacementDBWipeOnStart then
                                 begin
                                      if not OriginalAlreadyExists (protTCP, Socket1) then
                                         AddReplacementPair (protUDP, Socket1, Socket3);

                                      if not OriginalAlreadyExists (protTCP, Socket2) then
                                         AddReplacementPair (protUDP, Socket2, Socket4);
                                 end
                              else
                                  begin
                                       AddReplacementPair (protUDP, Socket1, Socket3);
                                       AddReplacementPair (protUDP, Socket2, Socket4);
                                  end;
                         end;
                end;
        end;

     SiMain.LeaveMethod ('TransferAddressesFromParameters');
end;

(* ------------------------------------------- *)

procedure tReplacementManager.storeCounter (const Name: string; const count: integer);

var Query : tFDQuery;
    help  : string;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     if count = 1 then
        Query.SQL.Text := 'INSERT INTO rdb_counters (name,counter) VALUES (:name,:counter);'
     else
         Query.SQL.Text := 'UPDATE rdb_counters set counter=:counter WHERE name=:name;';

     Query.ParamByName ('name').AsString := Name;
     Query.ParamByName ('counter').AsInteger := Count;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getCounter (const Name : string): integer;

var Query : tFDQuery;
    help  : string;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT counter FROM rdb_counters WHERE name=:name;';
     Query.ParamByName ('name').AsString := Name;
     Query.Open;

     if Query.RecordCount > 0 then
        result := Query.FieldByName ('counter').AsInteger
     else
         result := 0;

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getVLANCount: integer;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT COUNT(*) FROM rdb_vlanids';
     Query.Open;

     result := Query.RecordCount;

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const VLAN : tVLANID): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT originalvlanid FROM rdb_vlanids WHERE originalvlanid=:originalvlanid;';
     Query.ParamByName('originalvlanid').AsInteger := VLAN;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const MACAddress : tMACAddress): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT originalmac FROM rdb_ethernet WHERE originalmac=:originalmac;';
     Query.ParamByName('originalmac').AsString := MACAddress.asString;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const IPAddress : tIPv4Address): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT originalip FROM rdb_ipv4 WHERE originalip=:originalip;';
     Query.ParamByName('originalip').AsLongword := IPAddress.ValueAsLong;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const IPAddress : tIPv6Address): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT originalip FROM rdb_ipv6 WHERE originalip=:originalip;';
     Query.ParamByName('originalip').AsString := IPAddress.ValueAsString;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const Value : String; const Origin : tStringOriginType = qtAny): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     if Origin = qtAny then
        Query.SQL.Text := 'SELECT original FROM rdb_pcapngstring WHERE original=:original;'
     else
         begin
              Query.SQL.Text := 'SELECT replacement FROM rdb_pcapngstrings WHERE origin=:origin AND original=:original;';
              Query.ParamByName ('origin').AsInteger := getStringOriginIndexFromType (Origin);
         end;

     Query.ParamByName('original').AsString := Value;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const Protocol: byte; Socket: tIPv4Socket): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT originalport FROM rdb_tcpports WHERE originalport=:originalport AND originalip=:originalip;';
      protUDP : Query.SQL.Text := 'SELECT originalport FROM rdb_udpports WHERE originalport=:originalport AND originalip=:originalip;';
     end;

     Query.ParamByName('originalip').AsString := Socket.IPAddress.ValueAsString;
     Query.ParamByName('originalport').AsInteger := Socket.Port;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const Protocol: byte; Socket: tIPv6Socket): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT originalport FROM rdb_tcpports WHERE originalport=:originalport AND originalip=:originalip;';
      protUDP : Query.SQL.Text := 'SELECT originalport FROM rdb_udpports WHERE originalport=:originalport AND originalip=:originalip;';
     end;

     Query.ParamByName('originalip').AsString := Socket.IPAddress.ValueAsString;
     Query.ParamByName('originalport').AsInteger := Socket.Port;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const FQDN: tFQDN): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT originalfqdn FROM rdb_nameresolution WHERE originalfqdn=:originalfqdn;';
     Query.ParamByName('originalfqdn').AsString := FQDN.FullFQDN;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const Value: tGUID): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT oldguid FROM rdb_guids WHERE oldguid=:oldguid;';
     Query.ParamByName('oldguid').AsString := Value.ToString;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.OriginalAlreadyExists (const Value: int64; const Origin: tIntegerOriginType) : boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     if Origin = ioAny then
        Query.SQL.Text := 'SELECT original FROM rdb_genericintegers WHERE original=:original;'
     else
         begin
              Query.SQL.Text := 'SELECT replacement FROM rdb_genericintegers WHERE origin=:origin AND original=:original;';
              Query.ParamByName ('origin').AsInteger := getIntegerOriginIndexFromType (Origin);
         end;

     Query.ParamByName('original').AsLargeInt := Value;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const MACAddress: tMACAddress): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT replacementmac FROM rdb_ethernet WHERE replacementmac=:replacementmac;';
     Query.ParamByName('replacementmac').AsString := MACAddress.asString(':');
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const VLAN: tVLANID): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT replacementvlanid FROM rdb_vlanids WHERE replacementvlanid=:replacementvlanid;';
     Query.ParamByName('replacementvlanid').AsInteger := VLAN;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const IPAddress: tIPv4Address): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT replacementip FROM rdb_ipv4 WHERE replacementip=:replacementip;';
     Query.ParamByName('replacementip').AsLongword := IPAddress.ValueAsLong;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const IPAddress: tIPv6Address): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT replacementip FROM rdb_ipv6 WHERE replacementip=:replacementip;';
     Query.ParamByName('replacementip').AsString := IPAddress.ValueAsString;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const Value: String; const Origin : tStringOriginType = qtAny): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     if Origin = qtAny then
        Query.SQL.Text := 'SELECT replacement FROM rdb_pcapngstring WHERE replacement=:replacement;'
     else
         begin
              Query.SQL.Text := 'SELECT replacement FROM rdb_pcapngstrings WHERE origin=:origin AND replacement=:replacement;';
              Query.ParamByName ('origin').AsInteger := getStringOriginIndexFromType (Origin);
         end;

     Query.ParamByName('replacement').AsString := Value;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const Protocol: byte; Socket: tIPv4Socket): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT replacementport FROM rdb_tcpports WHERE replacementport=:replacementport AND originalip=:originalip;';
      protUDP : Query.SQL.Text := 'SELECT replacementport FROM rdb_udpports WHERE replacementport=:replacementport AND originalip=:originalip;';
     end;

     Query.ParamByName('originalip').AsString := Socket.IPAddress.ValueAsString;
     Query.ParamByName('replacementport').AsInteger := Socket.Port;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const Protocol: byte; Socket: tIPv6Socket): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT replacementport FROM rdb_tcpports WHERE replacementport=:replacementport AND originalip=:originalip;';
      protUDP : Query.SQL.Text := 'SELECT replacementport FROM rdb_udpports WHERE replacementport=:replacementport AND originalip=:originalip;';
     end;

     Query.ParamByName('originalip').AsString := Socket.IPAddress.ValueAsString;
     Query.ParamByName('replacementport').AsInteger := Socket.Port;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const Protocol : byte; Port : word) : boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     // Find a port for IPv4 (which is identical to the IPv6 mappings)

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT replacementport FROM rdb_tcpports WHERE replacementport=:replacementport AND originalip="0.0.0.0";';
      protUDP : Query.SQL.Text := 'SELECT replacementport FROM rdb_udpports WHERE replacementport=:replacementport AND originalip="0.0.0.0";';
     end;

     Query.ParamByName('replacementport').AsInteger := Port;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const FQDN : tFQDN): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT replacementfqdn FROM rdb_nameresolution WHERE replacementfqdn=:replacementfqdn;';
     Query.ParamByName('replacementfqdn').AsString := FQDN.FullFQDN;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const Value: tGUID): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT newguid FROM rdb_guids WHERE newguid=:newguid;';
     Query.ParamByName('newguid').AsString := Value.ToString;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const OldVLAN, NewVLAN: tVLANID);

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_vlanids (originalvlanid,replacementvlanid) VALUES (:originalvlan,:replacementvlan);';
     Query.ParamByName('originalvlan').AsInteger := OldVLAN;
     Query.ParamByName('replacementvlan').AsInteger := NewVLAN;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getVLANReplacementFromDB (const VLAN: tVLANID): tVLANID;

var Query : tFDQuery;
    help  : string;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT replacementvlanid FROM rdb_vlanids WHERE originalvlanid=:originalvlan;';
     Query.ParamByName('originalvlan').AsInteger := VLAN;
     Query.Open;

     if Query.RecordCount = 0 then
        result := -1
     else
         result := Query.FieldByName('replacementvlanid').AsInteger;

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForVLAN (const OldVLAN: tVLANID): tVLANID;

var newVLAN     : tVLANID;
    isValid     : boolean;
    RepeatCount : integer;
    i           : integer;

begin
//   SiMain.EnterMethod ('tReplacementManager.getReplacementForVLAN');

     if priv_anonparams.VLANDoReplace then
        begin
             newVLAN := -1;

             // Get VLAN replacement from database, if available
             // Otherwise we get a -1

             // TODO: make sure that replacements are made as configured in the anonymization specs, and not just all the time
//           if priv_anonparams.VLANDoReplaceVLANByIDList then

             newVLAN := getVLANReplacementFromDB (OldVLAN);

             // If there is no replacement, generate random MAC

             if newVLAN = -1 then
                begin
                     RepeatCount := 0;

                     repeat
                           inc (RepeatCount);

                           if Repeatcount > 1 then
                              begin
                                   SiMain.LogWarning ('Warning: randmizing VLAN "' + intToStr (OldVLAN) + '" repeated ' + IntToStr (RepeatCount) + ' times');

                                   if RepeatCount > 10 then
                                      raise Exception.Create('Error: randmizing VLAN "' + intToStr (OldVLAN) + '" repeated more than 10 times. Something isn''t right.');
                              end;

                           // Generate a VLAN between 1 and 4095

                           newVLAN := 1 + random (4094);

                           // Check if is is unique and not the same as the original VLAN

                           isValid := not (ReplacementAlreadyExists (newVLAN) or (OldVLAN=NewVLAN));

                     until isValid;

                     // Store to database to keep replacement consistent

                     addReplacementPair (OldVLAN, newVLAN);
                end;

             result := newVLAN
        end
     else
         result := OldVLAN;

//   SiMain.LeaveMethod ('tReplacementManager.getReplacementForVLAN');
end;

(* ------------------------------------------- *)

function tReplacementManager.getTCPSequenceDelta (const Tuple: t5Tuple; const TCPSequence : tTCPSequence): integer;

var value  : integer;
    key    : string;
    entry  : tTCPSequenceAdjustmentDelta;
    marker : tTCPSequenceAdjustmentMarker;
    len    : integer;
    i      : integer;

begin
     value := 0;
     key := Tuple.DictionaryKey;

     if priv_TCPSequenceDeltaMarkers.ContainsKey (key) then
        begin
             marker := priv_TCPSequenceDeltaMarkers.Items[key];
             len := length (marker.Adjustments);

             for i := 1 to len do
                 begin
                      entry := marker.Adjustments[i-1];

                      // add all deltas for sequence numbers prior to the current

                      if (TCPSequence > entry.TCPSequence) then
                         value := value + entry.Delta;
                 end;
        end;

     result := value;
end;

(* ------------------------------------------- *)

function tReplacementManager.getMACCount: integer;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT COUNT(*) FROM rdb_ethernet';
     Query.Open;

     result := Query.RecordCount;

     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const OldMAC, NewMAC: tMACAddress);

var Query : tFDQuery;

begin
     priv_LookupMAC.Add (OldMAC, NewMAC);

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_ethernet (originalmac,replacementmac) VALUES (:originalmac,:replacementmac);';
     Query.ParamByName('originalmac').AsString := OldMAC.asString(':');
     Query.ParamByName('replacementmac').AsString := NewMAC.asString(':');
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getMACReplacementFromDB (const MAC: tMACAddress): tMACAddress;

var Query : tFDQuery;
    help  : string;

begin
     if priv_LookupMAC.ContainsKey (MAC) then
        result := priv_LookupMAC.Items[MAC]
     else
         begin
              Query := tFDQuery.Create(nil);
              Query.Connection := priv_database;
              Query.SQL.Text := 'SELECT replacementmac FROM rdb_ethernet WHERE originalmac=:originalmac;';
              Query.ParamByName('originalmac').AsString := MAC.asString(':');
              Query.Open;

              if Query.RecordCount = 0 then
                 result := VALUE_ZERO_MAC
              else
                  begin
                       help := Query.FieldByName('replacementmac').AsString;
                       result.setfromString(help, ':');

                       priv_LookupMAC.Add (MAC, result);
                  end;

              Query.Free;
         end;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForAddress (const MAC: tMACAddress; const Force : boolean = false): tMACAddress;

var address     : tMACAddress;
    RepeatCount : integer;
    isValid     : boolean;
    i           : integer;

begin
     // Get MAC Address replacement from database, if available
     // Otherwise we get a zero MAC

     address := getMACReplacementFromDB (MAC);

     // If there is no replacement, generate random MAC

     if (address = VALUE_ZERO_MAC) then
        begin
             if priv_anonparams.EthernetRandomizeMACAddress or Force then
                begin
                     RepeatCount := 0;

                     repeat
                           inc (RepeatCount);

                           if Repeatcount > 1 then
                              begin
                                   SiMain.LogWarning ('Warning: randmizing MAC address "' + MAC.asString + '" repeated ' + IntToStr (RepeatCount) + ' times');

                                   if RepeatCount > 10 then
                                      raise Exception.Create('Error: randmizing MAC address "' + MAC.asString + '" repeated more than 10 times. Something isn''t right.');
                              end;

                           // Check if the original MAC is zero or broadcast and exit if it is

                           if (MAC = VALUE_ZERO_MAC) or (MAC = VALUE_BROADCAST_DEST) then
                              Exit (MAC);

                           // Check if it is a multicast address

                           if (MAC.RangeBytes[1] = $33) and (MAC.RangeBytes[2] = $33) then
                              begin
                                   // Check if it is a well known multicast address, which we do by
                                   // checking if the octets 3 and 4 are zero

                                   if (MAC.RangeBytes[3] = 0) and (MAC.RangeBytes[4] = 0) then
                                      Exit (MAC)
                                   else
                                       begin
                                            // This is most likey a MAC for a solicited node MCast address

                                            address.RangeBytes[1] := $33;
                                            address.RangeBytes[2] := $33;

                                            for i := 3 to 6 do
                                                address.RangeBytes[i] := random (256);
                                       end;
                              end;

                           // Otherwise replace it, leaving the vendor part intact if required

                           if priv_anonparams.EthernetLeaveVendorIntact then
                              begin
                                   address.RangeBytes[1] := MAC.RangeBytes[1];
                                   address.RangeBytes[2] := MAC.RangeBytes[2];
                                   address.RangeBytes[3] := MAC.RangeBytes[3];

                                   i := 4;
                              end
                           else
                               begin
                                    // Set first octed to 0 with locally administered bit set
                                    // and use $f as higher nibble since it is not seen in manuf file

                                    address.RangeBytes[1] := $f2;

                                    i := 2;
                               end;

                           // Generate remaining bytes

                           repeat
                                 address.RangeBytes[i] := random (256);

                                 inc (i)
                           until i > 6;

                           // Check if result is valid, meaning no collision on replacement addresses and not
                           // the same replacement as the original

                           isValid := not (ReplacementAlreadyExists (address) or (MAC=Address));

                     until isValid;

                     // Store to database to keep replacement consistent

                     addReplacementPair (MAC, address);
                end
             else
                 begin
                      if priv_anonparams.EthernetBlackMarkerMACAddress then
                         address.initialize
                      else
                          address := MAC;
                 end;
        end;

     result := address;
end;

(* ------------------------------------------- *)

function tReplacementManager.getIPv4Count: integer;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT COUNT(*) FROM rdb_ipv4';
     Query.Open;

     result := Query.RecordCount;

     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const OldIP, NewIP: tIPv4Address);

var Query : tFDQuery;

begin
     priv_LookupIPv4.Add (OldIP, NewIP);

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_ipv4 (originalip,replacementip) VALUES (:originalip,:replacementip);';
     Query.ParamByName('originalip').AsLongword := OldIP.ValueAsLong;
     Query.ParamByName('replacementip').AsLongword := NewIP.ValueAsLong;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getIPv4AddressReplacementFromDB (const IPAddress: tIPv4Address): tIPv4Address;

var Query : tFDQuery;
    help  : integer;

begin
     result.initialize;

     if priv_LookupIPv4.ContainsKey (IPAddress) then
        result := priv_LookupIPv4.Items[IPAddress]
     else
         begin
              Query := tFDQuery.Create(nil);
              Query.Connection := priv_database;
              Query.SQL.Text := 'SELECT replacementip FROM rdb_ipv4 WHERE originalip=:originalip;';
              Query.ParamByName('originalip').AsLongword := IPAddress.ValueAsLong;
              Query.Open;

              if Query.RecordCount = 0 then
                 result.ValueAsLong := 0
              else
                  begin
                       help := Query.FieldByName('replacementip').AsInteger;

                       result.ValueAsLong := longword (help);

                       priv_LookupIPv4.Add (IPAddress, result);
                  end;

              Query.Free;
         end;
end;

(* ------------------------------------------- *)

function tReplacementManager.getIPv4AddressReplacementFromSubnet (const IPAddress: tIPv4Address): tIPv4Address;

var index : integer;
    net   : tIPv4SubnetDefinition;
    addr  : tIPv4Address;

begin
     addr.initialize;
     index := priv_IPv4Subnethandler.GetSubnetIndexForIPAddress (IPAddress);

     if index = 0 then
        addr.ValueAsLong := 0
     else
         begin
              // We found a matching subnet

              net := priv_IPv4Subnethandler.GetSubnetByIndex(index);

              // get a replacement IP - either by just replacing the net part
              // or by also randomizing the host part

              if priv_anonparams.IPv4KeepHostPart then
                 addr := net.getReplacementIPForSubnet (IPAddress)
              else
                  addr := net.getReplacementIPForSubnet(IPv4_NODE_UNSPECIFIED);
         end;

     result := addr;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForAddress (const IPAddress: tIPv4Address): tIPv4Address;

var replacement : tIPv4Address;
    RepeatCount : integer;
    help        : tIPv4ByteArray;
    isUnique    : boolean;
    isValid     : boolean;
    helprange   : integer;
    subrange    : longword;
    i           : integer;

begin
     // Check the database first

     replacement := getIPv4AddressReplacementFromDB (IPAddress);

     // If this is 0.0.0.0/8 or loopback, return it

     if ((IPAddress.ValueAsLong AND $ff000000) = 0) or IPAddress.isLoopback then
        exit (IPAddress);

     // if the database did not return an address, take a look
     // if there's a subnet rule. Otherwise generate an address

     if replacement.ValueAsLong = 0 then
        begin
             if priv_anonparams.IPv4ReplaceBySubnetlist then
                begin
                     replacement := getIPv4AddressReplacementFromSubnet (IPAddress);

                      // Store Subnet based IP to database if required, which is always
                      // the case for random IPs and optional for those where the Host
                      // part was kept

                      if replacement.ValueAsLong > 0 then
                         begin
                              if priv_anonparams.IPV4StoreGeneratedSubnetIPsToDB or not priv_anonparams.IPv4KeepHostPart then
                                 AddReplacementPair (IPAddress, replacement);
                         end;
                end;

             // Randomize, if configured

             if (replacement.ValueAsLong = 0) then
                begin
                     if priv_anonparams.IPv4RandomizeIPAddress then
                        begin
                             // Handle addresses we do not want to change, like broadcast, all zero, or APIPA

                             if IPAddress.IsBroadcast or (IPAddress = IPv4_NODE_UNSPECIFIED) then
                                Exit (IPAddress);

                             if IPAddress.IsAPIPA and not priv_anonparams.IPv4RandomizeAPIPA then
                                Exit (IPAddress);

                             if IPAddress.IsDocumentationAddress and not priv_anonparams.IPv4RandomizeDocumentation then
                                Exit (IPAddress);

                             if IPAddress.IsPrivate and (priv_anonparams.IPv4PrivateAddressRandomizeMode = parKeepOriginal) then
                                Exit (IPAddress);

                             RepeatCount := 0;

                             repeat
                                   inc (RepeatCount);

                                   if Repeatcount > 1 then
                                      begin
                                           SiMain.LogWarning ('Warning: randmizing IPv4 address "' + IPAddress.ValueAsString + '" repeated ' + IntToStr (RepeatCount) + ' times');

                                           if RepeatCount > 10 then
                                              raise Exception.Create('Error: randmizing IPv4 address "' + IPAddress.ValueAsString + '" repeated more than 10 times. Something isn''t right.');
                                      end;

                                   // Handle multicast

                                   if IPAddress.IsMulticast then
                                      begin
                                           if priv_anonparams.IPv4RandomizeMulticast then
                                              begin
                                                   help[1] := 224 + random (16);

                                                   for i := 2 to 3 do
                                                       help[i] := random (256);
                                              end
                                           else
                                               Exit (IPAddress);
                                      end
                                   else
                                       if IPAddress.IsPrivate and (priv_anonparams.IPv4PrivateAddressRandomizeMode = parRandomizeToPrivate) then
                                          begin
                                               // Randomize, but to a private range

                                               replacement.initialize (true, iprsPrivate);
                                          end
                                       else
                                           begin
                                                // Randomize and make sure the original is not kept by accident

                                                replacement.initialize (true, iprsAll);
                                           end;

                                   // Check if address is already a replacement

                                   isUnique := not (ReplacementAlreadyExists (replacement) or  (IPAddress = Replacement));

                             until isUnique;

                             // Add to database

                             AddReplacementPair (IPAddress, replacement);
                        end
                     else
                         if priv_anonparams.IPv4BlackMarkerIPAddress then
                            replacement.initialize
                         else
                             // Worst case, we found no replacement, so we return the original

                             replacement := IPAddress;
                end;
        end;

     result := replacement;
end;

(* ------------------------------------------- *)

function tReplacementManager.getIPv6Count: integer;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT COUNT(*) FROM rdb_ipv6';
     Query.Open;

     result := Query.RecordCount;

     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const OldIP, NewIP: tIPv6Address);

var Query : tFDQuery;

begin
     priv_LookupIPv6.Add (OldIP, NewIP);

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_ipv6 (originalip,replacementip) VALUES (:originalip,:replacementip);';
     Query.ParamByName('originalip').AsString := OldIP.ValueAsString;
     Query.ParamByName('replacementip').AsString := NewIP.ValueAsString;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getIPv6AddressReplacementFromDB (const IPAddress: tIPv6Address): tIPv6Address;

var Query : tFDQuery;
    help  : string;

begin
     result.initialize;

     if priv_LookupIPv6.ContainsKey (IPAddress) then
        result := priv_LookupIPv6.Items[IPAddress]
     else
         begin
              Query := tFDQuery.Create(nil);
              Query.Connection := priv_database;
              Query.SQL.Text := 'SELECT replacementip FROM rdb_ipv6 WHERE originalip=:originalip;';
              Query.ParamByName('originalip').AsString := IPAddress.ValueAsString;
              Query.Open;

              if Query.RecordCount = 0 then
                 result.ValueAsWordStruct := IPv6_NODE_UNSPECIFIED
              else
                  begin
                       help := Query.FieldByName('replacementip').AsString;
                       result.setValueFromString(help);

                       priv_LookupIPv6.Add (IPAddress, result);
                  end;

              Query.Free;
         end;
end;

(* ------------------------------------------- *)

function tReplacementManager.getIPv6AddressReplacementFromPrefix (const IPAddress: tIPv6Address): tIPv6Address;

var index : integer;
    net   : tIPv6SubnetDefinition;
    addr  : tIPv6Address;

begin
     // Check the subnet handler if it has a subnet for the orignal address

     index := priv_IPv6Subnethandler.GetSubnetIndexForIPAddress (IPAddress);

     if index > 0 then
         begin
              // We found a matching subnet

              net := priv_IPv6Subnethandler.GetSubnetByIndex(index);

              // get a replacement IP - either by just replacing the net part
              // or by also randomizing the host part

              if priv_anonparams.IPv6KeepHostPart and not IPAddress.IsUnspecifiedAddress then
                 addr := net.getReplacementIPForSubnet (IPAddress)
              else
                  begin
                       // Create a new address

//                     addr.initialize (true);

                       addr := getRandomIPv6Address (IPAddress);
                       addr := net.getReplacementIPForSubnet(addr);
                  end;
         end
     else
         addr.initialize;

     result := addr;
end;

(* ------------------------------------------- *)

function tReplacementManager.getIPv6PrefixReplacement (const IPAddress: tIPv6Address): tIPv6Address;

var replacement   : tIPv6Address;
    dummy         : tIPv6Address;
    maskbits      : integer;
    remainingbits : integer;
    fullblocks    : integer;
    RepeatCount   : integer;
    block         : word;
    isUnique      : boolean;
    Subnet        : tIPv6SubnetDefinition;
    i             : integer;

begin
     // Check if there is a replacement already

     replacement := getIPv6AddressReplacementFromPrefix (IPAddress);

     if not replacement.IsUnspecifiedAddress then
        exit (replacement);

     // Create a new prefix

     block := IPAddress.ValueAsWordStruct[0];

     replacement.initialize;
     replacement.ValueAsWordStruct[0] := 1 + random (65534);    // Safety fallback
     replacement.MaskBitCount := IPAddress.MaskBitCount;

     repeatcount := 0;

     repeat
           isUnique := true;

           inc (RepeatCount);

           if Repeatcount > 1 then
              begin
                   SiMain.LogWarning ('Warning: creating IPv6 prefix "' + IPAddress.ValueAsString + '" repeated ' + IntToStr (RepeatCount) + ' times');

                   if RepeatCount > 10 then
                      raise Exception.Create('Error: creating IPv6 prefix "' + IPAddress.ValueAsString + '" repeated more than 10 times. Something isn''t right.');
              end;

           // Link local shoudl stay link local

           if (block = $fe80) then
              begin
                   replacement.ValueAsWordStruct[0] := $fe80;
                   replacement.ValueAsWordStruct[1] := 0;
                   replacement.ValueAsWordStruct[2] := 0;
                   replacement.ValueAsWordStruct[3] := 0;
              end
           else
               begin
                    // Create random remaining full prefix blocks (16 bits each)

                    maskbits := IPAddress.MaskBitCount - 16;
                    fullblocks := maskbits div 16;
                    remainingbits := maskbits mod 16;

                    for i := 1 to fullblocks do
                        replacement.ValueAsWordStruct[i] := random (65536);

                    // If we have a partial block used for prefix bits, create them

                    if remainingbits > 0 then
                       begin
                            block := random (65536);

                            case remainingbits of
                              1 : block := (block and $8000);
                              2 : block := (block and $C000);
                              3 : block := (block and $E000);
                              4 : block := (block and $F000);
                              5 : block := (block and $F800);
                              6 : block := (block and $FC00);
                              7 : block := (block and $FE00);
                              8 : block := (block and $FF00);
                              9 : block := (block and $FF80);
                             10 : block := (block and $FFC0);
                             11 : block := (block and $FFE0);
                             12 : block := (block and $FFF0);
                             13 : block := (block and $FFF8);
                             14 : block := (block and $FFFC);
                             15 : block := (block and $FFFE);
                            end;

                            replacement.ValueAsWordStruct[fullblocks + 1] := block;
                       end;

                    // Determine & overwrite first block for specific ranges to keep the replacement similar

                    // Public range

                    if (block >= $2000) and (block <= $3fff) then
                       replacement.ValueAsWordStruct[0] := $2000 + random ($2000);

                    // Local Unicast ranges

                    if (block >= $fc00) and (block <= $fcff) then
                       replacement.ValueAsWordStruct[0] := $fc00 + random (256);

                    if (block >= $fd00) and (block <= $fdff) then
                       replacement.ValueAsWordStruct[0] := $fd00 + random (256);
               end;

           // Check if the new prefix doesn't already exist

           isUnique := (priv_IPv6Subnethandler.GetSubnetIndexForIPAddress (replacement) = 0)

     until isUnique;

     // Add a subnet replacement entry to make sure we replace IP addresses to the correct new subnet

     dummy.initialize;
     dummy.CopyPrefixFrom (IPAddress);

     Subnet := tIPv6SubnetDefinition.create (dummy);
     Subnet.ReplacementNetAddress := replacement;

     // Add the new subnet pair to the subnet store
     // TODO: maybe write to replacement DB as well?

     priv_IPv6Subnethandler.AddSubnet (Subnet);

     result := replacement;
end;

(* ------------------------------------------- *)

function tReplacementManager.getRandomIPv6Address (const OriginalAddress: tIPv6Address) : tIPv6Address;

var replacement     : tIPv6Address;
    OriginalMAC     : tMACAddress;
    ReplacementMAC  : tMACAddress;
    block           : word;

begin
     // Check if is EUI64, because then we replace based on MAC

     if OriginalAddress.IsEUI64Address then
        begin
             OriginalMAC := OriginalAddress.getEUI64baseMAC;

             // Get a replacement MAC, either existing or a new one (automatically stored to DB)
             // since sanitizing Ethernet may not be active we need to force it in this case

             ReplacementMAC := getReplacementForAddress (OriginalMAC, true);

             replacement.setValueFromMAC(OriginalAddress, ReplacementMAC);
        end
     else
         begin
              // Randomize all blocks first

              replacement.initialize (true);
         end;

     // Keep the IP in the same address range as the original

     block := OriginalAddress.ValueAsWordStruct[0];

     // Public range

     if (block >= $2000) and (block <= $3fff) then
        replacement.ValueAsWordStruct[0] := $2000 + random ($2000);

     // Local Unicast ranges

     if (block >= $fc00) and (block <= $fcff) then
        replacement.ValueAsWordStruct[0] := $fc00 + random (256);

     if (block >= $fd00) and (block <= $fdff) then
        replacement.ValueAsWordStruct[0] := $fd00 + random (256);

     // Link local

     if (block = $fe80) then
        begin
             replacement.ValueAsWordStruct[0] := $fe80;
             replacement.ValueAsWordStruct[1] := 0;
             replacement.ValueAsWordStruct[2] := 0;
             replacement.ValueAsWordStruct[3] := 0;
        end;

     result := replacement;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.CreateSolicitedNodeReplacement (const OriginalAddress, Replacement: tIPv6Address);

var OrigSolicited   : tIPv6Address;
    NewSolicited    : tIPv6Address;
    OrigMAC         : tMACAddress;
    NewMAC          : tMACAddress;
    help            : tMACAddress;

begin
     // Create Solicited node multicast entries in case we need them

     OrigSolicited := OriginalAddress.getSolicitedNodeAddress;
     NewSolicited := getIPv6AddressReplacementFromDB (OrigSolicited);

     // Check if it already exists and add it if it doesn't

     if NewSolicited.IsUnspecifiedAddress then
        begin
             // Need to base this on the MCast MAC address so it stays synchronized

             NewSolicited := replacement.getSolicitedNodeAddress;

             OrigMAC := OrigSolicited.getMCastBaseMAC;
             NewMAC := NewSolicited.getMCastBaseMAC;

             help := getMACReplacementFromDB (OrigMAC);

             if (help = VALUE_ZERO_MAC) then
                AddReplacementPair (OrigMAC, NewMAC)
             else
                 if (help <> NewMAC) then
                    raise Exception.Create('Error: Solicited Node MCast MAC mismatch.');

             AddReplacementPair (OrigSolicited, NewSolicited);
        end;

end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForAddress (const IPAddress: tIPv6Address) : tIPv6Address;

var replacement     : tIPv6Address;
    NewPrefix       : tIPv6Address;
    isNormalAddress : boolean;
    isUnique        : boolean;
    RepeatCount     : integer;

begin
     // Check if we're seeing all-zero or loopback

     if IPAddress.IsUnspecifiedAddress or IPAddress.IsLocalHost then
        exit (IPAddress);

     // Check if we're dealing with a prefix, not an address

     if IPAddress.IsPrefixInformation then
        begin
             // Generate a new prefix (it is automatically stored the the subnet replacement list in the process)

             replacement := getIPv6PrefixReplacement (IPAddress);
        end
     else
         replacement := getIPv6AddressReplacementFromDB (IPAddress);

     // if the database did not return an address, take a look
     // if there's a subnet rule. Otherwise generate an address

     if replacement.IsUnspecifiedAddress then
        begin
             // Check by subnet list

             if priv_anonparams.IPv6ReplaceBySubnetlist then
                begin
                     replacement := getIPv6AddressReplacementFromPrefix (IPAddress);

                     if not replacement.IsUnspecifiedAddress then
                        begin
                             // Store Subnet based IP to database if required, which is always
                             // the case for random IPs and optional for those where the Host
                             // part was kept

                             if priv_anonparams.IPV6StoreGeneratedSubnetIPsToDB or not priv_anonparams.IPv6KeepHostPart then
                                AddReplacementPair (IPAddress, replacement);
                        end;
                end;

             if replacement.IsUnspecifiedAddress then
                begin
                     // Check for documentation address

                     if IPAddress.IsDocumentationAddress and not priv_anonparams.IPv6RandomizeDocumentationAddr then
                        Exit (IPAddress);

                     // Handle Multicast

                     if IPAddress.IsMulticast then
                        begin
                             // Solicited node is handled by adding them on unicast replacements automatically,
                             // so a DB search will have them turned up already. All other multicast addresses
                             // are not touched

                             if IPAddress.isSolicitedNodeMCast then
                                begin
                                     // Safeguard: clear the trailing three bytes

                                     replacement := IPAddress;
                                     replacement.ValueAsWordStruct[6] := $ff00;
                                     replacement.ValueAsWordStruct[7] := 0;
                                end
                             else
                                 Exit (IPAddress);
                        end
                end;

             // Check if replacementmode is randomization with prefix prevention or do full randomization otherwise

             if replacement.IsUnspecifiedAddress and (priv_anonparams.IPv6RandomizeAddressMode <> iprNone) then
                begin
                     RepeatCount := 0;

                     repeat
                           // Randomize and make sure the original is not kept by accident

                           inc (RepeatCount);

                           if Repeatcount > 1 then
                              begin
                                   SiMain.LogWarning ('Warning: randmizing IPv6 address "' + IPAddress.ValueAsString + '" repeated ' + IntToStr (RepeatCount) + ' times');

                                   if RepeatCount > 10 then
                                      raise Exception.Create('Error: Randomizing IPv6 address "' + IPAddress.ValueAsString + '" repeated more than 10 times. Something isn''t right.');
                              end;

                           // Check if we have a fe80:: address with an EUI64 interface ID and
                           // randomize it. If we don't do it here, it will cause a loop when
                           // the IPv6RandomizeAddressMode is iprRandomizeAndKeepPrefix

                           if (IPAddress.IsEUI64Address and IPAddress.IsLinkLocalAddress) then
                              begin
                                   replacement := getRandomIPv6Address (IPAddress);

                                   // Now check if we don't have this replacement address for a different original address already

                                   isUnique := not (ReplacementAlreadyExists (replacement) or (IPAddress = Replacement));
                              end
                           else
                               begin
                                    // Normal address; handle it according to randomization setting

                                    case priv_anonparams.IPv6RandomizeAddressMode of
                                     iprRandomizeAndKeepPrefix     : begin
                                                                          // randomizes only the prefix and stores it to the subnet list so it will match for the
                                                                          // next address with the same original prefix

                                                                          replacement := IPAddress;

                                                                          // create new prefix (will be written to the database in the process)

                                                                          NewPrefix := getIPv6PrefixReplacement (IPaddress);

                                                                          // Write new prefix over the old one

                                                                          replacement.CopyPrefixFrom (NewPrefix, 64);

                                                                          // Now check if we don't have this replacement address for a different original address already

                                                                          isUnique := not (ReplacementAlreadyExists (replacement) or (not IPAddress.IsLinkLocalAddress and (IPAddress = Replacement)));
                                                                     end;
                                     iprRandomizeBothAndKeepPrefix : begin
                                                                          // randomizes both prefix and interface ID, and stores the prefix to the subnet list

                                                                          replacement.initialize (true);

                                                                          // create new prefix (will be written to the database in the process)

                                                                          NewPrefix := getIPv6PrefixReplacement (IPaddress);

                                                                          // Write new prefix over the old one

                                                                          replacement.CopyPrefixFrom (NewPrefix, 64);

                                                                          // Now check if we don't have this replacement address for a different original address already

                                                                          isUnique := not (ReplacementAlreadyExists (replacement) or (IPAddress = Replacement));
                                                                     end;
                                     iprFullStatelessRandomization : begin
                                                                          // random prefix, random IP

                                                                          replacement := getRandomIPv6Address (IPAddress);

                                                                          // Now check if we don't have this replacement address for a different original address already

                                                                          isUnique := not (ReplacementAlreadyExists (replacement) or (IPAddress = Replacement));
                                                                     end;
                                     iprBlackMarker                : begin
                                                                          replacement.initialize;
                                                                     end;
                                    end;
                               end;

                     until isUnique or (priv_anonparams.IPv6RandomizeAddressMode = iprBlackMarker);

                     if not (priv_anonparams.IPv6RandomizeAddressMode = iprBlackMarker) then
                        begin
                             // Add new replacement to database

                             AddReplacementPair (IPAddress, replacement);

                             // Create a matching solicited node address if it doesn't exist yet

                             CreateSolicitedNodeReplacement (IPAddress, replacement);
                        end;
                end
             else
                 replacement := IPAddress;
{
             else
                 begin
                      if priv_anonparams.IPv6BlackMarkerIPAddress then
                         replacement.initialize
                      else
                          replacement := IPAddress;
                 end;
}
        end;

     result := replacement;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const OldString,  NewString: string; const Origin: tStringOriginType);

var Query : tFDQuery;

begin
     if Origin = qtAny then
        priv_LookupStrings.Add (OldString, NewString);

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_pcapngstrings (origin,original,replacement) VALUES (:origin,:original,:replacement);';
     Query.ParamByName ('origin').AsInteger := getStringOriginIndexFromType (Origin);
     Query.ParamByName ('original').AsString := OldString;
     Query.ParamByName ('replacement').AsString := NewString;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForString (const OldString: string; const Origin: tStringOriginType): string;

var Query : tFDQuery;

begin
     if priv_LookupStrings.ContainsKey (OldString) then
        result := priv_LookupStrings.Items[OldString]
     else
         begin
              Query := tFDQuery.Create(nil);
              Query.Connection := priv_database;

              if Origin = qtAny then
                 begin
                      Query.SQL.Text := 'SELECT replacement FROM rdb_pcapngstrings WHERE original=:original;';
                      Query.ParamByName('original').AsString := OldString;
                 end
              else
                  begin
                       Query.SQL.Text := 'SELECT replacement FROM rdb_pcapngstrings WHERE origin=:origin AND original=:original;';
                       Query.ParamByName('original').AsString := OldString;
                       Query.ParamByName('origin').AsInteger := getStringOriginIndexFromType (Origin);
                  end;

              Query.Open;

              if Query.RecordCount = 0 then
                 begin
                      // For WiFi, we may have to generate a generic new SSID

                      if Origin = qt802dot11 then
                         begin
                              if priv_anonparams.WiFiReplaceRemainingSSIDs then
                                 begin
                                      inc (priv_SSIDIndex);

                                      result := priv_anonparams.WiFiReplaceRemainingSSIDsPrefix + priv_SSIDIndex.ToString;

                                      AddReplacementPair (Oldstring, result, qt802dot11);
                                 end;
                         end
                      else
                          result := ''
                 end
              else
                  result := Query.FieldByName('replacement').AsString;
         end;

     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair(const OldGUID, NewGUID: tGUID);

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_guids (OldGUID,NewGUID) VALUES (:original,:replacement);';
     Query.ParamByName ('original').AsString := GUIDToString (OldGUID);
     Query.ParamByName ('replacement').AsString := GUIDToString (NewGUID);
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

// IP-ID for IPv4

procedure tReplacementManager.AddReplacementPair (const SocketPair : tIPv4SocketPair; const OldIPID, NewIPID : tIPID);

var Query : tFDQuery;
    help  : tIPv4SocketPair;

begin
     help := SocketPair;
     help.normalize;

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_ipidentification (originalid,replacementid,sourceip,destinationip) VALUES (:original,:replacement,:sourceip,:destinationip);';
     Query.ParamByName ('original').AsInteger := OldIPID;
     Query.ParamByName ('replacement').AsInteger := NewIPID;
     Query.ParamByName ('sourceip').AsString := Help.Source.IPAddress.ValueAsString;
     Query.ParamByName ('destinationip').AsString := help.Destination.IPAddress.ValueAsString;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

// IP-ID for IPv6

procedure tReplacementManager.AddReplacementPair (const SocketPair : tIPv6SocketPair; const OldIPID, NewIPID : tIPID);

var Query : tFDQuery;
    help  : tIPv6SocketPair;

begin
     help := SocketPair;
     help.normalize;

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_ipidentification (originalid,replacementid,sourceip,destinationip) VALUES (:original,:replacement,:sourceip,:destinationip);';
     Query.ParamByName ('original').AsInteger := OldIPID;
     Query.ParamByName ('replacement').AsInteger := NewIPID;
     Query.ParamByName ('sourceip').AsString := Help.Source.IPAddress.ValueAsString;
     Query.ParamByName ('destinationip').AsString := help.Destination.IPAddress.ValueAsString;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForGUID (const OldGuid: tGUID): tGUID;

var Query : tFDQuery;
    help  : string;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'SELECT NewGUID FROM rdb_guids WHERE OldGUID=:original;';
     Query.ParamByName('original').AsString := GUIDToString (OldGuid);
     Query.Open;

     if Query.RecordCount = 0 then
        result := ZERO_GUID
     else
         begin
              help := Query.FieldByName('NewGUID').AsString;

              result := StringToGUID (help);
         end;

     Query.Free;
end;

(* ------------------------------------------- *)

// procedure tReplacementManager.AddReplacementPair (const Protocol : byte; const OldSocket: tIPv4Socket; NewPort: word);
procedure tReplacementManager.AddReplacementPair (const Protocol : byte; const OldSocket: tIPv4Socket; const NewSocket: tIPv4Socket);

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'INSERT INTO rdb_tcpports (originalip, originalport, replacementip, replacementport) VALUES (:originalip,:originalport,:replacementip,:replacementport);';
      protUDP : Query.SQL.Text := 'INSERT INTO rdb_udpports (originalip, originalport, replacementip, replacementport) VALUES (:originalip,:originalport,:replacementip,:replacementport);';
     end;

     Query.ParamByName ('originalip').AsString := OldSocket.IPAddress.ValueAsString;
     Query.ParamByName ('originalport').AsInteger := OldSocket.Port;
     Query.ParamByName ('ReplacementIP').AsString := NewSocket.IPAddress.ValueAsString;
     Query.ParamByName ('replacementport').AsInteger := NewSocket.Port;

     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const Protocol : byte; const OldSocket: tIPv6Socket; const NewSocket: tIPv6Socket);

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'INSERT INTO rdb_tcpports (originalip, originalport, replacementip, replacementport) VALUES (:originalip,:originalport,:replacementip,:replacementport);';
      protUDP : Query.SQL.Text := 'INSERT INTO rdb_udpports (originalip, originalport, replacementip, replacementport) VALUES (:originalip,:originalport,:replacementip,:replacementport);';
     end;

     Query.ParamByName ('originalip').AsString := OldSocket.IPAddress.ValueAsString;
     Query.ParamByName ('originalport').AsInteger := OldSocket.Port;
     Query.ParamByName ('ReplacementIP').AsString := NewSocket.IPAddress.ValueAsString;
     Query.ParamByName ('replacementport').AsInteger := NewSocket.Port;

     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const Protocol : byte; const OldPort, NewPort : word);

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'INSERT INTO rdb_tcpports (originalip, originalport, replacementip, replacementport) VALUES (:originalip,:originalport,:replacementip,:replacementport);';
      protUDP : Query.SQL.Text := 'INSERT INTO rdb_udpports (originalip, originalport, replacementip, replacementport) VALUES (:originalip,:originalport,:replacementip,:replacementport);';
     end;

     Query.ParamByName ('originalip').AsString := '0.0.0.0';
     Query.ParamByName ('originalport').AsInteger := OldPort;
     Query.ParamByName ('ReplacementIP').AsString := '0.0.0.0';
     Query.ParamByName ('replacementport').AsInteger := NewPort;

     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const OldFQDN, NewFQDN: tFQDN);

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_nameresolution (originalfqdn,replacementfqdn) VALUES (:originalfqdn,:replacementfqdn);';
     Query.ParamByName ('originalfqdn').AsString := OldFQDN.FullFQDN;
     Query.ParamByName ('replacementfqdn').AsString := NewFQDN.FullFQDN;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddReplacementPair (const OldInteger, NewInteger: int64; const Origin: tIntegerOriginType);

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;
     Query.SQL.Text := 'INSERT INTO rdb_genericintegers (origin,original,replacement) VALUES (:origin,:original,:replacement);';
     Query.ParamByName ('origin').AsInteger := getIntegerOriginIndexFromType (Origin);
     Query.ParamByName ('original').AsLargeInt := OldInteger;
     Query.ParamByName ('replacement').AsLargeInt := NewInteger;
     Query.ExecSQL;
     Query.Free;
end;

(* ------------------------------------------- *)

procedure tReplacementManager.AddTCPSequenceDelta (const tuple : t5Tuple; const TCPSequence : tTCPSequence; const delta : integer);

var key    : string;
    entry  : tTCPSequenceAdjustmentDelta;
    marker : tTCPSequenceAdjustmentMarker;

begin
     entry.TCPSequence := TCPSequence;
     entry.Delta := delta;

     key := tuple.DictionaryKey;

     if priv_TCPSequenceDeltaMarkers.ContainsKey (key) then
        begin
             marker := priv_TCPSequenceDeltaMarkers.Items[key];
             marker.addAdjustment (entry);

             // Store

             priv_TCPSequenceDeltaMarkers.Items[key] := marker;
        end
     else
         begin
              marker.initizalize (tuple);
              marker.addAdjustment (entry);

              priv_TCPSequenceDeltaMarkers.Add (key, marker);
         end;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForSocket (const Protocol : byte; const OldSocket : tIPv4Socket): tIPv4Socket;

var Query        : tFDQuery;
    NewSocket    : tIPv4Socket;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     // Check database first

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT replacementip,replacementport FROM rdb_tcpports WHERE originalport=:originalport AND originalip=:originalip;';
      protUDP : Query.SQL.Text := 'SELECT replacementip,replacementport FROM rdb_udpports WHERE originalport=:originalport AND originalip=:originalip;';
     end;

     Query.ParamByName ('originalport').AsInteger := OldSocket.Port;
     Query.ParamByName ('originalip').AsString := OldSocket.IPAddress.ValueAsString;

     Query.Open;

     if Query.RecordCount > 0 then
        begin
             NewSocket.IPAddress.setValueFromString (Query.FieldByName('replacementip').AsString);
             NewSocket.Port := Query.FieldByName('replacementport').AsInteger;
        end
     else
         NewSocket.initialize;

     Query.Free;

     result := NewSocket;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForSocket (const Protocol : byte; const OldSocket : tIPv6Socket) : tIPv6Socket;

var Query        : tFDQuery;
    NewSocket    : tIPv6Socket;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT replacementport FROM rdb_tcpports WHERE originalport=:originalport AND originalip=:originalip;';
      protUDP : Query.SQL.Text := 'SELECT replacementport FROM rdb_udpports WHERE originalport=:originalport AND originalip=:originalip;';
     end;

     Query.ParamByName ('originalport').AsInteger := OldSocket.Port;
     Query.ParamByName ('originalip').AsString := OldSocket.IPAddress.ValueAsString;

     Query.Open;

     if Query.RecordCount > 0 then
        begin
             NewSocket.IPAddress.setValueFromString (Query.FieldByName('replacementip').AsString);
             NewSocket.Port := Query.FieldByName('replacementport').AsInteger;
        end;

     Query.Free;

     result := NewSocket;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForPort (const Protocol : byte; const OldPort : word) : word;

var Query        : tFDQuery;
    Port         : word;
    isSystemPort : boolean;
    boRandomize  : boolean;
    RepeatCount  : integer;
    NewPort      : word;
    isUnique     : boolean;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     // Check database first

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT replacementport FROM rdb_tcpports WHERE originalport=:originalport AND originalip="0.0.0.0";';
      protUDP : Query.SQL.Text := 'SELECT replacementport FROM rdb_udpports WHERE originalport=:originalport AND originalip="0.0.0.0";';
     end;

     Query.ParamByName ('originalport').AsInteger := OldPort;

     Query.Open;

     if Query.RecordCount = 0 then
        begin
             // Randomize Ports if enabled

             isSystemPort := (OldPort > 0) and (OldPort < 1024);

             boRandomize := False;

             case Protocol of
              protTCP : begin
                             if priv_anonparams.TCPRandomizePorts then
                                boRandomize := (not isSystemPort) or priv_anonparams.TCPIgnoreSystemPorts;
                        end;
              protUDP : begin
                             if priv_anonparams.UDPRandomizePorts then
                                boRandomize := (not isSystemPort) or priv_anonparams.UDPIgnoreSystemPorts;
                        end;
             end;

             if boRandomize then
                begin
                     RepeatCount := 0;

                     repeat
                           inc (RepeatCount);

                           if Repeatcount > 1 then
                              begin
                                   SiMain.LogWarning ('Warning: randmizing port for "' + OldPort.ToString + '" repeated ' + IntToStr (RepeatCount) + ' times');

                                   if RepeatCount > 10 then
                                      raise Exception.Create('Error: Randomizing port for "' + OldPort.ToString + '" repeated more than 10 times. Something isn''t right.');
                              end;

                           // Create new port replacement

                           Port := 1 + Random (65535);

                           NewPort := Port;

                           isUnique := not (ReplacementAlreadyExists (Protocol, NewPort) or (OldPort = NewPort));

                     until isUnique;

                     AddReplacementPair (Protocol, OldPort, NewPort);
                end
             else
                 Port := OldPort;
        end
     else
         Port := Query.FieldByName('replacementport').AsInteger;

     Query.Free;

     result := Port;
end;

(* ------------------------------------------- *)
{
function tReplacementManager.getReplacementForPort (const Protocol : byte; const OldPort : word) : word;

var Query        : tFDQuery;
    Port         : word;
    isSystemPort : boolean;
    boRandomize  : boolean;
    RepeatCount  : integer;
    NewSocket    : tIPv6Socket;
    isUnique     : boolean;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     case Protocol of
      protTCP : Query.SQL.Text := 'SELECT replacementport FROM rdb_tcpports WHERE originalport=:originalport AND originalip="::");';
      protUDP : Query.SQL.Text := 'SELECT replacementport FROM rdb_udpports WHERE originalport=:originalport AND originalip="::");';
     end;

     Query.ParamByName ('originalport').AsInteger := OldSocket.Port;
     Query.ParamByName ('originalip').AsString := OldSocket.IPAddress.ValueAsString;

     Query.Open;

     if Query.RecordCount = 0 then
        begin
             // Randomize Ports if enabled

             isSystemPort := (OldSocket.Port > 0) and (OldSocket.Port < 1024);

             case Protocol of
              protTCP : begin
                             if priv_anonparams.TCPRandomizePorts then
                                boRandomize := (not isSystemPort) or priv_anonparams.TCPIgnoreSystemPorts;
                        end;
              protUDP : begin
                             if priv_anonparams.UDPRandomizePorts then
                                boRandomize := (not isSystemPort) or priv_anonparams.UDPIgnoreSystemPorts;
                        end;
             end;

             if boRandomize then
                begin
                     RepeatCount := 0;

                     repeat
                           inc (RepeatCount);

                           if Repeatcount > 1 then
                              begin
                                   SiMain.LogWarning ('Warning: randmizing port for "' + OldSocket.ValueAsString + '" repeated ' + IntToStr (RepeatCount) + ' times');

                                   if RepeatCount > 10 then
                                      raise Exception.Create('Error: Randomizing port for "' + OldSocket.ValueAsString + '" repeated more than 10 times. Something isn''t right.');
                              end;

                           // Create new port replacement

                           Port := 1 + Random (65535);

                           NewSocket := OldSocket;
                           NewSocket.Port := Port;

                           isUnique := not (ReplacementAlreadyExists (Protocol, NewSocket) or (OldSocket = NewSocket));

                     until isUnique;

                     AddReplacementPair (Protocol, OldSocket, NewSocket);
                end
             else
                 Port := OldSocket.Port;
        end
     else
         Port := Query.FieldByName('replacementport').AsInteger;

     Query.Free;

     result := Port;
end;
}
(* ------------------------------------------- *)

function tReplacementManager.getReplacementForIPID (const OldIPID : tIPID; const OldSocketPair: tIPv4SocketPair) : tIPID;

var Query      : tFDQuery;
    help       : tIPv4SocketPair;
    NewID      : tIPID;
    IDIsUnique : boolean;

begin
     help := OldSocketPair;
     help.normalize;

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     Query.SQL.Text := 'SELECT replacementid FROM rdb_ipidentification WHERE originalid=:originalid AND (sourceip=:sourceip and destinationip=:destinationip);';

     Query.ParamByName ('originalid').AsInteger := OldIPID;
     Query.ParamByName ('sourceip').AsString := help.Source.IPAddress.ValueAsString;
     Query.ParamByName ('destinationip').AsString := help.Destination.IPAddress.ValueAsString;

     Query.Open;

     if Query.RecordCount = 0 then
        begin
             if priv_anonparams.IPv4RandomizeIPID then
                begin
                     // Create new id replacement

                     repeat
                           NewID := Random (65536);

                           // Check that we do not have a replacement that already exists

                           Query.SQL.Text := 'SELECT replacementid FROM rdb_ipidentification WHERE replacementid=:newid AND (sourceip=:sourceip and destinationip=:destinationip);';
                           Query.ParamByName ('newid').AsInteger := NewID;
                           Query.ParamByName ('sourceip').AsString := help.Source.IPAddress.ValueAsString;
                           Query.ParamByName ('destinationip').AsString := help.Destination.IPAddress.ValueAsString;

                           Query.Open;

                           IDIsUnique := (Query.RecordCount = 0) and (OldIPID <> NewID);

                     until IDIsUnique;

                     AddReplacementPair (OldSocketPair, OldIPID, NewID);
                end
             else
                 NewID := OldIPID;
        end
     else
         NewID := Query.FieldByName('replacementid').AsInteger;

     Query.Free;

     result := NewID;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForInteger (const OldInteger: int64; const Origin: tIntegerOriginType) : int64;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     if Origin = ioAny then
        begin
             Query.SQL.Text := 'SELECT replacement FROM rdb_genericintegers WHERE original=:original;';
             Query.ParamByName('original').AsLargeInt := OldInteger;
        end
     else
         begin
              Query.SQL.Text := 'SELECT replacement FROM rdb_genericintegers WHERE origin=:origin AND original=:original;';
              Query.ParamByName('original').AsLargeInt := OldInteger;
              Query.ParamByName ('origin').AsInteger := getIntegerOriginIndexFromType (Origin);
         end;

     Query.Open;

     if Query.RecordCount = 0 then
        result := -1
     else
         result := Query.FieldByName('replacement').AsLargeInt;

     Query.Free;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForIPID (const OldIPID : tIPID; const OldSocketPair: tIPv6SocketPair) : tIPID;

var Query      : tFDQuery;
    help       : tIPv6SocketPair;
    NewID      : tIPID;
    IDIsUnique : boolean;

begin
     help := OldSocketPair;
     help.normalize;

     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     Query.SQL.Text := 'SELECT replacementid FROM rdb_ipidentification WHERE originalid=:originalid AND (sourceip=:sourceip and destinationip=:destinationip);';

     Query.ParamByName ('originalid').AsInteger := OldIPID;
     Query.ParamByName ('sourceip').AsString := help.Source.IPAddress.ValueAsString;
     Query.ParamByName ('destinationip').AsString := help.Destination.IPAddress.ValueAsString;

     Query.Open;

     if Query.RecordCount = 0 then
        begin
             if priv_anonparams.IPv4RandomizeIPID then
                begin
                     // Create new id replacement

                     repeat
                           NewID := Random (65536);

                           // Check that we do not have a replacement that already exists

                           Query.SQL.Text := 'SELECT replacementid FROM rdb_ipidentification WHERE replacementid=:newid AND (sourceip=:sourceip and destinationip=:destinationip);';
                           Query.ParamByName ('newid').AsInteger := NewID;
                           Query.ParamByName ('sourceip').AsString := help.Source.IPAddress.ValueAsString;
                           Query.ParamByName ('destinationip').AsString := help.Destination.IPAddress.ValueAsString;

                           Query.Open;

                           IDIsUnique := (Query.RecordCount = 0) and (OldIPID <> NewID);

                     until IDIsUnique;

                     AddReplacementPair (OldSocketPair, OldIPID, NewID);
                end
             else
                 NewID := OldIPID;
        end
     else
         NewID := Query.FieldByName('replacementid').AsInteger;

     Query.Free;

     result := NewID;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForMPLSLabel (const OldLabel: longword) : longword;

var value : longword;

begin
     if OriginalAlreadyExists (OldLabel, ioMPLSLabel) then
        value := getReplacementForInteger (OldLabel, ioMPLSLabel)
     else
         begin
              repeat
                    value := random (priv_anonparams.MPLSMaximumLabelValue - 16) + 16;          // 0-15 are reserved; 20 bit value

              until not ReplacementAlreadyExists (value, ioMPLSLabel);

              AddReplacementPair (OldLabel, value, ioMPLSLabel);

              // Check that we don't run out of possible values if the maximum range is set really low

              inc (priv_MPLSLabelReplacementCount);

              if priv_MPLSLabelReplacementCount >= (priv_anonparams.MPLSMaximumLabelValue - 16) then
                 priv_anonparams.MPLSMaximumLabelValue := priv_anonparams.MPLSMaximumLabelValue * 2;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForFQDNFromDB (const OldFQDN: tFQDN): tFQDNPair;

var Query : tFDQuery;
    help1 : tFQDN;
    help2 : tFQDN;
    sHelp : string;
    i     : integer;

begin
     // Prepare statement that looks for any part of the fqdn

     sHelp := '';

     for i := 1 to OldFQDN.ElementCount do
         begin
              if i = 1 then
                 sHelp := '"' + OldFQDN.getPartialFQDN(1) + '"'
              else
                  sHelp := sHelp + ' OR originalfqdn="' + OldFQDN.getPartialFQDN(i) + '"';
         end;

     if sHelp <> '' then
        begin
             Query := tFDQuery.Create(nil);
             Query.Connection := priv_database;
             Query.SQL.Text := 'SELECT * FROM rdb_nameresolution WHERE originalfqdn=' + sHelp + ' ORDER BY LENGTH(originalfqdn) DESC;';
             Query.Open;

             if Query.RecordCount = 0 then
                begin
                     help1.setFQDN('');
                     help2.setFQDN('');
                end
             else
                 begin
                      help1.setFQDN(Query.FieldByName('originalfqdn').AsString);
                      help2.setFQDN (Query.FieldByName('replacementfqdn').AsString);
                 end;

             Query.Free;

             result.Original := help1;
             result.Replacement := help2;
        end
     else
         result.initialize;
end;

(* ------------------------------------------- *)

function tReplacementManager.getReplacementForFQDN (const OldFQDN: tFQDN): tFQDN;

var helpfqdn : tFQDN;
    helppair : tFQDNPair;
    help2    : string;
    count    : integer;
    len      : integer;
    i        : integer;

begin
     if OldFQDN.FullFQDN <> '' then
        begin
             helppair := getReplacementForFQDNFromDB (OldFQDN);

             if helppair.Replacement.FullFQDN = '' then
                begin
                     // Create a new replacement

                     case priv_anonparams.DHCPv4ReplaceHostnameMode of
                      rhnSetToZeroString : begin
                                                // Simple clear the name

                                                helpfqdn.initialize;
                                           end;
                      rhnUsePlaceholders : begin
                                                // Simple x-ing the name parts

                                                help2 := '';

                                                for i := 1 to OldFQDN.ElementCount do
                                                    begin
                                                         count := length (OldFQDN.Elements[i-1]);

                                                         if i = 1 then
                                                            help2 := StringOfChar ('x', count)
                                                         else
                                                             help2 := help2 + '.' + StringOfChar ('x', count);
                                                    end;

                                                helpfqdn.setFQDN (help2);
                                           end;
                     end;

                     // Store new fqdn

                     AddReplacementPair (OldFQDN, helpfqdn);
                end
             else
                 begin
                      // check database result - it could be a perfect match, or just a partial

                      if helppair.Original.FullFQDN = OldFQDN.FullFQDN then
                         helpfqdn := helppair.Replacement
                      else
                          begin
                               // construct remaining part

                               len := length (OldFQDN.FullFQDN) - length (helppair.Original.FullFQDN);
                               help2 := copy (OldFQDN.FullFQDN, 1, len);

                               if pos ('.', help2) > 0 then
                                  help2 := StringOfChar ('x', len - 1)
                               else
                                   help2 := StringOfChar ('x', len);

                               helpfqdn.setFQDN (help2  + '.' +  helppair.Replacement.FullFQDN);
                          end;
                 end;
        end
     else
         helpfqdn.initialize;

     result := helpfqdn;
end;

(* ------------------------------------------- *)

function tReplacementManager.ReplacementAlreadyExists (const Value: int64; const Origin: tIntegerOriginType): boolean;

var Query : tFDQuery;

begin
     Query := tFDQuery.Create(nil);
     Query.Connection := priv_database;

     if Origin = ioAny then
        Query.SQL.Text := 'SELECT replacement FROM rdb_genericintegers WHERE replacement=:replacement;'
     else
         begin
              Query.SQL.Text := 'SELECT replacement FROM rdb_genericintegers WHERE origin=:origin AND replacement=:replacement;';
              Query.ParamByName ('origin').AsInteger := getIntegerOriginIndexFromType (Origin);
         end;

     Query.ParamByName('replacement').AsLargeInt := Value;
     Query.Open;

     result := (Query.RecordCount > 0);

     Query.Free;
end;

(* ------------------------------------------- *)

end.
