{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_tcpconversation_manager;

interface

uses System.Generics.Collections,
     base_structures,
     definition_ip_addresses,
     definition_5tuple,
     definition_tcp_structures;

type tOnNewISN = procedure (const FlowIndex : integer; OldISN, NewISN : longword) of object;
     tOnUpdateTCPState = procedure (const State : tTCPBasicConversationState) of object;

     tTCPFlow = class (tObject)

                      constructor create (const FlowIndex : integer; const ISN : longword; const VLANID : tVLANID; const InterfaceID : longword; const Tuple : t5Tuple);

                      private
                             priv_FlowIndex           : integer;                           // index within the tuple/anchor range, usually there's only 1
                             priv_ISN                 : tTCPSequence;
                             priv_SocketPair          : tIPSocketPair;
                             priv_ClientIP            : tIPAddress;
                             priv_FirstFrameTimeStamp : tFrameTimestamp;
                             priv_LastFrameTimeStamp  : tFrameTimestamp;
                             priv_basicstate          : tTCPBasicConversationState;
                             priv_VLANID              : tVLANID;                           // used for checks if seeing the same flow on different VLANs
                             priv_InterfaceID         : longword;                          // same, but for flows on different interfaces
                             priv_GlobalFlowIndex     : integer;                           // index assigned when creating the flow, counting up per flow (regardless of tuple/anchor, think Wireshark streamindex)

                             EventOnNewISN            : tOnNewISN;
                             EventOnUpdateTCPState    : tOnUpdateTCPState;

                             function getDictionaryKey : string;
                             procedure setNewISN (const NewISN : tTCPSequence);
                             procedure SetNewBasicState (const State : tTCPBasicConversationState);

                      public
                            property FlowIndex : integer read priv_FlowIndex;
                            property SocketPair : tIPSocketPair read priv_SocketPair;
                            property ClientIP : tIPAddress read priv_ClientIP;
                            property ISN : tTCPSequence read priv_ISN write setNewISN;
                            property FirstFrameTime : tFrameTimestamp read priv_FirstFrameTimeStamp;
                            property LastFrameTime : tFrameTimestamp read priv_LastFrameTimeStamp;
                            property BasicState : tTCPBasicConversationState read priv_basicstate write SetNewBasicState;
                            property VLANID : tVLANID read priv_VLANID;
                            property InterfaceID : longword read priv_InterfaceID;
                            property GlobalFlowIndex : integer read priv_GlobalFlowIndex write priv_GlobalFlowIndex;
                            property DictionaryKey : string read getDictionaryKey;

                            /// <summary>
                            ///   Checks if a sequence is realistic for the
                            ///   state, meaning that it is within the range of
                            ///   expected/seen sequence numbers
                            /// </summary>
                            /// <param name="Sequence">
                            ///   Sequence to be checked
                            /// </param>
                            /// <returns>
                            ///   true if the sequence is a realistic match
                            /// </returns>
                            function DetermineSequenceMatch (const Sequence : tTCPSequence) : boolean;

                            destructor Destroy; override;
                end;

     tTCPFlowList = class (tObject)

                          constructor create (const BaseTuple : tTCPBaseTuple; const ISN : longword);

                          private
                                 priv_BaseTuple           : tTCPBaseTuple;
                                 priv_FlowList            : tObjectList<tTCPFlow>;
                                 priv_FlowByISN           : tObjectDictionary<longword, tTCPFlow>;

                                 function getConversationCount : integer;

                                 procedure OnEventNewISNSet (const FlowIndex : integer; OldISN, NewISN : longword);
                                 procedure OnEventUpdateTCPState (const State : tTCPBasicConversationState);

                          public
                                property BaseTuple : tTCPBaseTuple read priv_BaseTuple;
                                property count : integer read getConversationCount;

                                function First : tTCPFlow;
                                function Last : tTCPFlow;
                                function getEntry (const index : integer) : tTCPFlow;
                                procedure AddFlow (Flow : tTCPFlow);

                                destructor Destroy; override;
                    end;


     tTCPConversationManager = class (tObject)

                                     constructor create;

                                     private
                                            priv_Conversations             : tObjectlist<tTCPFLowList>;                               // anchor/tuple based, can have multiple flows
                                            priv_dictAnchoredConversations : tObjectDictionary<string, tTCPFlowList>;                 // key is a tTCPBaseTuple (Anchor + 5Tuple) as string
                                            priv_dictTupleConversations    : tObjectDictionary<string, tTCPFlowList>;                 // key is just the 5Tuple as string
                                            priv_FlowEntryList             : tObjectList<tTCPFlow>;                                   // list of flows, like Wireshark streams

                                            priv_TotalFlowCount            : integer;                                                 // incremented per flow

                                            function GetAnchorBasedFlowList (const BaseTuple : tTCPBaseTuple) : tTCPFlowList;
                                            function Get5TupleBasedFlowList (const Tuple : t5Tuple) : tTCPFlowList;

                                     public
                                           property TotalFlowCount : integer read priv_TotalFlowCount;

                                           /// <summary>
                                           ///   Adds a new flow to the
                                           ///   conversation manager. This can
                                           ///   be another flow for an
                                           ///   existing tuple/anchor
                                           ///   combination or a completely
                                           ///   new set. The routine will
                                           ///   check for tuple/anchor and add
                                           ///   the new flow to the list if
                                           ///   the combination already
                                           ///   exists, or create a new list
                                           ///   and add the new flow as first
                                           ///   entry.
                                           /// </summary>
                                           /// <param name="Anchor">
                                           ///   L2 anchor
                                           /// </param>
                                           /// <param name="Tuple">
                                           ///   5-Tuple
                                           /// </param>
                                           /// <param name="ISN">
                                           ///   Initial Sequence Number
                                           /// </param>
                                           /// <returns>
                                           ///   The new flow object
                                           /// </returns>
                                           function AddFlow (const Anchor : tConversationAnchor; const Tuple : t5Tuple; const ISN : longword) : tTCPFlow;
                                           /// <summary>
                                           ///   Returns the latest flow for
                                           ///   the anchor/tuple combination,
                                           ///   usually to add more packet
                                           ///   details to it. For that,
                                           ///   sequence and acknowledge are
                                           ///   checked to be plausible for
                                           ///   the new packet to be part of
                                           ///   the flow, and if not, an new
                                           ///   flow is created and returned
                                           ///   (e.g. when the new packet is
                                           ///   heavily out of window)
                                           /// </summary>
                                           /// <param name="Anchor">
                                           ///   L2 anchor
                                           /// </param>
                                           /// <param name="Tuple">
                                           ///   5-tuple
                                           /// </param>
                                           /// <param name="Sequence">
                                           ///   TCP Sequence number
                                           /// </param>
                                           /// <param name="Acknowledge">
                                           ///   TCP Acknowledgement number
                                           /// </param>
                                           /// <returns>
                                           ///   the last flow in the list for
                                           ///   the anchor/tuple
                                           /// </returns>
                                           function GetLatestFlow (const Anchor : tConversationAnchor; const Tuple : t5Tuple; const Sequence, Acknowledge : tTCPSequence) : tTCPFlow;
                                           /// <summary>
                                           ///   Returns the number of flows in
                                           ///   the flow manager for the given
                                           ///   anchor and tuple
                                           /// </summary>
                                           function GetFlowCount (const Anchor : tConversationAnchor; const Tuple : t5Tuple) : integer;
                                           /// <summary>
                                           ///   Looks for a flow where the
                                           ///   given sequence is the best
                                           ///   match. This can be used to
                                           ///   find an older flow when a late
                                           ///   packet arrives.
                                           /// </summary>
                                           /// <param name="Anchor">
                                           ///   Base Anchor
                                           /// </param>
                                           /// <param name="Tuple">
                                           ///   5-Tuple of the FLow
                                           /// </param>
                                           /// <param name="Sequence">
                                           ///   Sequence of the current packet
                                           /// </param>
                                           function GetFlowForSequence (const Anchor : tConversationAnchor; const Tuple : t5Tuple; const Sequence : tTCPSequence) : tTCPFlow;
                                           /// <summary>
                                           ///   Returns a flow based on it's
                                           ///   absolute number (meaning, if
                                           ///   two flows have the same
                                           ///   anchor/tuple, they have
                                           ///   different indexes here)
                                           /// </summary>
                                           /// <param name="Index">
                                           ///   Absolute index
                                           /// </param>
                                           function GetFlowByAbsoluteIndex (const Index : integer) : tTCPFlow;

                                           destructor Destroy; override;

                               end;

implementation

uses SysUtils;

(* ------------------------------------------- *)
{ tTCPFlow }
(* ------------------------------------------- *)

constructor tTCPFlow.create (const FlowIndex : integer; const ISN : longword; const VLANID : tVLANID; const InterfaceID : longword; const Tuple : t5Tuple);

begin
     inherited create;

     priv_FlowIndex := FlowIndex;
     priv_ISN.initialize (ISN);
     priv_VLANID := VLANID;
     priv_InterfaceID := InterfaceID;

     if Tuple.IPVersion = 4 then
        priv_SocketPair.initialize (Tuple.IPv4SocketPair)
     else
         priv_SocketPair.initialize (Tuple.IPv6SocketPair);

     priv_GlobalFlowIndex := 0;
end;

(* ------------------------------------------- *)

destructor tTCPFlow.Destroy;

begin
     inherited;
end;

(* ------------------------------------------- *)

procedure tTCPFlow.setNewISN (const NewISN: tTCPSequence);

begin
     if assigned (EventOnNewISN) then
        EventOnNewISN (priv_FlowIndex, priv_isn.Value, NewISN.Value);

     priv_isn := NewISN;
end;

(* ------------------------------------------- *)

procedure tTCPFlow.SetNewBasicState (const State: tTCPBasicConversationState);

begin
     if State.InitialSequenceNo <> priv_ISN then
        SetNewISN (State.InitialSequenceNo);

     priv_FirstFrameTimeStamp := State.FirstFrameTimestamp;
     priv_LastFrameTimeStamp := State.LastFrameTimestamp;

     priv_ClientIP := State.ClientIPAddress;

     priv_basicstate := State;
end;

(* ------------------------------------------- *)

function tTCPFlow.getDictionaryKey: string;

begin
     result := priv_isn.DictionaryKey;
end;

(* ------------------------------------------- *)

function tTCPFlow.DetermineSequenceMatch (const Sequence: tTCPSequence): boolean;

var upperSeq : tTCPSequence;
    lowerSeq : tTCPSequence;
    value    : boolean;

begin
     value := false;

     // Check client sequence

     lowerSeq := priv_basicstate.InitialSequenceNo;
     upperSeq := priv_basicstate.LastSeenClientSeq + priv_basicstate.ServerWindow.CurrentSize;

     value := (Sequence >= lowerSeq) and (Sequence <= upperSeq);

     if not value then
        begin
             // Check server sequence

             lowerSeq := priv_basicstate.SYNACKSequenceNo;
             upperSeq := priv_basicstate.LastSeenServerSeq + priv_basicstate.ClientWindow.CurrentSize;


             value := (Sequence >= lowerSeq) and (Sequence <= upperSeq);
        end;

     result := value;
end;

(* ------------------------------------------- *)
{ tTCPFlowList }
(* ------------------------------------------- *)

constructor tTCPFlowList.create (const BaseTuple: tTCPBaseTuple; const ISN: longword);

var Flow : tTCPFlow;

begin
     inherited create;

     priv_BaseTuple := BaseTuple;

     priv_FlowList := tObjectList<tTCPFlow>.create (true);
     priv_FlowByISN := tObjectDictionary<longword, tTCPFlow>.create ([]);

     // Since we just created a new tuple, add the first conversation to it right away

     Flow := tTCPFlow.create (1, ISN, BaseTuple.Anchor.VLANId, BaseTuple.Anchor.InterfaceID, BaseTuple.Tuple);
     Flow.EventOnNewISN := OnEventNewISNSet;
     Flow.EventOnUpdateTCPState := OnEventUpdateTCPState;

     priv_FlowList.Add (Flow);
     priv_FlowByISN.Add (ISN, Flow);
end;

(* ------------------------------------------- *)

destructor tTCPFlowList.Destroy;

begin
     priv_FlowByISN.Free;
     priv_FlowList.Free;

     inherited;
end;

(* ------------------------------------------- *)

procedure tTCPFlowList.OnEventNewISNSet (const FlowIndex : integer; OldISN, NewISN : longword);

var Flow : tTCPFlow;

begin
     // A conversation got a new ISN, so we need to update the dictionary

     Flow := priv_FlowByISN.Items[OldISN];
     priv_FlowByISN.Remove (OldISN);
     priv_FlowByISN.Add (NewISN, Flow);
end;

(* ------------------------------------------- *)

procedure tTCPFlowList.OnEventUpdateTCPState (const State: tTCPBasicConversationState);

begin

end;

(* ------------------------------------------- *)

procedure tTCPFlowList.AddFlow (Flow: tTCPFlow);

begin
//   assert (not priv_FlowByISN.ContainsKey(Flow.priv_ISN.Value), 'Cannot add new flow: same ISN ' + Flow.priv_ISN.Value.ToString + ' as a previous flow (tTCPFlowList.AddFlow)');
{
     if priv_FlowByISN.ContainsKey(Flow.priv_ISN.Value) then
        sleep (1);
}
     priv_FlowList.Add (Flow);
     priv_FlowByISN.Add (Flow.priv_ISN.Value, Flow);
end;


(* ------------------------------------------- *)

function tTCPFlowList.getConversationCount: integer;

begin
     result := priv_FlowByISN.Count;
end;

(* ------------------------------------------- *)

function tTCPFlowList.First: tTCPFlow;

begin
     result := priv_FlowList.First;
end;

(* ------------------------------------------- *)

function tTCPFlowList.Last: tTCPFlow;

begin
     result := priv_FlowList.Last;
end;

(* ------------------------------------------- *)

function tTCPFlowList.getEntry (const index: integer): tTCPFlow;

begin
     assert ((index > 0) and (index <= priv_flowlist.Count));

     result := priv_flowlist[index-1];
end;

(* ------------------------------------------- *)
{ tTCPConversationManager }
(* ------------------------------------------- *)

constructor tTCPConversationManager.create;

begin
     inherited;

     priv_Conversations := tObjectlist<tTCPFLowList>.create (true);
     priv_dictAnchoredConversations := tObjectDictionary<string, tTCPFlowList>.create ([]);
     priv_dictTupleConversations := tObjectDictionary<string, tTCPFlowList>.create ([]);
     priv_FlowEntryList := tObjectList<tTCPFlow>.create (false);

     priv_TotalFlowCount := 0;
end;

(* ------------------------------------------- *)

destructor tTCPConversationManager.Destroy;

begin
     priv_dictAnchoredConversations.Free;
     priv_dictTupleConversations.Free;
     priv_Conversations.Free;
     priv_FlowEntryList.Free;

     inherited;
end;

(* ------------------------------------------- *)

function tTCPConversationManager.GetAnchorBasedFlowList (const BaseTuple : tTCPBaseTuple) : tTCPFlowList;

var key : string;

begin
     key := BaseTuple.DictionaryKey;

     // Check if we already have a conversation for this tuple

     if priv_dictAnchoredConversations.ContainsKey (key) then
        result := priv_dictAnchoredConversations.Items[key]
     else
         result := nil;
end;

(* ------------------------------------------- *)

function tTCPConversationManager.Get5TupleBasedFlowList (const Tuple: t5Tuple): tTCPFlowList;

var key : string;

begin
     key := Tuple.DictionaryKey;

     // Check if we already have a conversation for this tuple

     if priv_dictTupleConversations.ContainsKey (key) then
        result := priv_dictTupleConversations.Items[key]
     else
         result := nil;
end;

(* ------------------------------------------- *)

function tTCPConversationManager.AddFlow (const Anchor : tConversationAnchor; const Tuple: t5Tuple; const ISN: longword) : tTCPFlow;

var List  : tTCPFlowList;
    List2 : tTCPFlowList;
    Flow  : tTCPFlow;
    Base  : tTCPBaseTuple;
    count : integer;

begin
     inc (priv_TotalFlowCount);

     Base.initialize (Anchor, Tuple);

     List := GetAnchorBasedFlowList (Base);

     // If we have an existing entry we work with that, otherwise add a new one

     if assigned (List) then
        begin
             // existing tuple found, so we add a new conversation to it

             count := list.count;

             Flow := tTCPFlow.create (count + 1, ISN, Anchor.VLANId, Anchor.InterfaceID, Tuple);

             // set global flow index

             Flow.GlobalFlowIndex := priv_TotalFlowCount;

             // assign callback for ISN changes

             Flow.EventOnNewISN := list.OnEventNewISNSet;
             Flow.EventOnUpdateTCPState := list.OnEventUpdateTCPState;

             List.AddFlow (Flow);
        end
     else
         begin
              // New tuple, so determine the base and add a new entry, plus assign the new ISN callback function, all at once. Yup.

              List := tTCPFlowList.create (Base, ISN);

              priv_Conversations.Add (List);
              priv_dictAnchoredConversations.Add (Base.DictionaryKey, List);

              Flow := List.Last;
              Flow.GlobalFlowIndex := priv_TotalFlowCount;

              // also add to the tuple dictionary, if the tuple doesn't exist yet,
              // otherwise add the new flow to the existing list

              if not priv_dicttupleconversations.ContainsKey (Tuple.DictionaryKey) then
                 priv_dictTupleConversations.Add (Tuple.DictionaryKey, List);
         end;

     // Add the new flow to the "single flow entry list"

     priv_FlowEntryList.Add(Flow);

     result := Flow;
end;

(* ------------------------------------------- *)

function tTCPConversationManager.GetLatestFlow (const Anchor: tConversationAnchor; const Tuple: t5Tuple; const Sequence, Acknowledge : tTCPSequence) : tTCPFlow;

var list   : tTCPFlowList;
    Base   : tTCPBaseTuple;
    flow   : tTCPFlow;

begin
     result := nil;

     Base.initialize (Anchor, Tuple);
     list := GetAnchorBasedFlowList (Base);

     if assigned (list) then
        result := list.Last
     else
         begin
              // Check if we have a flow with a different anchor that matches, e.g.
              // when there's a FDX capture or asymetric routing. We do that by just
              // looking for the 5-Tuple with no anchor, and check the latest flow

              list := Get5TupleBasedFlowList (Tuple);

              if assigned (list) and (list.count > 0) then
                 begin
                      flow := list.Last;

                      // Check if it's on the same VLAN, in which case we check if the sequences are close enough to indicate that
                      // the packets belong to an existing state with a different ethernet pair

                      if (flow.VLANID = Anchor.VLANID) and (flow.InterfaceID = Anchor.InterfaceID) then
                         begin
                              // Check if the ack matches next expected ack

                              if (Acknowledge = flow.BasicState.NextExpectedClientSeq) or (Acknowledge = flow.BasicState.NextExpectedServerSeq) then
                                 begin
                                      // Add entry to anchor based dictionary to find it fast next time, and avoid confusion if the seq/ack doesn't match exactly

                                      priv_dictAnchoredConversations.Add (Base.DictionaryKey, list);

                                      result := list.Last;
                                 end;
                         end
                      else
                          begin
                               // check if the existing state is onesided. Otherwise we have to assume that we have the same conversation on different VLANs,
                               // which means we can't merge

                               if (flow.BasicState.ServerPackets = 0) then
                                  begin
                                       // Check if this packet is different from than what we already have, otherwise it's probably just a duplicate
                                       // we do that by checking if the sequence is in-window - if it's not, it must be the other part of the conversation

                                       if not SequenceInWindow (Sequence, flow.BasicState.ClientWindow) then
                                          begin
                                               // Add entry to anchor based dictionary to find it fast next time, and avoid confusion if the seq/ack doesn't match exactly

                                               priv_dictAnchoredConversations.Add (Base.DictionaryKey, list);

                                               result := list.Last;
                                          end;
                                  end
                               else
                                   if (flow.BasicState.ClientPackets = 0) then
                                      begin
                                           // Check if this packet is different from than what we already have, otherwise it's probably just a duplicate
                                           // we do that by checking if the sequence is in-window - if it's not, it must be the other part of the conversation

                                           if not SequenceInWindow (Sequence, flow.BasicState.ServerWindow) then
                                              begin
                                                   // Add entry to anchor based dictionary to find it fast next time, and avoid confusion if the seq/ack doesn't match exactly

                                                   priv_dictAnchoredConversations.Add (Base.DictionaryKey, list);

                                                   result := list.Last;
                                              end;
                                      end;
                          end;
                 end;
         end;
end;

(* ------------------------------------------- *)

function tTCPConversationManager.GetFlowCount (const Anchor: tConversationAnchor; const Tuple: t5Tuple) : integer;

var List  : tTCPFlowList;
    Base : tTCPBaseTuple;

begin
     Base.initialize (Anchor, Tuple);
     List := GetAnchorBasedFlowList (Base);

     result := list.count;
end;

(* ------------------------------------------- *)

function tTCPConversationManager.GetFlowForSequence (const Anchor: tConversationAnchor; const Tuple: t5Tuple; const Sequence: tTCPSequence) : tTCPFlow;

var list  : tTCPFlowList;
    Base  : tTCPBaseTuple;
    entry : tTCPFlow;
    count : integer;
    match : boolean;

begin
     Base.initialize (Anchor, Tuple);
     list := GetAnchorBasedFlowList (Base);

     // There must be a list, so verify it

     assert (assigned (list));

     count := list.count;
     entry := list.getEntry (count);

     match := entry.DetermineSequenceMatch (Sequence);

     // Find the entry where the sequence is a realistic match

     while not match and (count > 1) do
           begin
                dec (count);
                entry := list.getEntry (count);

                match := entry.DetermineSequenceMatch (Sequence);
           end;

     if match then
        result := entry
     else
         result := nil;
end;

(* ------------------------------------------- *)

function tTCPConversationManager.GetFlowByAbsoluteIndex (const Index: integer): tTCPFlow;

begin
     if (Index > 0) and (Index <= priv_TotalFlowCount) then
        result := priv_FlowEntryList[Index-1]
     else
         result := nil;
end;

(* ------------------------------------------- *)

end.
