{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_processchain;

interface

uses Contnrs,
     class_rawframe,
     class_tracewriter,
     class_frameprocessor_baseclass,
     class_frameprocessor_stub;

type tProcessChain = class (tObject)

                           constructor create;
                           procedure free;

                           private
                                  priv_taskprocessorlist : tObjectlist;

                                  priv_firstprocessor    : tFrameProcessor;
                                  priv_lastprocessor     : tFrameProcessor;

                                  priv_doNotAddStub      : boolean;
                                  priv_framediscarded    : boolean;

                                  procedure onFrameDiscard (Sender : tObject; Frame : tRawFrame);
                           public
                                 property DoNotAddStub : boolean read priv_doNotAddStub write priv_doNotAddStub;
                                 property FrameWasDiscarded : boolean read priv_framediscarded;

                                 {$REGION 'Documentation'}
                                 /// <summary>
                                 ///   <para>
                                 ///     Adds another processor to the chain.
                                 ///     New processors are added back to front,
                                 ///      so the first processor added will
                                 ///     output the final frame, either to a
                                 ///     tracewriter object or as output
                                 ///     parameter
                                 ///   </para>
                                 ///   <para>
                                 ///     The first processor added needs to
                                 ///     come with an existing tracewriter
                                 ///     object in case the chain should write
                                 ///     results to file directly.
                                 ///   </para>
                                 /// </summary>
                                 /// <param name="Processor">
                                 ///   FrameProcessor to be added to the chain
                                 /// </param>
                                 {$ENDREGION}
                                 procedure AddProcessor (Processor : tFrameProcessor);

                                 {$REGION 'Documentation'}
                                 /// <summary>
                                 ///   Processes a given frame and writes the
                                 ///   result directly to a tracewriter
                                 ///   component that must have been assigned
                                 ///   to the last frameprocessor in the chain
                                 ///   before creating the chain. Unless the
                                 ///   frame is discarded in a processor of
                                 ///   course, e.g. in a filtering processor.
                                 /// </summary>
                                 /// <param name="Frame">
                                 ///   Frame to be processed. This frame will
                                 ///   be freed after processing is complete,
                                 ///   so if the original frame is still needed
                                 ///   after processing a copy needs to be used
                                 ///   as parameter<br />
                                 /// </param>
                                 {$ENDREGION}
                                 procedure ProcessFrame (Frame : tRawFrame); overload;
                                 {$REGION 'Documentation'}
                                 /// <summary>
                                 ///   Processes a given frame and returns the
                                 ///   result as a new tRawFrame Object.
                                 /// </summary>
                                 /// <param name="Frame">
                                 ///   Frame to be processed. This frame will
                                 ///   be freed after processing is complete,
                                 ///   so if the original frame is still needed
                                 ///   after processing a copy needs to be used
                                 ///   as parameter
                                 /// </param>
                                 /// <param name="ProcessedFrame">
                                 ///   Result of the process chain
                                 /// </param>
                                 {$ENDREGION}
                                 procedure ProcessFrame (Frame : tRawFrame; out ProcessedFrame : tRawFrame); overload;

                     end;

implementation

(* ------------------------------------------- *)
{ tProcessChain }
(* ------------------------------------------- *)

constructor tProcessChain.create;

begin
     inherited create;

     priv_taskprocessorlist := tObjectlist.Create (false);
     priv_firstprocessor := nil;
     priv_lastprocessor := nil;

     priv_doNotAddStub := false;
end;

(* ------------------------------------------- *)

procedure tProcessChain.free;

var i : integer;

begin
     // Discard process chain

     for i := 1 to priv_taskprocessorlist.Count do
         begin
              (priv_taskprocessorlist[i-1] as tFrameProcessor).free;
              priv_taskprocessorlist[i-1] := nil;
         end;

     priv_taskprocessorlist.free;

     inherited free;
end;

(* ------------------------------------------- *)

procedure tProcessChain.onFrameDiscard (Sender: tObject; Frame: tRawFrame);

begin
     priv_framediscarded := true;
end;

(* ------------------------------------------- *)

procedure tProcessChain.AddProcessor (Processor: tFrameProcessor);

begin
     // If this is the first processor in the chain, add it

     if priv_firstprocessor = nil then
        begin
             priv_taskprocessorlist.Add (Processor);
             priv_firstprocessor := Processor;

             // If there is no writer assigned to the processor, add a frame output stub
             // instead which will give us the final frame back

             if (Processor.TraceWriter = nil) and not priv_doNotAddStub then
                begin
                     priv_lastprocessor := tFrameProcessorStub.create ('FrameOutputStub');
                     priv_firstprocessor.NextProcessor := priv_lastprocessor;

                     priv_taskprocessorlist.Add (priv_lastprocessor);
                end
             else
                 priv_lastprocessor := Processor;
        end
     else
         begin
              // Otherwise put it at the front and chain the rest behind it

              Processor.NextProcessor := priv_firstprocessor;
              priv_firstprocessor := Processor;

              priv_taskprocessorlist.Insert (0, Processor);
         end;
end;

(* ------------------------------------------- *)

procedure tProcessChain.ProcessFrame (Frame : tRawFrame);

begin
     priv_framediscarded := false;

     priv_firstprocessor.InitParseMaster;
     priv_firstprocessor.addFrame (Frame);

{$ifdef debug}
     assert (not priv_lastprocessor.ParseMasterExists);
{$endif}
end;

(* ------------------------------------------- *)

procedure tProcessChain.ProcessFrame (Frame : tRawFrame; out ProcessedFrame : tRawFrame);

begin
     priv_framediscarded := false;
     priv_firstprocessor.addFrame (Frame);

{$ifdef debug}
     assert (not priv_lastprocessor.ParseMasterExists);
{$endif}

     ProcessedFrame := (priv_LastProcessor as tFrameProcessorStub).OutputFrame;
end;

(* ------------------------------------------- *)

end.
