{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_taskprocessor_edit;

interface

uses Contnrs,
     System.Generics.Collections,
     base_structures,
     definition_ip_addresses,
     class_traceloader,
     class_tracewriter,
     class_writer_pcapng,
     class_tracefilelist,
     class_loader_pcapng,
     class_interfacelist,
     class_taskprocessor_baseclass,
     class_frameprocessor_baseclass;

type tEditParameters = record
                             TracefileListIndex               : integer;
                             ReorderFrames                    : boolean;

                             ReplaceCookedHdrWithEthernet     : boolean;
                             ReplaceEtherType                 : boolean;
                             OriginalEtherType                : integer;
                             ReplacementEtherType             : integer;
                             ReplacementSLLMACAddress         : tMACAddress;
                             ForceSLLMACReplacement           : boolean;

                             RemoveJuniperHeader              : boolean;
                             AddPseudoL2HeaderForJuniper      : tPseudoL2HeaderMode;
                             RemoveVLANs                      : boolean;
                             RemoveGREHeader                  : boolean;
                             RemoveGTPUHeader                 : boolean;
                             RemoveMPLSShims                  : boolean;
                             RemoveCAPWAP                     : boolean;
                             FixFrameSizeMetaData             : boolean;
                             FixFrameSizeMetaNonHardSliced    : boolean;
                             FixEtherTypes                    : boolean;
                             AddSHBComment                    : boolean;

                             ChangeIDBInterfaceName           : boolean;
                             IDBInterfaceNameValue            : string;
                       end;

     tTaskProcessorEdit = class (tTaskProcessor)

                                constructor create (InputFile : tTracefileEntry; Output : tOutputlocation; const Parameters: tEditParameters);
                                procedure free;

                                private
                                       priv_params                    : tEditParameters;
                                       priv_outputfile                : string;
                                       priv_comment                   : string;

                                       priv_traceloader               : tTraceLoader;
                                       priv_tracewriter               : tPCAPngWriter;

                                       priv_processchain              : tObjectList<tFrameProcessor>;
                                       priv_firstprocessor            : tFrameProcessor;

                                       procedure OnBeforeWriteInterfaceBlock (Sender : tObject; CaptureInterface : tCaptureInterface);
                                       procedure OnCopySHBString (StringType : tSHBStringType; var Value : string);

                                       procedure PrepareProcessChain;
                                       procedure InitTraceloaderAndWriter;

                                protected

                                public
                                      function Execute : tTaskErrorCode; override;
                          end;


implementation

uses SiAuto,
     SysUtils,
     toolkit,
     class_rawframe,
     class_loader_enc,
     class_loader_naicap,
     class_loader_pcap,
     class_frameprocessor_replaceSLL,
     class_frameprocessor_juniper,
     class_frameprocessor_vlan,
     class_frameprocessor_gre,
     class_frameprocessor_gtpu,
     class_frameprocessor_mpls,
     class_frameprocessor_capwap,
     class_frameprocessor_metaupdate,
     class_frameprocessor_fixethertypes;

(* ------------------------------------------- *)
{ tTaskProcessorSeparate }
(* ------------------------------------------- *)

constructor tTaskProcessorEdit.create (InputFile : tTracefileEntry; Output : tOutputlocation; const Parameters: tEditParameters);

begin
     inherited create (InputFile, Output);

     priv_params := Parameters;

     priv_processchain := TObjectList<tFrameProcessor>.Create (true);
     priv_outputfile := Output.getOutputfilename;

     priv_comment := '';
end;

(* ------------------------------------------- *)

procedure tTaskProcessorEdit.free;

begin
     inherited;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorEdit.PrepareProcessChain;

var FrameProcessor : tFrameProcessor;
    NextProcessor  : tFrameProcessor;
    i              : Integer;

begin
     // Determine Writer for the end of the chain

     if priv_params.ReorderFrames then
        begin
             FrameProcessor := tFrameProcessor.create ('Writer', priv_tracewriter, 5000);
             FrameProcessor.FrameStorage.SortByTime := true;
        end
     else
         begin
              FrameProcessor := tFrameProcessor.create ('Writer', priv_tracewriter, 5000);  // 5000 for testing, could be 10000 (aka parameter not needed)
         end;

     priv_processchain.Add (FrameProcessor);
     NextProcessor := FrameProcessor;

     // Chain the optional processors, from back to front, so the
     // higher layer processors come first

     if priv_params.ReplaceCookedHdrWithEthernet then
        begin
             FrameProcessor := tFrameProcessorReplaceSLL.create('ReplaceSLL');

             with (FrameProcessor as tFrameProcessorReplaceSLL) do
                  begin
                       ReplaceEtherType := priv_params.ReplaceEtherType;
                       OriginalEtherType := priv_params.OriginalEtherType;
                       ReplacementEtherType := priv_params.ReplacementEtherType;
                       ReplacementMACAddress := priv_params.ReplacementSLLMACAddress;
                       ForceMACReplacement := priv_params.ForceSLLMACReplacement;
                  end;

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.RemoveMPLSShims then
        begin
             FrameProcessor := tFrameProcessorMPLS.create ('MPLS');

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.RemoveGREHeader then
        begin
             FrameProcessor := tFrameProcessorGRE.create ('GRE');

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.RemoveGTPUHeader then
        begin
             FrameProcessor := tFrameProcessorGTPU.create ('GTPu');

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.RemoveJuniperHeader then
        begin
             FrameProcessor := tFrameProcessorJuniper.create ('Juniper');
             (FrameProcessor as tFrameProcessorJuniper).AddPseudoL2Header := priv_params.AddPseudoL2HeaderForJuniper;

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.RemoveCAPWAP then
        begin
             FrameProcessor := tFrameProcessorCAPWAP.create ('CAPWAP');

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.RemoveVLANs then
        begin
             FrameProcessor := tFrameProcessorVLAN.create ('VLAN');

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.FixEtherTypes then
        begin
             FrameProcessor := tFrameProcessorFixEthertypes.create ('FixEtherTypes');

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     if priv_params.FixFrameSizeMetaData then
        begin
             FrameProcessor := tFrameProcessorMetaUpdate.create ('MetaData');

             with (FrameProcessor as tFrameProcessorMetaUpdate) do
                  begin
                       FixFrameLengthMeta := priv_params.FixFrameSizeMetaData;
                       FixFrameLengthWhenNotHardSliced := priv_params.FixFrameSizeMetaNonHardSliced;
                  end;

             FrameProcessor.NextProcessor := NextProcessor;
             NextProcessor := FrameProcessor;

             priv_processchain.Insert (0, FrameProcessor);
        end;

     // Update indexes

     for i := 1 to priv_processchain.Count do
         priv_processchain[i-1].ChainIndex := i;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorEdit.InitTraceloaderAndWriter;

var Linktype : longword;

begin
     if priv_params.AddSHBComment then
        begin
             // Prepare comment

             priv_comment := '';

             if priv_params.ReplaceCookedHdrWithEthernet then
                priv_comment := priv_comment + ' - Replacing Linux Cooked header with Ethernet header' + CRLF;

             if priv_params.ReorderFrames then
                priv_comment := priv_comment + ' - Reordering frames by timestamp' + CRLF;

             if priv_params.RemoveVLANs then
                priv_comment := priv_comment + ' - Removing all VLAN headers' + CRLF;

             if priv_params.RemoveGREHeader then
                priv_comment := priv_comment + ' - Removing the GRE header, including the preceding IP layer' + CRLF;

             if priv_params.RemoveGTPUHeader then
                priv_comment := priv_comment + ' - Removing the GTP-U header, including the preceding IP layer' + CRLF;

             if priv_params.RemoveJuniperHeader then
                priv_comment := priv_comment + ' - Removing the Juniper header' + CRLF;

             if priv_params.RemoveMPLSShims then
                priv_comment := priv_comment + ' - Removing MPLS shims' + CRLF;

             if priv_params.RemoveCAPWAP then
                priv_comment := priv_comment + ' - Removing CAPWAP header and preceeding layers' + CRLF;

             if priv_params.FixFrameSizeMetaData then
                priv_comment := priv_comment + ' - Fixing Frame size metadata' + CRLF;

             // Finalize comment

             if priv_comment <> '' then
                priv_comment := 'TraceWrangler v' + GetExeVersion (ParamStr(0)) + ' performed the following editing steps:' + CRLF + priv_comment;
        end
     else
         priv_comment := '';

     // Open reader and writer

     priv_tracewriter := tPCAPngWriter.create (priv_outputfile, true);
     priv_tracewriter.Open;

     if priv_params.ReplaceCookedHdrWithEthernet or priv_params.RemoveJuniperHeader then
        Linktype := LINKTYPE_ETHERNET
     else
         Linktype := LINKTYPE_NULL;

     case prot_InputFile.TracefileType of
      cfENC    : begin
                      priv_traceloader := tENCLoader.create (prot_InputFile.FullName, priv_params.TracefileListIndex, true);

                      priv_traceloader.openFile;
                      priv_traceloader.readTraceFileheader;

                      if priv_comment <> '' then
                         priv_tracewriter.HeaderCommentList.Add (priv_comment);

                      if LinkType = LINKTYPE_NULL then
                         LinkType := (priv_traceloader as tENCLoader).LinkType;

                      priv_tracewriter.WriteGenericHeader (LinkType);
                 end;
      cfCAP    : begin
                      priv_traceloader := tCAPLoader.create (prot_InputFile.FullName, priv_params.TracefileListIndex, true);

                      priv_traceloader.openFile;
                      priv_traceloader.readTraceFileheader;

                      if priv_comment <> '' then
                         priv_tracewriter.HeaderCommentList.Add (priv_comment);

                      priv_tracewriter.WriteGenericHeader ((priv_traceloader as tCAPLoader).CaptureInterfaces);
                 end;
      cfPCAP   : begin
                      priv_traceloader := tPCAPLoader.create (prot_InputFile.FullName, priv_params.TracefileListIndex, true);

                      priv_traceloader.openFile;
                      priv_traceloader.readTraceFileheader;

                      if priv_comment <> '' then
                         priv_tracewriter.HeaderCommentList.Add (priv_comment);

                      if LinkType = LINKTYPE_NULL then
                         LinkType := (priv_traceloader as tPCAPLoader).LinkType;

                      priv_tracewriter.WriteGenericHeader (Linktype);
                 end;
      cfPCAPng : begin
                      priv_traceloader := tPCAPNGLoader.create (prot_InputFile.FullName, priv_params.TracefileListIndex, true);

                      // open Loader

                      priv_traceloader.openFile;                        // TODO: handle failure of opening trace files
                      priv_traceloader.readTraceFileheader;

                      // Prepare Writer and transfer header details

                      priv_tracewriter.onBeforeWriteInterfaceBlock := OnBeforeWriteInterfaceBlock;
                      priv_tracewriter.onCopySHBString := OnCopySHBString;      // this will add the comment to the header on callback

                      (priv_tracewriter as tPCAPNGWriter).CopyPCAPngHeaderDetailsFromLoader (priv_traceloader as tPCAPNGLoader);
                 end;
     end;

     priv_traceloader.MaxFramesInMem := 10000;

     // Update file status and report it if possible

     prot_Inputfile.ProcessingStatus := psTaskProcessing;

     if assigned (EventOnStartProcessingOfFile) then
        EventOnStartProcessingOfFile (prot_Inputfile);
end;

(* ------------------------------------------- *)

procedure tTaskProcessorEdit.OnBeforeWriteInterfaceBlock (Sender : tObject; CaptureInterface: tCaptureInterface);

 begin
     if priv_params.ReplaceCookedHdrWithEthernet then
        begin
             if CaptureInterface.LinkType = LINKTYPE_LINUX_SLL then
                CaptureInterface.LinkType := LINKTYPE_ETHERNET;
        end;

     if priv_params.RemoveJuniperHeader then
        begin
             if CaptureInterface.LinkType = LINKTYPE_JUNIPER then
                CaptureInterface.LinkType := LINKTYPE_ETHERNET;
        end;

     if priv_params.ChangeIDBInterfaceName then
        begin
             CaptureInterface.Name := priv_params.IDBInterfaceNameValue;

        end;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorEdit.OnCopySHBString (StringType: tSHBStringType; var Value: string);

begin
     // Add a comment with the editing parameters

     if (StringType = sstComment) and (priv_comment <> '') then
        begin
             if Value = '' then
                Value := priv_comment
             else
                 Value := Value + CRLF + priv_comment;
        end;
end;

(* ------------------------------------------- *)

function tTaskProcessorEdit.Execute : tTaskErrorCode;

var Frame             : tRawFrame;
    fileframesscanned : integer;
    FrameOffset       : uint64;
    boAbort           : boolean;
    i                 : integer;

begin
//   SiMain.EnterMethod ('tTaskProcessorEdit.execute');

     InitTraceloaderAndWriter;

     // Prepare Process Chain

     PrepareProcessChain;

     priv_firstprocessor := priv_processchain[0];

     // Read Frames

     try
        repeat
              SiMain.LogVerbose ('Reading batch of frames from disk');

              try
                 priv_traceloader.readBatchFromDisk;
              except
                    if assigned (onErrorProcessingFrame) then
                       onErrorProcessingFrame (prot_Inputfile, fileframesscanned, ateLoader, boAbort);

                    if boAbort then
                       Exit (ecFailure);
              end;

              for i := 1 to priv_traceloader.FramesAvailable do
                  begin
                       Frame := priv_traceloader.getRawFrame(i);
                       FrameOffset := Frame.FileOffset;

                       // Call first processor in chain for each frame. It will pass the frame through the rest of
                       // the chain and finally write it to disk

                       try
                          priv_firstprocessor.addFrame (Frame);
                       except
                            if assigned (onErrorProcessingFrame) then
                               onErrorProcessingFrame (prot_Inputfile, Frame.FileFrameID, ateFrameProcessing, boAbort);

                            inc (prot_ParseErrorCount);

                            if boAbort then
                               Exit (ecFailure);
                       end;

                       inc (fileframesscanned);

                       // Trigger report if possible

                       ReportProgress (FrameOffset);
                  end;

        until priv_traceloader.FramesAvailable = 0;

        // Discard process chain

        for i := 1 to priv_processchain.Count do
            priv_processchain[i-1].free;

     finally
            // Write remaining frames to disk if any

            if priv_tracewriter.FramesInMem > 0 then
               (priv_tracewriter as tPCAPNGWriter).Flush;

            if prot_InputFile.TracefileType = cfPCAPng then
               begin
                    // Copy statistics

                    (priv_tracewriter as tPCAPNGWriter).CopyPCAPngStatisticsDetailsFromLoader (priv_traceloader as tPCAPNGLoader);
               end;

            // Close the loader/writer

            priv_tracewriter.Close;
            priv_tracewriter.free;

            // Clean up

            priv_traceloader.closeFile;
            priv_traceloader.free;
     end;

//   SiMain.LeaveMethod ('tTaskProcessorEdit.execute');
end;

(* ------------------------------------------- *)

end.

