{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_taskprocessor_merge;

interface

uses Contnrs,
     base_structures,
     definition_filter_structures,
     class_traceloader,
     class_tracewriter,
     class_writer_pcapng,
     class_tracefilelist,
     class_loader_pcapng,
     class_interfacelist,
     class_taskprocessor_baseclass,
     class_frameprocessor_baseclass,
     class_processchain,
     class_interfacemapper;

type tMergeMode = (mmConcatenate, mmMergeByTimestamp);
     tMergeSHBComment = (mscListFiles, mscFileCount, mscNone);

     tMergeParameters = record
                              MergeMode                        : tMergeMode;
                              MergeTargetFilename              : string;
                              InterfaceMapper                  : tInterfaceMapper;

                              FilterFrames                     : boolean;
                              FilterSet                        : tFilterSet;

                              AddSHBComment                    : tMergeSHBComment;
                              AddSourceFilePacketComment       : boolean;

                              AdjustFileTimestamp              : boolean;
                              AdjustTimestampMode              : tFileTimestampMode;
                              ManualTimestamp                  : tDateTime;
                        end;

     tTaskProcessorMerge = class (tObject)                     // Not derived from the usual baseclass because it is not a per file task

                                constructor create (const Parameters: tMergeParameters);
                                procedure free;

                                private
                                       priv_params                    : tMergeParameters;
                                       priv_outputfile                : string;
                                       priv_comment                   : string;

                                       priv_tracewriter               : tPCAPngWriter;
                                       priv_outputinterfaces          : tCaptureInterfaceList;
                                       priv_processchain              : tProcessChain;

                                       priv_firstfiletimestamp        : tDateTime;
                                       priv_lowesttimestamp           : tFrameTimeStamp;
                                       priv_highesttimestamp          : tFrameTimeStamp;

                                       {$REGION 'Documentation'}
                                       /// <summary>
                                       ///   Prepares an aggregated list of
                                       ///   interfaces that the original
                                       ///   interfaces are mapped to. This is
                                       ///   a subset of the target interface
                                       ///   list created in the form, and
                                       ///   created from the MappedToInterface
                                       ///   property of the original
                                       ///   interfaces of each file
                                       /// </summary>
                                       {$ENDREGION}
                                       procedure PrepareOutputInterfaceList;
                                       procedure CreateProcessChain;
                                       procedure PrepareTargetFile;
                                       function OpenReaderForFile (Entry : tTraceFileEntry) : tTraceLoader;
                                       function SetFileTimestamp : integer;

                                       procedure ConcatenateFiles;
                                       procedure MergeFilesByTimestamp;

                                protected

                                public
                                      function Execute : tTaskErrorCode;
                          end;


implementation

uses SiAuto,
     SysUtils,
     toolkit,
     class_rawframe,
     class_loader_enc,
     class_loader_naicap,
     class_loader_pcap,
     class_frameprocessor_filter;

(* ------------------------------------------- *)
{ tTaskProcessorSeparate }
(* ------------------------------------------- *)

constructor tTaskProcessorMerge.create (const Parameters: tMergeParameters);

begin
     inherited create;

     priv_processchain := tProcessChain.create;

     priv_params := Parameters;

     priv_comment := '';

     priv_firstfiletimestamp := 0;
     priv_lowesttimestamp.initialize;
     priv_highesttimestamp.initialize;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorMerge.free;

begin
     inherited;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorMerge.PrepareOutputInterfaceList;

var fileentry : tTracefileEntry;
    ifentry   : tCaptureInterface;
    ifxentry  : tCaptureInterface;
    i,j,k     : integer;
    index     : integer;
    boFound   : boolean;

begin
     index := 0;

     // Create the list for all target interfaces

     priv_outputinterfaces := tCaptureInterfaceList.create (true);

     priv_params.InterfaceMapper.getTargetInterfaceList (priv_outputinterfaces);
end;

(* ------------------------------------------- *)

procedure tTaskProcessorMerge.PrepareTargetFile;

var fileentry : tTraceFileEntry;
    i         : integer;

begin
     priv_tracewriter := tPCAPngWriter.create (priv_params.MergeTargetFilename, true);

     priv_tracewriter.Open;

     if priv_params.AddSHBComment <> mscNone then
        begin
             // Prepare comment, mimicking Wireshark's mergecap except for the mentioning of TraceWrangler

             priv_comment := 'File created by TraceWrangler by merging:' + CRLF;

             if priv_params.AddSHBComment = mscListFiles then
                begin
                     for i := 1 to priv_params.InterfaceMapper.TraceFileList.Count do
                         begin
                              fileentry := priv_params.InterfaceMapper.TraceFileList.getFile(i);

                              priv_comment := priv_comment + 'File' + IntToStr (i) + ': ' + fileentry.FileName + CRLF;
                         end;
                end
             else
                 priv_comment := priv_comment + 'Number of files merged:' + IntToStr (priv_params.InterfaceMapper.TraceFileList.Count) + CRLF;

             priv_tracewriter.HeaderCommentList.Add (priv_comment);
        end;

     // Write header

     priv_tracewriter.WriteGenericHeader (priv_outputinterfaces);
end;

(* ------------------------------------------- *)

procedure tTaskProcessorMerge.CreateProcessChain;

var processor : tFrameProcessor;

begin
     processor := tFrameProcessor.create ('Output', priv_tracewriter);

     priv_processchain.AddProcessor(processor);

     // Check if filters are enabled and at least one filter in the list
     // if so, add a filter processor to the chain

//   if priv_params.FilterFrames and (length (priv_params.FilterSet) > 0) then
     if priv_params.FilterFrames and (priv_params.FilterSet.count > 0) then
        begin
             processor := tFrameProcessorFilter.create ('FilterSet', priv_params.FilterSet);

             priv_processchain.AddProcessor (processor);
        end;
end;

(* ------------------------------------------- *)

function tTaskProcessorMerge.OpenReaderForFile (Entry : tTraceFileEntry) : tTraceLoader;

var loader : tTraceLoader;

begin
     case entry.TracefileType of
      cfENC    : loader := tENCLoader.create (Entry.FullName);
      cfCAP    : loader := tCAPLoader.create (Entry.FullName);
      cfPCAP   : loader := tPCAPLoader.create (Entry.FullName);
      cfPCAPng : loader := tPCAPngLoader.create (Entry.FullName);
     end;

     loader.openFile;
     loader.readTraceFileheader;

     result := loader;
end;

(* ------------------------------------------- *)

function tTaskProcessorMerge.SetFileTimestamp: integer;

var Filename  : string;
    FileEntry : tTraceFileEntry;
    Timestamp : tDateTime;
    rCode     : integer;

begin
     Filename := priv_params.MergeTargetFilename;

     case priv_params.AdjustTimestampMode of
      tsmOriginalFile : Timestamp := priv_firstfiletimestamp;
      tsmFirstFrame   : Timestamp := priv_lowesttimestamp.datetime;
      tsmLastFrame    : Timestamp := priv_highesttimestamp.datetime;
     end;

     if timestamp <> 0 then
        rcode := FileSetDate (filename, DateTimeToFileDate(timestamp));

     result := rCode;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorMerge.ConcatenateFiles;

var FileEntry   : tTracefileEntry;
    IfaceEntry  : tCaptureInterface;
    TraceLoader : tTraceLoader;
    Frame       : tRawFrame;
    Index       : integer;
    i,j         : integer;
    comment     : string;
    writecmnt   : boolean;

    test        : tMD5HashString;

begin
     // Create Target file

     PrepareTargetFile;
     CreateProcessChain;

     // Process all files

     for i := 1 to priv_params.InterfaceMapper.TraceFileList.count do
         begin
              FileEntry := priv_params.InterfaceMapper.TraceFileList.getFile(i);

              // Keep the file timestamp of the first file for later

              if i = 1 then
                 priv_firstfiletimestamp := FileEntry.FileDate;

              // Open the file and read it

              TraceLoader := OpenReaderForFile (FileEntry);

              comment := FileEntry.FullName;
              writecmnt  := priv_params.AddSourceFilePacketComment;

              repeat
                    Traceloader.readBatchFromDisk;

                    // Process available frames

                    for j := 1 to Traceloader.FramesAvailable do
                        begin
                             Frame := Traceloader.getRawFrame(j);

                             // Assign the interface to the frame. InterfaceLinkID is used as a failsafe to make sure the frame
                             // is written with the correct interface index

                             IFaceEntry := FileEntry.CaptureInterfaces.getInterfaceByIndex (Frame.InterfaceID);

                             Frame.InterfaceID := IFaceEntry.MappedToInterface.InterfaceID;
                             Frame.InterfaceLinkID := IFaceEntry.MappedToInterface.PacketLinkID;

                             if writecmnt then
                                Frame.Commentlist.Add (comment);

                             // Add frame to chain

                             priv_processchain.ProcessFrame (Frame);

                             // if chain discarded frame (because of filter), keep comment, otherwise discard

                             if not priv_processchain.FrameWasDiscarded then
                                begin
                                     comment := '';
                                     writecmnt := False;
                                end;
                        end;

              until Traceloader.EOF;
         end;

     // Release process chain, which will transfer the remaining frames
     // from the frame storage to the writer

     priv_processchain.free;

     // Finalize Writer

     priv_tracewriter.Flush;

     // Get timestamp values and clean up

     priv_lowesttimestamp := priv_tracewriter.FirstFrameTimestamp;
     priv_highesttimestamp := priv_tracewriter.LastFrameTimestamp;

     priv_tracewriter.Close;
     priv_tracewriter.free;
     priv_tracewriter := nil;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorMerge.MergeFilesByTimestamp;

begin



end;

(* ------------------------------------------- *)

function tTaskProcessorMerge.Execute : tTaskErrorCode;

begin
     // create the list of interfaces that will end up in the merged file

     PrepareOutputInterfaceList;

     // Merge depending on mode

     case priv_params.MergeMode of
      mmConcatenate      : ConcatenateFiles;           // Simple. Sort of :-)
      mmMergeByTimestamp : MergeFilesByTimestamp;      // Hard. Well, complicated. :-)
     end;

     if priv_params.AdjustFileTimestamp then
        SetFileTimestamp;

     priv_outputinterfaces.free;
end;

(* ------------------------------------------- *)

end.

