{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_taskprocessor_separate;

interface

uses Classes,
     base_structures,
     class_tracefilelist,
     class_taskprocessor_baseclass,
     pcapng_structures;

type tSeparationParameters = record
                                   OverwriteFileByDefault : boolean;

                             end;

     tOnOutputFileExists = procedure (const Filename : string; var Overwrite : boolean) of object;
     tOnExtractFileSection = procedure (const Filename : string; const SectionIndex : integer) of object;

     tTaskProcessorSeparate = class (tTaskProcessor)

                                    constructor create (InputFile : tTracefileEntry; Output : tOutputlocation; const Parameters: tSeparationParameters);
                                    procedure free;

                                    private
                                           priv_params             : tSeparationParameters;
                                           priv_Outputfile         : string;

                                           EventOnOutputFileExists : tOnOutputFileExists;
                                           EventOnExtractSection   : tOnExtractFileSection;

                                           {$REGION 'Documentation'}
                                           ///	<summary>
                                           ///	  Finds the next section node in
                                           ///	  the block chain, skipping over
                                           ///	  all others
                                           ///	</summary>
                                           ///	<param name="node">
                                           ///	  Current Node to start the
                                           ///	  search from. Should not be a
                                           ///	  section node unless it is the
                                           ///	  first in the chain.
                                           ///	</param>
                                           {$ENDREGION}
                                           function getNextSectionNode (node : pBlockChainEntry) : pBlockChainEntry;

                                           {$REGION 'Documentation'}
                                           ///	<summary>
                                           ///	  Extracts a section from the
                                           ///	  combined file.
                                           ///	</summary>
                                           ///	<param name="node">
                                           ///	  Section Block header
                                           ///	  containing start and end
                                           ///	  offsets of the current section
                                           ///	</param>
                                           {$ENDREGION}
                                           procedure ExtractSection (node : pBlockChainEntry);

                                    protected

                                    public
                                          {$REGION 'Documentation'}
                                          ///	<summary>
                                          ///	  Reports when a new section is
                                          ///	  exported. Filename and index of
                                          ///	  the current section is given in
                                          ///	  the parameters.
                                          ///	</summary>
                                          {$ENDREGION}
                                          property OnExtractFileSection : tOnExtractFileSection read EventOnExtractSection write EventOnExtractSection;
                                          property OnOutputFileExists : tOnOutputFileExists read EventOnOutputFileExists write EventOnOutputFileExists;

                                          function execute : tTaskErrorCode; override;
                              end;


implementation

uses SysUtils;

(* ------------------------------------------- *)
{ tTaskProcessorSeparate }
(* ------------------------------------------- *)

constructor tTaskProcessorSeparate.create (InputFile : tTracefileEntry; Output : tOutputlocation; const Parameters: tSeparationParameters);

begin
     inherited create (InputFile, Output);

     priv_params := Parameters;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorSeparate.free;

begin
     inherited;
end;

(* ------------------------------------------- *)

function tTaskProcessorSeparate.getNextSectionNode (node: pBlockChainEntry): pBlockChainEntry;

begin
     while (node <> nil) and (node^.BlockTypeID <> BLOCKTYPE_SECTION_HEADER) do
           node := node^.NextBlockType;

     result := node;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorSeparate.ExtractSection(node: pBlockChainEntry);

var StartOffset : integer;
    EndOffset   : integer;
    length      : integer;
    Source      : tFileStream;
    Destination : tFileStream;

begin
     StartOffset := node^.SectionStartOffset;        // TODO: deal with 64 bits
     EndOffset := node^.SectionEndOffset;
     length := (EndOffset - StartOffset) + 1;

     Source := TFileStream.Create (prot_inputfile.FullName, fmOpenRead, fmShareDenyNone);
     Destination := TFileStream.Create(priv_outputfile, fmCreate, fmShareDenyWrite);

     // Copy from start offset to end offset

     Source.Seek (StartOffset, soFromBeginning);
     Destination.CopyFrom(Source, length);

     Source.Free;
     Destination.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorSeparate.execute: tTaskErrorCode;

var iCount           : integer;
    Stat             : tPCAPNGStatistic;
    CurrentChainNode : pBlockChainEntry;
    boOverwrite      : boolean;
    boSkipFile       : boolean;
    outputpath       : string;

begin
     Stat := prot_Inputfile.PCAPngStatistics;
     CurrentChainNode := prot_Inputfile.BlockChainRoot;
     iCount := 0;

     if Stat.SectionCount > 1 then
        begin
             repeat
                   inc (iCount);
                   boSkipFile := false;

                   // Generate output file name

                   prot_OutputLocation.SectionID := iCount;
                   priv_Outputfile := prot_OutputLocation.getOutputfilename;
                   outputpath := ExtractFilePath (priv_Outputfile);

                   // Check if it already exists

                   if FileExists (priv_Outputfile) then
                      begin
                           if assigned (EventOnOutputFileExists) then
                              EventOnOutputFileExists (priv_Outputfile, boOverwrite)
                           else
                               boOverwrite := priv_params.OverwriteFileByDefault;

                           if boOverwrite then
                              try
                                 DeleteFile (priv_Outputfile);
                              except
                                    boSkipFile := true;
                              end
                           else
                               boSkipFile := true;
                      end
                   else
                       begin
                            // Check if the output directory exists and create it if not

                            if not directoryexists (outputpath) then
                               begin
                                    try
                                       ForceDirectories (outputpath);
                                    except
                                          // TODO: handle I/O errors
                                    end;
                               end;
                       end;

                   // Extract Section

                   CurrentChainNode := getNextSectionNode (CurrentChainNode);

                   if (CurrentChainNode <> nil) and not boSkipFile then
                      begin
                           if assigned (EventOnExtractSection) then
                              EventOnExtractSection (priv_Outputfile, iCount);

                           ExtractSection (CurrentChainNode);
                      end;

                   // Prepare for next node

                   CurrentChainNode := CurrentChainNode.NextBlockType;

             until iCount >= Stat.SectionCount;
        end;
end;

(* ------------------------------------------- *)

end.
